//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package eg;

import la.util.*;        // x2String and error
import static la.util.Util.*;
import la.la.Type;
import static la.la.Type.*;
import la.la.Value;
import static la.la.Value.*;
import la.maths.Vector;  // A data-set is a Vector of Values
import mml.*;            // For Model, UPModel, Normal, Mixture, Tree, etc.
import graph.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.
 *
 * Reference: 
 *   Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.
 *
 * This file is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (GPL) as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this file. If not, see <http://www.gnu.org/licenses/>.
 */
// ----------------------------------------------------------------------------

/** Using various {@link mml.README MML} Models of {@link Graphs},
    analyse a given {@link Graph Graph} defined in a file.
    Also see
    [<A HREF="https://doi.org/10.1007/978-3-319-76433-7" TARGET="BOOKA">the&nbsp;book</A>]
    and the
    [<A HREF="https://www.cantab.net/users/mmlist/MML/A/" TARGET="mmlist">home&nbsp;page</A>]. */
public class Graphing
 {
   private static la.util.Timer timer = new la.util.Timer();

   private static void markTime()
    { long t = timer.markTime();
      System.out.println( "-- " + (double)t/1000 + " secs (total="
                        + (double)timer.totalTime()/1000 + ") --" );
    }
   // -------------------------------------------------------------------------

   /** 'CHARS&nbsp;&times;&nbsp;CHARS' Type, i.e.,
        a pair of {@link Type#CHARS CHARS} (Strings). */
   public static Type.Tuple Chars2 =
      new Type.Tuple.GP("string^2", new Type [] {Type.CHARS, Type.CHARS});

   /** The Type is {@link Directed#unlabelled} in this case.  */
   public static graph.Type Gtype =                 // Gtype
      Directed.unlabelled;  // at the moment ?

   /** Poisson1 =
       {@link MML#Poisson0 Poisson0}.{@link Discretes#shifted shifted(1)}.  */
   public static Discretes Poisson1 = MML.Poisson0.shifted(1);    // for n > 0

   /** Some possible motifs (subgraphs).
       'C'&nbsp;cyclic,
       'K'<SUB>|V|</SUB>&nbsp;complete,
       'K<SUB>mn</SUB>'&nbsp;complete bipartite,
       'D'&nbsp;bi<U>d</U>irectional,
       FFL&nbsp;feed-forward loop, STAR<SUB>|V|</SUB>&nbsp;~hub,
       M-shaped and W-shaped.  */
   public static final Graph
         K5  = Directed.dense(new int[][]{ {0, 1, 1, 1, 1},  // komplete
                                           {1, 0, 1, 1, 1},
                                           {1, 1, 0, 1, 1},
                                           {1, 1, 1, 0, 1},
                                           {1, 1, 1, 1, 0} } ),
         K22 = Directed.dense(new int[][]{ {0, 0, 1, 1},  // bipartite
                                           {0, 0, 1, 1},  // aka bifan
                                           {0, 0, 0, 0},
                                           {0, 0, 0, 0}}),
         K23 = Directed.dense(new int[][]{ {0, 0, 1, 1, 1},
                                           {0, 0, 1, 1, 1},
                                           {0, 0, 0, 0, 0},
                                           {0, 0, 0, 0, 0},
                                           {0, 0, 0, 0, 0}}),
         K33 = Directed.dense(new int[][]{ {0, 0, 0, 1, 1, 1},
                                           {0, 0, 0, 1, 1, 1},
                                           {0, 0, 0, 1, 1, 1},
                                           {0, 0, 0, 0, 0, 0},
                                           {0, 0, 0, 0, 0, 0},
                                           {0, 0, 0, 0, 0, 0} } ),
         KD23 = Directed.dense(new int[][]{{0, 0, 1, 1, 1}, // dbl bipartite
                                           {0, 0, 1, 1, 1},
                                           {1, 1, 0, 0, 0},
                                           {1, 1, 0, 0, 0},
                                           {1, 1, 0, 0, 0} } ),
         KD33 = Directed.dense(new int[][]{{0, 0, 0, 1, 1, 1},
                                           {0, 0, 0, 1, 1, 1},
                                           {0, 0, 0, 1, 1, 1},
                                           {1, 1, 1, 0, 0, 0},
                                           {1, 1, 1, 0, 0, 0},
                                           {1, 1, 1, 0, 0, 0} } ),
         C3  = Directed.dense(new int[][]{ {0, 1, 0},        // single cycle 3
                                           {0, 0, 1},
                                           {1, 0, 0} } ),
         C4  = Directed.dense(new int[][]{ {0, 1, 0, 0},     // single cycle 4
                                           {0, 0, 1, 0},
                                           {0, 0, 0, 1},
                                           {1, 0, 0, 0} } ),
         CD4 = Directed.dense(new int[][]{ {0, 1, 0, 1},     // double cycle 4
                                           {1, 0, 1, 0},
                                           {0, 1, 0, 1},
                                           {1, 0, 1, 0} } ),
         CD5 = Directed.dense(new int[][]{ {0, 1, 0, 0, 1},  // double cycle 5
                                           {1, 0, 1, 0, 0},
                                           {0, 1, 0, 1, 0},
                                           {0, 0, 1, 0, 1},
                                           {1, 0, 0, 1, 0} } ),
         CD6 = Directed.dense(new int[][]{ {0, 1, 0, 0, 0, 1},  // dbl cycle 6
                                           {1, 0, 1, 0, 0, 0},
                                           {0, 1, 0, 1, 0, 0},
                                           {0, 0, 1, 0, 1, 0},
                                           {0, 0, 0, 1, 0, 1},
                                           {1, 0, 0, 0, 1, 0} } ),
         STAR3 = Directed.dense(new int[][]{{0,1,1},  // aka hub
                                            {0,0,0},
                                            {0,0,0}}),
         STAR4 = Directed.dense(new int[][]{{0,1,1,1},
                                            {0,0,0,0},
                                            {0,0,0,0},
                                            {0,0,0,0}}),
         STAR5 = Directed.dense(new int[][]{{0,1,1,1,1},
                                            {0,0,0,0,0},
                                            {0,0,0,0,0},
                                            {0,0,0,0,0},
                                            {0,0,0,0,0}}),
         STAR6 = Directed.dense(new int[][]{{0,1,1,1,1,1},
                                            {0,0,0,0,0,0},
                                            {0,0,0,0,0,0},
                                            {0,0,0,0,0,0},
                                            {0,0,0,0,0,0},
                                            {0,0,0,0,0,0}}),
         W5    = Directed.dense(new int[][]{{0,0,0,1,0}, // letter 'W'
                                            {0,0,0,1,1},
                                            {0,0,0,0,1},
                                            {0,0,0,0,0},
                                            {0,0,0,0,0}}),
         M5    = Directed.dense(new int[][]{{0,0,1,1,0}, // letter 'M'
                                            {0,0,0,1,1},
                                            {0,0,0,0,0},
                                            {0,0,0,0,0},
                                            {0,0,0,0,0}}),
         FFL = Directed.dense(new int[][]{{0,1,1}, {0,0,1}, {0,0,0}}), // ~Alon
         FFL4 = Directed.dense(new int[][]{ {0, 1, 1, 0},  // Maayan 2005
                                            {0, 0, 0, 1},
                                            {0, 0, 0, 1},
                                            {0, 0, 0, 0} } );
   // -------------------------------------------------------------------------

   /** Perform some elementary tests of Graph Models on
       random Graphs of 'V' vertices. */
   public static void tests(int V)                                  // tests(V)
    {
      int Ngraphs = 1;  // #Graphs per data-set
      System.out.println( "-- tests(|V|=" + V + ") "
                        + " x" + Ngraphs + " --" );
      la.util.Timer timer = new la.util.Timer(); timer.start();

      Value.Int V_int = new Value.Int(V);  // V as an Int
      Value.Cts V_cts = Value.real(V);     // V as a  Cts
      Discretes V_upm = new Discretes.Uniform(pair(V_int, V_int));  // !

      double pEdge = 0.07;      // density, pr(Edge(v,w)) exists
      Vector odds_E = Vector.doubles(new double [] {1 - pEdge, pEdge}),
             alpha  = Vector.doubles(new double [] {1 / pEdge, 1});

      // A number of UnParameterised and Parameterised Models of Graphs...
   // (i)
      Graphs.GERfixed     UGERf = new Graphs.GERfixed(pair(Gtype, V_upm));
      Graphs.GERfixed.M    GERf = UGERf.sp2Model(0, 0, pair(triv, odds_E));

   // (ii)
      Graphs.GERadaptive  UGERa = new Graphs.GERadaptive(
                            triple(Gtype, V_upm, alpha) );
      Graphs.GERadaptive.M GERa = UGERa.sp2Model(0, 0, triv);
   // (iii)
      Graphs.Skewed         USk = new Graphs.Skewed(pair(Gtype, V_upm));
      Graphs.Skewed.M        Sk = USk.sp2Model(0, 0, triv);

      Vector      motifs = Vector.values(new Graph [] {KD23, K5});
      Value.Tuple bounds = pair(three, six);  // on Motif (subgraph) sizes
   // (iv)
      MotifD    UMD = new MotifD(triple(Gtype, V_upm, bounds));
      MotifD.M   MD = UMD.sp2Model(0, 0, pair(triv, motifs));

   // (v)
      MotifA    UMA = new MotifA(triple(Gtype, V_upm, bounds));
      MotifA.M   MA = UMA.sp2Model(0, 0, pair(triv, motifs));

      Graphs   [] UGMs = new Graphs[]  {UGERf, UGERa, USk, UMD, UMA};
      Graphs.M []  GMs = new Graphs.M[]{ GERf,  GERa,  Sk,  MD,  MA};

      for( int i = 0; i < GMs.length; i ++ )
       { Graph [] Gs = new Graph [Ngraphs];
         int randomised = 0;
         while( randomised < 1 && Gs != null )
          { System.out.print(
                    (randomised == 0 ? "\n random source"
                                     : " randomised" + randomised )
                  + " #" + i + ", |E|=" );
            try{ for( int g = 0; g < Gs.length; g ++ )
                  { if( randomised == 0 )
                       Gs[g] = GMs[i].random();
                    else // i.e., 0
                       Gs[g] = Gs[g].randomised();
                    System.out.print( Gs[g].eSize() + "," );
                  }
               }catch(Exception e)
                 { Gs = null; System.out.println( e ); continue; }
            System.out.print( ":\n   " );
            Vector ds = Vector.values(Gs);      // data-set, ds
            for( int j = 0; j < GMs.length; j ++ )
             { Graphs.M  m = GMs[j];
               double ds_m = m.nlLH(m.stats(ds));  // (ds|m), nits
               System.out.print( j+":"+x2String(ds_m / MML.log2) + " bits, " );
             }//for j
            System.out.println();
            randomised ++ ;
          }
       }//for i

   // for( int i = 3; i < 6; i ++ )
   //  { Graph.SubGraphs sg = G.subGraphs(i, i);
   //    la.util.Timer timer = new la.util.Timer(); timer.start();
   //    while( sg.hasSome() )
   //     { Graph subg = sg.elt(); sg.advance(); } // either or...
   //     { Graph subg = sg.elt().canonical(); sg.advance(); }
   //    int n = sg.position();  // NB. enumeration only
   //    double t = (double)timer.markTime()/1000;
   //    System.out.println( n + " subgraphs, |V|=" + i +", " + t + "secs" );
   //  }//for i
      // -----                enumerate only;    canonical() each one    ver.2
      //      6750 subgraphs, |V|=3,   0.0secs;    ( 0.0 secs)             0.1
      //     95939 subgraphs, |V|=4,   0.0secs;    ( 1.0 secs)             1.5
      //   1401655 subgraphs, |V|=5,   4.0secs;    (77.0 secs ~ 4x19)     45.6
      //  20317666 subgraphs, |V|=6,   70.0secs;
      // 287391754 subgraphs, |V|=7, 1004.0secs;
      // n and time grow by about x 14,  2 x 2 Ghz iMac.

      System.out.println( "\n-- done tests(" + V + "), "
                        + (timer.markTime()/1000) + " secs --" );
    }//tests()
   // -------------------------------------------------------------------------

   /** Get the raw String data (two columns: v<SUB>0</SUB>&nbsp;v<SUB>1</SUB>)
       defining a Graph from the named file using {@link la.maths.Vector#csv}.
       Used in {@link #getGraph(String) getGraph(fn)}.  */
   public static Vector getData(String filename)                 // getData(fn)
    { java.io.FileInputStream inp = null;
      try{ inp = new java.io.FileInputStream(filename);
         }catch(Exception e){ Util.error("getData("+filename+"): "+e); }
      return Vector.csv(false, false, /* separator= */ '\t', Chars2, 0, inp);
    }//getData(filename)

   /** Get the graph from 'filename'.  Uses {@link #getData getData(...)}. */
   public static Graph getGraph(String filename)                // getGraph(fn)
    { Vector raw = getData( filename );
      final int E = raw.nElts();                                           // E
      System.out.println( E + " lines" );

      Value [] ids = raw.col(0).append( raw.col(1) ) .uniqueElts();
      System.out.println( Vector.values( ids ).nElts() + " distinct names" );

      final int V = ids.length;                                            // V
      String [] idStrings = new String[ V ];
      for( int i = 0; i < V; i ++ )
         idStrings[i] = ((Chars) ids[i]) . s;
      Type.Enum vLabelType = new Type.Enum("nodeLabel", idStrings);

      final int [] [] edges = new int [E] [2];
      int lastPrinted = -2;
      for( int i = 0; i < E; i ++ )
       { String s0 = ((Chars) raw.elt(i,0)).s,
                s1 = ((Chars) raw.elt(i,1)).s;
         int u = vLabelType.string2n(s0);
         edges[i][0] = u;
         int v = vLabelType.string2n(s1);
         edges[i][1] = v;
         boolean loop = u == v;
         if( i < 5 || i >= E-5 || loop )
          { System.out.println( "Edge " + i + ": ("
                              + edges[i][0] + ", " + edges[i][1] + ") ~ ("
                              + s0 + ", " + s1 + ")"
                              + (loop ? " --loop" : "") );
            lastPrinted = i;
          }
         else if( i == (lastPrinted + 1) ) System.out.println( " ..." );
       }//for i
      System.out.println();

      return new Directed.Sparse(edges);
    }//getGraph(filename)
   // -------------------------------------------------------------------------

   /** For UnParameterised {@link mml.Graphs.Motifs Motifs} Model of
       Graphs 'UMM', perform a
       {@link mml.Graphs.Motifs#greedy(Value,Vector,Value) greedy(,,)}
       search to select zero or more of several given candidate
       'motifs' (subGraphs) to form a fully-parameterised Model.
       'Gs'&nbsp;is the data-set of Graphs being modelled. */
   public static void greedy(Graphs.Motifs UMM, Graph [] motifs,
                             Value.Cts V_cts, Graph [] Gs)   // g1
    { Graphs.Motifs.M mdl =
         UMM.greedy( V_cts,  // ps for estimator
                     Vector.values(motifs),
                     Vector.values( Gs ) );
      System.out.println( "greedy, best= " + mdl + ", "
         + mdl.msgBits() + "=" + mdl.msg1bits() + "+" + mdl.msg2bits()
         + " bits" );
      System.out.println( "............................" );
    }//greedy(...)

   /** For each of several UnParameterised Motifs Models of Graphs
       UMMs[0], UMMs[1],&nbsp;...,&nbsp; and data-set of Graphs 'Gs' do
   {@link #greedy(Graphs.Motifs,Graph[],Value.Cts,Graph[]) greedy(,,,Gs)}. */
   public static void greedy(Graphs.Motifs [] UMMs, Graph [] motifs,
                             Value.Cts V_cts, Graph [] Gs)   // g2
    { for( int i = 0; i < Gs.length; i ++ )
       { greedy(UMMs[i], motifs, V_cts, Gs );  // calls g1 above
         System.out.println( "----------------------------" );
       }
    }//greedy(...)

   /** For each of several UnParameterised Motifs Models of Graphs
       UMMs[0], UMMs[1],&nbsp;...,&nbsp; and <I>one</I> Graph 'G' do
   {@link #greedy(Graphs.Motifs,Graph[],Value.Cts,Graph[]) greedy(,,,[G])}. */
   public static void greedy(Graphs.Motifs [] UMMs, Graph [] motifs,
                             Value.Cts V_cts, Graph G)    // g3
    { greedy( UMMs, motifs, V_cts, new Graph [] {G} ); }  // calls g2 above

    /** For UnParameterised {@link mml.Graphs.Motifs Motifs} Model of
       Graphs 'UMM' and <I>one</I> Graph 'G' do
    {@link #greedy(Graphs.Motifs,Graph[],Value.Cts,Graph[]) greedy(,,,[G])}. */
   public static void greedy(Graphs.Motifs UMM, Graph [] motifs,
                             Value.Cts V_cts, Graph G)    // g4
    { greedy( UMM, motifs, V_cts, new Graph [] {G} ); }   // calls g1 above
   // -------------------------------------------------------------------------

   /** Analyse the Graph defined in the nominated file. &nbsp;
       <NOBR>'<TT>Graphing [-r] [filename]</TT>'</NOBR> &nbsp;
       <NOBR>'-r': randomise the Graph.</NOBR>  */
   public static void main(String[] argv)                               // main
    { System.out.print("-- run Graphing " );
      for( int i = 0; i < argv.length; i ++ )
            System.out.print( argv[i] + " " );
      System.out.println( "--" );

      tests(60);  // random sources v. Models

      boolean randomised = false;  // ? scramble the input Graph ?
      int argvPosn = 0;
      String filename = "";
      if( argv.length > 0 && argv[0].equals( "-r" ) )
       { randomised = true; argvPosn = 1; }
      if(      argvPosn <  argv.length - 1 ) error( "|argv|=" + argv.length );
      else if( argvPosn == argv.length - 1 ) filename = argv[argvPosn];
      else  // argvPosn >  argv.length - 1
         filename =
      //  "../data/YeastCellCycle";      //  8K, V=296, E=550
      //  "../data/EcoliTransRegNoAuto"; //  6K, V=436, E=530 (raw + 3 loops)
          "../data/YeastDiauxic";        // 19K, V=783, E=1217
      //  "../data/Maayan-SigTransNet-edges"; // 10K, |V|=594, |E|=1422
      //  "../data/YeastStatic";         // 110K, V=3459, E=7053, too BIG
                                      //   (raw V=3459, E=7074 & loops)
      System.out.println( "\nfilename = " + filename );

// "null test" ...
// Graphs.GERadaptive     UGen = new Graphs.GERadaptive(
//       triple(Gtype, Poisson1, Vector.doubles(new double[]{100,1})));
//    Graphs.GERadaptive.M Gen = UGen.sp2Model(0, 0, Value.real(100));
// (ii)
// Graphs.Skewed        Usk = new Graphs.Skewed(pair(Gtype, Poisson1));
//    Graphs.Skewed.M   Gen = Usk.sp2Model(0, 0, Value.real(100));
//    Graph G = Gen.random().byDegree();

      Graph G = getGraph(filename);  // G
      if( randomised ) G = G.randomised();

      final int V = G.vSize();                    // |V| as int
      final Value.Cts V_cts = Value.real(V);      // |V| as cts
      Value.Int V_int = new Value.Int(V);         // |V| as Int

      System.out.println( "   " + filename
                        + (randomised ? " * randomised * " : "")
                        + ": |V|=" + G.vSize()
                        + ", |E|=" + G.eSize() );

// for( int i = 3; i < 6; i ++ )
//  { Graph.SubGraphs sg = G.subGraphs(i, i);
//    la.util.Timer timer = new la.util.Timer(); timer.start();
//    while( sg.hasSome() )
//     { Graph subg = sg.elt(); sg.advance(); } // either or...
//     { Graph subg = sg.elt().canonical(); sg.advance(); }
//    int n = sg.position();  // NB. enumeration only
//    double t = (double)timer.markTime()/1000;
//    System.out.println( n + " subgraphs, |V|=" + i +", " + t + "secs" );
//  }//for i
// -----                  enumerate only;    canonical() each one    ver.2
//      6750 subgraphs, |V|=3,   0.0secs;    ( 0.0 secs)             0.1
//     95939 subgraphs, |V|=4,   0.0secs;    ( 1.0 secs)             1.5
//   1401655 subgraphs, |V|=5,   4.0secs;    (77.0 secs ~ 4x19)     45.6
//  20317666 subgraphs, |V|=6,   70.0secs;
// 287391754 subgraphs, |V|=7, 1004.0secs;
// n and time grow by about x 14,  2 x 2 Ghz iMac.
      // ................................................................

      Graph [] Gs = new Graph [] { G };                    // Gs = [G]
      Vector ds = Vector.values( Gs );            // "dataset", ds=[G]

      double pEdge = 0.01;      // Edge density, pr(Edge(v,w)) exists
      Vector odds_E = Vector.doubles(new double [] {1-pEdge, pEdge}),
            odds_E2 = Vector.doubles(new double [] {1-2*pEdge, 2*pEdge});

      Graphs.GERfixed  UGER = new Graphs.GERfixed(pair(Gtype, Poisson1));
      Graphs.GERfixed.M GER = UGER.sp2Model(0, 0, pair(V_cts, odds_E));  // GER

      Graphs.GERadaptive  UGERa = new Graphs.GERadaptive(
         triple(Gtype, Poisson1, Vector.doubles(new double[]{100,1})));
      Graphs.GERadaptive.M GERa = UGERa.sp2Model(0, 0, V_cts);          // GERa

      Graphs.Skewed  USk = new Graphs.Skewed(pair(Gtype, Poisson1));
      Graphs.Skewed.M Sk = USk.sp2Model(0, 0, V_cts);                     // Sk

      Graph [] motifs  = new Graph []                       // candidate motifs
                       {STAR3, STAR4, STAR5, C3, C4,
                        K22, K23,
                        W5, M5,
                        FFL, FFL4};

      Value.Tuple bounds = pair(three, five);          // range of Motifs sizes

      MotifD UMD = new MotifD(triple(Gtype, Poisson1, bounds));          // UMD
      MotifA UMA = new MotifA(triple(Gtype, Poisson1, bounds));          // UMA

timer.start();
      System.out.println( GER .nl2Pr(G) + " bits by GER\n"
                        + GERa.nl2Pr(G) + " bits by GERa\n"
                        + Sk  .nl2Pr(G) + " bits by Sk"
                        );
markTime();
System.out.println();

      greedy( UMD, motifs, V_cts, Gs );
markTime();
      greedy( UMA, motifs, V_cts, Gs );
markTime();

      MotifD.M MDest = (MotifD.M) UMD.estimator(V_cts).ds2Model(ds);   // MDest
markTime();
      MotifA.M MAest = (MotifA.M) UMA.estimator(V_cts).ds2Model(ds);   // MAest
markTime();

      System.out.println( "\n|ds| = " + ds.nElts() + "\n"
       + MDest.msgBits() + "=" + MDest.msg1bits() + "+" + MDest.msg2bits() +
            " bits\n   by Estimated: " + MDest + ",\n"
       + MAest.msgBits() + "=" + MAest.msg1bits() + "+" + MAest.msg2bits() +
            " bits\n   by Estimated: " + MAest + ",\n"
       + GER .nl2LH(GER .stats(ds)) + " by " + GER  + ",\n"
       + GERa.nl2LH(GERa.stats(ds)) + " by " + GERa + ",\n"
       + Sk  .nl2LH(Sk  .stats(ds)) + " by " + Sk );

markTime();
      System.out.println("-- done Graphing --");
    }//main

 }//Graphing class

// (c) L.Allison, Dec 2014, Mon 24 July 2017. Licence to 'Graphing' is Gnu GPL.
