//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package graph;

import la.util.Series;
import la.la.Value;
import static la.la.Value.*;
import la.maths.Vector;
import mml.MML;
import mml.Graphs;
import mml.MotifD;
import mml.Estimator;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** Run a few(!), simple(!) tests on {@link Graph Graphs}. */
public class Test                                                       // Test
 {
   private static void show(Graph g)
    { show( "", g); }

   /** Print some information about 'g'. */
   private static void show(String title, Graph g)
    { System.out.println( "--" + title + " " + g.getClass().getName() + " " 
              + g + " : " + g.type()
              + ", |V|=" + g.vSize() + ", |E|=" + g.eSize()
              + ", |A|=" + g.nAutomorphisms()
              );
      System.out.println( "   .A()=" + g.A() );

      int [] ns = new int [g.vSize()];  // 'ns', a random permutation...
      (new mml.Permutation.Uniform(new Int(ns.length))) .Mdl .random(ns);
      Graph g2 = g.renumbered(ns);      // We should have g~g2
      if( g.isomorphic(g.renumbered(ns)) ) System.out.print( "   pass" );
      else System.out.println( "*** fail, g : " + g.A()
                          + "\n    & : " + g.canonical().A()
                          + "\n    g2: " + g2.A()
                          + "\n    & : " + g2.canonical().A() );

      Directed gd = g.isDirected() ? g.toDirected() : null;
      System.out.print( "   " );
      for( int v = 0; v < g.vSize(); v ++ )
         System.out.print( "v" + v + "(--" + g.degree(v)
         + (gd != null ? "(=" + gd.degree(v) + ")"
                       + " <-"+gd.inDegree(v)+" ->"+ gd.outDegree(v)
                       : "")
         + "), " );
      System.out.println();

      Series j0 = g.joinedTo(0);
      System.out.print( "   v0--" );
      while( j0.hasSome() )
       { System.out.print( "v" + j0.elt() + "," ); j0.advance(); }

      if( gd != null )
       { Series s0 = gd.directSuccessors(0);
         System.out.print( "   v0->" );
         while( s0.hasSome() )
          { System.out.print( "v" + s0.elt() + "," ); s0.advance(); }

         Series p0 = gd.directPredecessors(0);
         System.out.print( "   v0<-" );
         while( p0.hasSome() )
          { System.out.print( "v" + p0.elt() + "," ); p0.advance(); }
       }//if

      System.out.println();
    }//show(g)

   private static void showSub(Graph g, int hi) { showSub(g, 1, hi); }

   private static void showSub(Graph g, int lo, int hi)
    { System.out.println( "-- " + g + ".subGraphs("
                        +  lo + ".." + hi + ") :" );
      Series ss = g.subGraphs(lo, hi);
      while( ss.hasSome() )
       { System.out.println( "   " + (ss.position()+1) + ": " + ss.elt() );
         ss.advance();
       }
      System.out.println( "-- length=" + ss.position() + " -------" );
    }
   // .........................................................................

   /** Run a few, very(!) simple tests. */
   public static void main(String[] argv)                               // main
    { System.out.println("-- run graph.Test --");

      Directed box = Directed.dense(new int[][]{{0, 1, 1, 0},  // 0-->-1
                                                {0, 0, 1, 1},  // |  / |
                                                {0, 0, 0, 1},  // v /  v
                                                {0, 0, 0, 0}});// 2-->-3
      show(box);
      show(box.byDegree());
      showSub(box, 2, 3);

      Undirected  K5 = new Undirected.K(5);
      show(K5);
      show(K5.induced(new int[]{0,2,3}));

      Directed KD23 = Directed.dense(new int[][]{{0, 0, 1, 1, 1}, // dbl 
                                                 {0, 0, 1, 1, 1},
                                                 {1, 1, 0, 0, 0},
                                                 {1, 1, 0, 0, 0},
                                                 {1, 1, 0, 0, 0} } );

      System.out.println( "KD23:" );
      show(KD23);
      System.out.println( "KD23.canonical():" );
      show(KD23.canonical());
      Graph KD23rn = KD23.renumbered( new int [] {4, 3, 2, 1, 0 } );
      System.out.println( "KD23rn:" );
      show(KD23rn);
      System.out.println( "KD23rn.canonical():" );
      show(KD23rn.canonical());
      System.out.println( "KD23rn is "
                        + (KD23rn.isomorphic(KD23) ? "really" : "NOT")
                        + " isomorphic to KD23" );
      System.out.println();

      Directed    C4 = new Directed.C(4);
      show(C4);
      Directed empty = Directed.dense(new int[][]{ {0, 0, 0, 0, 0},
                                                   {0, 0, 0, 0, 0},
                                                   {0, 0, 0, 0, 0},
                                                   {0, 0, 0, 0, 0},
                                                   {0, 0, 0, 0, 0} } );
      showSub(C4,    2, 3); // ZZZ
      showSub(empty, 3, 4); // ZZZ
      show(C4.induced(new int[]{0,2,3}));
      show(C4.canonical());
      show(C4.asUndirected());

      Graph  C4rn = C4.renumbered(new int[]{3,2,1,0});
      show(C4rn );

      Graph C4rnC = C4rn.canonical().toDirected();
      show(C4rnC);

      Directed   d1 =   Directed.dense(new int[][]{ {0, 0, 1, 1},
                                                    {1, 0, 0, 1},
                                                    {1, 0, 0, 0},
                                                    {0, 1, 1, 0} } ); 
      show(d1);

      Undirected u1 = Undirected.dense(new int[][]{ {1, 0, 1, 0},
                                                       {0, 1, 1},
                                                          {0, 1},
                                                             {1} } );
      show(u1);

      int [] [] es1 = new int[][]{{2,4}, {1,2}, {5,1}, {2,1}};
      show( "sparse", new   Directed.Sparse( es1 ) );
      int [] [] es2 = new int[][]{{2,4}, {1,2}, {5,1}, {2,5}};
      show( "sparse", new Undirected.Sparse( es2 ) );

      // show(  Directed.dense(new int[][]{ {0} } ));     // |V|=1 !
      // show(Undirected.dense(new int[][]{ } ));         // |V|=1 !

System.out.println();
Graph K10 = new Directed.C(10);
// Graph K10 = new Undirected.K(10);
Graph.Contraction Con = K10.contraction(
   new int[]{ 1, 3,4, 7,8 } );
show(Con);
for( int i = 0; i < Con.vSize(); i ++ )
   System.out.print( " " + i + "~p" + Con.v2pv(i) );
System.out.println();
System.out.println();

      mml.Discretes Poisson1 = mml.MML.Poisson0.shifted(1); // ints n >= 1

// Directed...
      mml.Graphs.GERfixed GER_dir =
          new mml.Graphs.GERfixed(pair(Directed.unlabelled,
                                         Poisson1));
      mml.Estimator  GER_dir_est = GER_dir.estimator( ten );
      mml.Test.exercise(GER_dir_est,
                        Vector.values(new Graph[]{C4, new Directed.C(5)}));
// Undirected...
      mml.Graphs.GERfixed GER_und =
          new mml.Graphs.GERfixed(pair(Undirected.unlabelled,
                                         Poisson1));
      mml.Estimator  GER_und_est = GER_und.estimator( ten );
      mml.Test.exercise(GER_und_est,
                        Vector.values(new Graph[]{K5, u1}));

// some mml Models of random Graphs
      graph.Type gType = Directed.unlabelled;
      Cts meanV = real(25.0);  // about this many Vertices
      double pEdge = 0.1;      // density, pr(Edge(v,w)) exists
      Vector alpha = Vector.doubles(new double [] {1-pEdge, pEdge}),
            alpha2 = Vector.doubles(new double [] {1-2*pEdge, 2*pEdge});

      Graphs.GERfixed  UGER = new Graphs.GERfixed(pair(gType, Poisson1));
      Estimator       UGERest = UGER.estimator( meanV );
      Graphs.GERfixed.M GER = UGER.sp2Model(0, 0, pair(meanV, alpha2));  // GER

      Graphs.GERadaptive  UGER2 = new Graphs.GERadaptive(
         triple(gType, Poisson1, Vector.doubles(new double[]{16,4})));
      Estimator       UGER2est = UGER2.estimator(meanV);
      Graphs.GERadaptive.M GER2 = UGER2.sp2Model(0, 0, meanV);          // GER2

      // a small world -ish Model...
      Graphs.Skewed  USW = new Graphs.Skewed(pair(gType, Poisson1));
      Estimator   USWest = USW.estimator(meanV);
      Graphs.Skewed.M SW = USW.sp2Model(0, 0, meanV);                     // SW

      Graph [] motifs = new Graph [] {
        Directed.dense(new int[][]{{0,1,1,1}, {1,0,1,1}, {1,1,0,1}, {1,1,1,0}})
        };
      MotifD  UMM = new MotifD(triple(gType, Poisson1, pair(three, four)));
      Estimator UMMest = UMM.estimator(meanV);
      MotifD.M MM = UMM.new M(0, 0,
                      pair(meanV, Vector.values(motifs)));                // MM
// say...
      mml.Test.compete(/* trials= */ 10, /* N= */ 3, SW, USWest, UMMest); 
      mml.Test.compete(/* trials= */ 10, /* N= */ 3, MM, USWest, UMMest); 

      System.out.println("-- done graph.Test --");
    }//main

 }//Test class

// (c) Lloyd Allison July 2014, Wed 8 Apr 2015. Licence to 'Test' is Gnu GPL.
