//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package graph;

import la.util.Util;
import la.util.Series;
import la.la.Value;
import la.maths.Vector;
import la.maths.Matrix;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The class of Undirected Graphs.
    Operations on Undirected Graphs, and special Undirected Graphs.
    Note that
    <NOBR>&lang;v0, v1&rang;</NOBR> and
    <NOBR>&lang;v1, v0&rang;</NOBR> denote the same Edge.
    Also see {@link Directed} and {@link Graph}. */
public abstract class Undirected extends Graph                  // * Undirected
 {
   /** The standard type of Undirected, unlabelled, no self-loops Graphs. */
   public final static Type unlabelled = new Type("undirected", false);

   /** Type of Undirected, unlabelled, self-loops (O) <I>allowed</I> Graphs. */
   public final static Type unlabelled_O = new Type("undirected", false, true);

// from Graph:
// public Type type()                abstract                                 *
// public boolean   isDirected()
// public boolean isUndirected()
// public int vSize()                abstract                                 *
// public int eSize()
// public boolean selfLoops()        ? v -- v allowed?
// public int degree(v)              abstract                                 *
// public boolean isEdge(v0, v1)     ? v0 -- v1 ?  abstract                   *
// public boolean vLabelled()
// public boolean eLabelled()
// public Value vLabel(v)
// public Value eLabel(v0, v1)

   /** Equivalent to {@link #isEdge isEdge(v0,v1)} for an Undirected Graph
       (but not for a {@link Directed#adjacent Directed} Graph). */
   public boolean adjacent(int v0, int v1)                   // adjacent(v0,v1)
    { return isEdge(v0, v1); }

   /** Also see {@link Directed#degree}.
       A sub-class may have a more efficient implementation. */
   public int degree(int v)                                        // degree(v)
    { Series.Int ws = joinedTo(v);  ws.exhaust();
      return ws.position() + (isEdge(v,v) ? 1 : 0); // NB. loop
    }//degree(v)

   /** Return the upper-right triangular part of the adjacency (array) Matrix
       of this Undirected, and hence symmetric, Graph. If self-loops are
       prohibited the main (&lang;v,v&rang;) diagonal is omitted.  Also see
       {@link #A() A()}, and {@link Matrix#ints_UR(boolean,int[][],int)}. */
   public int [] [] upperRightA()
    { final int V = vSize(), d = selfLoops() ? 0 : 1;
      final int Vd = V - d;  // |V| or |V|-1
      int [] [] rows = new int [Vd] [];
      for( int v0 = 0; v0 < Vd; v0 ++ )
       { int [] rows_v0 = new int[Vd - v0];
         for( int v1 = v0+d; v1 < V; v1 ++ )  // [v0, V), or [v0+1, V)
            rows_v0[v1-v0-d /* =0,1,... */] = isEdge(v0, v1) ? 1 : 0;
         rows[v0] = rows_v0;
       }//for v0
      return rows;
    }//upperRightA()
   // .........................................................................

   /** Identity function; see {@link Graph#toUndirected}. */
   public Undirected toUndirected() { return this; }

   /** Convenience function. */
   public Directed asDirected()                                 // asDirected()
    { return new AsDirected(); }

   /** Treat every Undirected Edge of 'this' Undirected Graph as two
       {@link Directed Directed} Edges  (a self-loop remains a single
       self-loop, now Directed).  There is no loss of information.
       Also see {@link Directed#asUndirected}. */
   public class AsDirected extends Directed            // Undirected.AsDirected
                           implements SubGraph
    { final Type t = Undirected.this.type().clone(/*directed*/true);
      public Undirected parent() { return Undirected.this; }
      public int v2pv(int v) { return v; }
      public Type type() { return t; }
      public int vSize() { return Undirected.this.vSize(); }
      public boolean isEdge(int v0, int v1)
       { return Undirected.this.isEdge(v0, v1); }
      public boolean adjacent(int v0, int v1)
       { return Undirected.this.isEdge(v0, v1); } // efficient
      public Value vLabel(int v) { return Undirected.this.vLabel(v); }
      public Value eLabel(int v0, int v1)
       { return Undirected.this.eLabel(v0, v1); }
      public String toString()
       { return Undirected.this + ".asDirected()"; }
    }//Undirected.AsDirected class
   // .........................................................................

   /** A Dense, Undirected Graph with Type t and adjacency Matrix A.
       It is not necessarily "dense", that's just its intended use.
       Also see {@link Sparse} and {@link Directed.Dense}.  */
   public static class Dense extends Undirected             // Undirected.Dense
                             implements Graph.Dense
    { private final Type   t;
      private final Matrix A;
      public Dense(Type t, Matrix A)
       { this.t = t;  this.A = A;
         if( ! A.isSquare() || ! A.nearlySymmetric() )
            error( ", A not square symmetric" );
       }//constructor
      public Type type() { return t; }
      public int vSize() { return A.nElts(); }
      public boolean isEdge(int v0, int v1)
       { return A.elt(v0, v1).n() != 0; }
      public String toString()
       { return "Undirected.Dense(" + type() + ","
              + Matrix.ints_UR(selfLoops(), upperRightA(), 0) + ")";
       }//toString()
    }//Undirected.Dense class


   /** Return an Undirected Graph of Type 't' and square
       {@link la.maths.Matrix#ints_UR(boolean,int[][],int) symmetric}
       adjacency Matrix 'A'.
       Also see {@link Directed#dense} and {@link Graph#dense}. */
   public static Undirected dense(Type t, Matrix A)
    { return new Dense(t, A); }

   /** {@link #dense(Type,Matrix) dense(t,A)}.  'selfLoops'&nbsp;specifies
       if Graphs are <I>allowed</I> to have self-loops, <I>not</I> if
       this Graph actually has any.  */
   public static Undirected dense(boolean selfLoops, Matrix A)
    { return dense(selfLoops ? unlabelled_O : unlabelled, A); }

   /** {@link #dense(boolean,la.maths.Matrix) dense(false, A)}
       -- no self-loops. */
   public static Undirected dense(Matrix A) { return dense(false, A); }


   /** {@link #dense(Type,Matrix) dense(t,ints(.,A,.)}; A must be upper-right
       {@link la.maths.Matrix#ints_UR(boolean, int[][], int) triangular}. */
   public static Undirected dense(final Type t, final int [] [] A)
    { return dense(t, Matrix.ints_UR(/*incDiag*/t.selfLoops, A, /*diagV*/0)); }

   /** {@link #dense(Type,int[][]) dense(t,A)}.  A must be upper-right
       {@link la.maths.Matrix#ints_UR(boolean,int[][],int) triangular}.  */
   public static Undirected dense(boolean selfLoops, final int [] [] A)
    { return dense(selfLoops ? unlabelled_O : unlabelled, A); }

   /** {@link #dense(boolean,int[][]) dense(false,A)} -- no self-loops.
       A must be upper-right
       {@link la.maths.Matrix#ints_UR(boolean,int[][],int) triangular}.  */
   public static Undirected dense(final int [] [] A)
    { return dense(false, A); }
   // .........................................................................

   /** Convenience function. */
    public static Sparse sparse(Type t, int V, int [] [] es)
     { return new Sparse(t, V, es); }

   /** The class of Sparse Undirected Graphs.
       Also see {@link Dense} and {@link Directed.Sparse}.  */
   public static class Sparse extends Undirected           // Undirected.Sparse
                             implements Graph.Sparse
    { private Type t;
      public Type type() { return t; }

      /** The number of Vertices (V) and Edges (E). */
      private final int V, E;

      /** Given an Edge (v0,&nbsp;v1), v1 is in (ascending) joined[v0],
          and v0 is in joined[v1].  */
      public final int [] [] joined;

      /** The Edges, as int[|E|][2], defining 'this' Undirected
          Sparse Graph. */
      public final int [] [] es;

      /** Assume that the max Vertex mentioned in es is the last Vertex. */
      public Sparse(int [] [] es)
       { this(unlabelled, Directed.Sparse.maxV(es)+1, es); } // ...

      /** Undirected Edges <NOBR>es = {{v00, v01}, {v10, v11}, ...}</NOBR>. */
      public Sparse(Type t, int V, int [] [] es)
       { this.t  = t;
         this.E  = es.length;
         this.es = es;
         this.V  = V;
         int [] nJoined = new int[V];
         for( int v = 0; v < V; v ++ ) nJoined[v] = 0;
         for( int i = 0; i < es.length; i ++ )  // find Vertex in/out-degrees
          { final int v0 = es[i][0], v1 = es[i][1];
            nJoined[v0] ++ ;
            if( v0 == v1 ) continue;
            nJoined[v1] ++ ;
          }
         joined = new int [V] [];
         for( int v = 0; v < V; v ++ )
            joined[v] = new int [nJoined[v]];
         for( int i = 0; i < es.length; i ++ )
          { final int v0 = es[i][0], v1 = es[i][1];    // (v0, v1)
            int [] vs = joined[v0];                    // v0 -- v1
            int n = vs.length - nJoined[v0];  // 0, 1, ...
               for( ; n > 0 && v1 < vs[n-1]; n -- ) vs[n] = vs[n-1];
               // n == 0 || v1 >= vs[n-1], i.e., insertion sort
               vs[n] = v1; nJoined[v0] -- ;
               if( n > 0 && v1 == vs[n-1] )
                  error( ", duplicate Edge " + new Edge(v0, v1) );
            if( v0 == v1 /* loop */ ) continue; // else...
            vs = joined[v1];                           // v1 -- v0
            n = vs.length -  nJoined[v1];  // 0, 1, ...
               for( ; n > 0 && v0 < vs[n-1]; n -- ) vs[n] = vs[n-1];
               // n == 0 || v0 >= vs[n-1]
               vs[n] = v0;  nJoined[v1] -- ;
               if( n > 0 && v0 == vs[n-1] )
                  error( ", duplicate Edge " + new Edge(v0, v1) );
          }//for i
       }//constructor

      public int       vSize() { return V; }
      public int       eSize() { return E; }
      /** O(1)-time. */
      public int degree(int v) { return joined[v].length + (isEdge(v,v)?1:0); }

      public boolean isEdge(int v0, int v1)
       { final int [] j_v0 = joined[v0];  // v0 -- <somewhere>
         int lo = 0, hi = j_v0.length;
         if( hi <= 0 ) return false;
         while( lo < hi - 1 )
          { int mid = (lo + hi) / 2;  // lo < mid < hi, eg 0<1<2
            if( j_v0[mid] > v1 ) hi = mid;
            else       /* <= */  lo = mid;
          }// lo == hi-1
         return j_v0[lo] == v1;
       }//isEdge(v0,v1)

      public Series.Int joinedTo(int v)
       { return Series.fromInts(joined[v]); }

      public String toString()
       { return "Undirected.Sparse(" + t + "," + Matrix.ints(es) + ")"; }
      // .....................

      /** A renumbered Sparse Undirected Graph is Sparse and Undirected. */
      public Renumbered renumbered(int [] vs)
       { return new Renumbered(vs); }

      /** A Sparse Undirected Graph, renumbered according to
          {@link #vs vs}, is Sparse and Undirected. */
      public class Renumbered extends Sparse    // Undirected.Sparse.Renumbered
                              implements SubGraph
       { public Sparse parent() { return Sparse.this; }
         /** Vertex 'v' of 'this' is Vertex vs[v] of the parent Graph. */
         public final int [] vs;
         public Renumbered(int [] vs)
          { super(Sparse.this.t, Sparse.this.V /* ! */, renumbered_es(vs));
            this.vs = vs;
            if( vs.length != Sparse.this.vSize() )
               error( ", vs["+vs.length+"] need "+Sparse.this.vSize() );
          }//constructor
         public int     v2pv(int v) { return vs[v]; }
         public Value vLabel(int v) { return Sparse.this.vLabel(vs[v]); }
         public Value eLabel(int v0, int v1)
          { return Sparse.this.eLabel(vs[v0], vs[v1]); }
         public String toString()
          { return parent() + ".renumbered(" + Vector.ints(vs) + ")"; }
       }//Undirected.Sparse.Renumbered class

      private int [] [] renumbered_es(int [] vs)
       { int [] inv = new int [vs.length]; // inverse perm. to vs
         int [] [] es2 = new int [E][2];
         for( int i = 0; i < vs.length; i ++ ) inv[vs[i]] = i;
         // v of 'this' SubGraph corr. to  vs[v] of super graph and
         // w of super graph corr. to inv[w] of 'this' SubGraph.
         for( int e = 0; e < E; e ++ )
          { es2[e][0] = inv[es[e][0]]; es2[e][1] = inv[es[e][1]]; }
         return es2;
       }//renumbered_es(vs)
      // ..................

      /** An induced SubGraph of a Sparse Undirected Graph is
          Sparse and Undirected. */
      public Induced induced(final int [] vs) { return new Induced(vs); }

      /** An induced SubGraph of a Sparse Undirected Graph is
          Sparse and Undirected. */
      public class Induced extends Sparse          // Undirected.Sparse.Induced
                           implements SubGraph
       { public Sparse parent() { return Sparse.this; }
         public final int [] vs;
         public Induced(int [] vs)
          { super(Sparse.this.t, vs.length, induced_es(vs)); this.vs = vs;
            if( vs.length > Sparse.this.vSize() )
               error( ", vs["+vs.length+"] need <= "+Sparse.this.vSize() );
          }
         public int     v2pv(int v) { return vs[v]; }
         public Value vLabel(int v) { return Sparse.this.vLabel(vs[v]); }
         public Value eLabel(int v0, int v1)
          { return Sparse.this.eLabel(vs[v0], vs[v1]); }
         public String toString()
          { return parent() + ".induced(" + Vector.ints(vs) + ")"; }
       }//Undirected.Sparse.Induced class

      /** Work out which Edges of parent exist in subgraph,
          and renumber the Vertices in them. */
      private int [] [] induced_es(final int [] vs)
       { int [] inv = new int [V];  // map v in parent to v in subgraph
         for( int i = 0; i < vs.length; i ++ ) inv[vs[i]] = i;
         int E = 0;
         for( int e = 0; e < es.length; e ++ )  // count Edges to be...
            if( Util.search(vs, es[e][0]) && Util.search(vs, es[e][1]) )
               E ++ ; // <v,succ[v0][v1]> ok
         int [] [] es2 = new int [E] [2]; // subgraph's Edges...
         E = 0;
         for( int e = 0; e < es.length; e ++ )  // adjust Vertices...
            if( Util.search(vs, es[e][0]) && Util.search(vs, es[e][1]) )
             { es2[E][0] = inv[es[e][0]]; es2[E][1] = inv[es[e][1]]; E ++ ; }
         return es2;
       }//induced_es(vs)

    }//Undirected.Sparse class
   // .........................................................................

   /** Also see {@link Directed.Vertex}. */
   public class Vertex extends Graph.Vertex                // Undirected.Vertex
    { public Vertex(int v) { super(v); }
      public Undirected parent() { return Undirected.this; }
    }//Undirected.Vertex class

   /** Also see {@link Directed.Edge}. */
   public class Edge extends Graph.Edge                      // Undirected.Edge
    { public Edge(int v0, int v1) { super(v0, v1); }
      public Undirected parent() { return Undirected.this; }
      public String toString() { return "(" + v0 + "," + v1 + ")"; }
    }//Undirected.Edge class
   // .........................................................................

   /** The class of complete Undirected (unlabelled) Graphs, K<SUB>n</SUB>. */
   public static class K extends Dense                          // Undirected.K
    { /** The {@link #vSize number} of Vertices. */
      protected final int n;
      /** Construct the complete Undirected Graph on 'n' Vertices. */
      public K(final int n)
       { super( unlabelled,
                new Matrix.Ints(/* cols= */ n)
                { public int nElts() { return n; }
                  public Value elt(int r, int c)
                   { return new Int(r == c ? 0 : 1); }
                  public boolean nearlySymmetric() { return true; }
                } );
         this.n = n;
       }//constructor
      public String toString() { return "K_" + n; }
    }//K class
   // -------------------------------------------------------------------------

   /** See {@link Test#main}. */
   public static void main(String[] argv)                               // main
    { System.out.println("-- test Undirected --");
      Test.main(argv);
      System.out.println("-- done Undirected --");
    }//main

 }//Undirected class, https://www.cantab.net/users/mmlist/MML/A/

// (c) L.Allison July 2014, Fri 29 Aug 2014. Licence to 'Undirected' is Gnu GPL.
