//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package la.la;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** Semantically, <NOBR> Environment = Identifier &rarr; Value,</NOBR>
    an Environment, <NOBR> r : Environment</NOBR>
    ('r' is close to the Greek &rho;), is a function (in Java a class)
    mapping identifers onto the Values bound to them.
    Also see {@link #D} and {@link Declaration}. */
public class Environment
 {
   /** next links 'this' to the previous- (sub-) Environment, if any. */
   protected final Environment next;

   /** Create an empty Environment (with no sub-Environment). */
   protected Environment()
    { this(null); }//constructor

   /** Create an empty Environment linked to next. */
   protected Environment(Environment next)
    { this.next = next; }//constructor

   /** A (the) empty Environment -- might be useful for
       {@link Library Functions} that are complete unto themselves? */
   public static final Environment empty = new Environment();          // empty

   /** The initial Environment (standard {@link Library Library}). */
   public static final Environment initial = Library.env;            // initial

   /** Descend 'levels' of sub-Environments, and there return
       the Value bound to the 'offset'<SUP>th</SUP> Variable.  */
   public Value val(int levels, int offset)                         // val(l,o)
    { Environment e = this;
      for( ; levels > 0; levels -- )
       { assert e != null : "e==null in val(o," + offset + ")";
         e = e.next;
       }
      return e.val(offset);
    }//val(l,o)

   /** Return the Value bound to the 'offset'<SUP>th</SUP> Variable. */
   protected Value val(int offset)
    { throw new RuntimeException("empty.val(offset=" + offset + ")"); }

   /** Return the Value bound to Variable 'EId'. */
   public Value lookup(Expression.Ident EId)                     // lookup(EId)
    { if( EId.levels >= 0 ) return val(EId.levels, EId.offset);
      /* else */ return lookup(EId, 0);
    }

   /** Note, nothing is declared {@link #initial initially}. */
   protected Value lookup(Expression.Ident EId, int levels)
    { error("\nsadly, " + EId + " is not declared"); return null; }


   /** Extend 'this' Environment by binding a Variable 'id' to
       a Value 'v'; typically id is the formal
       {@link Expression.Application parameter} of a Function. */
    public Environment bind(String id, Value v)                   // bind(id,v)
     { return new subEnv1(id, v, this); }

   /** Extend 'this' Environment by binding Variables 'ids' to Values 'vs'. */
    public Environment bind(String [] ids, Value [] vs)         // bind(ids,vs)
     { return new subEnv(ids, vs, this); }


   /** Extend 'this' Environment as specified by Declaration(s) 'dec',
       semantically  <NOBR>D : Decs &rarr; Env &rarr; Env</NOBR>.
       If recursive ('let&nbsp;rec') the Values use
       the resulting Environment, otherwise 'this'.  */
   public Environment D(Declaration dec)                                // D(d)
    { final Value [] vals = new Value[dec.ids.length];
      final Environment r = bind(dec.ids, vals);
      final Environment local_r = dec.isRec ? r : this;   // let [rec] ?
      for( int i = 0; i < vals.length; i ++ )
         vals[i] = dec.es[i].defer(local_r);   // lazy
      return r;
    }//D(dec)


   /** For binding a <I>single</I> Variable to a <I>Value</I>.
       Also see {@link subEnv}. */
   private static class subEnv1 extends Environment                  // subEnv1
    { /** Bind 'id' to 'val. */
      private final String id;  private Value val;

      public subEnv1( String id, Value val, Environment r )
       { super(r); this.id = id; this.val = val; }//constructor

      protected Value val(int offset)
       { assert offset == 0 : "subEnv1.val(o=" + offset + ")";
         val = val.force(); return val;
       }//val(l,o)

      /** Return the Value bound to Variable 'EId' (counting levels). */
      protected Value lookup(Expression.Ident EId, int levels)
       { if( this.id.compareTo(EId.id) == 0 )
          { EId.levels = levels; EId.offset = 0; // quicker next time & ...
            val = val.force(); return val;       // ... overwrite & return
          }
         else return next.lookup(EId, levels+1);
       }//lookup

      /** For debugging. */
      public String toString() { return "[" + id + "]+" + next; }
    }//subEnv1 class


   /** Implements binding id[0] to vals[0], ids[1] to vals[i], etc,
       as in a {@link Declaration}. Also see {@link subEnv1}. */
   private static class subEnv extends Environment                    // subEnv
    { private final String [] ids;   // ids[] in 1-1 corr ...
      private final Value  [] vals;  // ... with vals[]

      public subEnv( String [] ids, Value [] vals, Environment r )
       { super(r); this.ids = ids; this.vals = vals; }//constructor

      /** Return the offet<SUP>th</SUP> Value in 'this' Environment. */
      protected Value val(int offset)
       { final Value v = vals[offset].force(); vals[offset] = v;
         return v;
       }

      /** Return the Value bound to Variable 'EId' (and count levels). */
      protected Value lookup(Expression.Ident EId, int levels)
       { for( int i = 0; i < ids.length; i ++ )
            if( ids[i].compareTo(EId.id) == 0 )
              { EId.levels = levels; EId.offset = i;  // faster val() next time
                final Value v = vals[i].force(); vals[i] = v;  // & overwrite v
                return v;
              }
         // failing that...
         return next.lookup(EId, levels+1);
       }//lookup

      /** For debugging. */
      public String toString()
       { StringBuilder sb = new StringBuilder("[");
         for( int i = 0; i < ids.length; i ++ ) sb.append(ids[i]+",");
         sb.append("]+" + next);
         return sb.toString();
       }
    }//subEnv class


   /** For debugging. */
   public String toString() { return "[]"; }

   void error(String msg)
    { throw new RuntimeException("Environment: " + msg); }

 }//Environment class

// (c) L. Allison, Oct 2007, 16 Nov 2016. Licence to 'Environment' is Gnu GPL.
