//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package la.la;

import la.maths.Vector;
import la.maths.Matrix;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018, 2019 Lloyd Allison.

   Reference:
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The class of Functions; note that a Function is a {@link Value}.
    Also see {@link Value.Lambda}, and the standard {@link Library}.<BR>
    (Classes {@link Integral Integral} and {@link Derivative Derivative}
    were modified and {@link CtsD2CtsD CtsD2CtsD} added in 2020.)  */
public abstract class Function extends Value                      // * Function
 {
   /** A Function, f, might have an inverse() Function.  If it does,
       it implements HasInverse and (f.inverse())(f(x))=x. */
   public static interface HasInverse                              // * Inverse
    { public Function inverse(); }

   /** This class exists so that one can create an anonymous
       {@link Function  Function} which (implements)
       {@link Function.HasInverse HasInverse}.
       An anonymous class can only instantiate one class or interface.  */
   public static abstract class WithInverse extends Function   // * WithInverse
                                         implements HasInverse
    { }

   /** Return {@link Type#FUNCTION}. */
   public Type type()
    { return Type.FUNCTION; }                                         // type()

   /** Apply 'this' Function to parameter 'p'. */
   public abstract Value apply(Value p);                          // * apply(p)

   /** Return the String "Function". */
   public String toString()                                       // toString()
    { return "Function"; }

   /** A unary operator on 'this' Function&nbsp;f, such that
       (&lt;op&gt;&nbsp;f)(x)&nbsp;=&nbsp;&lt;op&gt;&nbsp;(f(x)).<BR>
       That is,
       f.uOp(op).{@link #apply apply(x)} =
       f.apply(x).{@link Value#uOp uOp(op)}.
       Also see {@link Cts2Cts#uOp} and {@link Value#uOp}.  */
   public Function uOp(final int op)                                 // uOp(op)
    { return new Function()
       { public Value apply(Value x)
          { return Function.this.apply(x).uOp(op); }
       };//new Function
    }//uOp(op)

   /** A binary operator on Functions 'this' (f) and 'g', such that
   (f&nbsp;&lt;op&gt;&nbsp;g)(x)&nbsp;=&nbsp;f(x)&nbsp;&lt;op&gt;&nbsp;g(x).
       Also see {@link Value#bOp}.  */
   public Function bOp(final int op, final Function g)             // bOp(op,g)
    { return new Function()
       { public Value apply(Value x)
          { return Function.this.apply(x).bOp(op, g.apply(x)); }
       };//new Function
    }//bOp(op,g)
   // =========================================================================

   /** The class of Cts&nbsp;&rarr;&nbsp;Cts Functions.
       Define {@link #apply_x(double) apply_x(.)} and all else follows.
       A Cts2Cts has a {@link #d_dx() derivative} and
       an {@link #integral() integral}.
       Also see {@link Cts2Cts2Cts}.  */
   public static abstract class Cts2Cts extends Native             // * Cts2Cts
    {
      /** apply_x(x) is used by {@link #apply apply(c)}.
          Define apply_x(.) to make an instance of Cts2Cts.  */
      public abstract double apply_x(double x);                 // * apply_x(x)

      /** This class exists so that one can create an anonymous
          {@link Cts2Cts  Cts2Cts} which (implements)
          {@link Function.HasInverse HasInverse}.
          An anonymous class can only instantiate one class or interface.
          Also see {@link Function.WithInverse}.  */
      public static abstract class WithInverse extends Cts2Cts
                                            implements Function.HasInverse
       { }

      /** Calls {@link #apply_x apply_x(c.x())}.
          Note,&nbsp;if c.AoM()&gt;zero, apply(.) also calls
          the {@link #d_dx() derivative} to get an
          {@link Value.Cts#AoM() AoM()} for the result.  */
      public Cts apply(Value c)
       { Cts cc = (Cts) c;
         double aom = cc.AoM(), x = cc.x(), fx = apply_x(x);
         return aom <= 0 ? real(fx)
                         : cts(fx, Math.abs(aom*d_dx().apply_x(x)));
       }//apply(c)

      /** (Static) fromFunction(f) creates a Cts2Cts from a Function,
          'f', where f is not an instance of class Cts2Cts but f does
          in fact accept, and return, Cts values.  */
      public static Cts2Cts fromFunction(final Function f)   // fromFunction(f)
       { return new Cts2Cts()
          { public double apply_x(double x)
             { return f.apply(real(x)) . x(); }
       }; }//fromFunction(f)

      /** For example, sin.uOp(minus) satisfies
          (sin.uOp(minus))(x)&nbsp;=&nbsp;&minus;sin(x).
          Also see {@link Cts2Cts#uOp}.  */
      public Cts2Cts uOp(final int op)                               // uOp(op)
       { return new Cts2Cts()
          { public double apply_x(double x)
             { return real(Cts2Cts.this.apply_x(x)) .uOp(op) .x(); }
          };//new Cts2Cts
       }//uOp(op)

      /** Find a turning point of 'this' Cts2Cts, 'f', that is find x such
          that {@link Cts2Cts#d_dx() f'(x)}=0, given an initial guess, x0.
          Uses {@link #root root}. */
      public double turning(double x0)                           // turning(x0)
       { return this.d_dx().root(x0); }

      /** Solve f(x)&nbsp;=&nbsp;0, where 'f' is 'this' Cts2Cts, given
          an initial guess, x0.
          Currently it uses {@link #NewtonRaphson NewtonRaphson(x0)}. */
      public double root(double x0)                                 // root(x0)
       { return this.NewtonRaphson(x0); }

      /** Use Newton-Raphson to solve f(x)&nbsp;=&nbsp;0, where
          'f' is 'this' Cts2Cts, given an initial guess,&nbsp;x0.
          The central step is x1=x0&minus;f(x0)/df_dx_(x0);
          it helps if f has a closed-form derivative,
          {@link Cts2Cts#d_dx() d_dx}.
          (If Newton-Raphson fails, plan&nbsp;B is binary search.)  */
      public double NewtonRaphson(double x0)
       { final Cts2Cts f = this;       // the function in question
         final int    max   = 30;      // max iterations
         final double delta = 10E-7;   // find x to within one in 1/delta
         final double flat  = delta*delta;  // lower limit on the slope
         final Cts2Cts df_dx = f.d_dx();    // derivative of f
         double x1, y0 = 0, df_dx_x0,
                yp = Double.MAX_VALUE, ypx = 0, ym = -yp, ymx = 0;
         boolean p = false, m = false;
         for( int i = 0; i < max; i ++ )
          { y0 = f.apply_x(x0);
            switch( (int) Math.signum(y0) )
             { case  1: if( y0 < yp ) { p = true; yp = y0; ypx = x0; }; break;
               case -1: if( y0 > ym ) { m = true; ym = y0; ymx = x0; }; break;
               default: return x0;  // only in some weird universe
             }// (Note, ym < 0 < yp, but it is possible that ymx > ypx.)
            df_dx_x0 = df_dx.apply_x(x0);
            if( Math.abs(df_dx_x0) < flat ) // slope too flat
             { if( p && m ) break;
               /*else*/ return x0;
             }
            // the next line is the heart of the Newton-Raphson method...
            x1 = x0 - y0/df_dx_x0;  // i.e., project tangent onto x-axis
            double mag = Math.abs(x1 == 0 ? x0 : x1);
            if( Math.abs(x1 - x0) <= delta*mag ) return x1;          // success
            x0 = x1;
          }//for i

         // If Newton-Raphson did not work, there is a plan B but ...
         double power = 0.1;  // ... but we need both a +ve and -ve f(.)
         if( ! p )  // desperation (maybe f' was too flat?) ...
            for( int i = 0; i < max; i ++ )
             { ypx = x0 + power;      // seek an x s.t. f(x) >= 0
               y0 = f.apply_x(ypx);
               if( y0 >= 0 ) { p = true; yp = y0; break; }
               ypx = x0 - power;      // or this
               y0 = f.apply_x(ypx);
               if( y0 >= 0 ) { p = true; yp = y0; break; }
               power *= 3;
             }
         else /* p */ if( ! m )
            for( int i = 0; i < max; i ++ )
             { ymx = x0 + power;      // seek an x s.t. f(x) <= 0
               y0 = f.apply_x(ymx);
               if( y0 <= 0 ) { m = true; ym = y0; break; }
               ymx = x0 - power;      // or this
               y0 = f.apply_x(ymx);
               if( y0 <= 0 ) { m = true; ym = y0; break; }
               power *= 3;
             } /* else p && m */

         if( p && m )                        // N-R failed, can we ...
          { double mid = ymx, fmid;
            for( int i = 0; i < max; i ++ )  // ... give binary search a try
             { mid = (ymx + ypx) / 2.0;
               fmid = f.apply_x(mid);        // ymx < mid < ypx, or v.v.
               switch( (int) Math.signum(fmid) )
                { case  1: yp = fmid; ypx = mid; break;
                  case -1: ym = fmid; ymx = mid; break;
                  default: return mid;
                }
               if( Math.abs(ymx - ypx) <= delta ) return mid; // bleh
             }//for i
            return mid;  // best I could do ???
            // throw new RuntimeException( "x in [" + ymx + "," + ypx + "], "
            //                         +  "y in [" + ym  + "," + yp  + "]" );
          }// else !p || !m ...
         throw new RuntimeException( "root(" + f + "," + x0
            + "), p=" + p + ", m=" + m
            + ", x0="+x0+", y0="+y0 );  // failure
       }//NewtonRaphson(x0)
      // ......................................................................

      private Cts2Cts derivative = null; // cache; see below

      /** Return the derivative of 'this' function, d/dx&nbsp;this.
          Caches the result of {@link #make_d_dx()}.  */
      public final Cts2Cts d_dx()                                     // d_dx()
       { if( derivative == null) derivative = make_d_dx();
         return derivative;
       }//d_dx()

      /** The second derivative, d<SUP>2</SUP>/dx<SUP>2</SUP>&nbsp;this,
          that is {@link #d_dx() d_dx()} twice. */
      public final Cts2Cts d2_dx2()                                 // d2_dx2()
       { return d_dx().d_dx(); }

      /** Slave worker for {@link #d_dx() d_dx()}. This default returns
          new&nbsp;{@link Derivative Derivative()} which works by finite
          differences. Override make_d_dx if there is a closed form.  */
      protected Cts2Cts make_d_dx() { return new Derivative(); } // make_d_dx()


      private Cts2Cts2Cts integral = null; // cache; see below

      /** Return the integral of 'this' function,
          <FONT SIZE="+2">&int;</FONT>&nbsp;this(x)&nbsp;dx.
          Caches the result of {@link #make_integral()}.  Note that
          integral:<B>R</B>&rarr;<B>R</B>&rarr;<B>R</B> is a
          {@link Cts2Cts2Cts curried} function,
          integral(ko)(hi)=<SUB>lo</SUB>&int;<SUP>hi</SUP>this.  */
      public final Cts2Cts2Cts integral()                         // integral()
       { if( integral == null ) integral = make_integral();
         return integral;
       }//integral()

      /** Slave worker for {@link #integral() integral()}. This default
          implementation returns new&nbsp;{@link Integral Integral()}
          which uses numerical integretaion. Override make_integral if
          there is a closed form
          (as in {@link Derivative#make_integral() Derivative}).
          Also see {@link #exactIntegral exactIntegral(F)}.  */
      protected Cts2Cts2Cts make_integral() { return new Integral(); }
      // ----------------------------------------------------------------------

      /** This default Derivative of a given function '{@link #f f}'
          uses finite-differences &ndash; see {@link #apply_x  apply_x(x)} and
          {@link #apply_xx apply_xx(&delta;,x)}.
          A&nbsp;Derivative is essentialy a {@link Cts2Cts Cts2Cts} with
          an extra {@link #apply_xx apply_xx(&delta;,x)} (for
          finite-differences).
          Also see Cts2Cts.{@link Cts2Cts#d_dx() d_dx()} and
          Cts2Cts.{@link Cts2Cts#make_d_dx() make_d_dx()}.  */
      public class Derivative extends Cts2Cts                     // Derivative
       { /** "f" is a synonym for Cts2Cts.this, the Cts2Cts of
             which "this" is the Derivative,&nbsp;f'. */
         public final Cts2Cts f = Cts2Cts.this;

         /** Construct the Derivative of the Cts2Cts, '{@link #f f}'. */
         public Derivative() { }

         /** Calls {@link #apply_xx(double,double) apply_xx(&delta;,x)} for
             some "small" finite-difference &delta;. If you do not want to
             take your chances with &delta; being chosen for you, set your
             own with {@link #apply_xx(double,double) apply_xx(&delta;,x)};
             also see {@link #make_d_dx make_d_dx()}.  */
         public double apply_x(double x)                          // apply_x(x)
          { final double delta = 0.001,  // ??? Util.epsilon ???
                         xp    = Math.abs(x);  // = |x|
            return apply_xx( delta*(xp>1 ? 1 :  xp>delta ? xp : delta), x );
          }//apply_x(x)

         /** Compute the Derivative of {@link #f f} by finite-difference,
             &delta;.  Called by {@link #apply_x(double) apply_x(x)}.
             Return (f(x+&delta;/2)&minus;f(x-&delta;/2)/&delta;.
             Override it if there is something better (a closed form). */
         public double apply_xx(double delta, double x)
          { final double delta2 = delta * 0.5;
            return (f.apply_x(x+delta2) - f.apply_x(x-delta2)) / delta;
          }//apply_xx(delta,x)

         /** The Derivative of 'this' derivative is the second
             Derivative of {@link #f f}, by finite differences.  */
         protected Cts2Cts make_d_dx()
          { final Cts2Cts f0 = f;  // think about it!
            return new Derivative()
             { public double apply_xx(double delta, double x)
                { final double delta2 = delta * 0.5;
                  // ((f(x+d2)-f(x))/d2 - (f(x)-f(x-d2))/d2) / d2 = ...
                  return (f0.apply_x(x + delta2) + f0.apply_x(x - delta2)
                          - 2*f0.apply_x(x)) / (delta2*delta2);
                }//apply_xx(delta,x)
          }; }//make_d_dx()

         /** The indefinite {@link Integral Integral} of 'this'
             Derivative is&nbsp;{@link #f f}, of course.<BR>  So,
             <SUB>lo</SUB><FONT SIZE="+2">&int;</FONT><SUP>x</SUP> this(x) dx
             =&nbsp;f(x)&minus;f(lo).  */
         protected Cts2Cts2Cts make_integral()
          { return new Integral()
             { /** Override the default (and ignore n). */
               public double apply_nxx(int n, double lo, double x)
                { return f.apply_x(x) - f.apply_x(lo); }
          }; }//make_integral()
       }//Derivative class
      // ----------------------------------------------------------------------

      /** Integral: Cts&rarr;Cts&rarr;Cts (&plusmn;) integrates
          {@link #f f} from 'lo' to 'hi'. The key method is
          {@link #apply_nxx(int,double,double) apply_nxx(n,lo,x)}.
          For use by {@link Cts2Cts#make_integral Cts2Cts.make_integral()}
          and hence {@link Cts2Cts#integral Cts2Cts.integral()}.
          The Integral is performed numerically by Simpson's rule.
          If&nbsp;there is a closed-form, F, for the indefinite
          integral use {@link #exactIntegral exactIntegral(F)}.  */
      public class Integral extends Cts2Cts2Cts                     // Integral
       { /** 'f' is a synonym for 'Cts2Cts.this' and 'this'
             Integral is the (definite) Integral of&nbsp;'f'. */
         public final Cts2Cts f = Cts2Cts.this;

         /** Construct the definite {@link #apply2 Integral} of
             Cts2Cts&nbsp;{@link #f f}. */
         public Integral() { }

         /** Calls {@link #apply_x apply_x(lo.x())}.  An Integral has
             roughly the type Cts&rarr;Cts&rarr;Cts.  */
         public Cts2Cts apply(Value lo)                            // apply(lo)
          { return apply_x(lo.x()); }

         /** Returns a Cts&rarr;Cts which (i)&nbsp;will take hix &amp;
             call {@link #apply_xx apply_xx(lox,hix)}, and (ii)&nbsp;has
             the Derivative&nbsp;{@link #f f}. */
         public Cts2Cts apply_x(double lox)                     // apply_x(lox)
          { return new Cts2Cts()
             { public double apply_x(double hix)
                { return apply_xx(lox, hix); } // calls apply_xx
               public Cts2Cts make_d_dx() { return f; }
             };//new...
          }//apply_x(lox)

         /** Calls {@link #apply_nxx apply_nxx(n,lox,hix)} with a small
             default value of&nbsp;'n' to integrate {@link #f f} numerically.
             If you want to set 'n' yourself, override apply_xx and/or
             {@link #apply_nxx apply_nxx}.
             (See {@link #exactIntegral exactIntegral(F)} if there
             is a closed-form for the indefinite integral.)  */
         public double apply_xx(double lox, double hix)    // apply_xx(lox,hix)
          { return apply_nxx( 11 /* say */, lox, hix ); }

         /** Integrate {@link #f f} from 'lo' to 'x' numerically in 'n' steps
             using Simpson's rule, called by {@link #apply_xx apply_xx}.
             Note, 'n'&nbsp;is <I>forced</I> to be &ge;3, and odd.  See
             {@link #exactIntegral exactIntegral} if there is a
             closed-form for the indefinite integral.  */
         public double apply_nxx(int n, double lo, double x)   // apply_nxx(,,)
          { assert lo < x : (lo + ">=" + x + " in Integral");
            // NB. n should be odd (and >=3), so n-1 must be even (and >=2)!
            final int n_1 = n < 3 ? 2 : (n&1)==0 ? n : n-1;
            double sum = f.apply_x(lo) + f.apply_x(x);  // steps 0 and n-1
            int m = 4; // 4, 2, 4, ..., 2, 4.
            for( int i = 1; i < n_1; i ++ )             // steps 1 ... n-2
             { sum += m * f.apply_x((i*lo + (n_1-i)*x) / n_1); // interpolate
               m = 6 - m;  // 2, 4, 2, ...
             }//for i
 // System.out.print( "{" + "n="+n + ",lo="+lo + ",x="+x + ",sum="+sum + "}" );
            return (x - lo) * sum / (3*n_1);  // i.e. width * average
          }//apply_nxx(lo,x,n)
       }//Integral class
      // ........................................................

      /** Creates a definite integral from an indefinite one, F, <U>if</U>
          such a closed form is known. That is if there is an F such that
          <SUB>lo</SUB>&int;<SUP>hi</SUP>this=F(hi)&minus;F(lo).
          In such a case exactIntegral(F) can be used in
          {@link #make_integral make_integral()}.  */
      public Cts2Cts2Cts exactIntegral(Cts2Cts F)
       { return new Cts2Cts2Cts()
          {/** The integral of 'Cts2Cts.this' from lox to hix
               is F(hix)&minus;F(lox).  */
            public double apply_xx(double lox, double hix)
             { return F.apply_x(hix) - F.apply_x(lox); }
          };//new...
       }//exactIntegral(F)

    }//Cts2Cts class
   // -------------------------------------------------------------------------

   /** {@link Native2 Curried} continuous functions,
       <B>R</B>&rarr;<B>R</B>&rarr;<B>R</B> or more correctly
       Cts&rarr;Cts&rarr;Cts.
       {@link #apply_xx apply_xx} does the work; {@link #apply apply} and
       {@link #apply2   apply2} are "posh" overlays for it.  */
   public static abstract class Cts2Cts2Cts extends Native2      // Cts2Cts2Cts
    { /** Given 'v', return a {@link Cts2Cts Cts2Cts} which takes
          'w' and calls {@link #apply_xx apply_xx(v,w)}.  */
      public Cts2Cts apply(Value v)
       { return new Cts2Cts()
          { public double apply_x(double wx)
             { return apply_xx(v.x(), wx); }
          };//new...
       }//apply(x)

      /** <TT>apply2</TT> calls {@link #apply_xx apply_xx(x,y)}.
          Assumes that the AoM() of the result is zero or
          is handled elsewhere.  */
      public Value apply2(Value x, Value y)
       { return real(apply_xx(x.x(), y.x())); }

      /** Does the work, roughly  this(v)(w)=apply_xx(v,w).
          Used by {@link #apply apply} and {@link #apply2 apply2}.
          Define apply_xx to make a specific Cts2Cts2Cts.  */
      public abstract double apply_xx(double x, double y);   // * apply_xx(x,y)
    }//Cts2Cts2Cts class
   // -------------------------------------------------------------------------

   /** Functions of Vectors of D Continuous Values,
       R<SUP>D</SUP>&rarr;R<SUP>D</SUP>.  These Functions are useful in
       transforming {@link mml.R_D R_D} Models (probability distributions).
       An instance must define {@link #apply_Vec apply_Vec(.)},
       {@link #nlJ nlJ(.)} and {@link #J J(.)}.  See
       {@link Library#polar2cartesian polar2cartesian&lang;r,&theta;&rang;} and
       {@link Library#cartesian2polar cartesian2polar&lang;r,&theta;&rang;}
       as examples.  */
   public static abstract class CtsD2CtsD extends Native         // * CtsD2CtsD
    {
      /** Class CtsD2CtsD.WithInverse exists so that one can create an
          anonymous {@link CtsD2CtsD  CtsD2CtsD} which implements
          {@link Function.HasInverse HasInverse}.
          An anonymous class can only instantiate one class or interface.
          An example is
          {@link Library#polar2cartesian polar2cartesian&lang;r,&theta;&rang;}.
          Also see {@link Function.WithInverse}.  */
      public static abstract class WithInverse extends CtsD2CtsD           // *
                                   implements Function.HasInverse
       { }//CtsD2CtsD.WithInverse class

      /** Given a Vector xs:R<SUP>D</SUP>, return a Vector
          f(xs):R<SUP>D</SUP> where 'f' is 'this' Function.
          Calls {@link #apply_Vec apply_Vec(xs)} then
          if xs's AoM&gt;0 uses {@link #nlJ nlJ(xs)} and {@link #J J(xs)}
          to "adjust" the result's AoMs.  It is not entirely straightforward
          to distribute the result's AoM (inaccuracy) amongst its components
          so we do the best we can.  */
      public Vector apply(Value xs)
       { Vector v = (Vector) xs;         // input xs must be a Vector
         final int D = v.nElts();
         double [] ys = apply_Vec(v);    // the raw result
         if( v.constAoM() == 0 )         // here v is precise, so ...
            return Vector.doubles( ys ); // the result is precise
         else // adjust ys' AoM ...
          { double [] [] J = J(v);               // Jacobian matrix, J
            double nlVol = 0,  targetNlVol = v.nlAoM() + nlJ(v);
            double [] nlAoMs = new double [D];
            for( int i = 0; i < D; i ++ )
             { double AoM_i = 0;
               for( int j = 0; j < D; j ++ )     // proportional to ...
                  AoM_i += Math.abs( J[i][j] * v.AoM(j) );  // ys[i]'s AoM
               nlAoMs[i] = - Math.log( AoM_i );
               nlVol += nlAoMs[i];
             }//for i
            for( int i = 0; i < D; i ++ )        // scale volume to target
               nlAoMs[i] *= targetNlVol / nlVol; // is necessary!
            Cts [] ctss = new Cts [D];
            for( int i = 0; i < D; i ++ )
               ctss[i] = cts(ys[i], Math.exp( - nlAoMs[i] ));
            return Vector.values(ctss);
          }//fi
            // NB. the following is not sufficient to transform Models ...
            // return Vector.doubles(v.nlAoM()+nlJ(v), ys);
       }//apply(xs)

      /** Used by {@link #apply apply(xs)};
          ignores issues of the result's AoM.  */
      public abstract double [] apply_Vec(Vector xs);        // * apply_Vec(xs)

      /** Given a Vector xs:R<SUP>D</SUP>, return the negative log of the
          determinant of the {@link #J Jacobian} matrix
          (must be square).  This is needed to "adjust" the
          AoM of {@link #apply apply}'s result.
          (Note that I have not defined a default implementation
          using {@link Matrix#slowDeterminant}, say.) */
      public abstract double nlJ(Vector xs);                       // * nlJ(xs)

      /** The <I>Jacobian</I> is a matrix, J, where
          J<SUB>i,j</SUB>=&part;f<SUB>i</SUB>/&part;x<SUB>j</SUB>,
          and 'f' is 'this' Function.  If the matrix is square, the
          <I>Jacobian {@link #nlJ determinant}</I> is what you would expect.
          (All assuming 'this' Function is continuous and differentiable.)  */
      public abstract double [] [] J(Vector xs);                     // * J(xs)

    }//CtsD2CtsD class
   // =========================================================================

   /** Native, <TT>{@link Function}s</TT> specified by arbitrary Java
       code in {@link #apply apply(v)}. Also see {@link Library}.  */
   public static abstract class Native extends Function             // * Native
    {
      /** apply(v) does whatever you specify to implement 'this' Native
          Function <U>BUT</U> it must return a Value in at least
          <U>WHNF</U>; see {@link #force}. */
      public abstract Value apply(Value v);                          // * apply

      /** This class exists so that one can create an anonymous
          Function.{@link Native Native}  which (implements)
          {@link Function.HasInverse HasInverse}.
          An anonymous class can only instantiate one class or interface.
          Also see {@link Function.WithInverse}.  */
      public static abstract class WithInverse extends Native
                                            implements Function.HasInverse
       { }
    }//Function.Native class
   // -------------------------------------------------------------------------

   /** <I>Curried</I> native <TT>{@link Function}s</TT> specified by
       arbitrary Java code in {@link #apply2 apply2(v0,v1)}.
       Note, Native2 extends (i.e., is&nbsp;a) {@link Native}.  */
   public static abstract class Native2 extends Native             // * Native2
    {
      /** apply2(v0,v1) does whatever you specify, BUT it must
          return a Value in <U>WHNF</U> &ndash; see {@link #force}. */
      public abstract Value apply2(Value v0, Value v1);      // * apply2(v0,v1)

      /** Take the parameter 'v0' and return a {@link Native} which takes
          the parameter 'v1' to produce the result (this(v0))(v1).
          Hence 'this' is a curried Function.  */
      public Native apply(final Value v0)                          // apply(v0)
       { return new Native() // i.e., a Function
          { public Value apply(Value v1) { return apply2(v0, v1); } };
       }
    }//Function.Native2 class
   // -------------------------------------------------------------------------

   /** <I>3-Curried</I> native <TT>{@link Function}s</TT> specified by
       arbitrary Java code in {@link #apply3 apply3(v0,v1,v2)}.
       Note, Native3 extends (i.e., is&nbsp;a) {@link Native2} and
       hence {@link Native}. */
   public static abstract class Native3 extends Native2            // * Native3
    {
      /** apply3(v0,v1,v2) does whatever you specify, <U>BUT</U> it
          must return a Value in <U>WHNF</U>; see {@link #force}.  */
      public abstract Value apply3(Value v0, Value v1, Value v2);// * apply3

      /** Take the given parameters 'v0' and 'v1' and return a
          {@link Native} which takes 'v2' and produces a result.  */
      public Native apply2(final Value v0, final Value v1)
       { return new Native() // i.e., is a Function
          { public Value apply(Value v2) { return apply3(v0, v1, v2); } };
       }

      /** Take the given parameter 'v0' and return a {@link Native2} which
          takes parameters 'v1' and 'v2' and produces a result.  */
      public Native2 apply(final Value v0)
       { return new Native2() // NB. a Native2 is a Native is a Function
          { public Value apply2(final Value v1, final Value v2)
             { return apply3(v0, v1, v2); }
       }; }
    }//Function.Native3 class
   // -------------------------------------------------------------------------

   /** Return a Function, of one parameter, based on the unary operator
       'op' (not, -, hd, etc.). For example, (fromUop(-))x&nbsp;=&nbsp;-x.
       (Note that fromUop is a <TT>static</TT> method.)
       Also see {@link Function#uOp}, {@link Value#uOp},
       and {@link #fromBop}. */
   public static Function.Native fromUop(final int op)           // fromUop(op)
    { return new Function.Native()
       { public Value apply(Value v) { return v.uOp(op); }
    }; }//fromUop(op)

   /** Return a {@link Native2 Curried} Function based on the binary
       operator, 'op' (+, *, etc.).  For example,
       ((fromBop(+))x)y&nbsp;=&nbsp;x+y.
       (Note that fromBop is a <TT>static</TT> method.)
       Also see {@link Function#bOp}, {@link Value#bOp},
       and {@link #fromUop}. */
   public static Function.Native2 fromBop(final int op)          // fromBop(op)
    { return new Function.Native2()
       { public Value apply2(Value v1, Value v2) { return v1.bOp(op, v2); }
    }; }//fromBop(op)
   // -------------------------------------------------------------------------

   /** main() allows class Function to be <I>very slightly</I> tested,
       in isolation. */
   public static void  main(String[] argv)
    { System.out.println("-- test Function --");

      Value a=new Int(100), b=new Int(20), c=new Int(3);

      System.out.println( "  fromUop(-) 100 = "
         + fromUop(Lexical.minus).apply(a)
         + ", fromBop(+) " + a + " " + b + " = "
              + fromBop(Lexical.plus).apply(a).apply(b)
         );

      Native3 s3 = new Native3()  // s3 :: Int -> Int -> Int
       { public Int apply3(Value a, Value b, Value c)
          { return new Int(a.n() + b.n() + c.n()); }
       };
      System.out.println( "  on Native3: "
         + s3.apply3(a,b,c) + " = "         // all ...
         + s3.apply(a).apply2(b,c) + " = "  // partitions ...
         + s3.apply2(a,b).apply(c) + " = "  // of a,b,c ...
         + s3.apply(a).apply(b).apply(c)
         + "  -- s3(a,b,c), ..., s3 a b c"
         );

      // We seemingly must(?) go through the following contortions to
      //  1. let 'sin' forward-reference (fwd-ref) 'cos', and
      //  2. to deal with a 'final or effectively final' issue!
      Cts2Cts [] sc = new Cts2Cts [2];
      sc[0] =
         new Cts2Cts() // sin
          { public double  apply_x(double x) { return Math.sin(x); }
            public Cts2Cts make_d_dx()       { return sc[1]; } //fwd ref cos
            public Cts2Cts2Cts make_integral()
             { return exactIntegral(sc[1].uOp(Lexical.minus)); }
          };
      sc[1] =
         new Cts2Cts() // cos
          { public double apply_x(double x) { return Math.cos(x); }
            // If the following closed-forms are out,
            // they defaults to finite differences...
            public Cts2Cts  make_d_dx()
             { return sc[0].uOp(Lexical.minus); }
             // { return sin.uOp(Lexical.minus); }
            public Cts2Cts2Cts make_integral()
             { return exactIntegral(sc[0]); }
          };
      Cts2Cts sin = sc[0], cos = sc[1];

      Cts2Cts  d_cos = cos.d_dx(),
              d2_cos = d_cos.d_dx();
      Cts2Cts2Cts I_cos = cos.integral();

      System.out.println( "  on cos:"
         +  "\n    cos 0    = " + cos.apply(zeroR)
         +  "\n    cos pi/2 = " + cos.apply(PIby2)
         + ",\n    (d/dx cos) 0   = " + d_cos.apply(zeroR)  + " = (-sin 0)"
         + ",\n    (d2/dx2 cos) 0 = " + d2_cos.apply(zeroR) + " = (-cos 0)"
         + ",\n    I cos 0..pi/2  = " + I_cos.apply2(zeroR, PIby2)
         +  ",\n   (d/dx (I cos 0..))(pi/2) = "
                    +  I_cos.apply(zeroR).d_dx().apply(PIby2)
         + ",\n    cos.root(guess=1)    = " + cos.root(1)
         + ",\n    cos.turning(guess=1) = " + cos.turning(1)
         + ",\n"
         + " \n    sin 0    = " + sin.apply(zeroR)
         + ",\n    (d/dx sin) 0    = " + sin.d_dx().apply(zeroR) + " = (cos 0)"
         + ",\n    (d2/dx2 sin) 0  = " + sin.d2_dx2().apply(zeroR)
                   + " = (-sin 0)"
         + ",\n    I sin 0..pi/2   = " + sin.integral().apply2(zeroR, PIby2)
         );

      System.out.println("-- done Function --");
    }//main

 }//Function class

// (c) Lloyd Allison 2012, 2017, 11 Feb 2020. Licence to 'Function' is Gnu GPL.
