//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package la.la;
import la.la.Function.*;
import la.maths.Vector;
import la.maths.Matrix;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference:
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The standard library of predefined {@link Function Functions} and
    other {@link Value Values}; also see {@link Environment}. */
public class Library
 {
   /** The identity Function, <TT>id&nbsp;x&nbsp;=&nbsp;x.  */
   public static final Function id =                                    // id x
      new Function.Native.WithInverse()
       { /** Return x.  */
         public Value apply(Value x) { return x.force(); }//apply(x)
         /* The identity function is its own inverse. */
         public Function inverse() { return this; }
       };//id x

   /** Return the first element of a {@link Value.Tuple}. */
   public static final Function fst =                                    // fst
      new Function.Native()           // fst
       { public Value apply(Value p)
          { return ((Value.Tuple) p.force()) .fst(); }
       };//fst

   /** Return the second element of a {@link Value.Tuple}. */
   public static final Function snd =                                    // snd
      new Function.Native()
       { public Value apply(Value p)
          { return ((Value.Tuple) p.force()) .snd(); }
       };//snd

   /** Return the third element of a {@link Value.Tuple}. */
   public static final Function thrd =                                  // thrd
      new Function.Native()
       { public Value apply(Value p)
          { return ((Value.Tuple) p.force()) .thrd(); }
       };//thrd

   /** Return the fourth element of a {@link Value.Tuple}. */
   public static final Function frth =                                  // frth
      new Function.Native()
       { public Value apply(Value p)
          { return ((Value.Tuple) p.force()) .frth(); }
       };//frth


   /** Return a Function based on the unary operator,
       'op' (not, -, hd, etc.). Also see
       {@link Value#uOp} and {@link #bOp}. */
   public static Function.Native uOp(final int op)                   // uOp(op)
    { return new Function.Native()
       { public Value apply(Value v) { return v.uOp(op); }
    }; }//uOp(op)

   /** Return a curried Function based on the binary operator, 'op'
       (+, *, etc).  Also see {@link Value#bOp} and {@link #uOp}. */
   public static Function.Native2 bOp(final int op)                  // bOp(op)
    { return new Function.Native2()
       { public Value apply2(Value v1, Value v2) { return v1.bOp(op, v2); }
    }; }//bOp(op)


   /** curry: <TT>((t, u) &rarr; v) &rarr; (t &rarr; u &rarr; v)</TT>.
       <NOBR><TT>curry fu x y = fu(x, y)</TT>.</NOBR>
       Also see {@link #uncurry}.  */
   public static final Function.Native3 curry =                        // curry
      new Function.Native3()
       { public Value apply3(Value f, Value p0, Value p1)
          { return ((Function) f.force()) .apply(Value.pair(p0, p1));
          }//apply3(f,p0,p1)
       };//curry

   /** uncurry: <TT>(t &rarr; u &rarr; v) &rarr; ((t, u) &rarr; v)</TT>.
       <NOBR><TT>uncurry fc (x, y) = fc x y</TT>.</NOBR>
       Also see {@link #uncurry}. */
   public static final Function.Native2 uncurry =                  // uncurry f
      new Function.Native2()
       { public Value apply2(Value f, Value p)
          { if( ! Type.PAIR.ni(p) ) error( "uncurry(f,p), p:" + p.type() );
            final Value.Tuple p0p1 = (Value.Tuple) p.force();  // WHNF
            final Value p0 = p0p1.elt(0), p1 = p0p1.elt(1);
            return ((Function) // NB. all apply(.)s return Values in WHNF
                      ((Function) f.force()) .apply(p0)) .apply(p1);
          }//apply2(f,p)
       };//uncurry


   /** <TT>whnf p0 p1</TT> can be used to make a parameter (p0) strict, say:
       {@link Value#force force}&nbsp;p0 to weak head normal form (WHNF), but
       return p1 (also in WHNF). */
   public static final Function.Native2 whnf =                    // whnf p0 p1
      new Function.Native2()
       { public Value apply2(Value p0, Value p1)
          { p0.force(); return p1.force(); }
       };//whnf p0 p1


   /** EXPERIMENTAL, The class of Cts2Cts Functions,
       Power(c,p)(x)=c.x<SUP>p</SUP>.
       Better to use {@link #inv inv(x)} if c=1 and p=&minus;1.  */
   public static class Power extends Cts2Cts                           // Power
    { /** 'c' the multiplicative constant, and 'p' the
          power of 'x', as in c*x<SUP>p</SUP>.  */
      public final double c, p;
      public Power(Value c,  Value  p) { this(c.x(), p.x()); }
      public Power(          Value  p) { this(1,     p.x()); }
      public Power(          double p) { this(1,     p    ); }
      public Power(double c, double p) { this.c = c; this.p = p; }

      /** Returns {@link #c c}&nbsp;&nbsp;x<SUP>{@link #p p}</SUP>.  */
      public double apply_x(double x)
       { if( c ==  0 ) return 0;
         if( p ==  1 ) return c * x;
         if( p == -1 ) return c / x;
         if( p ==  0 ) return c;
         if( x <=  0 ) error( "Power("+c+","+p+").apply("+x+")" );  // ?
         /* else */ return c * Math.pow(x, p);  // = c * x^p
       }//apply_x(x)

      /** The derivative of c.x<SUP>p</SUP> is c.p.x<SUP>p-1</SUP>.  */
      public Cts2Cts make_d_dx() { return new Power(c*p, p-1); }

      /** The integral of c.x<SUP>p</SUP> is (c/(p+1))x<SUP>p+1</SUP>
          <I>unless</I> p=&minus;1 in which case the integral is c.log(x). */
      public Cts2Cts2Cts make_integral()
       { if( p == -1 )
            return exactIntegral( new Cts2Cts()  // TODO: use my log
             { public double apply_x(double x) { return c * Math.log(x); } } );
         else return exactIntegral( new Power(c/(p+1) , p+1) );
       }//make_integral()
    }//Power class

   /** The function, inv(x)=1/x.  It is its own
       {@link Function.WithInverse Inverse}-function,
       i.e., inv(inv(x))=x, provided x&ne;0.
       Its integral is {@link #log log}.  */
   public static final Cts2Cts.WithInverse inv =                       //inv(x)
   new Cts2Cts.WithInverse()
    { public double apply_x(double x) { return 1.0 / x; }

      /** The indefinite integral of inv(x)=1/x is {@link #log log(x)}. */
      public Cts2Cts2Cts make_integral() { return exactIntegral(log); }

      /** The inverse function of inv(x)=1/x is itself, 'this'. */
      public Cts2Cts.WithInverse inverse() { return this; }
    };//inv(x)


   /** exp is a {@link Function.Cts2Cts Cts2Cts} with an
       {@link Function.WithInverse Inverse}-Function,
       i.e., {@link #log log}. */
   public static final Cts2Cts.WithInverse exp =                      // exp(x)
   new Cts2Cts.WithInverse()
    { public double apply_x(double x) { return java.lang.Math.exp(x); }

      /** exp, e<SUP>x</SUP>, is its own derivative. */
      public Cts2Cts.WithInverse make_d_dx()     { return this; }

      /** exp(x), e<SUP>x</SUP>, is its own integral. */
      public Cts2Cts2Cts make_integral() { return exactIntegral(this); }

      /** The inverse function of exp(x) is {@link #log log(x)}. */
      public Cts2Cts.WithInverse inverse() { return log; }
    };//exp(x)

   /** log is a {@link Function.Cts2Cts Cts2Cts} with an
       {@link Function.WithInverse Inverse}-Function, i.e., {@link #exp exp}.
       Its derivative is {@link #inv inv}.  */
   public static final Cts2Cts.WithInverse log =                      // log(x)
   new Cts2Cts.WithInverse()
    { public double apply_x(double x) { return java.lang.Math.log(x); }

      /** The derivative of log(x) is {@link #inv inv(x)}=1/x. */
      public Cts2Cts.WithInverse make_d_dx() { return inv; }

      /** The integral of {@link #log log(x)} is (x.log(x)&minus;x). */
      public Cts2Cts2Cts make_integral()
       { return exactIntegral(
            new Cts2Cts()
             { public double apply_x(double x)
                { return x*Math.log(x) - x; }
             }               );
       }//make_integral()

      /** The inverse function of log(x) is {@link #exp exp(x)}. */
      public Cts2Cts.WithInverse inverse() { return exp; }
    };//log(x)
   // .........................................................................

   /** This Function, <B>R</B><SUP>2</SUP>&rarr;<B>R</B><SUP>2</SUP>,
       converts polar coordinates &lang;r,&theta;&rang;
       to Cartesian &lang;x,y&rang;.
       The Jacobian {@link CtsD2CtsD#J matrix} is
       {{cos(&theta;), &minus;r.sin(&theta;)},
        {sin(&theta;), r.cos(&theta;)}}.
       The negative log of its determinant, {@link CtsD2CtsD#nlJ nlJ(r_t)},
       equals &minus;log(r).
       These are  used by {@link CtsD2CtsD#apply apply(.)} to "adjust"
       the AoMs of the result.  The inverse Function is
       {@link #cartesian2polar cartesian2polar&lang;x,y&rang;}.<BR>
       Note that <PRE>
       {{cos(&theta;), &minus;r.sin(&theta;)}, 
        {sin(&theta;),  r.cos(&theta;)}}
     &times; {{x/r,   y/r}, 
        {&minus;y/r<SUP>2</SUP>, x/r<SUP>2</SUP>}}
     = {{x.cos&theta;/r+y.sin&theta;/r, y.cos&theta;/r&minus;x.sin&theta;/r},
        {x.sin&theta;/r&minus;y.cos&theta;/r, y.sin&theta;/r+x.cos&theta;/r}}
     = {{1,&nbsp;0},
        {0,&nbsp;1}}. </PRE>  */
   public static final CtsD2CtsD.WithInverse polar2cartesian =
   new CtsD2CtsD.WithInverse()                           // polar2cartesian(rt)
    {
      /** Given &lang;r,&theta;&rang;, return &lang;x,y&rang; where
          x=r.cos&theta; and y=r.sin&theta;.  */
      public double [] apply_Vec(Vector r_t)
       { int n = r_t.nElts();
         if( n != 2 ) error( "polar2cartesian, n=" + n );
         final double r = r_t.x(0), theta = r_t.x(1);
         final double sin_t = Math.sin(theta), cos_t = Math.cos(theta);
         return new double [] {r*cos_t, r*sin_t};
       }//apply_Vec(r_t)

      /** The negative log of the determinant of the
          {@link #J(double,double,double) Jacobian} matrix
          equals &minus;log(r) in this case.
          (So it can be calculated directly.)  */
      public double nlJ(Vector r_t)
       { int n = r_t.nElts();
         if( n != 2 ) error( "polar2cartesian, n=" + n );
         return - Math.log( r_t.x(0) );
       }//nlJ(r_t)

      /** The Jacobian matrix,
          x=r.cos(theta),   y=r.sin(theta), so
          dx/dr=cos(theta), dx/dtheta=&minus;r.sin(theta),
          dy/dr=sin(theta), dy/dtheta=r.cos(theta).
          Used by {@link #J(Value) J(&lang;r,&theta;&rang;)}.
          Note, the determinant is just 'r'; see {@link #nlJ nlJ(r_t)}.  */
      // check: https://en.wikipedia.org/wiki/Jacobian_matrix_and_determinant
      protected double [] [] J(double r, double cos_t, double sin_t)
       { return new double [] [] {{cos_t, -r*sin_t}, {sin_t, r*cos_t}};
       }//J(r,c,s)

      /** The Jacobian matrix; uses
          {@link #J(double,double,double) J(r,cos(t),sin(t))}.  */
      public double [] [] J(Vector r_t)
       { int n = r_t.nElts();
         if( n != 2 ) error( "polar2cartesian, n=" + n );
         final double r = r_t.x(0), theta = r_t.x(1);
         return J(r, Math.cos(theta), Math.sin(theta));
       }//J([r_t])

      /** The inverse of polar2cartesian is {@link #cartesian2polar}.  */
      public CtsD2CtsD inverse() { return cartesian2polar; }
    };//polar2cartesian
   // .........................................................................

   /** This Function, <B>R</B><SUP>2</SUP>&rarr;<B>R</B><SUP>2</SUP>,
       converts Cartesian coordinates &lang;x,y&rang;
       to polar coordinates &lang;r,&theta;&rang;.
       The Jacobian-{@link CtsD2CtsD#J matrix} is
       {{x/r,&nbsp;y/r},
        {&minus;y/r<SUP>2</SUP>, x/r<SUP>2</SUP>}}.
       The negative log of its determinant,
       {@link CtsD2CtsD#nlJ nlJ(xy)}, equals +log(r).
       These are  used by {@link CtsD2CtsD#apply apply(.)} to "adjust"
       the AoMs of the result.  The inverse Function is
       {@link #polar2cartesian polar2cartesian&lang;r,&theta;&rang;}.  */
   public static final CtsD2CtsD.WithInverse cartesian2polar =
   new CtsD2CtsD.WithInverse()                           // cartesian2polar(xy)
    { /** Given &lang;x,y&rang;, return &lang;r,&theta;&rang; where
          r=&radic;(x<SUP>2</SUP>+y<SUP>2</SUP> and &theta;=arctan2(y,x). */
      public double [] apply_Vec(Vector xy)
       { int n = xy.nElts();
         if( n != 2 ) error( "polar2cartesian, n=" + n );
         final double x = xy.x(0), y = xy.x(1);
         final double r = Math.hypot(x, y),  // sqrt(x^2+y^2)
                  theta = Math.atan2(y, x);  // beats arctan(y/x), -pi..+pi
         return new double [] {r, theta};
       }//apply_Vec(xy)

      /** The negative log of the determinant of the
          {@link #J(double,couble,double) Jacobian} matrix
          equals +log(r) in this case where
          r=&radic;(x<SUP>2</SUP>+y<SUP>2</SUP>).
          (So it can be calculated directly.)  */
      public double nlJ(Vector xy)
       { int n = xy.nElts();
         if( n != 2 ) error( "polar2cartesian, n=" + n );
         final double x = xy.x(0), y = xy.x(1);
         return 0.5 * Math.log(x*x + y*y);  // = log(sqrt(r))
       }//nlJ(r_t)

      /** The Jacobian matrix,
          {{x/r, y/r}, {&minus;y/r<SUP>2</SUP>, x/<SUP>2</SUP>}}
          where r=x<SUP>2</SUP>+y<SUP>2</SUP>.
          Used by {@link #J(Value) J(&lang;r,&theta;&rang;)}.
          Note, the determinant is just&nbsp;1/r; see {@link #nlJ nlJ(xy)}. */
      protected double [] [] J(double r, double x, double y)
       { final double r2 = r*r;
         return new double[][] {{x/r, y/r}, {-y/r2, x/r2}};
         // {{x/r, y/r},{-y/(x^2(1+(y/x)^2)), 1/(x(1+(y/x)^2))}}
         // {{x/r, y/r},{-y/r^2,              x/r^2           }}
         // so, det = (x^2+y^2)/r^3 = 1/r
       }//J(r,x,y)

      /** The Jacobian matrix; uses {@link #J(double,double,double) J(r,x,y)}.
          Note the Inverse function theorem: the matrix inverse of the
          Jacobian matrix of an invertible function is the
          Jacobian matrix of the inverse function.  */
      public double [] [] J(Vector xy)
       { int n = xy.nElts();
         if( n != 2 ) error( "polar2cartesian, n=" + n );
         final double x = xy.x(0), y = xy.x(1);
         return J(Math.hypot(x, y), x, y);   // hypot(x,y) = sqrt(x^2+y^2)
       }//J(xy)

      /** The inverse of cartesian2polar is {@link #polar2cartesian}.  */
      public CtsD2CtsD inverse() { return polar2cartesian; }
    };//cartesian2polar
   // -------------------------------------------------------------------------

   private final static String [] ids  = new String[]                    // ids
      { "id",  "fst", "snd", "thrd", "frth", "whnf", "curry", "uncurry",
        "inv", "exp", "log",
        "polar2cartesian", "cartesian2polar" };

   private final static Value  [] vals = new Value []                   // vals
      {  id,   fst,   snd,   thrd,   frth,   whnf,   curry,   uncurry,
         inv,  exp,   log,
         polar2cartesian,   cartesian2polar };

   /** An Environment binding names (Strings) to standard Functions,
       for example, "id"&rarr;{@link #id id},
       "fst"&rarr;{@link #fst fst}, etc.
       Also see {@link Environment#lookup lookup}  */
   public final static Environment env =                                 // env
      Environment.empty.bind(ids, vals);
   // -------------------------------------------------------------------------

   /** Run a few simple tests on Library. */
   public static void main(String[] argv)
    { System.out.println("-- test Library --");

      Value a=new Value.Int(10), b=new Value.Int(2);
      Value.Tuple ab = Value.pair(a, b);

      Function.Native u2 = new Function.Native() // u2 :: (Int, Int) -> Int,
       { public Value.Int apply(Value ab)        // i.e., uncurried
          { Value.Tuple abT = (Value.Tuple) ab.force();
            return new Value.Int(abT.fst().n() + abT.snd().n());
       }  };

      Function.Native2 c2 = curry.apply(u2);     // curried

      System.out.println( "  currying: "
                        + u2.apply(ab)  + " = "
                        + c2.apply(a).apply(b) + " = "
                        + c2.apply2(a, b) + " = "
                        + uncurry.apply(c2).apply(ab)
                        );

      System.out.println( "  uOp, bOp: "
                        + "(-) " + a + " = " + uOp(Lexical.minus).apply(a)
                        + ", (+) " + a + " " + b + " = "
                           + bOp(Lexical.plus).apply(a).apply(b)
                        );

      System.out.println("-- done Library --");
    }//main

 }//Library class

// (c) Lloyd Allison 2012, 2016, 11 Feb 2020. Licence to 'Library' is Gnu GPL.
