//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package la.la;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The parser, also see {@link Lexical} and {@link Expression}. */
public class Syntax
 { private final Lexical lex;

   public static final int applicPriority = 7;  // i.e. of   f x

   public static final long                               // useful Symbol Sets
   arithOprs = (1L << Lexical.plus) | (1L << Lexical.minus)
             | (1L << Lexical.times)| (1L << Lexical.over),

   logicOprs = (1L << Lexical.andSy) | (1L << Lexical.orSy)
             | (1L << Lexical.notSy),

   comparisonOprs =
               (1L << Lexical.eq) | (1L << Lexical.ne) | (1L << Lexical.le)
             | (1L << Lexical.lt) | (1L << Lexical.ge) | (1L << Lexical.gt);

   /* and */ public static final long
   unOprs    = (1L << Lexical.minus)
             | (1L << Lexical.hdSy)   | (1L << Lexical.tlSy)
             | (1L << Lexical.nullSy) | (1L << Lexical.notSy),

   binOprs   = (1L << Lexical.consSy)
             | arithOprs
             | comparisonOprs
             | (1L << Lexical.andSy) | (1L << Lexical.orSy),

   rightAssoc = (1L << Lexical.consSy);

   /* and */ public static final long
   startsExp = unOprs
             | (1L << Lexical.word)
             | (1L << Lexical.numeral) | (1L << Lexical.numeralR)
             | (1L << Lexical.charLiteral) | (1L << Lexical.triv)
             | (1L << Lexical.nilSy)
             | (1L << Lexical.trueSy) | (1L << Lexical.falseSy)
             | (1L << Lexical.open)   | (1L << Lexical.letSy)
             | (1L << Lexical.ifSy)   | (1L << Lexical.lambdaSy);

   public static final int[] oprPriority = new int[Lexical.eofSy];

   void init()
    { for(int i = 0; i < oprPriority.length; i++) oprPriority[i] = 0;
      oprPriority[Lexical.consSy] = 1;
      oprPriority[Lexical.orSy]   = 2;
      oprPriority[Lexical.andSy]  = 3;
      for(int i = Lexical.eq; i <= Lexical.ge; i++) oprPriority[i] = 4;
      oprPriority[Lexical.plus]  = 5;  oprPriority[Lexical.minus] = 5;
      oprPriority[Lexical.times] = 6;  oprPriority[Lexical.over]  = 6;
    }//init


   /** Construct a Syntax analyser of a given {@link Lexical} source. */
   public Syntax(Lexical lex)                                     //constructor
    { this.lex = lex; init(); }
   // -------------------------------------------------------------------------

   private void insymbol()                                          // insymbol
    { lex.insymbol();
    }//insymbol, local version

   private boolean syIs(int sym)
    { if( lex.sy() == sym ) { insymbol(); return true; }
      return false;
    }//syIs

   private void check(int sym) // check and skip a particular symbol
    { if( lex.sy() == sym ) insymbol();
      else error( Lexical.Symbol[sym] + " expected, actually have "
                + Lexical.Symbol[lex.sy()] );
    }//check


   public Expression exp()                                             // exp()
    { Expression e = exp(1); check(Lexical.eofSy); return e; }


   private Expression exp(int priority)                             // exp(...)
    { if( priority < applicPriority )
       { Expression e = exp(priority+1);
         int sym = lex.sy();

         if( member(sym, binOprs) && member(sym, rightAssoc) // eg 1::2::3::nil
          && oprPriority[sym] == priority )
          { insymbol();
            return new Expression.Binary(sym, e, exp(priority));
          }
         else
         while( member(sym, binOprs) && !member(sym, rightAssoc)
             && oprPriority[sym] == priority ) // e.g. 1+2+3
          { insymbol();
            e = new Expression.Binary(sym, e, exp(priority+1));
            sym = lex.sy();
          }
         return e;
       }// < applicPriority

      else if( priority == applicPriority )                     // e.g. f g h x
       { Expression e = exp(priority+1);
         int sym = lex.sy();
         while( member(sym, startsExp) && ! member(sym, binOprs) )
          { e = new Expression.Application(e, exp(priority+1));
            sym = lex.sy();
          }
         return e;
       }

      else if( member(lex.sy(), unOprs) )               // e.g. not p,  e.g. -3
       { int sym = lex.sy(); insymbol();
         return new Expression.Unary(sym, exp(priority));
       }

      else                                                           // operand
       { Expression e = null;
         switch( lex.sy() )
          { case Lexical.word:
               e = new Expression.Ident(lex.theWord()); insymbol(); break;
            case Lexical.numeral:  // e.g., 123
               e = new Expression.Const(new Value.Int(lex.theInt()));
               insymbol(); break;
           case Lexical.numeralR:  // e.g., 31.4E-1 to 0.01 inferred
               e = new Expression.Const(Value.cts(lex.theDbl(), lex.theAcc()));
               insymbol(); break;
            case Lexical.charLiteral:
               e = new Expression.Const(
                      new Value.Char(lex.theWord().charAt(0)));
               insymbol(); break;
            case Lexical.triv:
               e = Expression.triv;   insymbol(); break;
            case Lexical.nilSy:
               e = Expression.nilCon; insymbol(); break;
            case Lexical.trueSy:
               e = Expression.ttrue;  insymbol(); break;
            case Lexical.falseSy:
               e = Expression.ffalse; insymbol(); break;
            case Lexical.open:                                           // (e)
               insymbol();
               e = exp(1);
               if( lex.sy() == Lexical.comma ) e = tuple(e, 1);  // (e0,e1,...)
               check(Lexical.close);
               break;
            case Lexical.letSy:                             // let [rec] d in e
               insymbol();
               Declaration d = dec();
               check(Lexical.inSy);
               e = new Expression.Block(d, exp(1));
               break;
            case Lexical.ifSy:                          // if e then eT else eF
               insymbol();
               e = exp(1);
               check(Lexical.thenSy); Expression eT = exp(1);
               check(Lexical.elseSy); Expression eF = exp(1);
               e = new Expression.IfExp(e, eT, eF);
               break;
            case Lexical.lambdaSy:                           // e.g. lambda x.e
               insymbol();
               Expression fp = param(); check(Lexical.dot);
               e = new Expression.LambdaExp(fp, exp(1));
               break;
            default:  error("bad operand");
          }//switch
         return e;
       }//if
    }//exp(...)


   private Expression.Tuple tuple(Expression e, int k)
    { Expression.Tuple ans;
      if( lex.sy() == Lexical.comma )           // ","
       { insymbol();
         ans = tuple(exp(1), k+1);              // e ...
       }
      else // we now know the k of the k-Tuple (& hopefully at ")")
         ans = new Expression.Tuple(new Expression[k]);
      ans.subExps[k-1] = e;
      return ans;
    }//tuple


   private Expression param()
    { Expression e = null;
      if( lex.sy() == Lexical.word )            // lambda x.e
         e = new Expression.Ident(lex.theWord());
      else if( lex.sy() == Lexical.triv )       // lambda ().e
         e = Expression.triv;
      else error("Lambda expression: bad formal parameter");
      insymbol();
      return e;
    }//param


   private Declaration dec()
    { boolean isRec = false;
      if( lex.sy() == Lexical.recSy ) { isRec = true; insymbol(); }
      return dec1(isRec, 0);
    }//dec

   private Declaration dec1(boolean isRec, int n)
    { if( lex.sy() == Lexical.word )
       { String id = lex.theWord(); insymbol();   // id
         check( Lexical.eq );                     // =
         Expression e = exp(1);                   // e
         Declaration d;
         if( lex.sy() == Lexical.comma )          // ,
          { insymbol();
            d = dec1(isRec, n+1);                 // ...
          }
         else// no more, i.e., 'in ...', so create d
            d = new Declaration(isRec, new String[n+1], new Expression[n+1]);
         d.ids[n] = id; d.es[n] = e;              // fill in n-th 'id=e'
         return d;
       }
   /* else */ error("no identifier in declaration"); return null;
    }//dec1


   private boolean member(int n, long s)    // ? is n a member of the "set" s ?
    { return ((1L << n) & s) != 0; }

   /** Print an error message, msg, and stop. */
   public void error(String msg)                                       // error
    { lex.error( "Syntax: " + msg ); }
   // -------------------------------------------------------------------------

   /** 'main' allows Syntax to be tested on its own, for example,
       <NOBR> java ...Syntax &lt; someFPsourceFile </NOBR> */
   public static void main(String[] argv)                               // main
    { System.out.println("-- Syntax --");
      for(int i=0; i < argv.length; i++) // command line params if any
         System.out.print("argv[" + i + "]=" + argv[i] + "\n");

      Lexical lex = argv.length <= 0
                  ? new Lexical(System.in)       // std input, or ...
                  : Lexical.forFile(argv[0]);    // ... named file
      Syntax syn = new Syntax( lex );
      Expression e = syn.exp();
      System.out.println( e.toString() );

      System.out.println("\n-- end Syntax --");
    }//main

 }//Syntax class

// Lambda, (c) Lloyd Allison, Oct 2007, Fri 9 May 2014. Licence is Gnu GPL.
