//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package la.maths;

import la.la.*;               // for Value etc.
import static la.la.Value.*;  // constants and static methods

import static la.util.Util.x2String;

import la.maths.*;
import static la.maths.Maths.*;
import static la.maths.Vector.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** Test.{@link Test#main main()} runs some simple(!)
    tests on package {@link la.maths.README la.maths}. */
public class Test
 {
   private static String show(Value v) { return v + ":" + v.type(); }
   // -------------------------------------------------------------------------

   /** Run some simple tests on package {@link la.maths.README la.maths},
       that is on {@link Vector}, {@link Matrix} and {@link Q Q}. */
   public static void  main(String[] argv)
    { System.out.println("-- test maths --");

      System.out.println(
        "  logSum({9,10})=" + x2String(logSum(new double[]{9.0, 10.0}))
      + ", logSum(9,9)=" + x2String(logSum(9,9))
      + ", logSum(-9,-10)=" + x2String(logSum(-9.0, -10.0)) + " nits" );

      System.out.print( "\n  Catalans: " );
      for( int i = 0; i < 7; i++ )
       { System.out.print( "C["+i+"]="+catalan(i)+", " );
         // System.out.print( "("+cummulativeCatalans(i)+"), " );
       }
      System.out.println( "..." );

      System.out.println( "\n  logGamma(i) etc.:" );
      java.text.DecimalFormat fmt = new java.text.DecimalFormat( "#0.000" );
      for( int i = 0; i < 6; i ++ )
       { double lfi = logFactorial(i);
         System.out.println( "  "
          + i + "! = " + fmt.format(Math.exp(lfi))
          + " = exp(" + fmt.format(lfi) + ")"
          + ",  logGamma(" + (i+1) + ") = " + fmt.format(logGamma(i+1)) );
       }

      System.out.println( "\n  nCk by log_nCk(n,k):" );
      for( int n = 0; n < 7; n ++ )
       { System.out.print( "  " );
         for( int k = 0; k <= n; k ++ )
            System.out.print( n + "C" + k + "="
                            + fmt.format(Math.exp(log_nCk(n, k))) + " " );
         System.out.println();
       }

      System.out.println( "\n  Bessel_I(alpha, x):" );
      for( double alpha = 0; alpha <= 3; alpha += 0.5 )
       { System.out.print( "  " );
         for( int x = 0; x <= 4; x ++ )
            System.out.print( "I" + alpha + "(" + x + ")="
                        + fmt.format(Bessel_I(alpha,x))
                        + ", " );
         System.out.println();
       }


      Value one5 = Value.real(1.5),
            two1 = Value.real(2.1),
            half = Value.real(0.5),
            one2 = Value.real(1.2);

      Value oneC   = Value.cts(1, 0.1),
            twoC   = Value.cts(2, 0.1),
            threeC = Value.cts(3, 0.1);

      Vector vec = Vector.values( new Value[]{three,two,one} ),
             O   = Vector.doubles(new double[]{0, 0, 0}),
             x   = Vector.doubles(new double[]{1, 0, 0}),
             y   = Vector.doubles(new double[]{0, 1, 0}),
             z   = Vector.doubles(new double[]{0, 0, 1}),
             d   = Vector.doubles(new double[]{1, 1, 1});

      Matrix xyz = values(new Value[]{O, x, y, z}).asMatrix(),
             yzx = values(new Value[]{O, y, z, x}).asMatrix();

      Vector vec2= Vector.doubles(2.0, new double[]{1, 2, 3});
      // ----------------------------------------------------------------------

      System.out.println( "\n  Vector operations:"
            + "   vec=" + show(vec)
            + ", vec.[1,3)=" + vec.slice(1,3)
            + ", vec.[1,3).[0,1)=" + vec.slice(1,3).slice(0,1)
            + ", vec.{2,1,2}=" + vec.elts(new int[]{2,1,2})
         + "\n   vec.sorted=" + vec.sorted()
            + ", (vec,vec).append=" + values(new Vector[]{vec, vec}).append()
         + "\n   vec zip vec=" + vec.zip(vec)
         + "\n   -vec=" + vec.uOp(Lexical.minus)
            +  ", vec+vec=" + vec.bOp(Lexical.plus, vec)
            +  ", vec.foldr + ="
               + vec.foldr(Function.fromBop(Lexical.plus))
         + "\n   vec2=" + show(vec2)
         + "\n     as_nx1=" + vec2.as_nx1()
         +       " as_1xn=" + vec2.as_1xn()
         + "\n     asDiagonal=" + vec2.asDiagonal(zeroR)
         + ",\n   " + show(Vector.values(new Value[]{
                      Value.cts(1.8,0.1), Value.cts(1.7,0.1)} ))
         + ",\n   x=" + x + ", x.rotate(2*pi/3,d)="  // 120 degrees about d
                             + x.rotate(2*Math.PI/3, d)      // x -> y axes
         );

      Vector vecC = Vector.values(new Value[]{oneC, twoC, threeC}),
             vecC2 = vecC.scaled(2),
             vecCn = vecC.normalised();
      double vecCnorm = vecC.norm();          // ||vecCn|| = 1, now 2 deg.f.
      Vector vecC1 = vecC.scaled(1/vecCnorm); // ||vecC1|| = 1, still 3 deg.f.
      double vecC1_nlAoM = vecC1.nlAoM();
      System.out.println( "\n  Vector operations:"
           + "\n   vecC = " + vecC + " (" + vecC.nlAoM() + " nits)"
           + "\n     3 log 2 = " + 3.0*Math.log(2)
           + "\n     scaled(2) = " + vecC2 + " (" + vecC2.nlAoM() + ")"
           + "\n     scaled(1/norm) = " + vecC1+ " (" + vecC1_nlAoM + ")"
           + "\n     " + vecC1_nlAoM + " * 2/3 = " + 2*vecC1_nlAoM/3
           + "\n     normd() = "   + vecCn + " (" + vecCn.nlAoM() + ")"
         );
      // ----------------------------------------------------------------------

      java.util.Random RNG = new java.util.Random();
      double[][] ds = new double[5][5];
      for( int r = 0; r < ds.length; r ++ )
         for( int c = 0; c < ds[0].length; c ++ )
            ds[r][c] = RNG.nextDouble();

      Matrix M =
         // doubles( ds );  // use the random ds[][] above
         Matrix.doubles( new double[][] { {1,2,1}, {3,2,1}, {3,0,4} } );
      Matrix Mi   = M.inverse();
      Matrix M2x3 = Matrix.values( new Value[][]{{one5, two1, half},
                                          {half, one2, one5}} );
      Matrix M2x3t= M2x3.transpose();
      Matrix P1   = M2x3 .times(M2x3t);
      Matrix P2   = M2x3t.times(M2x3 );
      System.out.println( "\n  Matrix operations:"
                        + "\n   M=" + M
                        + "\n     vec M = " + vec.times(M)
                             + ", M vec = " + M.times(vec)
                        + "\n     Mi = " + Mi
                        + "\n     M*Mi = " + M.times(Mi)
                        + "\n     Mi*M = " + Mi.times(M)
                        + "\n   M2x3  = " + show(M2x3)
                        + ", shape=" + ints(M2x3.shape())
                        + "\n   M2x3' = " + show(M2x3t)
                        + "\n   M2x3*M2x3' = " + show(P1)
                        + "\n   M2x3'*M2x3 = " + show(P2)
         );

      Matrix S = Matrix.doubles( new double[][]{{1,2,3},{2,1,1},{3,1,1}} );
         Value.Tuple sE = S.eigen();
         Vector  sEvals = (Vector) sE.fst();
         Matrix  sEvecs = (Matrix) sE.snd();
      System.out.println( "\n  Eigen things:"
                        + "\n  S=" + S
                        + "\n    E_Vals=" + sEvals
                        + "\n    E_Vecs=" + sEvecs
                        + "\n  (S*Evecs')'="
                               + S.times(sEvecs.transpose()).transpose()
                        + "\n    c.f.      "
                               + sEvals.asDiagonal(zero).times(sEvecs)
                        + "\n            = E_Vals.asDiag() * E_Vecs"
         );

      // System.out.println( "   file 'test_dbls': "
      //                   + fromLexical(Lexical.forFile("test_dbls")) );
      // 1.2   1.23 1.2
      // 2.345 2.4  2.5
      // 3.1   3.2  3.3
      // 4.1   4.2  4.3       -- test_dbls, say.

      // ----------------------------------------------------------------------

      Q qr = xyz.match(yzx);
      System.out.println( "\n  Rotations by qr=" + qr
                        + "\n    qr("+x+") = " + x.rotate(qr)
                        + "\n    qr("+y+") = " + y.rotate(qr)
                        + "\n    qr("+z+") = " + z.rotate(qr)
         );

      System.out.println("-- done maths --");
    }//main

 }//class Test

// (c) Lloyd Allison 2013, 29 April 2014.  Licence to 'Test' is Gnu GPL.
