//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package la.maths;

import java.util.Comparator;
import java.io.InputStream;

import la.util.*;
import static la.maths.Maths.*;
import la.la.*;
import mml.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference:
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The class of <I>homogeneous</I> Vectors of arbitrary length. The
    principal operations are {@link #nElts() nElts()} the number of
    <I>top-level</I> elements, {@link #elt(int) elt(i)}, and
    {@link #wt(int) wt(i)} the weight of elt(i). The default
    assumptions are that &forall;&nbsp;i&nbsp;wt(i)=1
    (see {@link #constWt() constWt()}), and that, if applicable,
    elt(i).{@link Value#AoM() AoM()} may vary with&nbsp;i
    (see {@link #constAoM() constAoM()}.  Also see
    {@link Type#VECTOR}, {@link la.la.Value.Tuple} and
    {@link Matrix}. */
public abstract class Vector extends Value.Structured               // * Vector
                             implements Value.Scannable
 {
   /** Class Derived can be useful when making a modified Vector.
       For a Vector v, <NOBR>v.new Derived(){ nix }</NOBR> is just
       like v, but a sub-class of Derived probably overides something. */
   public class Derived extends Vector                               // Derived
    {    /** The Vector from which 'this' is Derived. */
      public Vector     parent() { return Vector.this; }

         /** The Type of the original Vector. */
      public Type         type() { return Vector.this.type(); }

         /** The element Type of the original Vector. */
      public Type      eltType() { return Vector.this.eltType(); }

         /** The number of elements in original Vector. */
      public int         nElts() { return Vector.this.nElts();    }

         /** elt(i) of the original Vector. */
      public Value    elt(int i) { return Vector.this.elt(i);     }

         /** wt(i) of the original Vector. */
      public double    wt(int i) { return Vector.this.wt(i);      }

         /** constWt() of the original Vector. */
      public double    constWt() { return Vector.this.constWt();  }

         /** constAoM() of the original Vector. */
      public double   constAoM() { return Vector.this.constAoM(); }

    }//Derived class
   // ............................

   /** The empty Vector with no elements at all. */
   public static final Vector empty =                                  // empty
    new Vector()
     { public int      nElts() { return 0; }
       public Value elt(int i) { throw RTE( ".elt(" + i + ")" ); }
       public double wt(int i) { throw RTE( ".wt("  + i + ")" ); }
     };

   /** "["; see {@link #print} and {@link #toString()}. */
   protected String  opens() { return "["; }
   /** "]"; see {@link #print} and {@link #toString()}. */
   protected String closes() { return "]"; }


   /** {@link Type Type}.{@link Type#VECTOR VECTOR}. */
   public Type type() { return Type.VECTOR; }                         // type()

   /** The number of elements (rows) in 'this' Vector.
       Also see {@link #wts() wts()}.  */
   public abstract int nElts();                                    // * nElts()

   /** Element 'i' of 'this' Vector. Also see {@link #n(int) n(i)},
       {@link #x(int) x(i)}, and {@link #wt}.  */
   public abstract Value elt(int i);                                // * elt(i)

   /** Return this.elt(i).n() &mdash; <I>assuming</I> this is a
       Vector of Int.  A subclass might have a faster way to do this.  */
   public int n(int i) { return elt(i).n(); }                           // n(i)

   /** Return this.elt(i).x() &mdash; <I>assuming</I> this is a
       Vector of Cts.  A subclass might have a faster way to do this.  */
   public double x(int i) { return elt(i).x(); }                        // x(i)

   /** The type of all(!) elements of 'this' Vector.
       Note that a Vector must be homogeneous. */
   public Type eltType()                                           // eltType()
    { if( nElts() <= 0 ) error( ".eltType(), dunno, no elements!" );
      return elt(0).type();
    }//eltType()

   /** <TT>wt(i)</TT>, the weight, of {@link #elt(int) elt(i)} in
       'this' data-set (Vector); this <I>default</I> is 1.0.  We
       may have fractional {@link Weighted#wt(int) weights} for
       {@link mml.Mixture mixture}-modelling, say, &gt;1 for multiple copies
       of a datum, and zero to turn a datum "off".  Note that wt(i),
       and {@link #constWt() constWt()} must be consistent!
       For some operations on Vector, such as append, it is pretty
       clear what wt(i) should be in the result, but not for some
       others, such as zip, in which case you get something plausible.
       Also see {@link #wts()}.  */
   public double wt(int i) { return 1.0; }                             // wt(i)

   /** By default, return 1.0 indicating that every element has that
       {@link #wt weight}. Return a negative to indicate that weights may
       vary. Used by {@link #wts() wts()}, {@link #wtNlAoM() wtNlAoM()}, etc..
       (Also see {@link #constAoM() constAoM()}.)  */
   public double constWt() { return 1.0; }                         // constWt()

   /** By default, return &minus;1 (any negative) indicating the elements' AoMs
       (if appropriate) may vary from element to element. If they all equal
       aom&ge;0, return that value. If the elements are precise, return zero.
       Used by {@link #nlAoM() nlAoM()} and {@link #wtNlAoM() wtNlAoM()}.
       Also see {@link #constNlAoM() constNlAoM()} and
       {@link #constWt() constWt()}. */
   public double constAoM() { return -1; /*i.e., false*/ }        // constAoM()

   /** If {@link #constAoM constAoM()} is &gt;0, return
       its negative log, otherwise an error. */
   public double constNlAoM()                                   // constNlAoM()
    { final double cAoM = constAoM();
      if( cAoM > 0 ) return - Math.log( cAoM ); // else...
      throw RTE( ".nlAoM(), .constAoM()=" + cAoM + " <= 0" );
    }//constNlAoM()

   /** The total -ve log AoM of the whole Vector (if appropriate); it
       looks at {@link #constAoM() constAoM()} and acts accordingly.
       This is particularly for {@link mml.Model#pr pr(v)} and nlPr(v)
       where v is a Cts Vector. Also see {@link #wtNlAoM() wtNlAoM()}. */
   public double nlAoM()                                             // nlAoM()
    { return nlAoM(0, nElts()); }

   /** Return the sum {@link #nlAoM() nlAoM} of elements [lo,&nbsp;hi).
       Also see {@link #nlAoM(int) nlAoM(i)}.  */
   public double nlAoM(int lo, int hi)                          // nlAoM[lo,hi)
    { final double cAoM = constAoM();
      if( cAoM >  0 ) return (hi-lo) * constNlAoM(); // for all elts
      if( cAoM == 0 ) error( ".nlAoM(), constAoM()==0, all elts precise" );
      // else constAoM() < 0 ...
      double ans = 0;  // cAoM < 0, AoM may vary elt to elt ...
      for( int i = lo; i < hi; i ++ ) ans += elt(i).nlAoM();
      return ans;
    }//nlAoM[lo,hi)

   /** Return this.elt(i).AoM() &mdash; <I>assuming</I> this is a
       Vector of Cts. A subclass might have a faster way to do this.  */
   public double AoM(int i) { return elt(i).AoM(); }

   /** Return this.elt(i).nlAoM() &mdash; <I>assuming</I> this is a
       Vector of Cts. A subclass might have a faster way to do this.  */
   public double nlAoM(int i) { return elt(i).nlAoM(); }

   /** Return <NOBR>&sum;<SUB>i</SUB> {@link #wt(int) wt(i)}.</NOBR>
       It looks at {@link #constWt() constWt()} and acts accordingly.
       <NOBR>wts() is the</NOBR> real 'N' of a data-set. Many Vectors can
       do wts() in O(1)-time, but {@link Weighted#wts() not&nbsp;all}! */
   public double wts()                                                 // wts()
    { return wts(0, nElts()); }

   /** Return the total {@link #wts() weights} of elements [lo,&nbsp;hi). */
   public double wts(int lo, int hi)                              // wts[lo,hi)
    { final double cW = constWt();
      if( cW >= 0 ) return (hi-lo) * cW;
      double ans = 0;  // else wts may vary elt to elt ...
      for( int i = lo; i < hi; i ++ ) ans += wt(i);
      return ans;
    }//wts[lo,hi)

   /** Return the index, m, of the median element on the basis of
       {@link #wt(int) weights}.  If {@link #constWt() constWt()}&ge;0
       <NOBR>m = nElts()&div;2.</NOBR> Otherwise it is the smallest 'm'
       s.t. <NOBR>2&times;&Sum;<SUB>0..m</SUB> wt(m)
                  &ge; {@link #wts() wts()}</NOBR>. It assumes the
       Vector is sorted somehow and does not check this.  */
   public int wtMedian()                                          // wtMedian()
    { if( constWt() >= 0 ) return nElts()/2; // e.g., 5 elts, [0..4] -> 2
      int m; double w = 0; final double wts = wts();
      if( wts <= 0 ) error( ".wtMedian(), wts()=" + wts );
      for( m = 0; 2*w < wts; m ++ ) w += wt(m);
      // 2*w >= wts, and 0 < m <= nElts()
      return m-1;
    }//wtMedian()

   /** The total -ve log AoM of the whole Vector (if appropriate),
       taking into account <I>element {@link #wt(int) weights}</I>.
       This is particularly for {@link mml.Model#nlLH nlLH(ss)} of a
       Cts data-set and so on. It looks at {@link #constWt() constWt()}
       and at {@link #constAoM() constAoM()} and acts accordingly.
       Also see {@link #nlAoM() nlAoM()}.  */
   public double wtNlAoM()                                         // wtNlAoM()
    { return wtNlAoM(0, nElts()); }

   /** Return the total weighted {@link #wtNlAoM() nlAoM} of
       elements [lo,&nbsp;hi).  */
   public double wtNlAoM(int lo, int hi)                      // wtNlAoM[lo,hi)
    { final double cAoM = constAoM();
      if( cAoM >  0 ) return wts(lo,hi) * constNlAoM();  // all elt AoMs same
      if( cAoM == 0 ) error( ".wtNlAoM(), cAoM==0" );
      double ans = 0;  // else cAoM < 0, so AoMs may vary...
      final double cW = constWt();
      if( cW >= 0 )    // but wts all the same
         return nlAoM(lo, hi) * cW;
      // else... both wt(i) and elt(i).AoM() may vary with i ...
      for( int i = lo; i < hi; i ++ ) ans += wt(i) * elt(i).nlAoM();
      return ans;
    }//wtNlAoM[lo,hi)

   /** Return a sub-Vector of 'this', being made up of the
       positions, ps.  Also see {@link #cols(int[]) cols(cs)}
       and {@link la.la.Value.Tuple#elts(int[])}.  */
   public Derived elts(final int [] ps)                             // elts(ps)
    { final Vector v = this;  final int nE = nElts();
      for( int i = 0; i < ps.length; i ++ ) // check...
       { final int ps_i = ps[i];
         if( ps_i < 0 || ps_i >= nE )
            error( ".elts(ps), ps[" + i + "]=" + ps[i] );
       }//...AOK
      return this.new Derived()
       { public int      nElts() { return ps.length;    }
         public Value elt(int i) { return v.elt(ps[i]); }
         public double wt(int i) { return v.wt( ps[i]); }
    }; }//elts(ps)

   /** Return an array containing the unique elements of
       'this' Vector in ascending order. */
   public Value [] uniqueElts()                                 // uniqueElts()
    { java.util.TreeSet <Value> ts = new java.util.TreeSet <Value> ();
      for( int i = 0; i < nElts(); i ++ ) ts.add(elt(i)); // NB. no duplicates
      return ts.toArray( new Value[ts.size()] );
    }//uniqueElts()

   /** Return a Vector consisting of a single element, 'e'. */
   public static Vector singleton(final Value e)                // singleton(e)
    { return new Vector()
       { public int       nElts() { return 1; }
         public Value  elt(int i)
          { if( i != 0 ) error( ".elt(i=" + i + ")" ); /*else*/ return e; }
         public String toString() { return "singleton(" + e + ")"; }
    }; }//singleton(v)


   /** Make a {@link la.util.Series#fromVector Series}, the elements
       of 'this' {@link Vector Vector}, one at a time. */
   public Series toSeries()                                       // toSeries()
    { return Series.fromVector(this); }

   /** Make a Vector from the given Scannable Value sv.
       (If sv <I>is</I> a Vector, return sv.) In general the returned
       Vector is inefficient for non-sequential random access.
       The length of sv.toSeries() must be deterministic!<BR>
       Why do we have fromScannable(sv) and not 'fromSeries(s)' &ndash;
       well a {@link Series Series} is used up and cannot be "rewound"
       but a Scannable can be "re-&nbsp;toSeries()&nbsp;-ed".  */
   public static Vector fromScannable(final Scannable sv)    // fromScannable()
    { if( sv instanceof Vector ) return (Vector) sv; // else...
      return new Vector()
       { private int length = -1; // i.e., length is unknown as yet
         private Series s = sv.toSeries();
         public int nElts()
          { if( length < 0 ) // i.e., unknown
             { s.exhaust(); length = s.position(); }
            return length;
          }//nElts()
         public Value elt(int i) { moveTo(i); return s.elt(); }
         public double wt(int i) { moveTo(i); return s .wt(); }
         public double constWt() { return -1; /* NB. false */ }
         private void moveTo(int i)
          { if( i < s.position() ) s = sv.toSeries(); // i >= pos = 0
            while( i > s.position() ) s.advance();    // i = pos
          }//moveTo(i)
    }; }//fromScannable(sv)


   /** Return 'this' Vector, sorted on {@link Value#comparator}.
       Also see {@link #sorted(Comparator) sorted(Comparator)}. */
   public Derived sorted()                                          // sorted()
    { return sorted( Value.comparator ); }

   /** Return 'this' Vector sorted on column 'col'. */
   public Derived sorted(final int col)                          // sorted(col)
    { return sorted( new Comparator <Value> ()
          { public int compare(Value e1, Value e2)
             { return e1.elt(col).compareTo( e2.elt(col) ); }
          } );
    }//sorted(col)

   /** Return 'this' Vector sorted on 'cmp'.
       Also see {@link #sorted() sorted()}. */
   public Derived sorted(Comparator <Value> cmp)                 // sorted(cmp)
    { final int N = nElts();
      int [] src = new int[N], tgt = new int[N];
      for( int i = 0; i < N; i ++ )
       { src[i] = i; tgt[i] = i; }
      merge(0, N, src, tgt, cmp);  // see merge(...) below
      return elts(tgt);
    }//sorted(cmp)

   /** MergeSort tgt[lo,hi) on the basis of
       <NOBR>cmp(elt(tgt[.]), elt(tgt[.])),</NOBR>,
       possibly using src[lo,mid) and src[mid,hi)
       where <NOBR>mid = (lo + hi)&div;2</NOBR>. */
   protected void merge(int lo, int hi, int[] src, int[] tgt,      // merge(..)
                        Comparator <Value> cmp)
    { if( hi <= lo+1 ) return;        // hi-lo < 2, else ...
      final int mid = (lo + hi) / 2;  // hi-lo >= 2, lo < mid < hi,
      // e.g., [0,2) -> [0,1) & [1,2),  [0,3) -> [0,1) & [1,3), etc.
      merge(lo, mid, tgt, src, cmp); merge(mid, hi, tgt, src, cmp);
      int a = lo, b = mid, t = lo;
      while( true )
         if( cmp.compare( elt(src[a]), elt(src[b]) ) <= 0 ) // <=, stable
          { tgt[t] = src[a]; a ++ ; t ++ ;
            if( a >= mid ) // src[lo,mid) is exhausted, copy src[b,hi)
             { for( int i = b; i < hi;  i ++ ) { tgt[t] = src[i]; t ++ ; }
               return;
          }  }
         else // ... [a] ... [b]... > 0
          { tgt[t] = src[b]; b ++ ; t ++ ;
            if( b >= hi  ) // src[mid,hi) is exhausted, copy src[a,mid)
             { for( int i = a; i < mid; i ++ ) { tgt[t] = src[i]; t ++ ; }
               return;
          }  }//if
    }//merge(..)
   // ---------------------------------------------------

   /** 'This' Vector (data-set?) of elements, each element
       explicitly {@link #wt(int) weighted}, somehow.  */
// ??? To work around Apple Java 1.6...
// public abstract class Weighted extends Derived                 // * Weighted
   public abstract class Weighted extends Vector                  // * Weighted
    {
      /** The weight of elt(i) in this Vector (data-set).  */
      public abstract double wt(int i);                           // * wt(i)

      /** Default -1, indicating elts' weights may vary from elt to elt. */
      public double constWt() { return -1; /* false */ }          // constWt()

      /** ??? The following are copied from 'class Derived' to
          ??? work around Apple Java 1.6 (Feb 2013). */
      public Type      type() { return Vector.this.type(); }
      public Type   eltType() { return Vector.this.eltType(); }
      public int      nElts() { return Vector.this.nElts();   }
      public Value elt(int i) { return Vector.this.elt(i);    }

      /** As per {@link Vector#constAoM() parent} Vector. */
      public double   constAoM() { return Vector.this.constAoM();   }

    }//class Weighted
   // ......................

   /** Return a Vector, based on the data in 'this' Vector, but with weights
       adjusted &times;&nbsp;wts[i].  Also see {@link #weight(double)}. */
   public Weighted weight(final double[] wts)                    // weight(wts)
    { final double cW = constWt();
      if( nElts() != wts.length ) error( ".weight(wts), lengths" );
      return cW >= 0 ?
         this.new Weighted()        // 'this' has constant weight cW
          { public double wt(int i) { return cW * wts[i]; } }
       : this.new Weighted()
          { public double wt(int i) { return Vector.this.wt(i) * wts[i]; } };
    }//weight(wts)

   /** Return a Vector, based on the data in 'this' Vector, but with
       weights adjusted &times;&nbsp;'wt'&nbsp;&ge;&nbsp;0.
       Also see {@link #weight(Vector)}. */
   public Weighted weight(final double wt)                        // weight(wt)
    { assert wt >= 0 : "wt = " + wt + " < 0";
      final double cW = constWt() * wt;
      return cW >= 0 ?
         this.new Weighted()
          { public double wt(int i) { return cW; } // !
            public double constWt() { return cW; } // !
          }
       : this.new Weighted()
          { public double wt(int i) { return Vector.this.wt(i) * wt; } };
    }//weight(wt)

   /** Re-weight 'this' Vector adjusting weights by
       &times;&nbsp;v.elt(i).x().
       Also see {@link #weight(double[])}. */
   public Vector weight(final Vector v)                            // weight(v)
    { if( nElts() != v.nElts() ) error( ".weight(v), lengths" );
      final double cW1 = constWt();
      return cW1 >= 0 ?  // ? 'this' has a constant weight ?...
         this.new Weighted()        // cW1 >= 0 and cW2 < 0
          { public double wt(int i) { return cW1 * v.elt(i).x(); } }
       : this.new Weighted()        // cW1 < 0  and cW2 < 0
          { public double wt(int i)
             { return Vector.this.wt(i) * v.elt(i).x(); }
          };
    }//weight(v)

   /** Re-weight 'this' Vector setting all
       weights <I>equal to</I> w&nbsp;&ge;&nbsp;0. */
   public Vector weightLike(double w)                          // weightLike(w)
    { assert w >= 0 : "w = " + w + " < 0";
      return this.new Weighted()
          { public double wt(int i) { return w; }
            public double constWt() { return w; }
          };//new Weighted
    }//weightLike(w)

   /** Re-weight 'this' Vector setting weights, wt(i),
       <I>equal to</I>&nbsp;v.wt(i).  Also see {@link #weight(double[])}. */
   public Vector weightLike(final Vector v)                    // weightLike(v)
    { if( nElts() != v.nElts() ) error( ".weightLike(v), lengths" );
      final double cWv = v.constWt();
      return cWv >= 0 ?
         this.new Weighted()
          { public double wt(int i) { return cWv; }
            public double constWt() { return cWv; }
          }
       : this.new Weighted()  // cWv < 0
          { public double wt(int i) { return v.wt(i); }
            public double constWt() { return cWv; }
          };
    }//weightLike(v)
   // -----------------------------------------------

   /** Compare 'this' Vector to v, returning -ve&nbsp;(&lt;),
       0&nbsp;(=), or +ve&nbsp;(&gt;). */
   public int compareTo(Value /*Vector*/ v)                     // compareTo(v)
    { if( ! Type.VECTOR.ni(v) ) error( ".compareTo(v), v not a Vector" );
      int c = 0;  final int N = Math.min(nElts(), v.nElts());
      for( int i = 0; i < N; i ++ )
       { c = elt(i).compareTo(v.elt(i));
         if( c != 0 ) return c;
       }
      return nElts() - v.nElts(); // equal prefix, shorter is lesser
    }//compareTo(v)

   /** The Euclidean <I>norm</I>, ||v||=&radic;(v<B>.</B>v), of 'this'
       Vector <U>of Cts</U>, that is, &radic;{@link #sumSq() sumSq()}. */
   public double norm()                                               // norm()
    { return Math.sqrt(sumSq()); }

   /** 'This' Vector's Euclidean {@link #norm() norm} as a
       {@link la.la.Value.Cts} with nlAoM() being the Vector's
       divided by nElts().  Also see {@link #normalised normalised}.  */
   public Cts norm_Cts()
    { return cts(norm(), Math.exp( - nlAoM() / nElts()) ); }

   /** Provided 'this' is a Vector of Cts, return the sum of squares,
       <NOBR>&sum;<SUB>i</SUB> elt(i).x()<SUP>2</SUP></NOBR>.  This
       default iterates over the elements. Also see {@link #norm()}.  */
   public double sumSq()                                             // sumSq()
    { double sum = 0;
      for( int i = 0; i < nElts(); i ++ )
       { final double eix = elt(i).x(); sum += eix * eix; }
      return sum;
    }//sumSq()

   /** The dot- (inner-) product of Vectors of Cts, 'this' and 'v'. */
   public double dot(Vector v)                                        // dot(v)
    { if( nElts() != v.nElts() )
         error( ".dot(v), shapes, " + nElts() + "!=" + v.nElts() );
      double sum = 0;
      for( int i = 0; i < nElts(); i ++ )
         sum += elt(i).x() * v.elt(i).x();
      return sum;
    }//dot(v)

   /** 'this' Vector of Cts, v, as a row-Vector, times {@link Matrix} M,
       as in <NOBR>v M.</NOBR> */
   public Vector times(Matrix M)                                  // v.times(M)
    { return as_1xn().times(M).elt(0); }

   /** Return 'this' Vector scaled by a factor, s. <B>Note</B>,&nbsp;AoM
       changes by &times;s <I>per element</I>, e.g., if v.elt(i).x() is
       known to 0.001, v.scaled(10).elt(i).x() is known to 0.01. */
   public Vector scaled(final double s)                            // scaled(s)
    { final boolean aomZero = AoM() <= 0;
      final Vector v = this;
      return this.new Derived()
       { /** The element, both .x() and .AoM(), scaled. */
         public Value elt(int i)
          { Cts vi = (Cts) v.elt(i);
            return aomZero ? real(vi.x()*s)
                           : cts( vi.x()*s, vi.AoM()*s);
          }//elt(i)
         /** AoM() of the whole Vector. */
         public double AoM()
          { return aomZero ? 0 : Math.exp( - nlAoM() ); }
         /** nlAoM() of the whole Vector. This is changed by scaling! */
         public double nlAoM()
          { return v.nlAoM() - v.nElts()*Math.log(s); }
            // if s > 1, log(AoM) > 0, then...
            // AoM() & log(AoM()) increase, so nlAoM() decreases
         public double constAoM()
          { final double cAoM = v.constAoM();
            return cAoM >= 0 ? cAoM * s : cAoM;
          }
    }; }//scaled(s)

   /** Return 'this' Vector <U>of Cts</U> normalised to 1.0, which has
       implications for {@link #AoM} and {@link #nlAoM}. <B>Note</B>
       (i)&nbsp;AoM changes by &times;1/norm() <I>per element</I>, <I>and</I>
       (ii)&nbsp;degrees of freedom reduce by one and hence
       nlAoM() reduces by a factor of (nElts-1)/nElts. */
   public Vector normalised()                                   // normalised()
    { return normalised1(); /* say */ }

   /** Return 'this' Vector <U>of Cts</U> {@link #normalised normalised}
       by making a Vector of new (scaled) elements. */
   protected Vector normalised1()
    { final double nrm = norm(), aom = AoM();
      if( aom == 0 && Math.abs(nrm - 1.0) < epsilon )
         return this;  // ok as is, else ...
      final int nE = nElts();
      final double scl = 1.0 / nrm;
      final double[] ds = new double[nE];
      for( int i = 0; i < nE; i ++ )
         ds[i] = elt(i).x() * scl;  // scaling
      return aom <= 0
         ? doubles( ds )
         : doubles( (nlAoM() - nE*Math.log(scl)) * (nE-1.0) / nE, ds );
    }//normalised1()

   /** Return 'this' Vector <U>of Cts</U> {@link #normalised normalised}
       but <I>without</I> copying the (scaled) elements. */
   protected Vector normalised2()
    { final int nE = nElts();  // the dimension, D, of R^D space
      final boolean aomZero = AoM() <= 0;
      final double scl = 1.0 / norm();
      final Vector v = this;
      return this.new Derived()
       { public Value elt(int i)
          { return Value.real( v.elt(i).x()*scl ); }
         /** AoM of the normalised Vector as a whole. */
         public double   AoM()
          { return aomZero ? 0.0 : Math.exp( -nlAoM() ); }
         public double constAoM()
          { final double cAoM = v.constAoM();
            return cAoM >= 0 ? cAoM * scl : cAoM;
          }
         /** nlAoM of the normalised Vector as a whole. */
         public double nlAoM()
          { if( aomZero ) error( ".nlAoM(), but AoM() is 0, exact" );
            return (v.nlAoM() - nE * Math.log(scl)) * (nE-1.0) / nE;
          }//nlAoM()
    }; }//normalised2()

   /** The Vector <NOBR>{v, v, ..., v},</NOBR> n&nbsp;times. */
   public static Vector repeat(final Value v, final int n)       // repeat(v,n)
    { return new Vector()
       { public int      nElts() { return n; }
         public Value elt(int i) { return v; }
    }; }//repeat(v,n)

   /** Vector.Slice, a subVector of 'this' Vector; see {@link #slice slice}.
       It can be none, or all, of the {@link #parent() parent()} Vector.
       See {@link #slice slice}, {@link #take take}, {@link #drop drop}. */
   // To work around Apple, java 1.6 v. java 1.7 ...
   // public class Slice extends Derived  // Apple:-(
   public class Slice extends Vector                                   // Slice
    { /** The range [lo, hi), lo inclusive to hi <I>exclusive</I>, of
          the {@link #parent() parent()}'s elements.
          <NOBR> lo &le; hi. </NOBR> */
      public final int lo, hi;

      /** Note, [lo,&nbsp;hi), lo inclusive to hi <I>exclusive</I>. */
      public Slice(int lo, int hi)
       { if( lo < 0 || hi > Vector.this.nElts() )
            error( ".slice[" + lo + "," + hi + ") bounds" );
         this.lo = lo;
         this.hi = Math.max(lo, hi); // paranoia, see nElts()
       }//constructor

      /** {@link #hi hi} - {@link #lo lo}. */
      public int      nElts() { return hi - lo; /* NB. >= 0 */ }
      /** {@link #parent() parent()}.elt({@link #lo lo}+i). */
      public Value elt(int i) { return Vector.this.elt(lo+i);  }
      /** {@link #parent() parent()}.wt({@link #lo lo}+i). */
      public double wt(int i) { return Vector.this.wt( lo+i);  }

      // next slab should come from Derived...
      /** The Vector of which 'this' Slice is a Slice. */
      public Vector  parent() { return Vector.this; }
      public Type      type() { return Vector.this.type();     }
      public Type   eltType() { return Vector.this.eltType();  }
      public double constWt() { return Vector.this.constWt();  }
      public double constAoM(){ return Vector.this.constAoM(); }

      /** A Slice of a Slice is just a Slice,
          {@link #parent() parent()}.slice({@link #lo lo}+lo2,lo+hi2). */
      public Vector slice(int lo2, int hi2)
       { return Vector.this.slice(lo + lo2, lo + hi2); } // !

      /** 'this' appended to v2 but be efficient if they have
          the same {@link #parent() parent} and abut. */
      public Vector append(Vector v2)
       { if(    nElts() <= 0 ) return v2;
         if( v2.nElts() <= 0 ) return this;
         if( v2 instanceof Slice )
          { Slice v2s = (Slice) v2;
            if( Vector.this == v2s.parent() && hi == v2s.lo )
               return Vector.this.slice(lo, v2s.hi);
               // i.e., [lo,mid) ++ [mid,hi) = [lo,hi); this...
               // ...can happen where stats(ds) = ds itself, say.
          }//else...
         return super.append(v2);
       }//append(v2)

      /** Return 'this' Slice with elements p.[loP,hiP) deleted; it requires
          (and checks) that <NOBR>p == {@link #parent() parent()}.</NOBR>
          The elements being dropped must in fact be in 'this' Slice.
          <NOBR>Note, [loP, hiP).</NOBR> Also see {@link #addMore addMore}. */
      public Vector delete(Vector p, int loP, int hiP)
       { if( p != /* parent */ Vector.this )
            error( ".drop(p," + loP + "," + hiP + "), p:"
                        + p.getClass().getName() + " != parent():"
                 + parent().getClass().getName() );
         return drop(loP - lo, hiP - lo);
       }//drop(,,)

      /** Return 'this' Slice with elements p.[loP,hiP) prepended or
          appended as appropriate; it requires (and checks) that
          <NOBR>p == {@link #parent() parent()}.</NOBR>
          It is an error if 'this' and p overlap.
          <NOBR> Note, [loP, hiP). </NOBR>  */
      public Vector addMore(Vector p, int loP, int hiP)
       { if( p != /* parent */ Vector.this )
            error( ".drop(p,loP,hiP), p:" + p.getClass().getName()
                 + " != parent():" + parent().getClass().getName() );
         if( loP >= hi ) return this.append(p.slice(loP, hiP));
         if( hiP <= lo ) return p.slice(loP, hiP).append(this);
         throw RTE( ".addMore(p," + loP + "," + hiP + "), to ["
                                  + lo +  "," + hi +  ")" );
       }//addMore(,,)

    }//Slice class
   // -----------------------------------

   /** Return a {@link Slice Slice} (section), <NOBR>[lo, hi),</NOBR> of
       'this' Vector, that is, the elements between lo inclusive, and hi
       <I>exclusive</I>.  Note that a copy is <I>not</I> made.  Also see
       {@link #take take}, {@link #drop drop} and {@link Slice#slice}. */
   public Vector slice(final int lo, final int hi)              // slice[lo,hi)
    { return new Slice(lo, hi); }

   /** Return the first 'm' elements of 'this' Vector;
       see {@link #drop} and {@link #slice}. */
   public Vector take(int m)                                         // take(m)
    { return slice(0, m); }
    // NB. use of 'slice', not new Slice, the former can be overridden.

   /** Return the last 'm' elements of 'this' Vector; see {@link #slice}. */
   public Vector takeLast(int m)                                 // takeLast(m)
    { final int n = nElts();  return slice(n-m, n); }

   /** Return <I>all but</I> the first 'm' elements of 'this' Vector;
       see {@link #take} and {@link #slice}. */
   public Vector drop(int m)                                         // drop(m)
    { return slice(m, nElts()); }

   /** Drop elements [lo,hi) of 'this' Vector, lo&nbsp;inclusive,
       hi&nbsp;<I>exclusive</I>. Return take(lo).append(drop(hi)),
       e.g., <NOBR>[0,1,2,3,4].drop[1,3) &rarr; [0,3,4].</NOBR> */
   public Vector drop(int lo, int hi)                            // drop(lo,hi)
    { if( hi < lo ) error( ".drop[" + lo + "," + hi + ")" );
      return take(lo).append(drop(hi));
    }//drop(lo,hi)

   /** Return <I>all but</I> the last 'm' elements of 'this' Vector;
       see {@link #slice}. */
   public Vector dropLast(int m)                                 // dropLast(m)
    { return slice(0, nElts()-m); }

   /** Either {@link #addMore addMore} (add=true) Vector 'v' to this
       or {@link #delete delete} (add=false) v from this Vector. */
   public Vector combine(boolean add, Slice v)                // combine(add,v)
    { return add ? addMore(v) : delete(v); }

   /** {@link #delete(Vector,int,int) delete(s.parent(),s.lo,s.hi)}. */
   public Vector delete(Slice s)                                   // delete(s)
    { return delete(s.parent(), s.lo, s.hi); }

   /** Delete elements [lo,&nbsp;hi) of 'v' from 'this'; it requires
       (and checks) that <NOBR>v == this</NOBR>. The elements being
       deleted must in fact be in 'this' Vector.
       Also see {@link Slice#delete(Vector,int,int)}.
       <NOBR>delete(v,lo,hi) and</NOBR>
       {@link #addMore(Vector,int,int) addMore(v,lo,hi)} are used to
       create subsets of the elements of a Vector, e.g., sub-data-sets. */
   public Vector delete(Vector v, int lo, int hi)            // delete(v,lo,hi)
    { if( v != this ) error( ".delete(v,lo,hi), v!=this" );
      return drop(lo, hi);
    }//delete(v,lo,hi)

   /** {@link #addMore(Vector,int,int) addMore(s,0,s.nElts())}. */
   public Vector addMore(Slice s)                                 // addMore(s)
    { return addMore(s.parent(), s.lo, s.hi); }

   /** It is impossible, an error, to add more than zero elements
       (of 'v'==this) to 'this' Vector as they are all already there,
       but to a {@link Slice#addMore Slice} is possible.  Requires
       (and checks) that <NOBR>v == this</NOBR>.
       <NOBR>addMore(v,lo,hi) and</NOBR>
       {@link #delete(Vector,int,int) delete(v,lo,hi)} are used to
       create subsets of the elements of a Vector, e.g., sub-data-sets. */
   public Vector addMore(Vector v, int lo, int hi)          // addMore(v,lo,hi)
    { if( v != this ) error( ".addMore(v,lo,hi), v!=this" );
      if( hi == lo ) return this;  // adding none *is* OK, else...
      throw RTE( ".addMore(v," + lo + "," + hi + "), this is already all" );
    }//addMore(v,lo,hi)

   /** Append Vectors 'this' and 'v2', of the <I>same</I> type! */
   public Vector append(final Vector v2)                          // append(v2)
    { final Vector v1 = this;
      final int n1 = nElts(), n = n1 + v2.nElts();
      if( n1 <= 0 ) return v2; // NB. can happen if start with ...
      if( n <= n1 ) return v1; //     ... [] and add sections.
      return new Vector() // note const<Blah> overridden...
       { public Type    type() { return v1.   type(); }       // type()
         public Type eltType() { return v1.eltType(); }    // eltType()
         public int nElts() { return n; }                    // nElts()
         /** Return an element from 'this', or 'v2', as appropriate. */
         public Value elt(int i)                              // elt(i)
          { return i < n1 ? v1.elt(i) : v2.elt(i - n1); }
         /** Return a weight from 'this', or 'v2', as appropriate. */
         public double wt(int i)                               // wt(i)
          { return i < n1 ? v1.wt( i) : v2.wt( i - n1); }
         public double constWt()                           // constWt()
          { final double cW1 = v1.constWt();
            return cW1 == v2.constWt() ? cW1 : -1;
          }//constWt()
         public double constAoM()                         // constAoM()
          { final double A1 = v1.constAoM();
            return A1 == v2.constAoM() ? A1 : -1;
          }//constAoM()
    }; }//append(v2)

   /** For 'this' non-empty Vector (data-set?) of Vectors,
       {@link #append(Vector) append} them all, that is
       flatten k dimensions to (k-1) dimensions. */
   public Vector append()                                           // append()
    { final int n = nElts();
      if( n >= 2 )
       { final int half_n = n/2;  // note, half_n >= 1; divide and conquer
         return take(half_n).append() .append( drop(half_n).append() );
       }// else base case ...
      if( n == 1 ) return (Vector) elt(0);
      else return empty; //  throw RTE( ".append([])" );
    }//append()

   /** Unary operator, 'op', on 'this' Vector, that is apply {@link Lexical op}
       element-wise, returning a new Vector of results. */
   public Value uOp(int op)                                          // uOp(op)
    { final Value[] ans = new Value[nElts()];
      for( int i = 0; i < ans.length; i ++ ) ans[i] = elt(i).uOp(op);
      return values(ans);
    }//uOp(op)

   /** Binary operator, 'op', on 'this' Vector, and 'v', that is apply
       {@link Lexical op} element-wise returning a Vector of results. */
   public Value bOp(int op, Value v)                               // bOp(op,v)
    { if( ! Type.VECTOR.ni(v) ) error( ".bOp(v), v not a Vector" );
      Vector vv = (Vector) v.force(); // v may be Defer-ed; see Defer.bOp()
      if( vv.nElts() != nElts() )
         error( ".bOp(" + Lexical.int2sy(op) + "," + v + ") shapes" );
      final Value[] ans = new Value[nElts()];
      for( int i = 0; i < ans.length; i ++ )
         ans[i] = elt(i).bOp(op, vv.elt(i));
      return values(ans);
    }//bOp(op,v2)


   /** <TT>zip</TT> 'this' Vector with another Vector 'v',
       of the same length, to produce a Vector of pairs as in,
       <NOBR><TT>zip(new Vector [] {this, v}</TT></NOBR>.
       Also see {@link #zip(Vector[]) zip(cols)} and {@link #unzip}.  */
   public Vector zip(Vector v)                                        // zip(v)
    { return zip(new Vector [] {this, v}); }

   /** <TT>zip</T> several Vectors to make a Vector of Tuples.
       Elements are weighted as in <TT>cols[0]</TT>.
       Used by {@link la.la.Value.Tuple#zip}.
       Also see {@link #zip(Vector) zip(v)} and {@link #unzip() unzip}. */
   public static Vector zip(final Vector [] cols)                  // zip(cols)
    { final int N = cols[0].nElts(),  // length
                W = cols.length;      // width
      for( int c = 1; c < cols.length; c ++ )
         if( cols[c].nElts() != N )
            Util.error( "zip(cols), unequal lengths, column " + c + ": "
                      + cols[c].nElts() + "!=" + N );
      return new Vector()
       { public int nElts() { return N; }
         public Tuple elt(int r)
          { Value[] row = new Value[W];
            for( int c = 0; c < W; c ++ ) row[c] = cols[c].elt(r);
            return tuple(row);
          }//elt(r)
         /** Weighted like col<SUB>0</SUB>. */
         public double  wt(int i) { return cols[0].wt(i); }
         public double  constWt() { return cols[0].constWt();  }
         /** Just like col<SUB>c</I> of
             <NOBR>zip [..., col<SUB>c-1</SUB>, col<SUB>c</SUB>, ...]</NOBR>
             but weighted like 'this' (and col<SUB>0</SUB>). */
         public Vector col(int c) { return cols[c].weightLike(cols[0]); }
    }; }//zip()

   /** <TT>unzip</TT> 'this' Vector of <I>Tuples</I> into a Tuple
       of Vectors. Also see {@link #zip(Vector) zip(v)},
       {@link #zip(Vector[]) zip(cols)} and
       <TT>Tuple.{@link la.la.Value.Tuple#zip() zip()}</TT>. */
   public Tuple unzip()                                              // unzip()
    { if( nElts() <= 0 ) error( ".unzip(), zero elements" );
      Type eT = eltType();
      if( ! (eT instanceof Type.Tuple) )
        error( ".unzip(), not Vector of Tuple" );
      return unzip( ((Type.Tuple) eT) .nElts() ); //see below
    }//unzip()

   /** unzip 'this' Vector of Tuples each of (must be) nCol elements. This
       variation on {@link #unzip() unzip()} allows for 'this'=empty. */
   public Tuple unzip(int nCol)                                  // unzip(nCol)
    { Vector [] vs = new Vector [nCol];
      for( int i = 0; i < nCol; i ++ ) vs[i] = col(i); // does wt(i)!
      return Value.tuple(vs);
    }//unzip(nCol)

   /** Apply <I>Function</I> 'f' to each element of 'this' Vector,
       returning a Vector of results. */
   public Vector map(Function f)                                      // map(f)
    { Value [] ans = new Value [nElts()];
      for( int r = 0; r < ans.length; r ++ )
         ans[r] = new Value.Defer.App(f, elt(r)); // ?lazy
         // ans[r] = f.apply( elt(r) );              // ?strict
      return values(ans).weightLike(this);   // !
    }//map(f)

   /** foldl (fold-left, reduce) 'this' Vector with curried Function 'fc'
       and zero (identity) 'z', in effect <NOBR>((((z fc e<SUB>0</SUB>)
       fc e<SUB>1</SUB>) ...&nbsp;) fc e<SUB>n-1</SUB>)</NOBR>.
       Also see {@link #foldr foldr(fx,z)}. */
   public Value foldl(Function fc, Value z)                      // foldl(fc,z)
    { Value ans = z;
      for( int i = 0; i < nElts(); i ++ )
         ans = fc .apply(ans) .apply(elt(i));
      return ans;
    }//foldl(fc,z)

   /** {@link #drop drop}(1)
       .{@link #foldl(Function,Value) foldl}(fc,elt(0)) */
   public Value foldl(Function fc)                                 // foldl(fc)
    { return drop(1).foldl(fc, elt(0)); }

   /** foldr (fold-right, reduce) 'this' Vector with curried Function
       'fc' and zero (identity) 'z', in effect <NOBR>(e<SUB>0</SUB> fc (
       ...&nbsp;(e<SUB>n-2</SUB> fc (e<SUB>n-1</SUB> fc z))))</NOBR>.
       Also see {@link #foldl foldl(fc,z)}. */
   public Value foldr(Function fc, Value z)                      // foldr(fc,z)
    { Value ans = z;
      for( int i = nElts()-1; i >= 0; i -- )
         ans = fc .apply(elt(i)) .apply(ans);
      return ans;
    }//foldr(fc,z)

   /** {@link #take take}(n_1)
       .{@link #foldr(Function,Value) foldr}(fc,elt(n_1)),
       where <NOBR>n_1 = nElts() - 1.</NOBR>  */
   public Value foldr(Function fc)                                 // foldr(fc)
    { final int n_1 = nElts()-1; return take(n_1).foldr(fc, elt(n_1)); }

   /** For a Vector of Vector of Vec..., return the "nElts()" of the leading
       (hyper-)rectangular dimensions, e.g., a simple Vector would give [m],
       an m&times;n {@link Matrix Matrix} [m,n], etc., and
       a jagged Vector of Vectors&nbsp;[m]. */
   public int[] shape()                                              // shape()
    { return shape(this, 0); }//see shape(v,d) below

   /** The (recursive) slave of {@link #shape() shape()}. */
   private static int[] shape(Vector v, int d)
    { final int[] s =
         v.nElts() > 0 && v.elt(0) instanceof Vector && v.isRectangular()
         ? shape((Vector) v.elt(0), d+1)
         : new int[d+1];
      s[d] = v.nElts();
      return s;
    }//shape(v,d)
   // .........................................................................

   /** <I>Provided</I> 'this' is a Vector of {@link la.la.Value.Structured},
       return {@link #nElts() nElts()} of <TT>elt(row)</TT>.
       Also see {@link #nCols()}.  */
   public int nElts(int row)                                      // nElts(row)
    { return ((Value.Structured) elt(row)) .nElts(); }

   /** <I>Provided</I> 'this' is a Vector of {@link la.la.Value.Structured},
       return the element at row 'r', <NOBR>column 'c'.</NOBR> */
   public Value elt(int r, int c)                                   // elt(r,c)
    { return ((Value.Structured) elt(r)). elt(c); }

   /** <I>Provided</I> 'this' is a Vector of {@link la.la.Value.Structured},
       return the c-th column, as a Vector. E.g., col(c) is used by
       {@link mml.Independent} to select a column of a multivariate data-set.
       Also see {@link #cols(int[]) cols(cs)}, columns plural.  */
   public Vector col(final int c)                                     // col(c)
    { return this.new Derived()
       { public Type    type() { return super.type(); }
         /** The element Type. */
         public Type eltType() { return super.eltType(); }
         /** The i-th element of 'this' column. */
         public Value elt(int i) { return Vector.this.elt(i,c); }
         // wt(i) & wts() from parent Vector, but...
         /** This column has const AoM iff parent does... */
         public double constAoM()
          { return Vector.this.constAoM() >= 0 ? elt(0).AoM() : -1; //variable
          }//constAoM()
    }; }//col(c)

   /** <I>Provided</I> 'this' is a Vector of {@link la.la.Value.Tuple}, return
       the Vector of Tuples made up of the columns specified by cs. Also
       see {@link #col(int) col(c)} and {@link #elts(int[]) elts(ps)}. */
   public Vector cols(final int[] cs)                               // cols(cs)
    { Vector [] vs = new Vector [cs.length];
      for( int i = 0; i < vs.length; i ++ ) // works even if this is empty
         vs[i] = col(cs[i]);
      return Value.tuple(vs).zip();
    }//cols(cs)
// { return unzip().elts(cs).zip(); } // (!-)  fails if this is empty

   /** <I>Provided</I> 'this' is a rectangular Vector of
       {@link la.la.Value.Structured Structured}, which property <I>is</I>
       checked within, return its number of columns, otherwise throw an error.
       Also see {@link Matrix#nCols()} and {@link Matrix#nElts(int)}. */
   public int nCols()                                                // nCols()
    { if( nElts() == 0 ) return 0; // ?, empty, else ...
      Value e0 = elt(0);
      final Type  et = e0.type();
      final int   nC = ((Structured) e0) .nElts();
      for( int i = 1; i < nElts(); i ++ )
       { Value ei = elt(i);
         if( ! et.ni(ei) ) error( ", type of row " + i );
         else if( nC != ((Structured) ei) .nElts() )
            error( ", not rectangular, at row " + i + " of " + nElts() );
       }
      return nC;
    }//nCols()
   // .........................................................................
                                                     // 2D Matrix Operations...

   /** Provided 'this' is a <I>rectangular</I> Vector of Vectors, return it
       as the equivalent {@link Matrix}; also see {@link Matrix#asMatrix()}.
       (Just a "type" change and no significant extra space is consumed.)  */
   public Matrix asMatrix()                                       // asMatrix()
    { if( ! isRectangular() ) error( " is not Rectangular" );
      return new Matrix(nCols())
       { public int             nElts() { return Vector.this.nElts(); }
         public Value elt(int r, int c) { return Vector.this.elt(r, c); }
         public Vector       elt(int r) { return (Vector) Vector.this.elt(r); }
         public double        wt(int r) { return Vector.this.wt(r); }
         public double        constWt() { return Vector.this.constWt(); }
         public double       constAoM() { return Vector.this.constAoM(); }
    }; }//asMatrix()

   /** Make a "row-Vector", i.e., a 1&times;n-{@link Matrix Matrix},
       of 'this' Vector (regardless of the element type).
       Note that wt(0), for <I>the</I> single row, is 1.0.  */
   public Matrix as_1xn()                                           // as_1xn()
    { return new Matrix(nElts())
       { public int nElts() { return 1; }
         public Value elt(int r, int c)
          { if( r != 0 ) error( ".elt(" + r + "," + c + ")" );
            /*else*/ return Vector.this.elt(c);
          }//elt(r,c)
         public Vector elt(int r)
          { if( r != 0 ) error(".elt(" + r + ")" ); return Vector.this; }
         public double wt(int r)
          { if( r != 0 ) error( ".wt(" + r + ")" ); return super.wt(r); }
    }; }//as_1xn()

   /** Make a "column Vector", i.e., a n&times;1-{@link Matrix Matrix},
       of 'this' Vector (regardless of the element type).  */
   public Matrix as_nx1()                                           // as_nx1()
    { return new Matrix(1)
       { public int nElts() { return Vector.this.nElts(); }
         public Value elt(int r, int c)
          { if( c != 0 ) error( ".elt(" + r + "," + c + ")" );
            /*else*/ return Vector.this.elt(r);
          }
         public double    wt(int r) { return Vector.this.wt(r); }
         public double    constWt() { return Vector.this.constWt(); }
         public double   constAoM() { return Vector.this.constAoM(); }
         public Vector col(int c)
          { if( c != 0 ) error( ".col(" + c + ")" ); return Vector.this; }
     }; }//as_nx1()

   /** Make a square, diagonal {@link Matrix Matrix} of 'this' Vector,
       using 'z' for all off-diagonal elements. */
   public Matrix asDiagonal(final Value z)                     // asDiagonal(z)
    { return new Matrix(nElts())
       { public int nElts() { return Vector.this.nElts(); }
         public Value elt(int r, int c)
          { return r==c ? Vector.this.elt(r) : z; }
         public double wt(int i) { return Vector.this.wt(i);     }
         public double constWt() { return Vector.this.constWt(); }
    }; }//asDiagonal(z)

                                                           // 2D predicates ...

   /** <I>Provided</I> 'this' is a Vector of <I>Vectors</I>, is it
       rectangular?  That is, is it non-"jagged"? Also see
       {@link #isSquare()}, and {@link Matrix}. */
   public boolean isRectangular()                            // isRectangular()
    { return rowLength() >= 0; }

   /** <I>Provided</I> this is a Vector of Vectors, if it is
       rectangular return its row length, else a negative int. */
   public int rowLength()                                        // rowLength()
    { final int nRows = nElts();
      if( nRows == 0 ) return 0; // ?
      final int nCols = ((Vector) elt(0)) .nElts();
      for( int r = 1; r < nRows; r ++ )
         if( ((Vector) elt(r)) .nElts() != nCols ) return -1; // jagged
      return nCols;
    }//isRectangular() ?

   /** <I>Provided</I> 'this' is a Vector of Vectors, is it square?
       Also see {@link #isRectangular()}.  */
   public boolean isSquare()                                      // isSquare()
    { final int nRows = nElts();
      if( nRows == 0 ) return true; // ?
      return nRows == rowLength();
    }//isSquare() ?

   /** <I>Provided</I> 'this' is a Vector of Vectors of Cts, is it square
       &amp;&amp; (nearly-)symmetric? That is, are x[r][c] and x[c][r]
       {@link Maths#nearlyEqual(double, double, double) nearlyEqual}
       for all r and&nbsp;c?  */
   public boolean nearlySymmetric(double epsilon)         // nearlySymmetric(e)
    { final int nRows = nElts(0);
      if( nRows == 0 ) return true; // ?, else ...
      for( int r = 0; r < nRows; r ++ )
       { if( nElts(r) != nRows ) return false; // not Square
         for( int c = r+1; c < nRows; c ++ )
          { final double xrc = elt(r,c).x(),
                         xcr = elt(c,r).x();
            if( ! nearlyEqual(xrc, xcr, epsilon) ) return false;
       }  }
      return true;
    }//nearlySymmetric(e) ?

   /** <I>Provided</I> 'this' is a Vector of Vectors of Cts, return
   {@link #nearlySymmetric(double) nearlySymmetric}({@link Maths#epsilon}). */
   public boolean nearlySymmetric()                        // nearlySymmetric()
    { return nearlySymmetric(epsilon); }
   // -------------------------------------------------------------------------

   /** A simple, general purpose (GP) implementation of
       class {@link Vector}. Also see {@link Matrix.GP2}.  */
   public static class GP extends Vector                           // Vector.GP
    { /** The array of Values held in 'this' GP. */
      protected final Value [] elts;
      private Type t0 = null;  // cached eltType
      /** Given an array [] of Values, create a Vector. */
      public GP(Value [] elts)
       { this.elts = elts;
         // if( elts.length > 0 )
          // { Type t0 = elts[0].type();
            // for( int i = 1; i < elts.length; i ++ )
               // if( elts[i].type() != t0 )  // ?ok or too strict ???
                  // error(", " + t0 + "!=" + elts[i].type());
          // }//if
       }//constructor
      public int     nElts() { return elts.length; }
      public Value [] elts() { return elts; }
      /** Return the i-th {@link #elts element} ({@link #force forced}). */
      public Value elt(int i)
       { Value ei = elts[i].force();
         elts[i] = ei;
         if( t0 == null ) t0 = ei.type();
         if( ! t0.contains( ei.type() ) )
            error( ".elt(" + i + ")=" + ei + " :" + ei.type() + " != " + t0 );
         return ei;
       }//elt(i)
    }//Vector.GP class
   // .........................................................................

   /** Convenience function, <TT>values : Value[] &rarr; {@link GP GP}</TT>.
       Also see the 2D {@link Matrix#values(Value[][])}.  */
   public static GP values(Value [] vs) { return new GP(vs); }

   /** A Vector containing {@link la.la.Value.Int Int} Values can
       probably arrange efficient storage and operations. */
   public static abstract class Ints extends Vector                   // * Ints
    { public Type.Vector type() { return Type.VECTOR_INT; }
      public Type     eltType() { return type().eltType; }
      public double  constAoM() { return 0; }
    }//Ints abstract class

   /** Convenience function, <TT>ints : int[] &rarr; Vector</TT>.
       Also see the 2D {@link Matrix#ints(int[][])}. */
   public static Vector ints(final int [] ns)
    { return new Ints()
       { public int      nElts() { return ns.length; }
         public Value elt(int i) { return new Value.Int(ns[i]); }
         public int     n(int i) { return ns[i]; }
    }; }//ints(ns)

   /** Convenience function to make a Vector with elements of
       {@link la.la.Type.Discrete Discrete&nbsp;Type} et, and
       integer codes, ns. */
   public static Vector ints(final Type.Discrete et, final int [] ns)
    { final Type t = new Type.Vector(et);
      return new Vector()
       { public Type      type() { return  t; }
         public Type   eltType() { return et; }
         public int      nElts() { return ns.length; }
         public Value elt(int i) { return et.int2value(ns[i]); }
         public int     n(int i) { return ns[i]; }
    }; }//ints(et,ns)


   /** A Vector containing doubles ({@link la.la.Value.Cts Cts} Values) can
       probably arrange efficient storage and operations.
       Also see the function(s) {@link Vector#doubles}. */
   public static abstract class Doubles extends Vector             // * Doubles
    { public Type.Vector type() { return Type.VECTOR_CTS; }
      public Type     eltType() { return type().eltType; }
    }//Doubles abstract class

   /** Convenience function <TT> : nlAoM &times; double[] &rarr; Vector</TT>,
       where we know the nlAoM of the Vector <I>as a whole</I>, but <I>not</I>
       of each elt(i). This may well be the case in estimating a
       {@link mml.Model}'s (Vector of) parameters, say. (And do not confuse
       it with {@link #doubles(double[],double)}. */
   public static Doubles doubles(final double nlAoM, final double[] xs)
    { return new Doubles()
       { public int      nElts() { return xs.length; }
         /** The AoM of the Vector <I>as a whole</I>. */
         public double     AoM() { return Math.exp( - nlAoM ); }
         /** The negative log AoM of the Vector <I>as a whole</I>.
             AoM is not known for an individual element.  */
         public double   nlAoM() { return nlAoM; }
         /** As per {@link #nlAoM() nlAoM()}. */
         public double wtNlAoM() { return nlAoM; }
         /** Returns an exact Real(!); also see {@link #nlAoM()}. */
         public Value.Cts elt(int i) { return Value.real(xs[i]); }
// Experiment below.   Not completely justifiable.
// public Value.Cts elt(int i)
//  { return Value.cts(xs[i], Math.exp(-nlAoM/xs.length)); }
         public double      x(int i) { return xs[i]; }
         public String closes()
          { return super.closes() + "{" + Util.x2String(nlAoM) + "e}"; }
    }; }//doubles(nlAoM,xs)

   /** Convenience function <TT> : double[] &times; AoM &rarr; Vector</TT>,
       where every element of the Vector has the <I>same</I>
       {@link Value#AoM AoM} (do not confuse it with
       {@link #doubles(double,double[])}.  Also see the 2D
       {@link Matrix#doubles(double[][],double)}. */
   public static Doubles doubles(final double[] xs, final double AoM)
    { return new Doubles()
       { public int       nElts() { return xs.length; }
         public Value  elt(int i) { return Value.cts(xs[i], AoM); }
         public double   x(int i) { return xs[i]; }
         public double constAoM() { return AoM; } // for all elts
    }; }//doubles(xs,AoM)

   /** Convenience function <TT> : double[] &rarr; Vector</TT>; note AoM=0.
       Also see the 2-D {@link Matrix#doubles(double[][])}. */
   public static Doubles doubles(final double[] xs)
    { return new Doubles()
       { public int       nElts() { return xs.length; }
         public double      AoM() { return 0; }
         public Value  elt(int i) { return Value.real(xs[i]); }
         public double   x(int i) { return xs[i]; }
         public double constAoM() { return 0; }
    }; }//doubles(xs)


   /** A Vector containing Strings, that is, having
       {@link la.la.Value.Chars Chars} elements. */
   public static abstract class Strings extends Vector             // * Strings
    { public Type.Vector type() { return Type.VECTOR_CHARS; }
      public Type     eltType() { return type().eltType; }
    }//Strings abstract class

   /** Convenience function <TT> : String[] &rarr; Vector</TT>. */
   public static Strings strings(final String [] ss)
    { return new Strings()
       { public int            nElts() { return ss.length; }
         public Value.Chars elt(int i) { return new Value.Chars(ss[i]); }
    }; }//charss(ss)
   // -------------------------------------------------------------------------

   /** Given an angle, 't', in radians, and an axis, 'n', rotate 'this'
       <I>3D</I> Vector by&nbsp;t about&nbsp;n.
       Calls {@link #rotate(Q) rotate(q)}. */
   public Vector rotate(double t, Vector n)
    { if( n.nElts() != 3 ) // NB. rotate(Q) checks 'this' similarly
         error( ".rotate(t,n), n.nElts()=" + n.nElts() + " != 3" );
      final double nn = n.norm(), t2 = t/2;
      final double sin_t2_nn = Math.sin(t2)/nn, cos_t2 = Math.cos(t2);
      return rotate( Q.doubles(cos_t2, sin_t2_nn * n.elt(0).x(),
                                       sin_t2_nn * n.elt(1).x(),
                                       sin_t2_nn * n.elt(2).x()) );
    }//rotate(t,n)

   /** As specified by the {@link Q quaternion}, 'q', rotate 'this'
       <I>3D</I> Vector by an angle about an axis. (Is called by
       {@link #rotate(double,Vector) rotate(t,n)}.)  */
   public Vector rotate(Q q)
    { if( nElts() != 3 ) error( ".rotate(q), nElts()=" + nElts() + " != 3" );
      // prefer RH rule...
      return q.times(this.asQrotn()) .times(q.inverse()) .asV3(); // RH rule?
   // return q.inverse().times(this.asQrotn()) .times(q) .asV3(); // LH rule?
    }//rotate(q)

   /** Make 'this' <I>3D</I> Vector of Cts, <NOBR>(x, y, z),</NOBR> a
       {@link Q quaternion} <NOBR>(0.0, x, y, z).</NOBR>  Note, asQrotn() is
       used in {@link #rotate(Q) rotate(q)}.  Also see {@link #asQ()}.  */
   public Q asQrotn()
    { if( nElts() != 3 ) error( ".asQrotn(), nElts()=" + nElts() + " != 3" );
      return new Q()
       { public Value elt(int i) { return i==0 ? zeroR : Vector.this.elt(i-1); }
    }; }//asQrotn()

   /** Make 'this' <I>4D</I> Vector of Cts a {@link Q quaternion}
       in the obvious way. */
   public Q asQ()
    { if( nElts() != 4 ) error( ", nElts()=" + nElts() );
      return new Q()
       { public Value elt(int i) { return Vector.this.elt(i); }
    }; }//asQ()
   // .........................................................................

   /** Find the rotation (as a {@link Q quaternion}), by some angle about
       some axis through the origin, that minimises the sum of squared errors
       between the Vectors of <I>3D</I> Vectors of Cts, 'this' and&nbsp;'t'.
       Also see
       [<A HREF="http://www.allisons.org/ll/AlgDS/Numerical/Superposition/"
           TARGET="LL">www</A>].  */
   public Q match(Vector t)                                         // match(t)
    { final Vector s = this;  // s and t are Vectors of 3D Vectors
      if( s.nElts() != t.nElts() ) error( ", != lengths" );
      if( s.nCols() != 3 || t.nCols() != 3 ) error( ", columns" );
      double xm2 = 0, xp2 = 0, ym2 = 0, yp2 = 0, zm2 = 0, zp2 = 0;
      double xmym = 0, xmyp = 0,  xmzm = 0, xmzp = 0,
             xpym = 0, xpyp = 0,  xpzm = 0, xpzp = 0,
             ymzm = 0, ymzp = 0,
             ypzm = 0, ypzp = 0;
      for( int i = 0; i < s.nElts(); i ++ )  // get the various sums
       { Vector si = (Vector) s.elt(i), ti = (Vector) t.elt(i);
         final double s0 = si.elt(0).x(), t0 = ti.elt(0).x(),
                      s1 = si.elt(1).x(), t1 = ti.elt(1).x(),
                      s2 = si.elt(2).x(), t2 = ti.elt(2).x();
         final double xm = s0 - t0, xp = s0 + t0,
                      ym = s1 - t1, yp = s1 + t1,
                      zm = s2 - t2, zp = s2 + t2;
         xm2 += xm*xm; xp2 += xp*xp;
         ym2 += ym*ym; yp2 += yp*yp;
         zm2 += zm*zm; zp2 += zp*zp;
         xmym += xm*ym; xmyp += xm*yp; xmzm += xm*zm; xmzp += xm*zp;
         xpym += xp*ym; xpyp += xp*yp; xpzm += xp*zm; xpzp += xp*zp;
         ymzm += ym*zm; ymzp += ym*zp;
         ypzm += yp*zm; ypzp += yp*zp;
       }//for i
      Matrix M = Matrix.doubles( new double[][]{
         { xm2+ym2+zm2,   ypzm-ymzp,   xmzp-xpzm,   xpym-xmyp },
         {   ypzm-ymzp, xm2+yp2+zp2,   xmym-xpyp,   xmzm-xpzp },
         {   xmzp-xpzm,   xmym-xpyp, xp2+ym2+zp2,   ymzm-ypzp },
         {   xpym-xmyp,   xmzm-xpzp,   ymzm-ypzp, xp2+yp2+zm2 } } );
      Tuple val_vec = M.eigen();  // Eigen-values and -vectors
      Vector e_val = (Vector) val_vec.fst();
      Matrix e_vec = (Matrix) val_vec.snd();
      // print( M2string(M,'\n','=M') );
      // print( 'E-vals=' + V2string(e_val) );
      // print( M2string(e_vec,'\n','=E-vecs') );
      int small = 0;
      for( int i = 1; i < e_val.nElts(); i ++ )      // find smallest...
         if( e_val.elt(i).x() < e_val.elt(small).x() ) // ...Eigen value
            small = i;
      // System.out.print( "[lambda=" + e_val.elt(small).x() + "]" );
      return e_vec.elt(small).asQ(); // corr. to smallest Eigen-val
    }//match(s,t)
   // -------------------------------------------------------------------------

   /** Input a data-set of "comma-"separated values from a given InputStream
       (often a {@link java.io.FileInputStream FileInputStream}). Tuple Type
       'rowType' gives the Type of one row, that is each Variable (column).
       Presently, allowed Variable Types are
       {@link la.la.Type.Discrete Discrete} (Int, Bool, Enum, ...),
       {@link la.la.Type.Cts Cts} and
       {@link la.la.Type#CHARS CHARS} (String).
       The separator is often&nbsp;',' but can be (almost) any character.
       Leading and/or trailing white space can be trimmed from values.
       Zero or more leading lines can be skipped, e.g., a title or a comment.
       Values, and rows, may <I>not</I> be split across input lines.
       Makes use of {@link la.util.Series.Lines} and
       {@link la.util.Series.Separator}. */
   public static Vector csv(boolean trimHead, boolean trimTail,     // csv(...)
                            char separator,
                            Type.Tuple rowType,
                            int skipLines,
                            InputStream inp )
    { final long pORm  = (1L << Lexical.plus) | (1L << Lexical.minus);
      int size = 1000,  // default starting size of Vector
         nCols = rowType.nElts();  // width of Vector
      int    [] [] ns = new int    [nCols] [0];  // note, by columns
      double [] [] xs = new double [nCols] [0];  // ...
      double []   aom = new double [nCols];      // ... AoMs for xs[]
      String [] [] ss = new String [nCols] [0];  // Strings (Chars)

      int r, c;  // row r, column c
      for( c = 0; c < nCols; c ++ )
       { aom[c] = Double.MAX_VALUE;
         final Type tc = (Type)rowType.elt(c);  // allow...
         if( tc instanceof Type.Discrete )      // Discrete (Int, Enum), or
            ns[c] = new int[size];
         else if( tc instanceof Type.Cts )      // Cts
            xs[c] = new double[size];
         else if( tc == Type.CHARS )            // Chars (String)
            ss[c] = new String[size];
         else throw Util.RTE( "csv(...,rowtype[" + c + "]=" + tc + ",...)" );
       }//for c

      Series.Lines lines = new Series.Lines (inp);
      lines.advance(skipLines);  // e.g., maybe over a title
      for( r = 0; lines.hasSome(); r ++ )               // for each row (line)
       { if( r >= size )  // need more space (rows)
          { final int twoSize = size * 2;   // Note, csv's time ...
            for( c = 0; c < nCols; c ++ )   // ... complexity is O(N).
             { if( xs[c].length > 0 )       // Cts
                { double [] x2 = new double[twoSize];
                  for( int s = 0; s < size; s ++ ) x2[s] = xs[c][s];
                  xs[c] = x2;
                }
               else if( ns[c].length > 0 )  // Discrete
                { int [] n2 = new int[twoSize];
                  for( int s = 0; s < size; s ++ ) n2[s] = ns[c][s];
                  ns[c] = n2;
                }
               else if( ss[c].length > 0 )  // Chars (String)
                { String [] ss2 = new String[twoSize];
                  for( int s = 0; s < size; s ++ ) ss2[s] = ss[c][s];
                  ss[c] = ss2;
             }  }
            size = twoSize;
          }//if

         Series.Separator cols =
            new Series.Separator(trimHead, trimTail, separator,
                                 ((Chars) lines.elt()).s );
         for( c = 0; cols.hasSome(); c ++ )                 // for each column
          { if( c > nCols )
               Util.error( "nCols=" + nCols + ", c=" + c + ", row r=" + r );
            final Type tc = (Type)rowType.elt(c);
            String s = ((Chars) cols.elt()).s;

            if( tc instanceof Type.Discrete )         // Discrete (inc. Int)
             { if( tc instanceof Type.Enum )          // inc. Enum
                { if( s.length() > 2 && s.charAt(0)=='"'
                                     && s.charAt(s.length()-1) == '"' )
                  // special dispensation to Enums, e.g., "true" -> true ...
                     s = s.substring(1, s.length()-1);
                }
               ns[c][r] = ((Type.Discrete) tc).string2n(s);
             }//Discrete
            else if( tc instanceof Type.Cts ) // Cts
             { // Note, a Cts has both the value itself and also the AoM
               Lexical lex = Lexical.forString(s);
               int sy = lex.sy();
               int sign = sy == Lexical.minus ? -1 : 1;
               if( (1L << sy & pORm) > 0 ) lex.insymbol();  // [+|-]<?>
               if( lex.sy() == Lexical.numeralR )           // e.g., 1.23e-4
                { xs[c][r] = sign * lex.theDbl();
                  aom[c] = Math.min(aom[c], lex.theAcc());
                }
               else if( lex.sy() == Lexical.numeral )       // e.g., 123
                { xs[c][r] = sign * lex.theInt();
                  aom[c] = Math.min(aom[c], 1.0);
                }
               else lex.error( "bad cts value in csv(...)" ); // shows where
             }//Cts
            else if( tc == Type.CHARS )
             { if( s.length() > 2 && s.charAt(0)=='"'
                                  && s.charAt(s.length()-1) == '"' )
                  s = s.substring(1, s.length()-1); // "abc" -> abc
               ss[c][r] = s;
             }//Chars (String)
         // no other possibility

            cols.advance();
          }//for c

         if( c != nCols )
            Util.error( "nCols=" + nCols + ", row r=" + r );
         lines.advance();
       }//for r

      if( r < size )  // get rid of any excess length
       { size = r;
         for( c = 0; c < nCols; c ++ )
          { if( xs[c].length > 0 )         // Cts
             { double [] x2 = new double[size];
               for( r = 0; r < size; r ++ ) x2[r] = xs[c][r];
               xs[c] = x2;
             }
            else if( ns[c].length > 0 )    // Discrete
             { int [] n2 = new int[size];
               for( r = 0; r < size; r ++ ) n2[r] = ns[c][r];
               ns[c] = n2;
             }
            else if( ss[c].length > 0 )    // Chars (String)
             { String [] ss2 = new String[size];
               for( r = 0; r < size; r ++ ) ss2[r] = ss[c][r];
               ss[c] = ss2;
       }  }  }

      Vector [] cols = new Vector [nCols];
      for( c = 0; c < nCols; c ++ )
       { final Type tc = (Type)rowType.elt(c);
         if( tc instanceof Type.Discrete )    // Discrete
            cols[c] = Vector.ints((Type.Discrete) tc, ns[c]);
         else if( tc instanceof Type.Cts )    // CTS
            cols[c] = Vector.doubles(xs[c], aom[c]);
         else if( tc == Type.CHARS )          // Chars (String)
            cols[c] = Vector.strings(ss[c]);
      // no other possibility
       }//for c

      return Vector.zip(cols);
    }//csv(...)

   /** {@link #csv csv(...)} using some common default parameter values.
       'inp'&nbsp;may be a {@link java.io.FileInputStream FileInputStream}. */
   public static Vector csv(Type.Tuple rowType, InputStream inp)
    { return csv(true, true, ',', rowType, 0, inp); }
   // -------------------------------------------------------------------------

   private static String show(Value v) { return v + ":" + v.type(); }

   /** main() allows Vector to be (slightly) tested, in isolation. */
   public static void  main(String[] argv)
    { System.out.println("-- test Vector --");

InputStream ip = null;
Type.Tuple.GP ICB =
   new Type.Tuple.GP("ICB", new Type[]{Type.INT, Type.CTS, Type.BOOL, Type.CHARS});
try{ ip = new java.io.FileInputStream("z.csv");
} catch(Exception e) { System.out.println( "doh " + e ); }
Vector ds = csv( ICB, ip );
System.out.println( ds + " (" + ds.nElts() + ") " );
Vector u = Vector.values(ds.uniqueElts());
System.out.println(  u + " (" +  u.nElts() + ") " );

      System.out.println("   see Test.main(.)");
      System.out.println("-- done Vector --");
    }//main

 }//Vector class   http://www.allisons.org/ll/

// (c) Lloyd Allison 2007, 2017, 11 Feb 2020. Licence to 'Vector' is Gnu GPL.
