//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package la.util;

import java.util.Comparator;

import la.la.Type;
import la.la.Value;
import la.maths.Vector;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** Series, of {@link la.la.Value Values}, is like the standard Java
    Enumeration and Iterator but with simpler semantics. The most
    important methods are {@link #elt()}, {@link #hasSome()} and
    {@link #advance()}. A Series has state, and is eventually used up.
    Also see {@link la.la.Value.Scannable} &ndash; something that
    implements Scannable can produce a Series, <I>repeatedly</I>.  */
public abstract class Series                                    // class Series
 {
   /** The EMPTY Series always {@link #hasNone() hasNone()}. */
   public static Series EMPTY = new Series()                    // Series.EMPTY
    { public boolean hasSome() { return false; }
      public int    position() { return 0; }
      public Value       elt() { throw Util.RTE( this + ".elt()" ); }
      public void    advance() { error( ".advance()" ); }
      public String toString() { return "Series.EMPTY"; }
    };//EMPTY
   // -------------------------------------------------------------------------

   /** Is there at least a current element, ({@link #elt()})?
       Also see {@link #hasNone()}. */
   public abstract boolean hasSome();                         // * hasSome()

   /** For convenience -- return not&nbsp;{@link #hasSome() hasSome()}. */
   public boolean hasNone()                                     // hasNone()
    { return ! hasSome(); }

   /** What position in the Series are we at (starting at zero)?
       If {@link #hasNone()}, the series was finite and position()
       is now its length. */
   public abstract int position();                           // * position()

   /** Return the current element, if there is one, but do
       <I>not</I> {@link #advance()}. Should throw an Exception
       if {@link #hasNone() hasNone()}. */
   public abstract Value elt();                                   // * elt()

   /** Return the weight of the current element; this default
       gives&nbsp;1.0.  Also see {@link Vector#wt(int)}.  */
   public double wt() { return 1.0; }                                // wt()

   /** Advance past the current element -- there might or might not
       be any more; also see {@link #hasSome()}. */
   public abstract void advance();                            // * advance()

   /** Advance past 'n' elements (an error if impossible). */
   public void advance(int n)                                  // advance(n)
    { if( n < 0 ) error( ".advance(" + n + ")" );
      for( int i = 0; i < n; i ++ )
       { if( hasNone() ) error( ".advance(" + n + "), i=" + i );
         advance();
       }
    }//advance(n)

   /** Advance to position 'n' (an error if impossible). */
   public void advanceTo(int n)                              // advanceTo(n)
    { advance(n-position()); }

   /** {@link #advance() Advance} until {@link #hasNone() hasNone},
       presumably for side-effects. */
   public void exhaust()                                        // exhaust()
    { while( hasSome() ) advance(); }

   /** Return a {@link la.la.Value.List List} of elements in 'this' Series;
       <I>beware, do not</I> share the Series because a Series has
       side-effects but a List does not. */
   public Value.List asList()                                    // asList()
    { if( hasSome() )
       { Value h = elt(); advance();
         return new Value.List.Cell(
            h,
            new Value.Defer()
             { public Value.List force()
                { if( v == null ) v = Series.this.asList(); // v, cache
                  return (Value.List) v;
             }  } );
       }
      else /* hasNone() */ return Value.List.NIL;
    }//asList()

   /** Return a short String description of 'this' Series,
       say for use in debugging. */
   public String toString()                                    // toString()
    { return this.getClass().getName(); }
   // ...................................

   /** Series producing Int Values. */
   public static abstract class Int extends Discrete            // * Series.Int
    { public Int() { super(Type.INT); }
      public Series.Int merge(Series s2) { return merge(true, s2); }
      public Series.Int merge(boolean keepShared, Series s2)
       { return super.merge(keepShared, s2).asInt(); }
    }//Series.Int class


   /** Series producing Discrete Values. */
   public static abstract class Discrete extends Series    // * Series.Discrete
    { private Type.Discrete et;
      public Discrete(Type t)
       { this.et = (Type.Discrete) t; }

      /** The current element's int "code". */
      public abstract int          elt_n();                // *

      /** See {@link #elt_n() elt_n()}. */
      public Value.Discrete elt() { return et.int2value(elt_n()); }

      public Series.Discrete merge(Series s2) { return merge(true, s2); }

      public Series.Discrete merge(final boolean keepShared, Series s)
       { final Series.Discrete s1 = this, s2 = (Series.Discrete) s;
         return new Series.Discrete(this.et)
          { private int pos = 0, n1 = 666, n2 = 666;
            private boolean known = false;
            private void makeKnown()
             { if( known ) return;
               n1 = s1.elt_n(); n2 = s2.elt_n(); known = true;
             }//makeKnown()
            public boolean hasSome() { return s1.hasSome() || s2.hasSome(); }
            public int    position() { return pos; }
            public int elt_n()
             { if( s1.hasNone() ) return s2.elt_n();
               if( s2.hasNone() ) return s1.elt_n(); // else...
               makeKnown();
               return n1 <= n2 ? n1 : n2;
             }//elt_n()
            public void advance()
             { if(      s1.hasNone() ) s2.advance();
               else if( s2.hasNone() ) s1.advance();
               else
                { makeKnown();
                  if( n1 <= n2 )
                   { s1.advance();
                     if( (n1 == n2) && ! keepShared ) s2.advance();
                   }
                  else s2.advance();
                }
               known = false;
               pos ++ ;
             }//advance()
          };//new Series.Int()
       }//merge(k,s)

      /** Treat 'this' Series.Discrete as a {@link Series.Int}. */
      public Series.Int asInt()
       { return new Series.Int()
          { public boolean hasSome() { return Discrete.this.hasSome();  }
            public int       elt_n() { return Discrete.this.elt_n();    }
            public int    position() { return Discrete.this.position(); }
            public void    advance() { Discrete.this.advance(); }
       }; }

    }//Series.Discrete class

   public static Series.Int fromInts(final int [] ns)      // fromInts(ns)
    { return new Series.Int()
       { private int pos = 0;
         public int    position() { return pos; }
         public boolean hasSome() { return pos < ns.length; }
         public int       elt_n() { return ns[pos]; }
         public void    advance() { pos ++ ; }
         public String toString() { return "fromInts(ns[" + ns.length + "])"; }
    }; }//fromInts(ns)
   // ...................................

   public static Series fromVector(final Vector v)           // fromVector()
    { return new Series()
       { private int pos = 0;
         public int    position() { return pos; }
         public boolean hasSome() { return pos < v.nElts(); }
         public Value elt()
          { if( hasNone() ) error( ".elt()" ); // else ...
            return v.elt(pos);
          }
         public double wt()
          { if( hasNone() ) error( ".wt()" ); // else ...
            return v.wt(pos);
          }
         public void    advance() { pos ++ ; }
         public String toString() { return "fromVector(" + v + ")"; }
    }; }//fromVector()

   /** Series of elements from 'this' until exhausted, then from 's2'. */
   public Series append(final Series s2)                       // append(s2)
    { final Series s1 = this;
      return new Series()
       { public boolean hasSome() { return s1.hasSome() || s2.hasSome();    }
         public int    position() { return s1.position() + s2.position();   }
         public Value       elt() { return (s1.hasSome() ? s1 : s2) .elt(); }
         public double       wt() { return (s1.hasSome() ? s1 : s2) .wt();  }
         public void    advance() { (s1.hasSome() ? s1 : s2) . advance();   }
         public String toString() { return s1 + ".append(" + s2 + ")";      }
    }; }//append(s2)


   /** {@link #merge(boolean,Series) merge(true,s2)}, i.e.,
       keep elements shared by 'this' and 's2'. */
   public Series merge(final Series s2) { return merge(true, s2); }

   /** Merge the outputs of 'this' and 's2' in <I>ascending</I> order.
       Both 'this' and s2 must be ascending. If 'not&nbsp;keepShared',
       and 'this' and 's1' return equal elements, then advance()
       advances them both.  */
   public Series merge(final boolean keepShared, final Series s2)
    { final Series s1 = this;
      return new Series ()
       { private boolean smallerIsKnown = false;  // cache results ...
         private Value e1 = null, e2 = null;      // ...
         private int pos = 0,           // note keepShared
                     comparison = 666;  // QI says it should be 616

         public boolean hasSome() { return s1.hasSome() || s2.hasSome(); }

         public int position() { return pos; }

         /** Note, it <I>requires</I> that s1.hasSome() && s2.hasSome() */
         private void makeSmallerKnown()
          { assert s1.hasSome() && s2.hasSome() : "s1 or s2 is exhausted";
            if( smallerIsKnown ) return;     // else...
            if( e1 == null ) e1 = s1.elt();  // else...
            if( e2 == null ) e2 = s2.elt();
            // comparison = ((Value)e1).compareTo((Value)e2); // no warn
            comparison = e1.compareTo(e2);   // warning: [unchecked] ...
            smallerIsKnown = true;
          }//makeSmallerKnown()

         public Value elt()
          { if( s1.hasNone() ) return s2.elt();
            if( s2.hasNone() ) return s1.elt();
            // else both of them hasSome()
            makeSmallerKnown();
            return comparison <= 0 ? e1 : e2;
          }//elt()

         public double wt()
          { if( s1.hasNone() ) return s2.wt();
            if( s2.hasNone() ) return s1.wt();
            // else both of them hasSome()
            makeSmallerKnown();             // Note, pref given ...
            return (comparison <= 0 ? s1 : s2) .wt();  // ... to s1
          }//wt()

         public void advance()
          { if(      s1.hasNone() ) s2.advance();                 // s2
            else if( s2.hasNone() ) s1.advance();                 // s1
            else // both of them hasSome()
             { makeSmallerKnown();
               if( comparison <= 0 )
                 { s1.advance(); e1 = null;                        // s1
                   if( comparison == 0 && ! keepShared )
                      { s2.advance(); e2 = null; }                 // s2
                 }
               else /*comparison>0*/ { s2.advance(); e2 = null; }  // s2
               smallerIsKnown = false;  // !
             }
            pos ++ ;
          }//advance()

         public String toString() { return s1 + ".merge(" + s2 + ")"; }
    }; }//merge(s2)


   /** The Series of just one element. Also see {@link #values values(arr)} */
   public static Series singleton(final Value v)                // singleton(v)
    { return new Series()
       { private Value e = v;
         public boolean hasSome() { return e != null; }
         public int    position() { return e == null ? 1 : 0; }
         public Value       elt()
          { if( hasNone() ) error( ".elt()" ); /* else */ return e; }
         public void    advance() { e = null; }
         public String toString() { return "Series.singleton(v)"; }
       };
    }//singleton(v)

   /** Series of elements from an array of Values, one at a time.
       Also see {@link Vector#toSeries}. */
   public static Series values(final Value[] arr)             // values(arr)
    { return new Series()
       { private int counter = 0;
         public int    position() { return counter; }
         public boolean hasSome() { return counter < arr.length; }
         public Value elt()
          { if( hasNone() ) error( ".elt()" ); // else ...
            return arr[counter];
          }
         public void    advance() { counter ++ ; }
         public String toString() { return "values(arr)"; }
    }; }//values(arr)


   /** Series of Ints from 'fst' inclusive, to 'ovr' <I>exclusive</I>, with
       an optional 'step' (default is&nbsp;1) which may be negative. */
   public static class Range extends Series.Int                        // Range
    { public final int fst, ovr, step;
      private int p = 0, n;
      public Range(Value.Int fst, Value.Int ovr)
       { this(fst.n(), ovr.n(), 1); }//constructor
      public Range(Value.Int fst, Value.Int ovr, Value.Int step)
       { this(fst.n(), ovr.n(), step.n()); }//constructor
      public Range(int fst, int ovr)
       { this(fst, ovr, 1); }//constructor
      /** The principal constructor for [fst, fst+step, fst+2*step, ...].
          Note, step may be positive or negative. */
      public Range(int fst, int ovr, int step)
       { this.fst=fst; this.ovr=ovr; this.step=step; n=fst; }//constructor
      public int position() { return p; }
      public boolean hasSome()
       { return step >= 0 ? n < ovr : n > ovr; };
      public Value.Int elt()
       { return new Value.Int(elt_n() /* see below */); }
      public int elt_n()
       { if( hasNone() ) error( ".elt()" ); // else ...
         return n;
       }
      public void advance() { p ++ ; n += step; }
      public String toString()
       { return "Range(" + fst + "," + ovr + "," + step + ")"; }
    }//class Series.Range


   /** Series of lines ({@link la.la.Value.Chars}, strings) from an
       input stream of byte such as a FileInputStream, say.
       Also see {@link Separator} if interested in CSV files.  */
   public static class Lines extends Series                            // Lines
    { /** The InputStream that 'this' Lines is based on. */
      public final java.io.InputStream inp;
      private int  lineNo = 0;
      private Value.Chars line = null;
      private byte[] buff = new byte [1];
      private boolean crPrevious = false;  // for ...\n, ...\r, or ...\r\n

      public Lines(java.io.InputStream inp)
       { this.inp = inp; advance(); }//constructor

      /** The current line number, starting at zero.
          Or, the number of lines processed if none remain. */
      public int    position() { return lineNo; }
      public boolean hasSome() { return line != null; }
      /** The current line (a Chars Value). Note, it
          excludes any terminating newline character. */
      public Value.Chars elt() { return line; }

      public void advance()
       { if( line != null ) lineNo ++ ;  line = null;
         try{ if( inp.available() <= 0 ) return;
              StringBuilder sb = new StringBuilder();
              do{ if( inp.read(buff) == 0 ) break;
                  char c = (char) buff[0];
                  if( c == '\n' )
                     if( crPrevious )   // "\r" then "\n" = one line
                      { crPrevious = false; continue; }
                     else break;        // lone "\n"
                  else if( c == '\r' )  // "\r", maybe "\n" to come?
                   { crPrevious = true;  break; }
                  else crPrevious = false;
                  sb.append(c);
              }while( inp.available() > 0 );
              // c == '\n' || inp.available() == 0
              line = new Value.Chars(sb.toString());
         }catch(Exception e ) { error( ": " + e ); }
       }//advance()

      public String toString() { return "Lines(" + inp + ")"; }
    }//class Series.Lines


   /** A Series of "comma"-separated variables (CSV) as Chars (strings) out
       of a String, s.  In fact the {@link #separator} can be any character.
       The terminology "variable" is used in statistics and refers to
       a column of values in a spreadsheet (.csv, data-set), say.
       You may also be interested in {@link Lines}.  */
   public static class Separator extends Series                    // Separator
    { /** The String being scanned for variables. */
      public final String s;

      /** The separator between variables, often but not always&nbsp;','. */
      public final char separator;

      /** To trim white space from heads and tails, or not? */
      public final boolean trimHead, trimTail;

      private String whiteSpace = " \t",      // i.e., space and tabs
                     quotes = "\'\"";
      private int varNo = 0, lo = 0, tl, hi = -1;  // csv ~ s[lo..tl)
      private Value.Chars csv = null;         // current Separator, if any

      /** Construct a Separator with the most common choice,
          separator&nbsp;=&nbsp;','.  */
      public Separator(String s) { this(',', s); }

      /** Construct a Separator with a given separator character.
          Note, {@link #trimHead} = {@link #trimTail} = true.  */
      public Separator(char separator, String s)
       { this(true, true, separator, s); }

      /** Note, if separator&nbsp;='&nbsp;', trimming makes no sense. */
      public Separator(boolean trimHead, boolean trimTail,
                       char separator, String s)
       { this.trimHead = trimHead;   this.trimTail = trimTail;
         this.separator = separator; this.s = s;
         switch( separator )
          { case '\t': whiteSpace = " "; break;  // tab separated, plausible
            case  ' ': whiteSpace = "";  break;  // space sep., conceivable
            case  '"': quotes = "\'"; break;
            case '\'': quotes = "\""; break;
          }
         advance();
       }//constructor

      public int    position() { return varNo; }
      public boolean hasSome() { return lo < s.length(); }
      public Value.Chars elt() { return csv; }

      /** Advance to the next variable, if any. */
      public void advance()
       { if( csv != null ) varNo ++ ;  csv = null;
         lo = hi+1;  // skip the separator
         if( trimHead )
            while( lo < s.length() && whiteSpace.indexOf(s.charAt(lo)) >= 0 )
               lo++;
         if( lo >= s.length() ) return;
         hi = lo;
         while( hi < s.length() )
          { char c = s.charAt(hi);
            if( c == separator ) break;
            hi ++ ;
            if( quotes.indexOf(c) >= 0 )                     // "...
             { while( hi < s.length() && s.charAt(hi) != c ) //  ..."
                  hi ++ ;                                  // or ...<eoln>
               if( hi < s.length() ) hi ++ ;
          }  }
         tl = hi;
         if( trimTail )
            while( tl > lo && whiteSpace.indexOf(s.charAt(tl-1)) >= 0 )
               tl -- ;
         csv = new Value.Chars(s.substring(lo, tl));  // [lo, hi)
       }//advance()

      public String toString()
       { return "Separator(" + trimHead + "," + trimTail
              + ",'" + separator + "',...)";
       }//toString()
    }//class Separator


   /** Throw a RuntimeException. The exception includes the current
       {@link #position() position()}, the <TT>toString()</TT> of
       'this' Series and 'msg'. */
   public void error(String msg)
    { Util.error( "(position=" + position() + ") " + this + msg ); }
// -------------------------------------------------------------------------

   public static void main(String[] args)
    { System.out.println("-- test Series --");

      System.out.println("see la.util.Test");

      System.out.println("-- done Series --");
    }//main

 }//Series class   http://www.allisons.org/ll/

// (c) Lloyd Allison Oct 2011, Tue 14 Feb 2017. Licence to 'Series' is Gnu GPL.
