//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.la.Function;
import la.la.Function.Cts2Cts;
import la.maths.Vector;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference:
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** Continuous may be used to define (UnParameterised) Models over
    {@link la.la.Value.Cts Cts} Values by specifying a (-log) probability
    density function, {@link Continuous.M#nlPdf_x}, on a Value's 'x()'.
    The&nbsp;fully parameterised Models are {@link Continuous.M}.  */
public abstract class Continuous extends ByPdf                  // * Continuous
 { public Continuous(Value dp)
     { super(dp); }//constructor

   public abstract M sp2Model(double msg1, double msg2, Value sp);         // *
   // .........................................................................

   /** The (abstract) class of fully parameterised Continuous Models.
       The UnParameterised Model is {@link Continuous}. */
   public abstract class M extends ByPdf.M                    // * Continuous.M
    { public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp); }//constructor

      /** The negative log probability density of x;
          also see {@link #nlPdf}. */
      public abstract double nlPdf_x(double x);                 // * nlPdf_x(x)

      /** Calls <TT>{@link #nlPdf_x nlPdf_x}(d.x())</TT>. */
      public double nlPdf(Value d)
       { return nlPdf_x( ((Value.Cts) d) .x() ); }

      /** Call {@link #random_x()}, if implemented, to return
          an <I>exact</I> random {@link la.la.Value.Real}; it might be
          necessary to do something different if, for example,
          you want accuracy (AoM) to depend on the result's x().
          But if you want random Values with zero or other
          constant AoM, the present strategy is fine.  */
      public Value.Real random()                                    // random()
       { return Value.real( random_x() ); }

      /** Return a random
          {#link la.la.Value#cts Cts}({@link #random_x() random_x()},AoM). */
      public Value.Cts random(double AoM)
       { return Value.cts(random_x(), AoM); }                    // random(AoM)

      /** Return an array double[n] of {@link #random_x() random_x()}. */
      public double[] random_x(int n)
       { double[] ans = new double[n];
         for( int i = 0; i < n; i ++ ) ans[i] = random_x();
         return ans;
       }//random_x(n)

      /** Return n random Cts each with the specified AoM. */
      public Vector random(int n, double AoM)
       { return Vector.doubles(random_x(n), AoM); }

      /** Called by {@link #random()}; this default is
          not implemented (throws an Exception) but see
          {@link Continuous.Uniform.M#random_x()} and
          {@link NormalUPM.M#random_x()}, say.  Note, there is an issue
          about generating random Cts Values with non-zero AoM if AoM
          depends on&nbsp;x.  */
      public double random_x()                                  // * random_x()
       { throw UOE( ".random_x() is not yet(?) supported, sorry" ); }
      // ..............................

      /** This transform(f) is like {@link Model#transform Model.transform(f)}
          except that, instead of "just" returning a Model, <I>this</I>
          transform returns a (trivially) parameterised Continuous.M
          (actually a {@link M.Transform.MM}).<BR>
          Also see the related but different
          {@link Continuous#transform Continuous.transform(f)}. */
      public Continuous.M transform(Cts2Cts f)                  // transform(f)
       { return new Transform(f).Mdl; }

     /** A Continuous.M.Transform "is a" (extends) Continuous,
         i.e., is an UnParameterised Continuous model; the fully
         <U>trivially</U> parameterised Model is {@link Transform.MM}.
         Also see the related but different {@link Continuous.Transform}.<BR>
         <B>NB.</B> The preferred way to transform a (parameterised)
         Continuous.M is to use function {@link M#transform transform(f)}.  */
      public class Transform extends Continuous       // Continuous.M.Transform
       {
         /** The Function, f:Cts&rarr;Cts, doing the transforming of data.  */
         public final Function.Cts2Cts f;

         /** The "problem defining" parameter 'f' is a Cts2Cts
             and is saved in '{@link #f f}'.  */
         public Transform(Value f)
          { super(f); this.f = (Function.Cts2Cts) f; }//constructor

         /** As per the enclosing Continuous.M.this's stats(...) on the
             transformed data-set, ds.map({@link #f f}).  */
         public Value stats(Vector ds, int lo, int hi)
          { return Continuous.M.this.stats(ds.map(f), lo, hi ); }

         /** Combine {@link #stats statistics} ss0 and ss1 as
             as per the enclosing Continuous.M.this does.  */
         public Value stats(boolean add, Value ss0, Value ss1)
          { return Continuous.M.this.stats(add, ss0, ss1); }  // !

         /** Trival estimator, always "estimates" {@link #Mdl Mdl}. */
         public Estimator estimator(Value t)
          { return Mdl.asEstimator(t); }

         public MM sp2Model(double msg1, double msg2, Value sp)
          { return new MM(msg1, msg2, sp); }

         /** <I>The</I> fully (trivially) parameterised {@link MM} Model. */
         public final MM Mdl = new MM(0, 0, triv);
         // ......................................

         /** (Wanted to call this class M, as in Continuous.M.Transform.M,
             but the compiler (1.8.0_101) objects.)
             The&nbsp;fully, <I>trivially</I> parameterised...M.
             The UnParameterised model is {@link Transform};
             a Transform.MM "is a" (extends) Continuous.M.
             Also see the related but different
             {@link Continuous.Transform}.<BR>
             <B>NB.</B> The preferred way to transform a Continuous.M
             is to use function {@link M#transform transform(f)}.  */
         public class MM extends Continuous.M      // Continuous.M.Transform.MM
          { /** Note, msg1=0 and statistical parameter sp=triv, checked. */
            public MM(double msg1, double msg2, Value sp)
             { super(msg1, msg2, sp); zeroTriv(msg1, sp); }//constructor

            /** {@link Continuous.M#nlPdf_x M.nlPdf_x(f(x))}
                &minus; log(|{@link Cts2Cts#d_dx f.d_dx()(x)}|),
                the second term to "correct" the datum's nlAoM.  */
            public double nlPdf_x(double x)
             { return Continuous.M.this.nlPdf_x(f.apply_x(x))
                    - Math.log(Math.abs(f.d_dx().apply_x(x))); // NB neg log!
             }//nlPdf_x(x)

            /** Use the enclosing Continuous.M.this's nlLH(ss) but
                statistics 'ss' come from {@link Transform#stats}.  */
            public double nlLH(Value ss)
             { return Continuous.M.this.nlLH(ss); }

            /** Requires '{@link #f f}' to have an implemented inverse to work.
                Sample with the enclosing Continuous.M.this's random_x()
                and apply {@link #f f}'s inverse.  */
            public double random_x()  // requires f to have an inverse
             { return ((Cts2Cts) ((Function.HasInverse) f) .inverse() )
                      .apply_x(Continuous.M.this.random_x());
             }//random()

            public String toString()
             { return Continuous.M.this + ".transform(" + f + ")"; }
          }//Continuous.M.Transform.MM class, the Model

       }//Continuous.M.Transform class, the UPModel

    }//Continuous.M (abstract) class, the Model
   // .........................................................................

   /** This transform(f) is like {@link UPModel#transform UPModel.transform(f)}
       except that, instead of "just" returning a UPModel, <I>this</I>
       transform returns a (UnParameterised) Continuous (actually a
       {@link Continuous.Transform}).<BR>
       Also see the related but different
       {@link Continuous.M#transform Continuous.M.transform(f)}. */
   public Continuous transform(Cts2Cts f) { return new Transform(f); }

   /** A transformed UnParameterised Continuous model;
       the parameterised Model is {@link Transform.M}.
       See {@link MML#logNormal logNormal} for example.
       We would <I>like</I> Transform to extend both {@link Continuous}
       and {@link UPModel.Transform}, the former for what
       we want Transform to <U>be</U>, the latter to inherit
       some code, but it cannot so we go with the former and
       copy some code from the latter.
       Also see the related but different {@link Continuous.M.Transform}.<BR>
       <B>NB.</B> The preferred way to transform a Continuous is
       to use function {@link #transform transform(f)}.  */
   public class Transform extends Continuous            // Continuous.Transform
    { /** The continuous Function, f:Cts&rarr;Cts,
          doing the transforming of data.  */
      public final Function.Cts2Cts f;

      /** Problem defining parameter 'f' must be a {@link Cts2Cts}.  */
      public Transform(Value f)
       { super(f); this.f = (Function.Cts2Cts) f; }//constructor

      /** Return a fully parameterised {@link Transform.M}.  */
      public M sp2Model(double msg1, double msg2, Value sp)
       { return new M(msg1, msg2, sp); }

      /** The enclosing Continuous.this's
          stats(ds.map({@link #f f})), on a transformed data-set.  */
      public Value stats(Vector ds, int lo, int hi)
       { return Continuous.this.stats(ds.map(f), lo, hi ); }

      /** Combine {@link #stats statistics} ss0 and ss1 as per
          the enclosing Continuous.this.  */
      public Value stats(boolean add, Value ss0, Value ss1)
       { return Continuous.this.stats(add, ss0, ss1); }  // !

      public String toString()
       { return Continuous.this + ".transform(" + f + ")"; }

      /** Uses the enclosing Continuous.this's estimator but
          {@link #stats statistics} take the transforming
          function '{@link #f f}' into account.  */
      public Est estimator(Value ps)                           // estimator(ps)
       { return new Est(ps)             // i.e., Continuous.Est
          { final Estimator e = Continuous.this.estimator(ps);
            public Transform.M ss2Model(Value ss) // ss from Transform...
             { Model m = e.ss2Model(ss);
               return new Transform.M(m.msg1(), m.msg2(), m.statParams());
             }
          };//new Est(ps)...
       }//estimator(ps)
      // ................................

      /** The fully parameterised transformed Model; the
          UnParameterised model is {@link Continuous.Transform}.
          Statistical parameters are as per the enclosing Continuous.
          Note that a {@link UPModel.Transform.M} is just a UPModel.M but
          a Continuous.Transform.M is a (extends) Continuous.M.  Also see
          the related but different {@link Continuous.M.Transform.MM}.<BR>
          <B>NB.</B> The preferred way to transform a Continuous is
          to use function {@link Continuous#transform transform(f)}.  */
      public class M extends Continuous.M             // Continuous.Transform.M
       { /** The Continuous.M Model 'm' doing the work behind the scenes. */
         public final Continuous.M m;

         /** Statistical parameter(s) 'sp' are as per {@link Continuous.M}
             and are used to create&nbsp;'{@link #m m}'. */
         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp);
            this.m = Continuous.this.sp2Model(msg1, msg2, sp);
          }//constructor

         /** Negative log probability density as per {@link #m m} but
             corrected by {@link #f f}'s derivative.  */
         public double nlPdf_x(double x)
          { return m.nlPdf_x(f.apply_x(x))
                 - Math.log(Math.abs(f.d_dx().apply_x(x))); // NB neg log!
          }//nlPdf_x(x)

         /** {@link #m m}.nlLH(ss), but note that statistics
             'ss' comes from {@link #stats stats}. */
         public double nlLH(Value ss) { return m.nlLH(ss); }

         /** {@link #m m}.stats(ds.map({@link #f f}),...), i.e.,
             transformed data.  Also see {@link #nlLH nlLH(ss)}.  */
         public Value stats(Vector ds, int lo, int hi)
          { return m.stats(ds.map(f), lo, hi); }

         /** Combine {@link #stats statistics} 'ss0' and 'ss1'
             as '{@link #m m}' does.  */
         public Value stats(boolean add, Value ss0, Value ss1)
          { return m.stats(add, ss0, ss1); }

         /** Sample from {@link #m m} and apply {@link #f f}'s
             inverse (provided it is defined).  */
         public double random_x()
          { return ((Function.Cts2Cts) ((Function.HasInverse) f).inverse())
                   .apply_x(m.random_x());
          }//random_x()
       }//Continuous.Transform.M class, the Model

    }//Continuous.Transform class, the UPModel
   // -------------------------------------------------------------------------

   /** UnParameterised Bounded Continuous Models, over
       a range, <NOBR>[lwb, upb],</NOBR> of Cts Values. */
   public static abstract class Bounded extends Continuous      // * ...Bounded
    { public Bounded(Value dp)
       { super(dp); }//constructor
             /** Return (lwb, upb) of the data-space. */
      public abstract Tuple bounds();                             // * bounds()
             /** The lower bound, {@link #bounds() bounds()}.fst(). */
      public Cts    lwb()   { return (Cts) bounds().fst(); }
             /** The upper bound, {@link #bounds() bounds()}.snd(). */
      public Cts    upb()   { return (Cts) bounds().snd(); }
             /** The lower bound of the data-space as a double. */
      public double lwb_x() { return lwb().x(); }
             /** The upper bound of the data-space as a double. */
      public double upb_x() { return upb().x(); }
      // ............................

      /** Fully parameterised Bounded Continuous Models. */
      public abstract class M extends Continuous.M            // * ...Bounded.M
       { public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp); }//constructor
             /** Bounded.this.{@link Bounded#bounds() bounds()}. */
         public Tuple bounds() { return Bounded.this.bounds(); }
             /** Bounded.this.{@link Bounded#lwb() lwb()}. */
         public Cts      lwb() { return Bounded.this.lwb(); }
             /** Bounded.this.{@link Bounded#upb() upb()}. */
         public Cts      upb() { return Bounded.this.upb(); }
             /** Bounded.this.{@link #lwb_x() lwb_x()}. */
         public double lwb_x() { return Bounded.this.lwb_x(); }
             /** Bounded.this.{@link #lwb_x() upb_x()}. */
         public double upb_x() { return Bounded.this.upb_x(); }
       }//Bounded.M (abstract) class, the Model

    }//Bounded (abstract) class, the UPModel
   // ......................................

   /** The UnParameterised Uniform Continuous Model on the range
       [lwb,&nbsp;upb]. Also see the fully (trivially) parameterised
       {@link Uniform.M}, and the {@link Discretes.Uniform}.  */
   public static class Uniform extends Bounded            // Continuous.Uniform
    { /** The lower (upper) bound on the .x() of a datum. */
      private final double lwb_x, upb_x;
      /** The negative log probability density of a (every) data Value. */
      private final double nlPdf;

      /** Construct an UnParameterised Uniform Model
          (distribution) over the bounds=[lwb,upb].  */
      public Uniform(Value bounds)
       { super(bounds.isTuple(2));
         this.lwb_x = lwb().x(); this.upb_x = upb().x();
         final double range = upb_x - lwb_x;
         this.nlPdf = Math.log(range);
       }//constructor

      public Tuple bounds() { return (Tuple) defnParams(); }
      public double lwb_x() { return lwb_x;  }
      public double upb_x() { return upb_x;  }

      /** Return a fully (trivially) parameterised Uniform
          Continuous {@link M Model};
          note that the statistical parameter, t, is trivial. */
      public M sp2Model(double m1, double m2, Value t /* = triv */)
       { return new M(m1, m2, t); }

      /** Given a data-set, ds, return sufficient statistics, ss, that
          is the (weighted) number of elements and their nlAoM.
          More on stats {@link UPModel#stats here}.  */
      public Value.Tuple stats(Vector ds, int lo, int hi)
       { return Value.pair( Value.real(ds    .wts(lo,hi)),
                            Value.real(ds.wtNlAoM(lo,hi)) );
       }//stats(ds,lo,hi)

      public Value.Tuple stats(boolean add, Value ss0, Value ss1)
       { Value.Tuple ss0t = ss0.isTuple(2), ss1t = ss1.isTuple(2);
         return add
            ? Value.pair( Value.real(ss0t.fst().x() + ss1t.fst().x()),
                          Value.real(ss0t.snd().x() + ss1t.snd().x()) )
            : Value.pair( Value.real(ss0t.fst().x() - ss1t.fst().x()),
                          Value.real(ss0t.snd().x() - ss1t.snd().x()) );
       }

      /** Given sufficient statistics,
         <NOBR><TT>ss = {@link Poisson0UPM#stats stats(ds)}</TT>,</NOBR>
         of a data-set, ds, return the negative log likelihood of ds.
         The statistical parameter of {@link Uniform.M} is trivial,
         so nlLH does not depend on it and nlLH can be here in
         the UnParameterised Model. */
      public double nlLH(Value ss)
       { final Value.Tuple ssT = ss.isTuple(2);
         final double N = ssT.fst().x(), nlAoM = ssT.snd().x();
         return nlPdf * N + nlAoM;
       }//nlLH(ss)

      /** The trivial Estimator that is a Uniform's. */
      public Estimator estimator(final Value t /* = triv */)
       { return Mdl.asEstimator(t); }
      // ............................

      /** <I>The</I> fully parameterised Uniform continuous
          Model on [lwb,&nbsp;upb].  */
      public final M Mdl = new M(0, 0, triv);

      /** {@link #Mdl} should be sufficient for many purposes.
          M, the class of fully- (trivially) parameterised Uniform
          continuous Model(s) on the range <NOBR>[lwb, upb]</NOBR>.  */
      public class M extends Bounded.M                  // Continuous.Uniform.M
       { /** requires msg1=0, sp=triv. */
         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp); zeroTriv(msg1, sp); }//constructor

         public double nlPdf_x(double d) { return nlPdf; }

         /** Return {@link Uniform}.{@link Uniform#nlLH nlLH(ss)}. */
         public double nlLH(Value ss)
          { return Uniform.this.nlLH(ss); }

         /** Called by {@link #random()}; note, the generated
             {@link la.la.Value.Cts} is exact, its AoM=0. */
         public double random_x()
          { return lwb_x + RNG.nextDouble() * (upb_x - lwb_x); }

       }//Continuous.Uniform.M class, the Model

    }//Continuous.Uniform class, the UPModel

 }//Continuous (abstract) class, the Model

// (c) L.Allison 2012, Jan 2014, 15 Apr 2020. Licence to Continuous is Gnu GPL.
