//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.la.Type;
import la.maths.Vector;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference:
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The (abstract) sub-class of UnParameterised Model that may be helpful
    in defining Models of Discrete data-spaces (Values), such as
    {@link la.la.Value.Enum} data. The fully parameterised model is
    {@link Discretes.M}. Particularly see {@link M#nlPr_n} and
    {@link M#random_n()}.  */
public abstract class Discretes extends UPModel                  // * Discretes
 { public Discretes(Value sp)
    { super(sp); }//constructor

   public M apply(Value sp) { return sp2Model(0, 0, sp); }

   public abstract M sp2Model(double msg1, double msg2, Value sp);         // *
   // ...............................................

   /** The (abstract) class of fully parameterised Models of
       Discrete data-spaces.  You need to specify
       {@link #nlPr_n nlPr_n(n)}.  */
   public abstract class M extends UPModel.M                   // * Discretes.M
    { public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp); }//constructor

      /** Back-room works for {@link #pr}, returns
          <NOBR>Math.exp( - {@link #nlPr_n nlPr_n(n)}</NOBR>.  */
      public double pr_n(int n)                                      // pr_n(n)
       { return Math.exp( - nlPr_n(n) ); }

      /** Get pr(d) from <TT> {@link #pr_n pr_n}(d.n()). </TT> */
      public double   pr(Value d) { return   pr_n( d.n() ); }

      /** Get nlPr(d) from <TT> {@link #nlPr_n nlPr_n}(d.n()). </TT> */
      public double nlPr(Value d) { return nlPr_n( d.n() ); }

      /** Back-room works for negative log probability of datum int n,
          {@link #nlPr}; it must be implemented in an instance. */
      public abstract double nlPr_n(int n);                      // * nlPr_n(n)

      /** Back-room works for {@link #random}. There is a tricky semantic
          problem about turning it into a Value through random() -- e.g.,
          does '1' denote C:DNA, or true:boolean, and so on (we first need
          to know the {@link la.la.Type} of the data-space).  */
      public abstract int random_n();                           // * random_n()

      /** Shift 'this' fully parameterised Model of Discretes by
          +{@link Shifted#offset offset} (~shift data by -offset). Its nlPr(d)
          and stats(ds,lo,hi) subtract {@link Shifted#offset offset} from data
          before handing it to the unshifted Model.  Note that shifting
          a {@link Bounded Bounded} Model may reduce the data-space.
          Also see {@link Discretes#shifted}.  */
      public M shifted(final int offset)                     // shifted(offset)
       { return new M(msg1(), msg2(), statParams())
          { public double  nlPr_n(int n)
             { return M.this.nlPr_n(n - offset); }

            public double nlLH(Value ss)
             { return M.this.nlLH(ss); } // NB. ss=stats(ds,l,h)...

            public Value stats(Vector ds, int lo, int hi)
             { return M.this.stats( //   i.e., map ds (\d. d-offset) ...
                  ds.map( new Native()
                     { public Discrete apply(Value d)  // i.e., (\d.d-offset)
                        { return ((Type.Discrete) d.type())
                                 .int2value(d.n() - offset);
                     }  } ),
                  lo, hi );
             }//stats(ds)

            public Value stats(boolean add, Value ss0, Value ss1)
             { return M.this.stats(add, ss0, ss1); } // !

            public Discrete random()
             { final Value r = M.this.random();
               return ((Type.Discrete) r.type())
                                       .int2value(r.n() + offset);
             }//random()

            public int random_n()
             { return M.this.random_n() + offset; }

            public String toString()
             { return M.this + ".shifted(" + offset + ")"; }
          };//new M(,,)
       }//shifted(offset)

    }//Discretes.M (abstract) class, the fully parameterised Model
   // .........................................................................

   /** Some Discrete Models, such as {@link Geometric0UPM.M}, have
       (N,sum) as sufficients statistics, <NOBR>ss = stats(ds),</NOBR>
       of a data-set ds, in which case they may implement
       {@link Model#stats} by calling NandSum. Note the Values of N and
       sum are Cts because we may have fractionally weighted data,
       say in a {@link Mixture Mixture}. */
   public static Vector NandSum(Vector ds)
    { return NandSum(ds, 0, ds.nElts()); }

   /** {@link #NandSum(Vector) NandSum} for elements [lo,&nbsp;hi).  */
   public static Vector NandSum(Vector ds, int lo, int hi)
    { double N = 0, sum = 0;
      for( int i = lo; i < hi; i ++ )
       { double wi = ds.wt(i);
         N += wi;  sum += wi * ds.elt(i).n();
       }
      return Vector.doubles(new double[]{N, sum});
    }//NandSum(ds)
   // .........................................................................

   /** Convenience function for
       new&nbsp;{@link Shifted Shifted(new Int(offset))}, for example,
       {@link Poisson0UPM Poisson0}.shifted(1) is the unparameterised
       Poisson distribution for integers n&ge;1.
       Also see {@link Discretes.M#shifted}.  */
   public Discretes shifted(int offset)
    { return new Shifted(new Int(offset)); }

   /** Convenience function for new&nbsp;{@link Shifted Shifted(offset)},
       for example, {@link Poisson0UPM Poisson0}.shifted(1)
       is the unparameterised Poisson distribution for integers n&ge;1.
       Also see {@link Discretes.M#shifted}.  */
   public Discretes shifted(Int offset)
    { return new Shifted(offset); }

   /** This {@link Discretes UnParameterised} Model of Discretes shifted by
       {@link #offset +offset}. A shifted UnParameterised Model
       {@link #estimator estimates} or otherwise creates a
       {@link M#shifted shifted&nbsp;Model}. Position parameters,
       such as mean &mu;, are in effect moved to &mu;+offset.
       <NOBR>E.g., Poisson1 = Poisson0.shifted(1)</NOBR> models ints n&ge;1.
       Note that shifting a {@link Bounded Bounded} Model may reduce the
       data-space.  Also see {@link #shifted} and {@link M#shifted}.<BR>
       <B>2020</B>: See {@link UPModel.Transform} for a
       more general way of transforming <I>any</I> Model using
       a {@link la.la.Function}.  */
   public class Shifted extends Discretes                  // Discretes.Shifted
    { /** The offset adjustment, as an int. */
      public final int offset;

      /** The problem defining parameter, dp, is
          the {@link #offset offset} Int. */
      public Shifted(Value dp)
       { super(dp); offset = ((Int) dp) .n(); }//constructor

      /** Return sufficient statistics, ss=stats(ds,lo,hi), for a
          shifted (-{@link #offset offset}) data-set ds[lo,hi). */
      public Value stats(Vector ds, int lo, int hi)
       { return Discretes.this.stats( //   i.e., map ds (\d. d-offset) ...
            ds.map( new Native()
               { public Discrete apply(Value d)
                  { return ((Type.Discrete) d.type())
                           .int2value(d.n() - offset);
               }  } ),
            lo, hi );
       }//stats(ds)

      public Value stats(boolean add, Value ss0, Value ss1)
       { return Discretes.this.stats(add, ss0, ss1); } // !

      /** The shifted Estimator estimates a
         {@link M#shifted shifted&nbsp;Model}. */
      public Est estimator(Value ps)
       { final Estimator UE = Discretes.this.estimator(ps);
         return new Est(ps)
          { public Discretes.M ss2Model(Value ss)
             { return ((M) UE.ss2Model(ss)) . shifted(offset); }
          };
       }//estimator(ps)

      public Discretes.M sp2Model(double msg1, double msg2, Value sp)
       { return Discretes.this.sp2Model(msg1, msg2, sp).shifted(offset); }

      public String toString()
       { return Discretes.this + ".shifted(" + offset + ")"; }

    }//Discretes.Shifted class, UnParameterised Model
   // .........................................................................

   /** The class of UnParameterised Models over Bounded Discrete data
       such as <NOBR><TT>[3, 7]</TT>,</NOBR> or DNA, say.
       Also see {@link Bounded.M}. */
   public static abstract class Bounded extends Discretes          // * Bounded
    {
      public Bounded(Value sp) { super(sp); }//constructor

          /** The bounds, [{@link #lwb() lwb()}, {@link #upb() upb()}],
              on 'this' Model's dataspace. */
      public abstract Tuple bounds();                             // * bounds()
          /** The lower bound on 'this' Model's dataspace. */
      public Value lwb() { return bounds().fst(); }                    // lwb()
          /** The upper bound on 'this' Model's dataspace. */
      public Value upb() { return bounds().snd(); }                    // upb()
          /** <TT> {@link #lwb}.n() </TT> */
      public int lwb_n() { return lwb().n(); }                       // lwb_n()
          /** <TT> {@link #upb}.n() </TT> */
      public int upb_n() { return upb().n(); }                       // upb_n()

      /** For <I>probable</I> use by
          {@link #stats(Vector,int,int) stats(ds,lo,hi)}, return the
          frequency counts of elements [lo,&nbsp;hi) of data-set 'ds'.
          Note that data {@link Vector#wt(int) weights} are taken into
          account.  More on stats {@link UPModel#stats here}.  */
      public Vector freqs(Vector ds, int lo, int hi)         // freqs(ds,lo,hi)
       { final int lwb_n = lwb_n(), upb_n = upb_n();
         double [] freqs = new double [upb_n - lwb_n + 1];
         for( int i = 0; i < freqs.length; i ++ ) freqs[i] = 0;
         for( int i = lo; i < hi; i ++ )         // May have fractional wts...
            freqs[ds.elt(i).n()-lwb_n] += ds.wt(i); // ...in mixture modelling
         return Vector.doubles(freqs);
       }//freqs(ds,lo,hi)

      public Vector freqs(Vector ds) { return freqs(ds, 0, ds.nElts()); }

      /** For frequency counts 'ss0' and 'ss1', either combine
          ss0 and ss1 (add=true), or remove ss1 from ss0 (add=false).
          Note, it is only valid to remove from ss0 the contribution
          of elements previously counted in ss0.  Also see
          {@link #freqs(Vector,int,int) freqs(ds,lo,hi)}.
          For probable use by
          {@link #stats(boolean,Value,Value) stats(add,ss0,ss1)}.  */
      public Vector freqs(boolean add, Value ss0, Value ss1)
       { Vector ss0v = (Vector) ss0, ss1v = (Vector) ss1;
         double [] freqs = new double [ss0v.nElts()];
         boolean neg = false;
         if( add ) // add, ss0 + ss1
            for( int i = 0; i < freqs.length; i ++ )
               freqs[i] = ss0.elt(i).x() + ss1.elt(i).x();
         else      // remove, ss0 \ ss1
            for( int i = 0; i < freqs.length; i ++ )
             { freqs[i] = ss0.elt(i).x() + ss1.elt(i).x();
               neg |= freqs[i] < 0;
             }
         if( neg ) throw new RuntimeException(
                     "freqs(" + add + "," + ss0 + "," + ss1 + "), -ve freq" );
         return Vector.doubles(freqs);
       }//freqs(add,ss0,ss1)
      // .....................................

      /** The (abstract) class of fully parameterised
           Discrete Bounded Models. */
      public abstract class M extends Discretes.M      // * Discretes.Bounded.M
       {
         private final Type t;  // MODEL of ...
         /** This Model's Type is Model of the
             {@link #bounds() bounds'} Type. */
         public Type type() { return t; }

         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp);
            t = new Type.Model(null, lwb().type());
          }//constructor

         public Tuple bounds() { return Bounded.this.bounds(); } //bounds()
             /** Return Bounded.this.{@link Bounded#lwb() lwb()}. */
         public Value lwb() { return Bounded.this.lwb(); }             // lwb()
             /** Return Bounded.this.{@link Bounded#upb() upb()}. */
         public Value upb() { return Bounded.this.upb(); }             // upb()
             /** <TT> {@link #lwb}.n() <TT> */
         public int lwb_n() { return Bounded.this.lwb_n(); }
             /** <TT> {@link #upb}.n() <TT> */
         public int upb_n() { return Bounded.this.upb_n(); }

         /** Return a random Value in the range of 'this' Model.
             It does so by calling {@link #random_n}. */
         public Value random()                                      // random()
          { return ((Type.Discrete) lwb().type()) .int2value(random_n()); }

         /** Return a random <TT>int</TT> in the range of 'this' Model; this
             default is rather inefficient; is called by {@link #random}. */
         public int random_n()                                    // random_n()
          { int i, upb = upb_n();
            double x = RNG.nextDouble(); // x in [0,1)
            assert x >= 0 && x <= 1 : errMsg( " x=" + x + " not in [0,1]" );
            for( i = lwb_n(); i < upb; i ++ ) // NB. < upb
             { double pr = pr_n(i);
               if( x <= pr ) return i;
               x -= pr;
             }//for i
            return i;  // i = upb
          }//random_n()

        /** Calls {@link #random_n()} n times. */
         public Vector random(int n)                               // random(n)
          { int [] ns = new int [n];
            for( int i = 0; i < n; i ++ ) ns[i] = random_n();
            // Note ...type() below
            return Vector.ints((Type.Discrete) lwb().type(), ns);
          }//random(n)

         /** Return a {@link la.util.Series.Discrete} that repeatedly
             {@link la.util.Series.Int#elt_n samples} a
             {@link #random_n() random} Value, provided 'this' Model can. */
         public la.util.Series.Discrete randomSeries()        // randomSeries()
          { return new la.util.Series.Discrete((Type.Discrete) lwb().type())
             { private int pos = -1, elt_n;
                { advance(); /* ! */ }// constructor
               public boolean hasSome() { return true; }
               public int    position() { return pos; }
               public int       elt_n() { return elt_n; }
               public void    advance() { pos ++ ; elt_n = random_n(); }
          }; }//randomSeries()

         /** Return a {@link Bounded Bounded} that always produces 'this'
             Bounded.M.  Note, the result has {@link #lwb() lwb()} and
             {@link #upb() upb()}.  Also see {@link Model#asUPModel}. */
         public Bounded asUPModel()                              // asUPModel()
          { final M mdl = this; // enclosing Bounded.M
            return new Bounded(triv)
             { public M sp2Model(double msg1, double msg2, Value sp)
                { if(msg1 != 0) error(".sp2Model("+msg1+","+msg2+","+sp+")");
                  return (M) mdl.asGiven(msg2); // !
                }//sp2Model(,,)
               public Tuple bounds() { return mdl.bounds(); }
               public Value stats(Vector ds, int lo, int hi)
                { return mdl.stats(ds, lo, hi); }
               public Value stats(boolean add, Value ss0, Value ss1)
                { return mdl.stats(add, ss0, ss1); }
               public Estimator estimator(Value ps/* triv */)
                { return mdl.asEstimator(ps); } // "estimates" mdl
               public String toString() { return mdl + ".asUPModel()"; }
          }; }//asUPModel()

       }//Discretes.Bounded.M (abstract) class, the Model

    }//Discretes.Bounded class, the UPModel
   // .........................................................................

   /** The UnParameterised Uniform Model (distribution) on data in
       <NOBR>bounds = [lo, hi],</NOBR> pretty much the simplest
       Discrete Model having only the trivial
       "{@link Uniform.M#statParams() statistical} parameter."
       Also see the fully parameterised {@link Uniform.M},
       and {@link Continuous.Uniform}. */
   public static class Uniform extends Bounded             // Discretes.Uniform
    { private final int    cardinality;
      private final double pr,  nlPr;

      /** Construct an "UnParameterised" Uniform Discrete Model
          over <NOBR> bounds = [lwb, upb]. </NOBR>. */
      public Uniform(Value bounds)
       { super(bounds.isTuple(2));
         cardinality = (upb_n() - lwb_n() + 1);
         pr   = 1.0 / ((double) cardinality);
         nlPr = Math.log(cardinality);
       }//constructor

      public Tuple bounds() { return (Tuple) defnParams(); }

      /** sp2Model(0, msg2, ()) -- Uniform has no true stat params. */
      public M sp2Model(double m1, double m2, Value t)
       { return new M(m1, m2, t); }

      /** Given a data-set, ds, return sufficient statistics, ss,
          that is the (weighted) number of elements in ds, for use
          in {@link Estimator#ss2Model} and {@link M#nlLH}.
          More on stats {@link UPModel#stats here}. */
      public Value.Real stats(Vector ds, int lo, int hi)
       { return real( ds.wts(lo, hi) ); }

      public Value.Real stats(boolean add, Value ss0, Value ss1)
       { return real( add ? ss0.x() + ss1.x() : ss0.x() - ss1.x() ); }

      /** Given sufficient statistics,
          <NOBR><TT>ss = {@link #stats stats(ds)}</TT>,</NOBR> of a
          data-set, ds, return the negative log LikeliHood, nlLH(ss),
          that is the neg log prob of ds under a Uniform Model. */
      public double nlLH(Value ss)
       { return nlPr * ((Value.Cts) ss) .x(); }
       // for the Uniform, nlLH does not depend on any statistical params.

      /** The trivial Estimator that is a Uniform's. */
      public Estimator estimator(Value t)
       { return Mdl.asEstimator(t); }
      // ................

      /** <I>The</I> Uniform Model on <NOBR>[lwb, upb]</NOBR>. */
      public final M Mdl = new M(0, 0, triv);

      /** {@link #Mdl} should be sufficient for most purposes.
          M, the class fully- (trivially) parameterised Uniform
          Discrete Model(s) on <NOBR>[lwb, upb]</NOBR>. */
      public class M extends Bounded.M                   // Discretes.Uniform.M
       { /** Requires msg1=0, sp=triv. */
         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp); zeroTriv(msg1, sp); }//constructor

         /** Back-room works for {@link #pr}. */
         public double pr_n(int n)
          { assert n >= lwb_n() && n <= upb_n()
                 : errMsg(" " + n + " out of range");
            return pr;
          }//pr_n(n)

         /** Back-room works for {@link #nlPr}.
             (Note, this implementation does not
             check that 'n' is in range.)  */
         public double nlPr_n(int n)
          { assert n >= lwb_n() && n <= upb_n()
                 : errMsg(" " + n + " out of range");
            return nlPr;
          }//nlPr_n(n)

         /** nlLH(ss), see {@link Uniform#nlLH}. */
         public double nlLH(Value ss)
          { return Uniform.this.nlLH(ss); }

         /** Generate a Uniform random int in
             <NOBR>[{@link #lwb_n() lwb}, {@link #upb_n() upb}]</NOBR>. */
         public int random_n()
          { int n;
            do{ n = lwb_n() + (int)Math.floor(RNG.nextDouble()*cardinality);
              }while( n > upb_n() /* paranoid programming */ );
            return n;
          }//random_n()

       }//Discretes.Uniform.M class, the Model

    }//Discretes.Uniform class, the UPModel

 }//Discretes (abstract) class, the Model

// (c) Lloyd Allison 2012, Sat 28 Jan 2017. Licence to 'Discretes' is Gnu GPL.
