//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Lexical;
import la.la.Value;
import la.maths.Vector;
import la.maths.Vector.Slice;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The class of UnParameterised (negative-) Exponential Model(s).
    For most purposes {@link MML#Exponential} and {@link ExponentialUPM.M}
    should be enough. Also see {@link LaplaceUPM Laplace}.  */
public class ExponentialUPM extends Continuous                // ExponentialUPM
 { public ExponentialUPM(Value t /* = triv */ )
    { super(t); if( t != triv ) error( ", t=" + t ); }//constructor

   /** The sufficient statistics <NOBR>ss = stats(ds)</NOBR> of elements
       [lo,&nbsp;hi) of data-set ds; are
       <NOBR>[N, &sum;ds(i), &sum;nlAoM]</NOBR>.
       Also see {@link ExponentialUPM.M#nlLH}.
       More on stats {@link UPModel#stats here}. */
   public Vector stats(Vector ds, int lo, int hi)
    { double sum = 0;
      for( int i = lo; i < hi; i ++ ) sum += ds.wt(i) * ds.elt(i).x();
      return Vector.doubles( new double[]
        { ds.wts(lo, hi), /* i.e., N */
          sum,
          ds.wtNlAoM(lo, hi) } );
    }//stats(ds,lo,hi)

   public Vector stats(boolean add, Value ss0, Value ss1)
    { return (Vector)
         ((Vector) ss0).bOp(add?Lexical.plus:Lexical.minus, (Vector) ss1);
    }//stats(add,ss0,ss1)

   public M sp2Model(double m1, double m2, Value A)
    { return new M(m1, m2, A); }

   /** <NOBR> F = N / A<SUP>2</SUP>. </NOBR>
       Negative log likelihood, <NOBR>- log L = N log A + (SUM x_i)/A.</NOBR>
       <NOBR>d/d A {- log L} = N/A - (SUM x_i)/A^2.</NOBR>
       So max_LH estimate, <NOBR>A = (SUM x_i)/N</NOBR>, equals the mean.
       <NOBR>d^2/d A^2 {- log L} = - N/A^2 + 2(SUM x_i)/A^3.</NOBR>
       In expectation,
       <NOBR>F = E(d^2/d A^2 (- log L))</NOBR>
       <NOBR>  = - N/A^2 + 2 N A / A^3 = N/A^2.</NOBR>
       Also see {@link #estimator estimator(ps)}.  */
   public double logF(double N, double A)
    { return Math.log(N) - 2.0 * Math.log(A); }

   /** The prior is 1/A over [lwb,&nbsp;upb] and, with this prior, the
       MML estimate is the sample mean. The Estimator's parameters
       <NOBR> ps = &lang;lwb, upb&rang;, </NOBR> the range of A,
       WARNING, NOT YET TESTED(???).
       Also see {@link #logF logF(N, A)}.  */
   public Est estimator(final Value ps)
    { return new Est(ps)
       { final Tuple bnds = ps.isTuple(2);
         final double lwb = bnds.fst().x(), upb = bnds.snd().x(),
                       k1 = latticeConstant(1);
         final double R_A = upb - lwb;  // A's range;  h(A) ~ 1 / R_A uniform

         /** Given ss = stats(ds) of a data-set, ds, return an M. */
         public M ss2Model(Value ss)
          { final Tuple NSA = ss.isTuple(3);
            final double  N = NSA.fst().x(), sum = NSA.snd().x(),
                      nlAoM = NSA.thrd().x();
            final double lwb2 = Math.max(lwb, Math.min(upb,  // !
                                   N > 0 ? 2 * Math.exp(-nlAoM / N) // say
                                         : lwb));
            double Ax = N >= 1
                        ? Math.max(lwb2, Math.min(upb, sum/N))  // = mean
                        : Math.sqrt(lwb2 * upb);             // geo. mean ???
            Real A = real(Ax);
            M mdl = (M) sp2Model(0, 0, A);
            double msg1 = - Math.log( Math.log(upb/lwb) )  // - log h(Ax)
                        + Math.log(Ax)                     //   --"--
                        + 0.5*logF(N, Ax) + 0.5*Math.log(k1),
                   msg2 = mdl.nlLH(ss) + 0.5;
            assert msg1 >= 0 && msg2 >= 0 : "msg1=" + msg1 + ", msg2=" + msg2;
            return (M) mdl.asGiven(msg1, msg2);
          }//ss2Model(ss)

    }; }//estimator(ps)

   public String toString()
    { return this == Exponential ? "Exponential" : super.toString(); }
   // ..........................

   /** The fully parameterised (negative-) Exponential Model, has
       statistical parameter {@link #A A}, its mean.
       Its variance equals {@link #A A}<SUP>2</SUP>.
       (In some texts it is parameterised with &lambda;=1/A.) */
   public class M extends Continuous.M                      // ExponentialUPM.M
    { /** Statistical parameter 'A', the mean. */
      public  final double A;
      private final double invA, logA;

      /** Note that {@link #A A} is the mean (and standard deviation).  */
      public M(double msg1, double msg2, Value A)
       { super(msg1, msg2, A);
         this.A = ((Value.Cts) A).x();
         invA = 1.0 / this.A;  logA = Math.log(this.A);
       }//constructor

      /** The pdf is
          <NOBR><SUP>1</SUP>/<SUB>A</SUB> e<SUP>-x/A</SUP></NOBR>,
          for x&ge;0. Its negative log is <NOBR>x / A + log A</NOBR>. */
      public double nlPdf_x(double x)
       { if( x < 0 ) error( "nlPdf(" + x + "<0)" );
         return x * invA + logA;
       }
       // pdf = { return Math.exp( - x / A) / A; }

      /** See {@link ExponentialUPM#stats(Vector,int,int) stats(...)}
          of a data-set ds;  for N data, {x<SUB>i</SUB>},
          <NOBR> nlLH = N.log A + (1/A) &sum; x<SUB>i</SUB>. </NOBR>  */
      public double nlLH(Value ss)
       { final Vector ssV = (Vector) ss;
         return logA * ssV.elt(0).x() + invA * ssV.elt(1).x() + ssV.elt(2).x();
             //           N                       sum              sum nlAoM
       }//nlLH(ss)

      /** Return random double from this Model, that is
          {@link #A A}*times;log(x), where x is uniform in [0,1). */
      public double random_x()
       { return - A * Math.log(1 - RNG.nextDouble()); } // also see Laplace
       // cumm., INT_0..x exp(-x/A)/A = {-exp(-x/A)}_[0..x] = 1-exp(-x/A)
       // d = 1-exp(-x/A) => exp(-x/A) = 1-d =>
       // x = -A ln(1-d), d uniform in (0,1], so 1-d uniform in [0,1).

    }//Exponential.M class, the Model

 }//ExponentialUPM class, the UPModel

// (c) L. Allison, 2012, Tue 9 Feb 2016. Licence to ExponentialUPM is Gnu GPL.
