//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.maths.Vector;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The fully parameterised Function Model (aka regression) with
    input&nbsp;datum (independent variable), 'id', and output&nbsp;datum
    (dependent variable), 'od'.  Called 'Function Models' because they are
    like "noisy", statistical functions, id&nbsp;&rarr;&nbsp;od.
    Most instances of FunctionModel
    are likely to be instances of {@link UPFunctionModel.M}.
    The main action is in and around {@link #condModel condModel(id)}.
    Also see {@link Dependent}.  For a simple but useful example of a
    Function Model see {@link CPT} (conditional probability table).
    Note that 'id' and/or 'od' may be multivariate.
    (And, a FunctionModel is a Model, not a {@link la.la.Function Function}.-)
    Also see {@link Model Model} and {@link SeriesModel SeriesModel}.  */
public abstract class FunctionModel extends Model            // * FunctionModel
 { /** Construct a fully parameterised FunctionModel with 2-part message
       length msg1+msg2, and statistical parameter(s), sp.  */
   public FunctionModel(double msg1, double msg2, Value sp)
    { super(msg1, msg2, sp); }//constructor

   /** Return the Model for output datum, od, conditional upon the
       given input datum, id.  <CODE>condModel(id)</CODE> is the
       key thing to define for a new FunctionModel.
       Also see {@link #condNlPr condNlPr(id,od)}
       and {@link #condPr condPr(id,od)}.  */
   public abstract Model condModel(Value id);                // * condModel(id)

   /** Given <SPAN STYLE="white-space:nowrap">iod = &lang;id, od&rang;,</SPAN>
       return -log&nbsp;pr(od|id).
       The input datum, id, is <I>common knowledge</I> and thus "free"
       &ndash; everyone knows&nbsp;id particularly the receiver (MML).
       The output datum, od, is {@link #condPr conditional} upon id.
       (Also see {@link Dependent}.)  */
   public double nlPr(Value iod)                                   // nlPr(iod)
    { Value.Tuple io_od = iod.isTuple(2);
      return condNlPr(io_od.fst(), io_od.snd()); // NB. iod=(id,od)
    }//nlPr(iod)

   /** Given <SPAN STYLE="white-space:nowrap">iod = &lang;id, od&rang;,</SPAN>
       return pr(od|id).
       The input datum, id, is <I>common knowledge</I> and thus "free"
       &ndash; everyone knows&nbsp;id particularly the receiver (MML).
       The output datum, od, is {@link #condPr conditional} upon id.
       (Also see {@link Dependent}.)  */
   public double pr(Value iod)                                       // pr(iod)
    { Value.Tuple io_od = iod.isTuple(2);
      return condPr(io_od.fst(), io_od.snd());
    }//pr(iod)

   /** Return the negative log probability of the output datum, od,
       conditional upon the input datum, id.  This default calls
       <CODE>{@link #condModel condModel(id)}.nlPr(od)</CODE>
       but a sub-class might know a short cut?  */
   public double condNlPr(Value id, Value od)                // condNlPr(id,od)
    { return condModel(id).nlPr(od); }

   /** Base 2 (bits) version of {@link #condNlPr condNlPr}. */
   public double condNl2Pr(Value id, Value od)              // condNl2Pr(id,od)
    { return condNlPr(id, od) * invLog2; }

   /** Return the probability of the output datum, od, conditional upon
       the input datum, id.  Note the order of the parameters.  This default
       calls <CODE>{@link #condModel condModel(id)}.pr(od)</CODE>
       but a sub-class might know a short cut?  */
   public double condPr(Value id, Value od)                    // condPr(id,od)
    { return condModel(id).pr(od); }

   /** Calls {@link #asGiven(double,double) asGiven(0,&nbsp;msg2)}. */
   public FunctionModel asGiven(double msg2)
    { return (FunctionModel) super.asGiven(0, msg2); }

   /** Return a clone of 'this' FunctionModel but with msg1
       and msg2 as specified.
       Also see Model.{@link Model#asGiven(double,double) asGiven(,)}. */
   public FunctionModel asGiven(double msg1, double msg2)
    { return (FunctionModel) super.asGiven(msg1, msg2); }

   /** Treat 'this' fully parameterised <I>FunctionModel</I> as an
       UnParameterised one.
       Also see Model.{@link Model#asUPModel asUPModel()}.  */
   public UPFunctionModel asUPModel()
    { final FunctionModel fm = this; // enclosing FunctionModel
      return new UPFunctionModel(Value.triv)
       { public FunctionModel sp2Model(double msg1, double msg2, Value sp)
          { if( msg1 != 0 ) error( ".sp2Model("+msg1+","+msg2+","+sp+")" );
            return fm.asGiven(msg2); // !
          }//sp2Model(,,)
         public Value stats(Vector ds, int lo, int hi)
          { return fm.stats(ds, lo, hi); }
         public Value stats(boolean add, Value ss0, Value ss1)
          { return fm.stats(add, ss0, ss1); }
         public Estimator estimator(Value ps /* = triv */ )
          { return fm.asEstimator(ps); } // "estimates" fm
         public String toString() { return fm + ".asUPModel()"; }
    }; }//asUPModel()

   /** <CODE>random()</CODE> is inappropriate and throws an Exception
       but see {@link #random(Value) random(id)}.
       Note,&nbsp;that {@link #condModel condModel(id)}
       might implement random(), we hope.  */
   public Value random()
    { throw UOE( ".random() is inappropriate for a FunctionModel" ); }

   /** {@link #condModel condModel(id)}.{@link #random() random()}.
       <I>Given</I> an input datum 'id', generate a random output datum.
       Obviously requires condModel(id) to implement random().  */
   public Value random(Value id)                                  // random(id)
    { return condModel(id).random(); }

 }//FunctionModel (abstract) class

// (c) L.Allison 2012, Tue 15 Aug 2017. Licence to 'FunctionModel' is Gnu GPL.
