//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.util.Util;   // standard functions
import la.la.Value;
import static la.maths.Maths.*;
import la.maths.Vector;
import la.maths.Vector.Slice;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** For most purposes {@link MML#Gamma} and {@link GammaUPM.M} should
    be enough, with little need to call upon GammaUPM directly.
    The&nbsp;&gamma; (gamma) Model (probability distribution) is for
    positive (&gt;0) continuous data.  */
public class GammaUPM extends Continuous                               // Gamma
 { public GammaUPM(Value t)
    { super(t); if( t != triv ) error( ", t=" + t ); }//constructor

   /** Given two-part message lengths m1 and m2, and statistical parameters
       sp, return a fully parameterised {@link GammaUPM.M Model}. */
   public M sp2Model(double m1, double m2, Value sp)
    { return new M(m1, m2, sp); }

   /** Return the data-set, ds.[lo,hi).
       ???TODO: Can we do better? Looks like (&sum; log x, &sum; x)??? */
   public Vector stats(Vector ds, int lo, int hi)
    { return ds.slice(lo, hi); } // TODO: Is there a better one?

   public Vector stats(boolean add, Value ss0, Value ss1)
    { return ((Vector) ss0).combine(add, (Slice) ss1); }

   /** TODO, Gamma's Estimator is not yet implemented. */
   public Est estimator(Value ps)
    { throw UOE( " not yet implemented" ); }

   public String toString()
    { return this == Gamma ? "Gamma" : super.toString(); }
   // .......................................

   /** The fully parameterised &gamma; Model (probability distribution).
       Also see {@link MML#Gamma} and {@link GammaUPM}. */
   public class M extends Continuous.M                         // GammaUPM.M
    {
      /** Shape parameter 'k', and scale parameter &theta; (theta). */
      public final double k, theta;

      /** Negative log normalising constant for pdf. */
      private final double k_1, nlNormaliser;

      /** Statistical parameters, <NOBR>sp =
          &lang;{@link #k k}, {@link #theta &theta;}&rang;.</NOBR> */
      public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp);
         Value.Tuple sp2 = sp.isTuple(2);
         k = sp2.fst().x(); theta = sp2.snd().x();
         assert k > 0 && theta > 0 : "k=" + k + ", theta=" + theta;
         k_1 = k - 1;
         nlNormaliser = logGamma(k) + k*Math.log(theta);
       }//constructor

      /** <NOBR> pdf(x) = {1 / (&Gamma;(k) &theta;<SUP>k</SUP>)}
                          x<SUP>k-1</SUP> e<SUP>-x/&theta;</SUP>. </NOBR> */
      public double nlPdf_x(double x)
       { assert x > 0 : "x=" + x + " <= 0";
         return nlNormaliser - k_1*Math.log(x) + x/theta ;
       }//nlPdf_x(x)

      /** Assumes that {@link GammaUPM#stats(Vector,int,int) statistics}
          are the data-set itself. */
      public double nlLH(Value ss)
       { return sumNlPr((Vector) ss); }

      private M G_k1_1 = null;  // see random_x() below

      /** See Marsaglia &amp; Tsang,
          <I>A simple method for generating Gamma variables</I>,
          ACM Trans on Math Software, 26(3), 363-372, 2000.
          See especially pages 369 to&nbsp;371. */
      public double random_x()
       { if( k < 1 )              // see p.371 of M and T (2000).
          { if( G_k1_1 == null )  // i.e., Gamma(k+1, 1)
               G_k1_1 = new M(0, 0, pair(real(k+1), one/* 1.0 */));
            double U = 1.0 - RNG.nextDouble();  // in (0, 1]
            return theta * G_k1_1.random_x() * Math.exp(Math.log(U) / k);
          }
      // else, k >= 1 ...
         double d = k - 1.0/3.0,
                c = 1.0 / Math.sqrt(9.0 * d),
                u, v, x, x2;
         do{ do{ x = N01.random_x(); v = 1.0 + c*x;
               }while( v <= 0 );
             v = v*v*v;
             u = RNG.nextDouble();  // uniform in [0,1)
             x2 = x*x;
           }while( (u >= 1.0 - 0.0331*x2*x2)
                && (Math.log(u) >= 0.5*x2 + d*(1.0 - v + Math.log(v)) ) );
         // At this point  d*v ~ Gamma(k,1).
         // Now if X ~ Gamma(k, theta)  then cX ~ Gamma(k, c*theta), any c>0;
         // so  if X ~ Gamma(k, 1) then theta*X ~ Gamma(k, theta)
         // (... see http://en.wikipedia.org/wiki/Gamma_distribution).
         return theta * d * v;
       }//random_x()

    }//class GammaUPM.M, the fully paramaterised Model

 }//class GammaUPM, the UnParameterised Model

// (c) Lloyd Allison, Feb 2014, 30 Oct 2014. Licence to 'GammaUPM' is Gnu GPL.
