//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.util.Util;   // standard functions
import la.la.Value;
import static la.maths.Maths.*;
import la.maths.Vector;
import la.maths.Vector.Derived;
import la.maths.Vector.Slice;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** Intervals, the UnParameterised FunctionModel. For a data-set
    (id,&nbsp;od)<SUP>*</SUP>, where the data-space of the input
    datum id, is totally ordered, divide the id-space into one or more
    {@link M#interval intervals} (buckets) and give a conditional Model
    of od corresponding to each bucket.  Note that the input datum may
    be continuous in which case it is "discretized." However it is only
    required to have a total order. The output datum, od, can be of any type
    at all, including Multivariate.  The fully parameterised FunctionModel
    is&nbsp;{@link M M}. */
public class Intervals extends UPFunctionModel                     // Intervals
 {
   /** upm, the UnParameterised Model for the output (dependent)
       datum, od.  upm&nbsp;will be {@link M#mdls parameterised}
       differently for each interval of the input datum, id. */
   public final UPModel upm;

   /** The problem defining parameter, {@link #upm upm}, is an
       UnParameterised Model suitable for the ouput datum, od.
       upm will be parameterised differently for each interval
       (bucket) of the input datum, id. */
   public Intervals(Value upm)
    { super(upm); this.upm = (UPModel) upm; }//constructor

   /** Given two-part message lengths, msg1 and msg2, and statistical
       parameters, sp, return a fully parameterised {@link M M}. */
   public M sp2Model(double msg1, double msg2, Value sp)
    { return new M(msg1, msg2, sp); }

   /** Sufficient statistics, ss&nbsp;=&nbsp;stats(ds), for a
       data-set, ds, are ds itself, sorted on the input data. */
   public Vector stats(Vector ds, int lo, int hi)
    { return ds.slice(lo, hi).sorted(/* column = */ 0); }
    // Incrementally, can do ss0.parent().append(whatever).sorted() --
    // see Vector.Derived.

   public Vector stats(boolean add, Value ss0, Value ss1)
    { Vector ss0v = ((Derived) ss0).parent();
      Slice  ss1v = (Slice) ((Derived) ss1).parent();
      return ss0v.combine(add, ss1v).sorted(0);
    }

   /** Return an Estimator for {@link M M}.  Parameters, ps, are
       passed to <TT>{@link #upm upm}.estimator(ps)</TT>. */
   public Est estimator(final Value ps)
    { return new Est(ps)
       { public M ss2Model(Value ss)
          { Estimator est = upm.estimator(ps);
            Vector iods = (Vector) ss;    // the data-set sorted on id
            Vector  ids = iods.col(0),    // input  data, ids
                    ods = iods.col(1);    // output data, ods
            int   nElts = iods.nElts(),
                      N = 1;              // # candidate cut pts, [1, N),
            int [] p = new int[nElts+1];  // where ids[] increases...
            p[0] = 0;      // i.e., first data index
            for( int i = 1; i < nElts; i ++ )
               if( ids.elt(i-1).compareTo(ids.elt(i)) < 0 ) // an increase
                { p[N] = i; N ++ ; }
            p[N] = nElts;  // i.e., last data index + 1
            double [] [] lh = new double [N+1] [];  // costs, lh[to][from]
            for( int j = 1; j <= N; j ++ )
               lh[j] = new double[j];  // triangular
            for( int i = 0; i < N; i ++ )       // 0 <= i < j <= N
             { Value ss_ij = upm.stats(ods, i, i);

               for( int j = i+1; j <= N; j ++ ) // NB. [p[i], p[j])
                { // O(N**2) if upm does incremental sufficient statistics...
                  ss_ij = upm.stats(true, ss_ij, ods, j-1, j);    // ...well.
                  lh[j][i] = est.ss2Model(ss_ij).msg();
           // was lh[j][i] = est.ds2Model( ods.slice(p[i], p[j]) ) .msg();
             }  }
            int    [] prev = new int [N+1];    // tracing back
            double [] best = new double [N+1]; // partial costs
            double cp_cost = Math.log(Math.max(1, N)); // approx!
            for( int j = 1; j <= N; j ++ )     // dynamic prog...
             { prev[j] = 0;         // i.e.,
               best[j] = lh[j][0];  // [0, p[j]) direct v. ...
               for( int i = 0; i < j; i ++ ) // best[0..i) + [i..j)
                { final double m = cp_cost + best[i] + lh[j][i];
                  if( m < best[j] ) { prev[j] = i; best[j] = m; }
                }//for i
             }//for j

            int Ninterval = 0;
            for( int cut = N; cut > 0; cut = prev[cut] ) Ninterval ++ ;
            int Ncut = Ninterval - 1;         // Ninterval >= 1, Ncut >= 0
            Value [] cuts = new Value[Ncut],  sps = new Value[Ninterval];
            double msg1 = WallaceInt0.nlPr_n(Ncut)  // # cut points,
                        + log_nCk(N, Ncut),    // each equiprobable
                   msg2 = 0;
            int upb = N;  // elts [?, p[upb]), i.e., [?, nElts)
            for( int ni = Ncut; ni >= 0; ni -- )
             { final int lwb = prev[upb]; // [p[lwb], p[upb])
               Model m = est.ds2Model( ods.slice(p[lwb], p[upb]) );
               msg1 += m.msg1();  msg2 += m.msg2();
               sps[ni] = m.statParams(); // Models' stat.params.
               if( ni < Ncut ) cuts[ni] = ids.elt(p[upb]); // 1 fewer
               upb = lwb;
             }//for ni
            return new M(msg1, msg2,
                         pair(Vector.values(cuts), Vector.values(sps)));
          }//ss2Model(ss)
    }; }//estimator(ps)
   // .........................................................................

   /** A fully parameterised FunctionModel in which the data-space of
       the input datum, id, is cut into {@link #cutPts intervals}, and
       for each interval there is a (conditional) {@link #mdls Model}
       of the output datum, od.  The UnParameterised FunctionModel
       is {@link Intervals here}. */
   public class M extends UPFunctionModel.M                      // Intervals.M
    {
      /** The N-1 ordered cut-points, N&ge;1, dividing the data-space
          of the input datum, id, into N intervals (buckets),
          {&lt;cp<SUB>0</SUB>, [cp<SUB>0</SUB>,cp<SUB>1</SUB>), ...,
          &ge;cp<SUB>N-2</SUB>}.  Also see {@link #mdls mdls}. */
      public final Vector cutPts;
      /** The N Models of the output datum, od, one for
          each {@link #cutPts interval} of id, N&ge;1. */
      public final Model  [] mdls;
      /** M's statistical parameters,
      sp&nbsp;=&nbsp;({@link #cutPts cutPts},sps), where sps are the
          statistical parameters for instances of {@link #upm upm}. */
      public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp);
         Tuple cm = sp.isTuple(2);
         cutPts = (Vector) cm.fst();
         Vector sps = (Vector) cm.snd();
         mdls = new Model[ sps.nElts() ];
         for( int i = 0; i < mdls.length; i ++ )
            mdls[i] = upm.apply(sps.elt(i));  // parameterised
         if( cutPts.nElts() != mdls.length-1 ) error( " lengths" );
       }//constructor

      /** Assumes that {@link #stats(Vector,int,int) statistics} 'ss'
          are the data-set itself (actually sorted). */
      public double nlLH(Value ss)
       { return sumNlPr((Vector) ss); }//nlLH(ss)

      /** Return 'i' where input datum 'id' falls into {@link #cutPts interval}
          'i', 0&nbsp;&le;&nbsp;i&nbsp;&lt;&nbsp;N.  */
      public int interval(Value id)                             // interval(id)
       { int lo = 0, hi = mdls.length; // e.g.,
         int mid = (lo + hi) / 2;      // (0,0,1) or (0,1,2) or (0,1,3) etc
         while( lo < mid )  // binary search for right Model...
          { // lo < mid < hi.  NB. one fewer cut-pts than Models
            if( id.compareTo(cutPts.elt(mid-1)) < 0 ) hi = mid; // NB. mid-1
            else /* id >= cp[mid] */ lo = mid;
            mid = (lo + hi) / 2;
            assert lo <= mid && mid < hi : "lo="+lo+" mid="+mid+" hi="+hi;
          }//lo = mid = hi-1
         return mid;
       }//interval(id)

      /** Return {@link #mdls mdls[}{@link #interval interval(id)]}.
          That is, return the Model of od conditional on id.  */
      public Model condModel(Value id)
       { return mdls[interval(id)]; }

    }//Intervals.M class, the FunctionModel

 }//Intervals class, the UPFunctionModel

// (c) Lloyd Allison, 2012, Tues 3 December 2013. Licence is Gnu GPL.
