//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.maths.Vector;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The class of UnParameterised Function Models (regressions) from a
    {@link la.la.Value.Cts Continuous} input datum (independent variable),
    'x', to a Continuous output datum (dependent variable),&nbsp;'y',
    <B>R</B>&rarr;<B>R</B>, y=ax+b+N(0,&sigma;).
    Having the trivial problem defining parameter, we really only
    need one {@link #instance instance} of the UnParameterised Model.
    Also see the fully {@link Linear1.M parameterised} Linear Model.  */
public class Linear1 extends UPFunctionModel                         // Linear1
 {
   /** Having a trivial problem-defining parameter, we really
       only need one instance of the UnParameterised Linear1 Model.  */
   public static final Linear1 instance = new Linear1(triv);        // instance

   /** The problem defining parameter dp is {@link #triv triv}-ial.  */
   public Linear1(Value dp)
    { super(dp);
      if( dp != triv ) error( " constructor(" + dp + ")" );
    }//constructor

   /** Given a data-set, ds, calculate sufficient statistics of ds[lo,hi),
       that is the quantities 'N' and the sums over all data
       &lang;x,y&rang;<SUP>*</SUP> of each of the following,
       x, x<SUP>2</SUP>, xy, y, y<SUP>2</SUP> and y.nlAoM.  */
   public Vector.Doubles stats(Vector ds, int lo, int hi)
    { if( ds.nCols() != 2 )
        error( ".stats(ds,lo,hi), ds not 2-columns" );
      Vector xs = ds.col(0), ys = ds.col(1);
      double x = 0, x2 = 0, xy = 0, y = 0, y2 = 0; // initialse the sums
      for( int i = lo; i < hi; i ++ )
       { double wi = ds.wt(i), xi = xs.elt(i).x(), yi = ys.elt(i).x();
         x += xi*wi;  x2 += xi*xi*wi;  xy += xi*yi*wi;
         y += yi*wi;  y2 += yi*yi*wi;
       }//for i
      return Vector.doubles(
         new double [] {ds.wts(lo,hi),         // i.e., N,
                        x, x2, xy, y, y2,      // interesting things,
                        ys.wtNlAoM(lo, hi)});  // nlAoM of ys
    }//stats(ds,lo,hi)

   /** Combine sufficient {@link #stats(Vector,int,int) statistics}
       ss0 and ss1 into ss0&plusmn;ss1, either by addition (add=true) or
       by subtraction (add=false).  */
   public Vector stats(boolean add, Value ss0, Value ss1)
    { return (Vector)
        ((Vector) ss0) . bOp(add?la.la.Lexical.plus:la.la.Lexical.minus,
         (Vector) ss1    );
    }// stats(add,ss0,ss1)

   /** Return the fully parameterised Linear1 {@link M Model}
       having statistical parameters sp. */
   public M apply(Value sp) { return new M(0, 0, sp); }

   /** Return the fully parameterised Linear1 {@link M Model} having
       two-part message lengths m1 and m2 and statistical parameters sp. */
   public M sp2Model(double m1, double m2, Value sp)
    { return new M(m1, m2, sp); }

   /** Return the [{@link Est estimator}]. */
   public Est estimator(Value ps) { return new Est(ps); }

   // ...................................................
   /** An Estimator of <NOBR> a, b and &sigma; </NOBR> in
       <NOBR> y = a x +b + N(0,&sigma;)</NOBR>, where estimator parameter
       <NOBR> ps = &lang;b<SUB>min</SUB>, b<SUB>max</SUB>,
                   &sigma;<SUB>min</SUB>, &sigma;<SUB>max</SUB>&rang; </NOBR>.
       It is possible to change the {@link #nlPrior prior(a,b,&sigma;)} by
       overriding it in a subclass of Est.  */
   public class Est extends UPFunctionModel.Est                  // Linear1.Est
    { /** Bounds on 'b'; {@link #nlPrior prior}, pr(b) is uniform.  */
      public final double bMin, bMax;
      /** Bounds on &sigma;; {@link #nlPrior prior},
          <NOBR> pr(&sigma;) ~ 1/&sigma; </NOBR>.  */
      public final double sigmaMin, sigmaMax;
      /** bMax-bMin and log(sigmaMax)-Math.log(sigmaMin) resp.   */
      protected double R_b, R_s;
      /** {@link MML#latticeConstant latticeConstant(3)}. */
      protected double k3;

      /** Parameter ps = &lang;bMin, bMax, sigmaMin, sigmaMax&rang;.  */
      public Est(Value ps)
       { super(ps);
         final Tuple bds = ps.isTuple(4);
         bMin = bds.fst().x();  bMax = bds.snd().x();
         sigmaMin = bds.thrd().x(); sigmaMax = bds.frth().x();
         assert bMin < bMax && 0 < sigmaMin && sigmaMin < sigmaMax
              : errMsg( ", Est(ps)" );
         R_b = bMax - bMin;                             // pr(b) = 1/R_b
         R_s = Math.log(sigmaMax) - Math.log(sigmaMin); // pr(s) ~ 1/s
         k3  = latticeConstant(3);
       }//constructor

      /** The negative log prior on <NOBR> &lang;a, b, &sigma;&rang;</NOBR>.
          This is,
          uniform on the slope-angle arctan(a) in (-&pi;/2,&pi;/2),
          uniform on b over its range, and
          1/&sigma; on &sigma; over its range.
          A subclass can override the (-log)&nbsp;prior if necessary.  */
      public double nlPrior(double a, double b, double sigma)
       { final double phi  = Math.atan(a);  // slope angle, in (-pi/2, +pi/2)
         return logPI - 2*Math.log(Math.cos(phi)) // ... a,
              + Math.log(R_b * R_s);              // b & s
       }//nlPrior(,,)

      /** The log of the Fisher information from N, &sigma;,
          mean(x<SUB>i</SUB><SUP>2</SUP>), and mean(x<SUB>i</SUB>).
          (Remember that the x<SUB>i</SUB> are "common knowledge.")  */
      public double logF(double N, double sigma, double x2bar, double xbar)
       { return 3*Math.log(N) - 6*Math.log(sigma)
              + log2 + Math.log(x2bar - xbar*xbar);
       }//logF(...)

      /** Given sufficient statistics <NOBR> ss =
          {@link Linear1#stats(Vector,int,int) stats(ds)} </NOBR> of a
          data-set ds, estimate a fully parameterised {@link M Model}.  */
      public M ss2Model(Value ss)
       { final Vector.Doubles ssV = (Vector.Doubles) ss;
         double  N = ssV.elt(0).x(),  // unpack ss ...
                 x = ssV.elt(1).x(),
                x2 = ssV.elt(2).x(),
                xy = ssV.elt(3).x(),
                 y = ssV.elt(4).x(),
                y2 = ssV.elt(5).x(),
             nlAoM = ssV.elt(6).x();
         double xbar = x/N, ybar = y/N;
         double a  = (xy - ybar*x) / (x2 - xbar*x);
         double b  = ybar - a*xbar;
                b  = Math.max(bMin, Math.min(bMax, b)); // & bounds
         double s2 = (y2 + a*a*x2 + N*b*b - 2*a*xy - 2*b*y + 2*a*b*x)
                   / (N - 1);
         double s  = Math.sqrt(s2);
                s  = Math.max(sigmaMin, Math.min(sigmaMax, s)); // & bounds
         Tuple sp = triple(real(a), real(b), real(s));  // stat params
         M mdl = Linear1.this.sp2Model(0, 0, sp);
         double phi  = Math.atan(a);  // slope angle, in (-pi/2, +pi/2)
         double msg1 = nlPrior(a, b, s)              // neg log prior
                     + 0.5 * logF(N, s, x2/N, xbar)  // log(F) / 2
                     + Math.log(k3),                 // lattice k
                msg2 = mdl.nlLH(ss) + 1.5;  // NB. + 3 params * 0.5 nits
         assert msg1 >=0 && msg2 >= 0 : "msg1=" + msg1 + ", msg2=" + msg2;
         return new M(msg1, msg2, sp);
       }//ss2Model(ss)
    }//Est class

   // ................................................
   /** A fully parameterised Linear1 Model,
       <NOBR> y = {@link #a a} * x + {@link #b b}
            + {@link MML#Normal N}(0, {@link #sigma &sigma;}) </NOBR>.
       Also see the UnParameterised {@link Linear1} Model.  */
   public class M extends UPFunctionModel.M                        // Linear1.M
                  implements HasPdf
    { /** Statistical parameters of the Model
          <NOBR> y = a &times; x + b + N(0, &sigma;). </NOBR>  */
      public  final double a, b, sigma;
      /** Real version of sigma for condModel.  */
      private final Real sigmaR;
      /** log(2&pi;)/2+log(&sigma;), and
          2 &sigma;<SUP>2</SUP> (used by {@link #nlLH}).  */
      private final double log_root_2_pi_sigma, two_sigma_2;
      /** Two-part message lengths msg1 and msg2, and statistical parameters
          <NOBR>sps = &lang;{@link #a a}, {@link #b b},
                            {@link #sigma &sigma;}&rang;</NOBR>. */
      public M(double msg1, double msg2, Value sps)
       { super(msg1, msg2, sps);
         Tuple sps3 = sps.isTuple(3);
         a = sps3.fst().x(); b = sps3.snd().x();
         sigmaR = (Real) sps3.thrd();  sigma = sigmaR.x();
         log_root_2_pi_sigma = 0.5*(log2 + logPI) + Math.log(sigma);
         two_sigma_2         = 2.0*sigma*sigma;
       }//constructor

      // ??? TODO: can short-cut condNlPr(x,y) quicker than via condModel

      /** Conditional Model of output datum y given input datum x, i.e., the
          {@link NormalUPM.M Normal}, <NOBR>N(ax + b, &sigma;) </NOBR>.  */
      public NormalUPM.M condModel(Value x)
       { double xx = x.x();
         return Normal.sp2Model(0, 0, pair(real(a * xx + b), sigmaR));
       }//condModel(id)

      /** Negative log pdf of 'y' {@link #condModel given}&nbsp;'x'. */
      public double nlPdf(Value xy)
       { Tuple xyT = xy.isTuple(2);
         return condModel(xyT.fst()).nlPdf(xyT.snd());
       }//nlPdf(xy)

      /** pdf of 'y' {@link #condModel given}&nbsp;'x'. */
      public double pdf(Value xy)
       { Tuple xyT = xy.isTuple(2);
         return condModel(xyT.fst()).pdf(xyT.snd());
       }//nlPdf(xy)

      /** Given a data-set, <NOBR> ds = zip xs ys, </NOBR> having sufficient
          <NOBR> statistics ss = {@link Linear1#stats stats(ds)}, </NOBR>
          return the negative log likelihood of the ys given the xs.
          ({@link FunctionModel#pr Recall} that the input datum (variable),
          x, is common knowledge.)  */
      public double nlLH(Value ss)
       { final Vector.Doubles ssV = (Vector.Doubles) ss;
         double  N = ssV.elt(0).x(),  // unpack ss ...
                 x = ssV.elt(1).x(),
                x2 = ssV.elt(2).x(),
                xy = ssV.elt(3).x(),
                 y = ssV.elt(4).x(),
                y2 = ssV.elt(5).x(),
             nlAoM = ssV.elt(6).x();
         return N * log_root_2_pi_sigma
              + (y2 + a*a*x2 + N*b*b - 2*a*xy - 2*b*y + 2*a*b*x) / two_sigma_2
              + nlAoM;
       }//nlLH(ss)

    }//Linear1.M class, the fully parameterised FunctionModel

 }//Linear1 class, the UPFunctionModel

//(c) L.Allison Nov 2016 & Wed 10 May 2017. Licence to 'Linear1' is Gnu GPL.
