//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.maths.Vector;
import la.maths.Vector.Slice;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** Model {@link MML#logStar0} should be enough for most purposes,
    but here are the classes, UnParameterised (LogStar0UPM) and fully
    (trivially) {@link LogStar0UPM.M parameterised}&nbsp;(M).  */
public class LogStar0UPM extends Discretes                       // LogStar0UPM
 { public LogStar0UPM(Value dp)
    { super(dp);  if( dp != triv ) error( ", dp=" + dp); }//constructor

   /** Return <NOBR>ss = ds.[lo, hi)</NOBR> that is the data itself
       as sufficient statistics of ds.
       More on stats {@link UPModel#stats here}. */
   public Vector stats(Vector ds, int lo, int hi) { return ds.slice(lo, hi); }

   public Vector stats(boolean add, Value ss0, Value ss1)
    { return ((Vector) ss0).combine(add, (Slice) ss1); }

   /** Return {@link MML#logStar0 logStar0} essentially, with msg2 set.
       Note that msg1&nbsp;=&nbsp;0 and sp&nbsp;=&nbsp;triv. */
   public M sp2Model(double msg1, double msg2, Value sp)
    { return new M(msg1, msg2, sp); }

   /** Return the "trivial" Estimator of {@link MML#logStar0 logStar0}. */
   public Estimator estimator(Value t)
    { return Mdl.asEstimator(t); }

   public String toString() { return "logStar0upm"; }
   // .....................

   /** Note, {@link MML#logStar0 logStar0} is Mdl. */
   public final M Mdl = new M(0, 0, triv);

   /** Model {@link MML#logStar0 logStar0} should be enough for most purposes,
       but here are the classes, fully (trivially) parameterised&nbsp;(M)
       and UnParameterised ({@link LogStar0UPM LogStar0UPM}). */
   public class M extends Discretes.M                          // LogStar0UPM.M
    { /** Note, requires msg1=0 and sp=triv. */
      public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, Value.triv); zeroTriv(msg1, sp); }//constructor

      /** Negative log probability of datum int, <NOBR>n &ge; 0</NOBR>. */
      public double nlPr_n(int n)
       { if( n < 0 ) error( ".pr_n(" + n + ")" );
         double  n1 = n+1;    // n1 >= 1
         double sum = 2.865;  // Do it in bits ...
         while( n1 > 1 ) { n1 = Math.log(n1) * invLog2; sum += n1; }
         return sum * log2;   // lastly, bits to nits
       }//nlPr_n(n)

      /** Assumes that {@link LogStar0UPM#stats(Vector,int,int) statistics}
          are the data-set itself. */
      public double nlLH(Value ss)
       { return sumNlPr((Vector) ss); }

      /** random_n() is not supported. */
      public int random_n()
       { throw UOE( ".random_n() is not yet(?) supported" ); }

      public String toString()
       { return  this == logStar0 ? "logStar0" : super.toString(); }

    }//LogStar0UPM.M class, the Model

 }//LogStar0UPM class, the UPModel

// (c) Lloyd Allison 2012, Fri 1 Aug 2014. Licence to 'LogStar0UPM' is Gnu GPL.
