//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import java.util.Random; // for RNG, for use in Model.random()

import la.la.Value;
import la.maths.Vector;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference:
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The class of Minimum Message Length [MML]
    tools for statistical and inductive inference, and machine learning.
    The main business is in (fully parameterised) {@link Model Model},
    {@link FunctionModel FunctionModel} and {@link SeriesModel SeriesModel}.
    Each of these may be created by an {@link UPModel UnParameterised}
    Model or its {@link Estimator Estimator}.<BR>
    Also see package mml's {@link README},
    [<A HREF="https://doi.org/10.1007/978-3-319-76433-7"
        TARGET="BOOKA">the&nbsp;book</A>]
    and the
    [<A HREF="https://www.cantab.net/users/mmlist/MML/A/"
        TARGET="mmlist">home&nbsp;page</A>].
*/
public class MML                                                         // MML
 {                                      /** log<SUB>e</SUB>(2) */
   public final static double    log2 = Math.log( 2.0 );
                                        /** 1 / log<SUB>e</SUB>(2) */
   public final static double invLog2 = 1 / log2;
                                        /** log<SUB>e</SUB>(&pi;) */
   public final static double   logPI = Math.log(Math.PI);
                                        /** log<SUB>e</SUB>(2&pi;). */
   public final static double  log2PI = Math.log(2 * Math.PI);

   /** Standard practice is for those Models that do implement
       {@link Model#random()} to use RNG so that a run can be repeated
       by setting RNG's {@link Random#setSeed seed} to a known value. */
   public final static Random RNG = new Random();                        // RNG

   // NB. MML has no constructor.

   /** latticeConstant(D) (aka &kappa;(D)), the lattice constant for
       D&nbsp;dimensions, D&nbsp;&ge;&nbsp;1, i.e., D&nbsp;parameters,
       [<A HREF="http://www.allisons.org/ll/MML/Notes/Fisher/"
           TARGET="LL">www</A>].  */
   public static double latticeConstant(int D)
    { final double [] k = new double[]
         { 0, // "non-entry" for D=0 parameters
           1/12.0,   5/(36.0*Math.sqrt(3.0)),  // for D = 1..8
           0.078543, 0.076603, 0.075625,
           0.074244, 0.073116, 0.071682 };
      final double k100 = 0.06080, k1000 = 0.05891;  // actually lower bounds
      assert D >= 0 : "D=" + D + " < 0 in latticeConstant(D)";
      final int l1 = k.length-1;
      return D < k.length ? k[D] // else, D > l1, linear interpolation...
           : D <=  100 ? (k[l1]*(100-D) +  k100*(D-l1) ) / (100.0 - l1)
           : D <= 1000 ? (k100*(1000-D) + k1000*(D-100)) / 900.0
           : 1 / (2.0 * Math.PI * Math.E);  // = 0.05855
    }//latticeConstant(D)

   /** See p.180 Wallace (2005), the informative explanation
       (D&nbsp;parameters estimated) versus the uninformative explanation,
       <NOBR>I1 - I0 = (D/2)(1 + log(2 &pi; k[D])),</NOBR> nits. This begins
       <NOBR>(1/2)log(2 &pi; e / 12)</NOBR> when D=1, and note that
       <NOBR>(1 + log(2 &pi; k[D]) &ge; 0</NOBR> for all D&gt;0.
       There is a <I>small</I> price to pay for stating an opinion about
       the value of a Model's parameter(s) compared to a message that
       does not state any such opinion. */
   public static double informativeIncrement(int D)
    { return (D / 2.0) * (1.0 + Math.log(2.0*Math.PI*latticeConstant(D)) ); }
   // -------------------------------------------------------------------------

   /** A Model of given, known, common knowledge. That is,
       pr(d)=1, nlPr(d)=0, for all data d. */
   public final static Model Known = new KnownClass(0, 0, Value.triv); // Known

   /** The unParameterised Model version of {@link #Known}. */
   public final static UPModel KnownUPM = Known.asUPModel();

   /** <I>The</I> UnParameterised Wallace Model for integers
       n&nbsp;&ge;&nbsp;0.  {@link #WallaceInt0 WallaceInt0}
       is fully (trivially) parameterised. */
   public static final WallaceInt0UPM  WallaceInt0upm =
      new WallaceInt0UPM(Value.triv);

   /** <I>The</I> fully (trivially) parameterised Wallace Model
       for integers n&nbsp;&ge;&nbsp;0.  If needed,
       {@link #WallaceInt0upm WallaceInt0upm} is UnParameterised. */
   public static final WallaceInt0UPM.M  WallaceInt0 = WallaceInt0upm.Mdl;

   /** <I>The</I> UnParameterised log<SUP>*</SUP> Model for integers
       n&nbsp;&ge;&nbsp;0.  {@link #logStar0 logStar0} is fully
       (trivially) parameterised.
       Also see {@link #WallaceInt0upm WallaceInt0upm}. */
   public static final LogStar0UPM logStar0upm = new LogStar0UPM(Value.triv);

   /** <I>The</I> fully (trivially) parameterised log<SUP>*</SUP> Model,
       or "universal" probability distribution, for integers
       n&nbsp;&ge;&nbsp;0.  Note, the Model has infinite mean &ndash; does
       your data source? If needed, {@link #logStar0upm logStar0upm} is
       UnParameterised.  Also see {@link #WallaceInt0 WallaceInt0}. */
   public static final LogStar0UPM.M logStar0 = logStar0upm.Mdl;

   /** <I>The</I> UnParameterised Geometric Model (distribution)
       over integers in [<B>0</B>,&nbsp;&infin;), capable of
       {@link Geometric0UPM#estimator estimating} a fully parameterised
       {@link Geometric0UPM.M Geometric} Model.  Also see
       [<A HREF="http://www.allisons.org/ll/MML/Discrete/Integers/Geometric/"
           TARGET="LL">www</A>].  */
   public static final Geometric0UPM Geometric0 = new Geometric0UPM(Value.triv);

   /** <I>The</I> UnParameterised Poisson Model (distribution)
       over integers in [0,&nbsp;&infin;), capable of
       {@link Poisson0UPM#estimator estimating} a fully parameterised
       {@link Poisson0UPM.M Poisson} Model.  Also see
       [<A HREF="http://www.allisons.org/ll/MML/Discrete/Integers/Poisson/"
           TARGET="LL">www</A>].  */
   public static final Poisson0UPM Poisson0 = new Poisson0UPM(Value.triv);

   /** <I>The</I> UnParameterised Laplace probability distribution. If you
       really want its class, that is {@link LaplaceUPM here}. Also see
       fully parameterised, {@link LaplaceUPM.M M<SUB>&mu;,b</SUP>}. */
   public static final LaplaceUPM Laplace = new LaplaceUPM(Value.triv);    // !

   /** <I>The</I> UnParameterised {@link ExponentialUPM Exponential} Model. */
   public static final ExponentialUPM Exponential
     =  new ExponentialUPM(Value.triv);

   /** <I>The</I> UnParameterised {@link GammaUPM Gamma} Model. */
   public static final GammaUPM Gamma = new GammaUPM(Value.triv);      // Gamma

   /** <I>The</I> UnParameterised {@link BetaUPM Beta} Model. */
   public static final BetaUPM Beta = new BetaUPM(Value.triv);          // Beta

   /** <I>The</I> UnParameterised {@link NormalUPM Normal} Model (Gaussian
       distribution) capable of {@link NormalUPM#estimator estimating}
       both &mu; and &sigma; together to give a fully parameterised
       Normal {@link NormalUPM.M M Model}.  Also see {@link NormalMu}, and
       [<A HREF="http:www.allisons.org/ll/MML/Continuous/Normal/"
           TARGET="LL">www</A>].  */
   public static final NormalUPM Normal = new NormalUPM(Value.triv);  // Normal
   // NB. We do need an explicit 'NormalUPM' class to get javadoc-umentation!

   /** The fully parameterised Normal Model,
       N<SUB>&lang;&mu;=0,&sigma;=1&rang;</SUB>; it might be useful. */
   public static final NormalUPM.M N01 =                                 // N01
       Normal.apply(Value.pair(Value.zeroR, Value.oneR)); // i.e., N(0,1)

   /** <I>The</I> UnParameterised {@link NormalMu NormalMu} Model with
       &mu;&nbsp;=&nbsp;0, given, and &sigma; unset.
       This is a reasonably common, useful special case.
       Also see {@link #N01 N<SUB>0,1</SUB>}.  */
   public static final NormalMu NormalMu0 =                        // NormalMu0
      new NormalMu(/* mu = */ Value.zeroR);

   /** <I>The</I> UnParameterised {@link Model.Transform Transformed} Model
       that is the {@link la.la.Library#log log}&ndash;{@link #Normal Normal}
       Model.  It is for Cts data in (0,&nbsp;&infin;) whose log follows
       a {@link #Normal Normal} distribution.  Its statistical parameters
       and the parameters of its estimator are as per the Normal.  */
   public static final Continuous logNormal =                      // logNormal
      Normal.transform(la.la.Library.log);

   /** <I>The</I> {@link mml.Linear1#instance instance} of the
       UnParameterised {@link Linear1 Linear1} FunctionModel;
       (Cts&rarr;Cts, linear regression). Also see the
       fully parameterised {@link Linear1.M}.  */
   public static final Linear1 Linear1 = mml.Linear1.instance;       // Linear1

   /** <I>The</I> {@link vMF UnParameterised} von&nbsp;Mises - Fisher
       Model (distribution) of Directions in <B>R</B><SUP>3</SUP>. */
   public static final vMF vMF3 = new vMF(Value.three);                 // vMF3
   // -------------------------------------------------------------------------

   /** Runs {@link Test}.main(argv).  */
   public static void main(String[] argv)                               // main
    { System.out.println("-- test MML --");
      Test.main(argv);
      System.out.println("-- done MML --");
    }//main

 }//MML class   https://www.cantab.net/users/mmlist/MML/A/

// (c)Lloyd Allison, 2012, May 2017, 10 Jan 2020. Licence to 'MML' is Gnu GPL.
