//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Lexical;
import la.la.Value;

import la.maths.Vector;
import la.maths.Matrix;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** A Markov Model of a given order &mdash; the UnParameterised Series Model.
    The fully parameterised Series Model is {@link Markov.M Markov.M}.  */
public class Markov extends UPSeriesModel.Length                      // Markov
 {
   /** The Markov Models are of a given 'order', over Series of
       data, [lwb,&nbsp;upb]<SUP>*</SUP> (bounds as ints).  */
   public final int order, lwb, upb;

   private final UPModel lenUPM;      // U P Model of lengths of data Series

   private final Value.Tuple bnds;
   private final MultiState MS;       // UPModel over [lwb,upb]*
   private final int range, range_o;  // range_o = range**order

   /** The problem defining parameter(s)
       dp = ({@link #lenUPM lenUPM}, {@link #order order},
             {@link #lwb lwb}, {@link #upb upb}). */
   public Markov(Value dp)
    { super(dp); Tuple ob = dp.isTuple(4);
      lenUPM  = (UPModel) ob.fst();               // length Model
      order = ob.snd().n();                     // the order
      lwb = ob.thrd().n(); upb = ob.elt(3).n(); // bounds
      bnds = pair(ob.thrd(), ob.elt(3));
      MS = new MultiState(bnds);
      range = upb - lwb + 1;
      int ro = 1;
      for( int i = 0; i < order; i ++ ) ro *= range;
      range_o = ro; // e.g. lwb=2, upb=5, order=2, range_o=16, say
    }//constructor

   public UPModel lenUPM() { return lenUPM; /* lengths */ }

   /** Given two part message lengths, msg1 and msg2, and
       statistical parameters, sp, return a fully
       {@link M parameterised} Markov Model. */
   public M sp2Model(double msg1, double msg2, Value sp)
    { return new M(msg1, msg2, sp); }

   /** Collect statistics, ss&nbsp;=&nbsp;stats(seqs,lo,hi), of
       a data-set, that is of a Vector of Vectors, seqs.
       Calls upon {@link #lenUPM} vai {@link #lenStats}.  */
   public Value stats(Vector seqs, int lo, int hi)
    { double [] [] freq = new double [range_o] [range];
      for( int i = 0; i < range_o; i ++ )
         for( int j = 0; j < range; j ++ ) freq[i][j] = 0;
      double uniformPart = 0;
      for( int i = lo; i < hi; i ++ )
       { int cntxt = 0;
         final Vector vi = (Vector) seqs.elt(i);
         final double wi = seqs.wt(i);
         for( int j = 0; j < vi.nElts(); j ++ )
          { final int vij = vi.elt(j).n();
            if( j < order ) uniformPart += wi; // uniform part
            else freq[cntxt][vij - lwb] += wi; // context Models part
            cntxt = cntxt(cntxt, vij);
          }//for j
       }//for i
      return triple( lenStats(seqs, lo, hi),   // lengths
                     real(uniformPart),      // for UniformMdl
                     Matrix.doubles(freq) ); // for context Models
    }//stats(seqs,lo,hi)

   /** Calls upon stats(add,.,) of {@link #lenUPM}.  */
   public Value stats(boolean add, Value ss0, Value ss1)
    { Value.Tuple ss0t = ss0.isTuple(3), ss1t = ss1.isTuple(3);
      return triple( lenUPM.stats(add, ss0t.fst(), ss1t.fst()),   // lengths
                     real(add?ss0t.snd().x()+ss1t.snd().x()
                             :ss0t.snd().x()-ss1t.snd().x()), // for UniformMdl
                     ((Matrix)ss0t.thrd()).bOp(add?Lexical.plus:Lexical.minus,
                                               (Matrix)ss1t.thrd()) );
    }//stats(add,ss0,ss1)

   /** Return the Markov Model Estimator. Parameters ps, if any,
       are passed to {@link #lenUPM lenUPM}. */
   public Est estimator(final Value ps)
    { return new Est(ps)
       { public M ss2Model(final Value ss)
          { Value.Tuple sst = ss.isTuple(3);
            Model  lenMdl = lenUPM.estimator(ps).ss2Model(sst.fst());
            double msg1 = lenMdl.msg1(), msg2 = lenMdl.msg2(); // so far
            Vector  fs = (Vector) sst.thrd();   // freqs
            msg2 += new Discretes.Uniform(bnds).Mdl.nlLH(sst.snd()); //Unif pts
            Model [] mdls = new Model [range_o];
            Value [] sp = new Value [range_o];  // for the MultiState.M s
            for( int i = 0; i < range_o; i ++ ) // per context...
             { final Model mi =
                  MS.estimator(Value.triv).ss2Model((Vector) fs.elt(i));
               mdls[i] = mi;
               sp[i] = mi.statParams();
               msg1 += mi.msg1();
               msg2 += mi.msg2();  // non Uniform parts
             }//for i
            return new M(msg1, msg2,
                         pair(lenMdl.statParams(), Vector.values(sp)));
          }//ss2Model(ss)
    }; }//estimator(ps)

   private int cntxt(int cntxt, int next)
    { return (cntxt*range + next-lwb) % range_o; }
   // .........................................................................

   /** Markov.M, the fully parameterised Markov Model, of a given
       {@link #order order}, for data [lwb,&nbsp;upb]<SUP>*</SUP>.
       Also see the UnParameterised {@link Markov Markov} Model.  */
   public class M extends UPSeriesModel.Length.M                    // Markov.M
    {
      /** mdls, MultiState Models, one <I>per</I> context. */
      public  final MultiState.M [] mdls;

      /** Model of lengths of series.  */
      private final Model lenMdl;

      /** A Uniform Model is used for the first
          {@link #order order} elements. */
      private final Discretes.Uniform.M uniformMdl;

      /** sp = ({@link #lenMdl() lenMdl()}'s,&nbsp;{@link #mdls mdls}').  */
      public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp);
         Value.Tuple l_ps = sp.isTuple(2);
         lenMdl = lenUPM.apply(l_ps.fst());  // lengths
         uniformMdl = new Discretes.Uniform(bnds).Mdl; // 1st order elts
         Vector prs = (Vector) l_ps.snd();
         if( prs.nElts() != range_o ) error( ".constructor" );
         mdls = new MultiState.M [range_o];            // most elts...
         for( int i = 0; i < range_o; i ++ )
            mdls[i] = (MultiState.M) MS.apply(prs.elt(i));
       }//constructor

      public final Model lenMdl() { return lenMdl; }

      /** Collect statistics, ss&nbsp;=&nbsp;stats(seqs,lo,hi), of
          a data-set, that is of a Vector of Vectors, seqs. 
          Calls upon {@link #lenMdl} vai {@link #lenStats}.  */
      public Value stats(Vector seqs, int lo, int hi)
       { double [] [] freq = new double [range_o] [range];
         for( int i = 0; i < range_o; i ++ )
            for( int j = 0; j < range; j ++ ) freq[i][j] = 0;
         double uniformPart = 0;
         for( int i = lo; i < hi; i ++ )
          { int cntxt = 0;
            final Vector vi = (Vector) seqs.elt(i);
            final double wi = seqs.wt(i);
            for( int j = 0; j < vi.nElts(); j ++ )
             { final int vij = vi.elt(j).n();
               if( j < order ) uniformPart += wi; // uniform part
               else freq[cntxt][vij - lwb] += wi; // context Models part
               cntxt = cntxt(cntxt, vij);
             }//for j
          }//for i
         return triple( lenStats(seqs, lo, hi),   // lengths
                        real(uniformPart),      // for UniformMdl
                        Matrix.doubles(freq) ); // for context Models
       }//stats(seqs,lo,hi)

      /** Calls upon stats(add,.,) of {@link #lenMdl}.  */
      public Value stats(boolean add, Value ss0, Value ss1)
       { Value.Tuple ss0t = ss0.isTuple(3), ss1t = ss1.isTuple(3);
         return triple(
                   lenMdl.stats(add, ss0t.fst(), ss1t.fst()),   // lengths
                   real(add?ss0t.snd().x()+ss1t.snd().x()
                           :ss0t.snd().x()-ss1t.snd().x()), // for UniformMdl
                   ((Matrix)ss0t.thrd()).bOp(add?Lexical.plus:Lexical.minus,
                                                  (Matrix)ss1t.thrd()) );
       }//stats(add,ss0,ss1)

      /** Given statistics, ss&nbsp;=&nbsp;{@link #stats stats(ds)}, of
          a data-set, ds, return the negative log likelihood of ds. */
      public double nlLH(Value ss)
       { final Value.Tuple sst = ss.isTuple(3);
         double nlLH = lenMdl.nlLH(sst.fst())        // lengths
                     + uniformMdl.nlLH(sst.snd()); // Uniform parts
         for( int i = 0; i < range_o; i ++ )       // Markov parts
            nlLH += mdls[i].nlLH( ((Vector) sst.thrd()) .elt(i) );
         return nlLH;
       }//nlLH(ss)

      /** Return the Analysis of a datum Series sv. */
      public Analysis analysis(Scannable sv)
       { return new Analysis(sv)
          { /** The context (index) applicable to the current eltD(). */
            private int cntxt = 0;
            public Model eltM()
             { return position() < order ? uniformMdl : mdls[cntxt]; }
            public void advance()
             { cntxt = cntxt(cntxt, eltD().n()); ds.advance(); }
       }; }//analysis(d)

    }//Markov.M class, the Series Model

 }//Markov class, the UP Series Model

// (c) Lloyd Allison, 2012, Thu 7 Dec 2017. Licence to 'Markov' is Gnu GPL.
