//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.maths.Vector;
import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The UnParameterised <CODE>Missing</CODE> Model of "missing data" &ndash;
    data that <I>may&nbsp;be</I> {@link la.la.Value.Maybe known} (present)
    or may be missing (absent, unknown).  The fully parameterised Model
    is {@link Missing.M}. */
public class Missing extends UPModel                                 // Missing
 {
   /** The UnParameterised 2-state Model of whether a data value
       is known (present, true) or missing (absent, false).
       Also see {@link M#presentMdl}. */
   public final MultiState presentUPM;

   /** The UnParameterised Model of those data that are known (present).
       It is set by the constructor and can be any UPModel of the
       (present) data.  Also see {@link M#valueMdl}.  */
   public final UPModel    valueUPM;

   /** Problem definition parameter 'dp' sets {@link #valueUPM valueUPM}. */
   public Missing(Value dp)
    { super(dp);
      presentUPM = new MultiState( pair(ffalse, ttrue) );
      valueUPM   = (UPModel) dp;
    }//constructor

   /** Given first- and second-part message lengths, msg1 and msg2, and
       statistical parameter(s), sp, return a fully parameterised
       {@link M}-Model. */
   public M sp2Model(double msg1, double msg2, Value sp)
    { return new M(msg1, msg2, sp); } // for the sake of the type

   /** For a Vector 'ds' of Value.{@link la.la.Value.Maybe Maybe}, return a
       Vector of {@link la.la.Value.Bool Bool}.  That is of ffalse for a
       missing value 'None', and ttrue for a 'Just&nbsp;v'.  */
   public static Vector present(Vector ds)
    { return new Vector.Ints()  // shadow Vector
       { public int          nElts() { return ds.nElts();   }
         public double     constWt() { return ds.constWt(); }
         public double     wt(int i) { return ds.wt(i);     }
         public Bool elt(int i)
          { return ds.elt(i).n() == 0 ? ffalse : ttrue; }
       };
    }//present(ds)

   /** For a Vector 'ds' of Value.{@link la.la.Value.Maybe Maybe}, return a
       Vector of those 'v' that are present in ds. That is values
       'Just&nbsp;v' that are not missing. In general the result is
       shorter than ds.
       Note, it would be rather nice to return a shadow Vector, the
       same length as 'ds', and set the weights of missing values to
       zero, but is it guaranteed that all stats() etc. avoid even
       looking at such (non-)values???  */
   public static Vector maybe_J(Vector ds)
    { int Npresent = 0;
      for( int i = 0; i < ds.nElts(); i ++ )
         Npresent += ((Maybe) ds.elt(i)) .n();  // roll call
      int [] ptr = new int [Npresent];
      int j = 0;
      for( int i = 0; i < ds.nElts(); i ++ )
         if( ds.elt(i).n() == 1 ) { ptr[j] = i; j ++ ; }
      return new Vector()
       { public int       nElts() { return ptr.length;      }
         public double  constWt() { return ds.constWt();    }
         public double constAoM() { return ds.constAoM();   }
         public double  wt(int i) { return ds.wt( ptr[i] ); }
         public Value  elt(int i)
               { return ((Maybe.Just) ds.elt( ptr[i] )) .v; } // !
       };
    }//maybe_J(ds)

   /** Use {@link #presentUPM presentUPM}.stats(...)
       and {@link #valueUPM     valueUPM}.stats(...) to
       calculate the  sufficient statistics of ds[lo,hi).
       Also see {@link M#stats(Vector,int,int) M.stats(...)}.  */
   public Value.Tuple stats(Vector ds, int lo, int hi)
    { return pair( presentUPM .stats(present(ds)),
                   valueUPM   .stats(maybe_J(ds)) );
    }//stats(ds,lo,hi)

   /** Use {@link #presentUPM presentUPM}.stats(...)
       and {@link #valueUPM     valueUPM}.stats(...).
       Combine statistics ss0 and ss1, either ss0 "+" ss1, or ss0 "-" ss1,
       depending on 'add' being true or false.
       Also see {@link M#stats(boolean,Value,Value) M.stats(...)}.  */
   public Value.Tuple stats(boolean add, Value ss0, Value ss1)
    { Value.Tuple ss0T = ss0.isTuple(2), ss1T = ss1.isTuple(2);
      return pair( presentUPM .stats(add, ss0T.fst(), ss1T.fst()),
                   valueUPM   .stats(add, ss0T.snd(), ss1T.snd()) );
    }//stats(add,ss0,ss1)

   /** Return an Estimator of a fully parameterised {@link M}-Model.
       Its parameter 'ps' is a pair, one for the estimators of
       {@link #presentUPM} and one for {@link #valueUPM}. */
   public Est estimator(Value ps)
    { return new Est(ps)
       { final Tuple psT = ps.isTuple(2);  // one per UPModel
         /** Given sufficient statistics, ss, return a
             fully parameterised {@link Missing.M} Model. */
         public M ss2Model(Value ss)
          { final Tuple ssT = ss.isTuple(2);
            Model
            presentMdl = presentUPM.estimator(psT.fst()).ss2Model(ssT.fst()),
            valueMdl   =   valueUPM.estimator(psT.snd()).ss2Model(ssT.snd());
            return Missing.this.
                   sp2Model( presentMdl.msg1() + valueMdl.msg1(),
                             presentMdl.msg2() + valueMdl.msg2(),
                        pair(presentMdl.statParams(),
                               valueMdl.statParams()) );
          }//ss2Model(ss)
       };//new Estimator
    }//estimator(ps)

   // .........................................................................

   /** The fully parameterised Model;
       {@link Missing} is the UnParameterised Model.  */
   public class M extends UPModel.M                                // Missing.M
    {
      /** The fully parameterised Model of whether a Value is present
          (known) or missing (unknown).  Also see {@link #presentUPM}. */
      public final MultiState.M presentMdl;

      /** The fully parameterised Model all present (known) Values.
          Also see {@link #valueUPM}.  */
      public final Model valueMdl;

      /** Two part message lengths, msg1 and msg2,
          and statistical parameter(s) 'sp'.
          'sp'&nbsp;is a pair of parameters &mdash;
          one for {@link #presentUPM presentUPM} and
          one for {@link #valueUPM   valueUPM}.  */
      public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp);
         Tuple spT  = sp.isTuple(2);  // i.e., a pair of statistical params
         presentMdl = presentUPM.sp2Model(0, 0, spT.fst());
         valueMdl   =   valueUPM.sp2Model(0, 0, spT.snd());
       }//constructor

      /** Return the negative log probability of datum&nbsp;'d'. This uses
          {@link #presentMdl presentMdl} for <CODE>None</CODE>, and
          both <CODE>presentMdl</CODE> and {@link #valueMdl valueMdl}
          for <CODE>Just(v)</CODE>.  */
      public double nlPr(Value d)
       { switch( ((Value.Maybe) d).n() )
          { case 0: return presentMdl.nlPr(ffalse);
            case 1: return presentMdl.nlPr(ttrue)
                         +   valueMdl.nlPr(((Value.Maybe.Just) d).v);
          }//switch
         throw RTE( ".nlPr(d=" + d + "), d not Maybe..." );
       }//nlPr(d)

      /** Use {@link #presentMdl presentMdl}.stats(...)
          and {@link #valueMdl     valueMdl}.stats(...) to
          calculate the  statistics of ds[lo,hi).
          Note this <I>might</I> be the same as
          {@link Missing#stats(la.maths.Vector,int,int)}, or not!  */
      public Value.Tuple stats(Vector ds, int lo, int hi)
       { return pair( presentMdl .stats(present(ds)),
                      valueMdl   .stats(maybe_J(ds)) );
       }//stats(ds,lo,hi)

      /** Use {@link #presentMdl presentMdl}.stats(...)
          and {@link #valueMdl     valueMdl}.stats(...).
          Combine statistics ss0 and ss1, either ss0 "+" ss1, or ss0 "-" ss1,
          depending on 'add' being true or false.
          Note this <I>might</I> be the same as
          {@link Missing#stats(boolean,Value,Value)}, or not!  */
      public Value.Tuple stats(boolean add, Value ss0, Value ss1)
       { Value.Tuple ss0T = ss0.isTuple(2), ss1T = ss1.isTuple(2);
         return pair( presentMdl .stats(add, ss0T.fst(), ss1T.fst()),
                      valueMdl   .stats(add, ss0T.snd(), ss1T.snd()) );
       }//stats(add,ss0,ss1)

      /** Given sufficient statistics,
          {@link #stats(Vector) ss=stats(ds)}, of a data-set 'ds',
          return the negative log LikeliHood of 'ds'. */
      public double nlLH(Value ss)
       { Tuple ssT = ss.isTuple(2);
         return presentMdl.nlLH(ssT.fst()) + valueMdl.nlLH(ssT.snd());
       }//nlLH(ss)

      /** Return a random Value from 'this' Model &ndash; provided that
          {@link #valueMdl valueMdl} can do random(). */
      public Value random()
       { return presentMdl.random_n() == 0
                          ? Value.Maybe.None
                          : new Value.Maybe.Just( valueMdl.random() );
       }//random()

    }//Missing.M class, the fully parameterised Model

 }//Missing class, the UPModel

// (c) Lloyd Allison, 9/2017, 7 Dec 2017. Licence to 'Missing' is Gnu GPL.
