//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.la.Function;
import la.la.Function.HasInverse;
import la.la.Type;
import la.maths.Vector;
import la.maths.Vector.Slice;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference:
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The abstract class of fully parameterised statistical Models. The main
    job of a Model is to return the probability, {@link #pr pr(d)}, and
    negative-log probability, {@link #nlPr nlPr(d)}, of a datum,&nbsp;d.
    Most instances of Model are likely to be instances of {@link UPModel.M}
    and often the {@link UPModel UnParameterised}-Model's
    {@link UPModel#estimator Estimator} estimates a
    fully parameterised Model from a given data-set, for example,
    {@link MML#Normal} estimates a Normal (Gaussian) probability distribution,
    {@link NormalUPM.M N<SUB>&mu;,&sigma;</SUB>}. Among example Models,
    {@link MultiState.M} has non-trivial
    {@link #statParams statistical}-parameters but
    {@link Discretes.Uniform.M} does not.
    There are sub-classes of Model, such as {@link Discretes.M} and
    {@link Multivariate.M}, that make it easier to create certain kinds of
    Model.  Note that a Model is a first-class {@link Value Value}.<BR>
    Also see {@link FunctionModel FunctionModel} and
    {@link SeriesModel SeriesModel}.  */
public abstract class Model extends Value                            // * Model
 { /** Where appropriate, the lengths of the first (msg1),
       and second (msg2) parts of an
       [<A HREF="http://www.allisons.org/ll/MML/" TARGET="LL">MML</A>]
       message transmitting
       (i)&nbsp;a&nbsp;Model (parameter estimate) &theta;, and
       (ii)&nbsp;training data-set ds|&theta;.
       Failing that, e.g., if the Model is a given, such as N<SUB>0,1</SUB>,
       set msg1 and msg2 to zero.  */
   protected final double msg1, msg2;

   /** Holds the statistical parameter(s), if any, of 'this' Model;
       is returned by {@link #statParams()}.  Also see the
       problem-{@link UPModel#defnParams() defining} parameters.  */
   protected final Value sp;

   /** {@link #msg1() msg1} and {@link #msg2() msg2} are the lengths,
       in nits, of transmitting (i)&nbsp;the Model's statistical
       parameter(s), {@link #statParams() sp}, and
       (ii)&nbsp;training data-set, ds|sp, where D was the training-data
       used to estimate sp. If this Model is a given, set msg1=msg2=0.  */
   public Model(double msg1, double msg2, Value sp)
    { this.msg1 = msg1; this.msg2 = msg2; this.sp = sp;
      if( msg1 < 0 || msg2 < 0 ) error(" msg:" + msg1 + "+" + msg2);
    }//constructor

   /** Returns {@link Type#MODEL}.  */
   public Type type() { return Type.MODEL; }                          // type()

   /** The statistical parameters (possibly
       {@link UPModel#estimator estimated}),
       if any, of 'this' Model as stored in {@link #sp sp}.
       Also see {@link #m1m2sp()} and {@link UPModel#apply}.  */
   public Value statParams()                                    // statParams()
    { return sp; }

   /** Return the triple({@link #msg1() msg1()}, {@link #msg2() msg2()},
       {@link #statParams() statParams()}).
       That amounts to a specification of the Model.  */
   public Value.Tuple m1m2sp()                                      // m1m2sp()
    { return Value.triple(
         Value.real(msg1()), Value.real(msg2()), statParams());
    }//m1m2sp()

   /** Length of the 1st part, 'M', of a two-part MML message
       'M;&nbsp;(ds|M)', where ds was the <I>training data</I>.
       More details under [{@link #msg() msg()}].  */
   public double msg1() { return msg1; }

   /** Length of the 2nd part, '(ds|M)', of a two-part MML message
       'M;&nbsp;(ds|M)', where D was the <I>training-data</I>.
       More details under [{@link #msg() msg()}].*/
   public double msg2() { return msg2; }

   /** The length of a two-part MML message, 'M;&nbsp;(ds|M)', where ds was
       the <I>training-data</I> used when estimating 'this' Model 'M'.
       When&nbsp;MML87 is used, the <I>standard formulae</I>
       (Wallace 2005, &sect;5.2.8, p.235) are
       <UL>
       <LI>msg1 = - log(h&nbsp;&theta;') + (1/2)log(F&nbsp;&theta;')
                + (k/2)log&nbsp;&kappa;<SUB>k</SUP>,  and
       <LI>msg2 = - log(f(ds|&theta;') + k/2 &nbsp; nits,
       <LI>for k-dimensional parameter estimate&nbsp;&theta;',
         prior&nbsp;h(&theta;'), Fisher&nbsp;F(&theta;'), lattice
         constant&nbsp;{@link MML#latticeConstant &kappa;<SUB>k</SUB>},
         and likelihood&nbsp;f(ds|&theta;').
       <LI>The '+k/2' term in msg2 is the expected correction for
         rounding off &theta;' to optimal (limited) precision.
       <LI>Statistical parameter(s) &theta;' is chosen to
         minimise msg&nbsp;=&nbsp;msg1&nbsp;+&nbsp;msg2.
       </UL>
       It could be argued that msg is not a property of the Model
       <I>in isolation</I>, rather of the process of estimating the Model
       in some context but, given the way that Models are estimated and
       compared, it is very convenient in practice to have
       msg={@link #msg1() msg1}+{@link #msg2() msg2} easily available from
       the Model. Also note that if a Model is used twice in a message,
       it is only transmitted once, and must be "free" the second time.  */
   public double      msg() { return msg1() + msg2();  }

   /** {@link #msg1()} in bits.  */
   public double msg1bits() { return msg1() * invLog2; }

   /** {@link #msg2()} in bits.  */
   public double msg2bits() { return msg2() * invLog2; }

   /** {@link #msg()} in bits.  */
   public double  msgBits() { return msg()  * invLog2; }

   /** The probability of a datum, 'd', given 'this' Model; also see
       {@link #nlPr nlPr(d)}.  Note that for a large, complex datum, d,
       pr(d) may underflow to zero. Best stick with nlPr(d) in such cases.
       This default pr(d) calls nlPr(d).   */
   public double pr(Value d)                                           // pr(d)
    { return Math.exp( - nlPr(d) ); }

   /** The negative log<SUB>e</SUB> {@link #pr probability} of a datum,
       'd', in nits;  nlPr must be defined when implementing a Model.
       Also see {@link #nl2Pr nl2Pr(d)}.  */
   public abstract double nlPr(Value d);                           // * nlPr(d)

   /** &sum; negative log probability over all data elements
       in data-set ds; you might want {@link #nlLH nlLH(ss)}
       and {@link #stats stats(ds)} instead?  <TT>sumNlPr(ds)</TT>
       and {@link #nlLH nlLH(ss)} should be equal but the latter is
       often quicker (where ss={@link #stats stats(ds)}).
       Also see {@link Defaults#nlLH}.  */
   public final double sumNlPr(Vector ds)                      // * sumNlPr(ds)
    { double sum = 0;
      for( int i = 0; i < ds.nElts(); i ++ )
         sum += ds.wt(i) * nlPr(ds.elt(i));
         // e.g., wt(i)==2 is equivalent to two instances of ds[i].
      return sum;
    }//sumNlPr(ds)

   /** Return the negative log<SUB>2</SUB> probability, nl2Pr, of datum 'd',
       in bits. Calculations are done in {@link #nlPr(Value) nits} but
       people often like information to be displayed in bits.  */
   public double nl2Pr(Value d)                                     // nl2Pr(d)
    { return nlPr(d) * invLog2; }

   /** Return a random Value from the modelled population, if possible.
       <B>Note,</B> random() is not <TT>abstract</TT>, rather a Model
       <I>optionally</I> implements it, with this default random() throwing
       an Exception. (If implemented well, random() isn't very FP-unctional.-)
       Also see {@link #random(int) random(n)} and {@link MML#RNG}.  */
   public Value random()                                          // ? random()
    { throw UOE( ".random() is not yet(?) supported, sorry" ); }

   /** Return 'n' {@link #random() random()} Values, if possible.
       The result can be used as an artificial (test-) data-set, say.  */
   public Vector random(int n)                                     // random(n)
    { Value [] ans = new Value[n];
      for(int i = 0; i < n; i ++ ) ans[i] = random();
      return Vector.values(ans);
    }//random(n)


   /** Return a {@link la.util.Series} which repeatedly
       {@link la.util.Series#elt() returns} a
       {@link #random() random()} Value from 'this' Model,
       provided the Model can do so.  But don't forget to
       {@link la.util.Series#advance() advance()} or it won't be
       very random!  Also see {@link Adaptive.M#randomSeries}.  */
   public la.util.Series randomSeries()                       // randomSeries()
    { return new la.util.Series()
       { protected int   pos = -1;    // can't call random() in constr ...
         protected Value elt = null;  // ... see Adaptive!
         public boolean hasSome() { return true; }
         public int    position() { return pos; }
         public Value       elt() { if( pos < 0 ) advance();  return elt; }
         public void    advance() { pos ++ ; elt = random(); }
         public String toString() { return Model.this + ".RandomSeries"; }
       };
    }//randomSeries()


   /** Return sufficient statistics,
       {@link #stats(Vector,int,int) stats(ds,0,ds.nElts())}, of
       a data-set 'ds'.
       There is unlikely to be any need for a subclass (a Model)
       to override stats(ds).  Also see {@link #nlLH nlLH(ss)}.  */
   public Value stats(Vector ds)                                   // stats(ds)
    { return stats(ds, 0, ds.nElts()); }

   /** For 'this' Model, calculate sufficient statistics, 'ss', of elements
       [lo,&nbsp;hi) of 'ds', e.g., for use in {@link #nlLH nlLH(ss)}.
       An option is to return the data itself, as in {@link Defaults},
       but many Models can do better.
       <B>Note</B> that ss=stats(ds,lo,hi) and {@link #nlLH nlLH(ss)} must
       be consistent.  A&nbsp;UPModel.M's
       {@link UPModel.M#stats(Vector,int,int) stats(ds,lo,hi)} <I>might</I>
       be able to make use of an enclosing {@link UPModel UPModel}'s
       {@link UPModel#stats(Vector,int,int) stats(ds,lo,hi)} if the same
       stats(...) are used for estimating, but this is not always
       the case, e.g., see Model {@link BestOf.M}.  */
   public abstract Value stats(Vector ds, int lo, int hi);                 // *

   /** Combine {@link #stats(Vector,int,int) statisticses},
       ss0&nbsp;&plusmn;&nbsp;ss1.
       '+'&nbsp;if 'add' is true, else&nbsp;'&minus;'.
       An option is to return the data itself, as in {@link Defaults},
       but many Models can do better. Also see
       {@link #stats(boolean,Value,Vector,int,int) stats(add,ss0,ds,lo,hi)}. */
   public abstract Value stats(boolean add, Value ss0, Value ss1);         // *

   /** {@link #stats(boolean,Value,Value) Combine} statisticses ss0 and
       {@link #stats(Vector,int,int) stats(ds,lo,hi)}. This allows
       advantage to be taken of additive (incremental) statistics, e.g.,
       by {@link mml.Intervals}. There is unlikely to be a need for a
       subclass (a Model) to override stats(add,ss0,ds,lo,hi).  */
   public Value stats(boolean add, Value ss0, Vector ds, int lo, int hi)
    { return stats(add, ss0, stats(ds, lo, hi)); }

   /** Given sufficient statistics,
       ss&nbsp;=&nbsp;{@link #stats stats(ds)}, of a <I>data-set</I>, ds,
       return the negative log LikeliHood, nlLH(ss), of ds.  Make sure that
       nlLH(ss), ss&nbsp;=&nbsp;{@link #stats(Vector,int,int) stats(ds,lo,hi)},
       and any {@link Estimator#ss2Model Estimator} are consistent!  */
   public abstract double nlLH(Value ss);                         // * nlLH(ss)

   /** The {@link #nlLH nlLH(ss)} of (sufficient stats, ss, of)
       a data-set, ds, but in <I>bits</I>, instead of nits.  */
   public double nl2LH(Value ss)                                   // nl2LH(ss)
    { return nlLH(ss) * invLog2; }

   /** Show the details of 'this' Model.  */
   public String toString()                                       // toString()
    { return "(" + getClass().getName() + " " + statParams() + ")"; }
   // .........................................................................

   /** Transform 'this' already parameterised Model by Function&nbsp;'f',
       roughly transform:&nbsp;(a&rarr;b)&rarr;Model&nbsp;a&rarr;Model&nbsp;b.
       (Returns a Model, in fact a Model.Transform.M.)  Convenience function
       for 'new&nbsp;{@link Transform#Mdl Transform(f).Mdl}.
       Note that 'f' must be 1-to-1.<BR>
       Also see the related but different {@link UPModel#transform}; note
       that, as distributions, upm(sp).transform(f)=upm.transform(f)(sp).<BR>
       The result of this transform(f) is "just" a Model; if you need a
       more specific sub-class of Model, do as in
       {@link Continuous.M#transform Continuous.M.transform(f)} and
       {@link R_D.M#transform R_D.M.transform(f)}.  */
   public Model transform(Function f)                           // transform(f)
    { return new Transform(f).Mdl; }

   /** Transform (the data for) Model.this by Function&nbsp;f
       (which is the problem defining parameter).
       Function 'f' must be 1&ndash;to&ndash;1.
       This amounts to applying the inverse, f<SUP>&minus;1</SUP>,
       to the random variable modelled by the Model being transformed,
       e.g., {@link la.la.Library#log log<SUP>&minus;1</SUP>}=exp.
       Model.Transform is an UnParameterised model and its
       (trivially) parameterised Model is {@link Transform.M}.
       Note, there is a more specific {@link Continuous.M.Transform}
       for transforming Continuous.M Models, and similarly
       {@link R_D.M.Transform}.<BR>
       Transformed {@link Transform#random random()} depends on f having an
       {@link la.la.Function.HasInverse inverse()} <I>implemented</I>.<BR>
       <B>NB.</B> The preferred way to create a transformed Model
       is with function {@link Model#transform transform(f)}.<BR>
       Also see the related but different
       {@link UPModel.Transform   UPModel.Transform}   and
       {@link UPModel.Transform.M UPModel.Transform.M}.<BR>
       &mdash;&nbsp;April&nbsp;2020.  */
   public class Transform extends UPModel                    // Model.Transform
    {
      /** 'f' the function doing the transforming (of data).  */
      public final Function f;

      /** Note, Function '{@link #f f}' is the problem defining parameter. */
      public Transform(Value f)
       { super(f);  this.f = (Function) f; }

      /** The enclosing Model.this's stats(...) applied to the
          data-set transformed by map({@link #f f}).  */
      public Value stats(Vector ds, int lo, int hi)
       { return Model.this.stats(ds.map(f), lo, hi ); }

      /** Combine {@link #stats statistics} 'ss0' and 'ss1' in
          the way that Model.this does.  */
      public Value stats(boolean add, Value ss0, Value ss1)
       { return Model.this.stats(add, ss0, ss1); }  // !

      public M sp2Model(double msg1, double msg2, Value sp)
       { return new M(msg1, msg2, sp); }

      /** Trivial estimator, always "estimates" {@link #Mdl Mdl}. */
      public Estimator estimator(Value t)
       { return Mdl.asEstimator(t); }

      /** <I>The</I> instance of M, 'Mdl'.  msg1=msg2=0.  */
      public final M Mdl = new M(0, 0, triv);
      // ....................................

      /** The fully parameterised transformed Model has a trivial
          statistical parameter, sp. The UPModel is {@link Transform}.<BR>
          Also see {@link #transform transform(f)}.  */
      public class M extends UPModel.M                     // Model.Transform.M
       { /** Note, msg1=0 and statistical parameter sp=triv, checked. */
         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp); zeroTriv(msg1, sp); }

         /** The enclosing Model.this's nlPr({@link #f f(d)}).  */
         public double nlPr(Value d)
          { return Model.this.nlPr(f.apply(d)); }

         /** The enclosing Model.this's pr({@link #f f(d)}).  */
         public double pr(Value d) // need both nlPr & pr for efficiency!
          { return Model.this.pr(f.apply(d)); }

         /** The enclosing Model.this's nlLH applied to
             {@link Transform#stats statistics} 'ss'. */
         public double nlLH(Value ss)
          { return Model.this.nlLH(ss); } // same as Transform....

         /** f<SUP>-1</SUP>&nbsp;(Model.this.random()).  Requires
             {@link #f f} to have an implemented inverse to work.  */
         public Value random()  // requires f to have an inverse
          { return ((Function.HasInverse) f).inverse()
                   .apply(Model.this.random());
          }//random()

         public String toString()
          { return Model.this + ".transform(" + f + ")"; }
       }//Model.Transform.M class, the Model

    }//Model.Transform class, the UnParameterised Model
   // .........................................................................

   /** Return 'this' Model as a "given", with zero first-part
       message length, and a specified second-part, msg2,
       {@link #asGiven(double,double) asGiven(0,msg2)}.
       This may be necessary because the Model may be competing
       against others on some new "training" data-set. For example,
       {@link #asUPModel() asUPModel()} returns asGiven(msg2). Also see
       {@link UPModel.M#asGiven(double) UPModel.M.asGiven(msg2)}.  */
   public Model asGiven(double msg2)                           // asGiven(msg2)
    { return asGiven(0, msg2); }

   /** Enables setting the first- and second-part message lengths, msg1
       and msg2, after having estimated the statistical parameter(s) of
       a Model, say. Also see {@link #asGiven(double) asGiven(msg2)}.  */
   public Model asGiven(double msg1, double msg2)             // asGiven(m1,m2)
    { java.lang.reflect.Constructor <? extends Model> cr = null;
      Class dc = double.class;
      try { cr = this.getClass().getConstructor(dc, dc, Value.class);
            return cr.newInstance(msg1, msg2, this.statParams());
      }catch(Exception e)
        { throw RTE( ".asGiven(" + msg1 + "," + msg2 + ") failed, " + e ); }
    }//asGiven(msg1,msg2)

   /** Check msg1&nbsp;=&nbsp;0 and sp&nbsp;=&nbsp;(), that both are trivial.
       If not, throw an error.  This can be required of a Uniform Model,
       say, because the standard constructor for every Model has the
       form foobar(msg1,&nbsp;msg2,&nbsp;sp).  */
   protected void zeroTriv(double msg1, Value sp)
    { if( msg1!=0 || sp!=triv ) error( ", msg1=" + msg1 + ", sp=" + sp ); }
   // NB. zeroTriv cannot be static because it uses error... which isn't.

   /** It might be necessary, in some context, to treat 'this' fully
       parameterised Model as an {@link UPModel UnParameterised} Model,
       having trivial problem-definition parameter, that always produces
       (by {@link UPModel#apply apply}, {@link UPModel#estimator estimator},
       etc.) 'this' Model {@link #asGiven(double) asGiven}, so to say.
       Note that the result is just a UPModel so, if you want a subclass,
       override asUPModel suitably. If you want 'this' Model as an Estimator
       use {@link #asEstimator asEstimator(triv)}.  Also see
       {@link FunctionModel#asUPModel} and {@link SeriesModel#asUPModel}.  */
   public UPModel asUPModel()                                    // asUPModel()
    { final Model mdl = this; // enclosing Model
      return new UPModel(Value.triv)
       { // Specify sp2Model(m1,m2,sp), stats(ds), estimator(ps), toString()
         public Model sp2Model(double msg1, double msg2, Value sp)
          { if( msg1 != 0 ) error( ".sp2Model("+msg1+","+msg2+","+sp+")" );
            return mdl.asGiven(msg2); // !
          }//sp2Model(,,)
         public Value stats(Vector ds, int lo, int hi)
          { return mdl.stats(ds, lo, hi); }
         public Value stats(boolean add, Value ss0, Value ss1)
          { return mdl.stats(add, ss0, ss1); }
         public Estimator estimator(Value ps/* triv */)
          { return mdl.asEstimator(ps); } // "estimates" mdl
         public String toString() { return mdl + ".asUPModel()"; }
       };
    }//asUPModel()

   /** Return an Estimator that always "estimates" 'this' Model.
       Note the "estimated" Model is a given, and thus has msg1=0.
       Also see {@link #asUPModel() asUPModel()}. Note, ps=triv!  */
   public Estimator asEstimator(final Value ps)              // asEstimator(ps)
    { final Model mdl = Model.this;
      return new Estimator(ps)
       { { if( ps != triv ) error( ".asEstimator(" + ps + ")" ); }
         public Value    defnParams() { return Value.triv; }
         /** Return the enclosing Model's stats(ds).  */
         public Value stats(Vector ds, int lo, int hi)
          { return mdl.stats(ds, lo, hi); }
         public Value stats(boolean add, Value ss0, Value ss1)
          { return mdl.stats(add, ss0, ss1); }
         /** Return the enclosing Model, with msg1=0.  */
         public Model sp2Model(double msg1, double msg2, Value sp)
          { if( msg1 != 0 ) error( ".sp2Model("+msg1+","+msg2+","+sp+")" );
            return mdl.asGiven(msg2);
          }//sp2Model(,,)
         /** Return the enclosing Model, with msg1=0, msg2=nlLH(ss).  */
         public Model ss2Model(Value ss)
          { return sp2Model(0, mdl.nlLH(ss), Value.triv); }
         /** Return (ss2Model(ss),triv), triv because the estimation is.  */
         public Tuple ss2ModelSp(Value ss)
          { return pair(ss2Model(ss), triv /* NB. */ ); }
         public String toString() { return mdl + ".asEstimator()"; }
       };
    }//asEstimator(ps)
   //  ------------------------------------------------------------------------

   /** A subclass of {@link Model Model} that sets default methods
       {@link #stats(Vector,int,int)      stats(ds,lo,hi)},
       {@link #stats(boolean,Value,Value) stats(add,ss0,ss1)} and
       {@link #nlLH nlLH(ss)}, even if they are slow.  Only use (extend)
       Defaults if you cannot think of anything better, or
       you are very lazy.  */
   public static abstract class Defaults extends Model        // Model.Defaults
    {
      public Defaults(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp); }//constructor

      /** This default returns the data itself,
          ds.{@link Vector#slice slice(lo,hi)}, as
          statistics but many Models can do much better.
          <B>Note</B> that ss=stats(ds,lo,hi),
          {@link #stats(boolean,Value,Value) stats(add,ss0,ss1)} and
          {@link #nlLH nlLH(ss))} must be consistent.
          Also see {@link Model#stats(Vector,int,int)}.  */
      public Value stats(Vector ds, int lo, int hi)          // stats(ds,lo,hi)
       { return ds.slice(lo, hi); }

      /** {@link #stats(Vector,int,int) statistics}
          ss0&nbsp;&plusmn;&nbsp;ss1,
          '+'&nbsp;if 'add' is true othewise&nbsp;'&minus;'.
          This default implementation assumes statistics are the
          data themselves but many Models can do better.  */
      public Value stats(boolean add, Value ss0, Value ss1)        // stats(,,)
       { return ((Vector) ss0).combine(add, (Slice) ss1); }

      /** This default implementation assumes that statistics are the
          data themselves and uses {@link #sumNlPr sumNlPr} but many
          Models can do better.  Make sure that nlLH(ss),
          ss&nbsp;= {@link #stats(Vector,int,int) stats(ds,lo,hi)},
          and any {@link Estimator#ss2Model Estimator} are consistent! */
      public double nlLH(Value ss)                                  // nlLH(ss)
       { return sumNlPr((Vector) ss); }

   }//Model.Defaults class

 }//Model class

// (c) L. Allison, 2012, Aug 2017, 15 Apr 2020. Licence to 'Model' is Gnu GPL.
