//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import java.util.BitSet;

import la.util.*;
import la.la.Value;
import static la.la.Value.*;
import static la.maths.Maths.*;
import la.maths.Vector;
import la.maths.Vector.Slice;
import la.maths.Matrix;
import static mml.MML.*;
import graph.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The UnParameterised 'MotifD' Model of {@link Graph Graphs}.
    The fully parameterised Model is&nbsp;{@link MotifD.M M}.
    Also see {@link MotifA}.  */
public class MotifD extends Graphs.Motifs                             // MotifD
 {
   // protected final graph.Type gType; // from Graphs
   // also                       gType()

   // protected final UPModel upmV; // from Graphs
   // also                    upmV()

   // protected Discretes.M mdlNmotifs;  // from Graphs.Motifs,
   // protected Model mdlMotif;          // from Graphs.Motifs, and
   // public double msgMotifs(Graph [] ms)
   // public double msgMotifs(Vector vm)

   /** The UnParameterised ({@link Adaptive Adaptive}) Model over the
       probability of existence of possible Edges outside instances of
       {@link M#motifs motifs}. That is, the Model of entries in the
       adjacency matrix, A[,], that are not covered by any instance
       of a motif. Also see {@link M#mdlE mdlE}. */
   public final Discretes.Bounded upmE =                                // upmE
      new Adaptive(pair( pair(zero, one),  // bounds
                         Vector.doubles(new double[]{9, 1}) )); // ie. 10+/-%

   /** Bounds [sgMinV, sgMaxV] on |V| for the motifs (patterns) that
       the {@link #estimator estimator} will search for (but not
       necessarily find). Also see {@link #mdlMotif mdlMotif}.  */
   public final int sgMinV, sgMaxV;

   /** Problem definition parameters,
       <NOBR> dp = &lang;{@link #gType gType}, {@link #upmV  upmV},
           &lang;{@link #sgMinV sgMinV},
                 {@link #sgMaxV sgMaxV}&rang;&rang;. </NOBR>  */
   public MotifD(Value dp)
    { super(dp); Tuple dpT = dp.isTuple(3);
      gType = (graph.Type) dpT.fst();     // the population of Graphs
      upmV  = (Discretes)  dpT.snd();     // the UPModel of |V| >= 1
      Tuple bnds = dpT.thrd().isTuple(2); // [sgMinV, sgMaxV]
      sgMinV = bnds.fst().n(); sgMaxV = bnds.snd().n();

      mdlNmotifs = Geometric0.sp2Model(0, 0, twoR); // model of #motifs >= 0
      mdlMotif = new Graphs.GERadaptive(             // model of each motif
                  triple(gType,
                         new Discretes.Uniform(pair(new Int(sgMinV),
                                                    new Int(sgMaxV)) ),
                         Vector.doubles(new double[]{1, 1})  // alpha
                        )          )
              .sp2Model(0, 0, triv /* sp of Discretes.Uniform */ );
    }//constructor

   /** Sufficient statistics are elements [lo,&nbsp;hi)
       of the data-set ds itself. */
   public Vector stats(Vector ds, int lo, int hi)            // stats(ds,lo,hi)
    { return ds.slice(lo, hi); }

   public Vector stats(boolean add, Value ss0, Value ss1)          // stats(,,)
    { return ((Vector) ss0).combine(add, (Slice) ss1); }

   /** Return a fully parameterised MotifD {@link MotifD.M Model}. */
   public MotifD.M sp2Model(double msg1, double msg2, Value sp) // sp2Model(,,)
    { return new M(msg1, msg2, sp); }

   /** Return an Estimator for a MotifD {@link MotifD.M Model}. Parameters 'ps'
       are for {@link #upmV upmV}.estimator(ps). Estimating {@link M#mdlV mdlV}
       is standard but finding good {@link M#motifs motifs} is hard so
       heuristics are used.  The time complexity should be approximately
       O(#G.V.d<SUP>sgMaxV-1</SUP>), given a data-set of #G Graphs of
       average size V and average degree&nbsp;d.  */
   public Est estimator(final Value ps)                        // estimator(ps)
    { return new Est(ps)
       { public M ss2Model(Value ss)
          { final int topN = 20;  // # of candidate Motifs fed into greedy()
            Vector ds = (Vector) ss;           // i.e., data-set
            java.util.Vector <Graph>  mtfV = new java.util.Vector <Graph>  ();
            java.util.Vector <Int>    scrV = new java.util.Vector <Int>    ();
            java.util.Vector <BitSet> bitV = new java.util.Vector <BitSet> ();

            Discretes.M mdlV = ds2mdlV(ps, ds);  // of |V|

            int instances = 0;                      // next, seek motifs
            for( int i = 0; i < ds.nElts(); i ++ )  // each Graph g_i in ds
             { Graph g_i = (Graph) ds.elt(i);       // ... g_i
               final int g_i_V = g_i.vSize();
               Graph.SubGraphs sgs = g_i.subGraphs(sgMinV, sgMaxV);

               if( i > 0 ) // clear out any old coVers of previous g_i
                  for( int j = 0; j < bitV.size(); j ++ )
                     bitV.set(j, new BitSet(g_i_V));

               while( sgs.hasSome() )         // each sized subGraph of g_i
                { Graph.Induced sg = sgs.elt();         // sg
                  // ?could just skip low-density subGraphs?
                  instances ++ ;  // trace, stats
                  final Graph sgC = sg.canonical().flatten();
                  // sgC: if sg is novel we need sgC to add it and
                  //      if not we still need it to confirm isomorphism
                  boolean novel = true;  // tentatively
                  int j = 0;
                  for( ; j < mtfV.size(); j ++ )
                   { Graph mtfV_j = (Graph) mtfV.elementAt(j);
                     if( sgC.structurallyIdentical(mtfV_j) )
                      { novel = false; break; }         // seen sg before
                   }//for j

                  if( novel )  // add new instance and init info...
                   { mtfV.addElement(sgC);              // instance
                     scrV.addElement(zero);             // score = 0
                     BitSet bs = new BitSet(g_i_V);     // sg coVers {}
                     bitV.addElement(bs);               //  - " -
                   }//if;  note the novel motif is at position 'j'

                  // sg's "score" is E x #disjoint_instances ...
                  if( ! bitV.elementAt(j).intersects(sgs.vertices()) )
                   { scrV.setElementAt(new Int( scrV.elementAt(j).n()
                                              + sg.eSize()), j);
                     bitV.elementAt(j).or(sgs.vertices()); // + sg's Vs
                   }

                  sgs.advance(); // next subGraph of g_i
                }//while sgs...
             }//for i

            // Sort the counted subGraphs on decreasing score, 'scrV'.
            for( int i = 1; i < mtfV.size(); i ++ )
             { Graph  mtfV_i = mtfV.elementAt(i);
               Int    scrV_i = scrV.elementAt(i); int scrV_i_n = scrV_i.n();
               BitSet bitV_i = bitV.elementAt(i);
               // don't need bitV anymore
               int j = i;
               for( ; j > 0; j -- )
                  if( scrV.elementAt(j-1).n() < scrV_i_n )
                   { mtfV.setElementAt(mtfV.elementAt(j-1), j);
                     scrV.setElementAt(scrV.elementAt(j-1), j);
                     bitV.setElementAt(bitV.elementAt(j-1), j);
                   }
                  else break; // >=
               if( j == i ) continue; // else...
               mtfV.setElementAt(mtfV_i, j);
               scrV.setElementAt(scrV_i, j);
               bitV.setElementAt(bitV_i, j);
             }//for i

// ZZZ...
if( verbosity > 0 )
{ int lastShown = 0, maxExtras = 5;
  System.out.println( "   MotifD.estimator: "
                    + mtfV.size()+" motifs, " + instances+" instances}" );
  for( int i = 0; i <  mtfV.size(); i ++ )
   if( i < topN || i >= mtfV.size()-5 || 
       (maxExtras > 0 &&
        mtfV.elementAt(i).eSize() >= mtfV.elementAt(i).vSize() /* "dense" */ ) )
    { if( i > lastShown + 1 ) System.out.println( " . . . etc." );
      Graph m_i = mtfV.elementAt(i);
      System.out.println( i + ":" + m_i + ": " + scrV.elementAt(i)
                        + " " + m_i.eSize() + "e #" + m_i.hashCode() );
      lastShown = i;
      if( i >= topN && i < mtfV.size()-5 ) maxExtras -- ;
    }
}//if(...)

            // Choose a small number of the "best" motifs greedily ...
            Graph [] mtfArr = mtfV.toArray(new Graph [mtfV.size()]);
            Vector mtfVec = Vector.values(mtfArr); // not java.util.Vector!-)
            return (MotifD.M) greedy(
                        mtfVec.take(Math.min(topN, mtfArr.length)), mdlV, ds );
          }//ss2Model(ss)
       };//new Est(ps)
    }//estimator(ps)


   /** ?Obsolete;  'instnc' only uses Vertices in 'free'? */
   protected boolean clearV(Graph g, BitSet free,                // clearV(,,,)
                         Graph.Induced instnc, BitSet of_instnc)
    { boolean noOverlap = true;  // optimism
      of_instnc.clear();         // empty
      final int iV = instnc.vSize();
      for( int a = 0; a < iV; a ++ )
       { final int v_a = instnc.vs[a];  // i.e., in g
         if( free.get(v_a) ) of_instnc.set(v_a);     // +{v_a}
         else { noOverlap = false; break; }
       }//for a
      return noOverlap;
    }//clearV(...)

   /** ?Obsolete;  'instnc' only uses Vertex-pairs (V2) in 'free'? */
   protected boolean clearV2(Graph g, BitSet free,              // clearV2(,,,)
                         Graph.Induced instnc, BitSet of_instnc)
    { boolean noOverlap = true;  // optimism
      of_instnc.clear();         // empty
      final int iV = instnc.vSize();
      for( int a = 0; noOverlap && a < iV; a ++ )
       { final int v_a = instnc.vs[a];  // i.e., in g
         final int b0 = g.isDirected() ? 0 : a;
         for( int b = b0; b < iV; b ++ ) if( a != b )
          { final int a_b = g.vPair2n(v_a, instnc.vs[b]);
            if( free.get(a_b) ) of_instnc.set(a_b);  // +{a_b}
            else { noOverlap = false; break; }
          }//for b
       }//for a
      return noOverlap;
    }//clearV2(,,,)
   // .........................................................................

   /** The fully parameterised MotifD Model of Graphs.  See
       {@link #mdlV mdlV}, {@link #mdlE mdlE}, {@link #motifs motifs} and
       {@link #nlPr nlPr(g)}. The UnParameterised Model is {@link MotifD}.
       Also see {@link MotifA.M}.  */
   public class M extends Graphs.Motifs.M                           // MotifD.M
    {
      // protected final Model mdlV; // from Graphs.M
      // also                  mdlV()

      /** The fully parameterised ({@link Adaptive.M Adaptive}) Model of
          that part of the Adjacency Matrix not covered by instances of
          {@link #motifs motifs[.]}. Also see {@link #upmE}. */
      public final Discretes.Bounded.M mdlE;                            // mdlE

      /** The motifs (patterns, templates), sorted on |V|,
          that can be used to compress a given Graph. There will be good
          compression if there are many instances of motifs[.] in the Graph.
          The motifs may be "given" or {@link MotifD#estimator estimated}. */
      protected final Graph [] motifs;                              // motifs[]

      /** Returns {@link #motifs motifs[]};
          see {@link Graphs.Motifs.M#motifs()}. */
      public Graph [] motifs() { return motifs; }

      /** |V| for the smallest and largest of the {@link #motifs motifs[.]}.
          Also see the <I>different</I> {@link MotifD#sgMinV} and sgMaxV. */
      public final int sgMinV, sgMaxV;

      /** sp = &lang;sp<SUB>{@link #mdlV mdlV}</SUB>,
                     {@link #motifs motifs}&rang;.  */
      public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp); Value.Tuple spT = sp.isTuple(2);
         mdlV = upmV.apply(spT.fst());     // (a)
         mdlE = ((Adaptive) upmE).Mdl;
         Vector sgV = (Vector) spT.snd();  // (b)
         motifs = new Graph [sgV.nElts()];
         for( int i = 0; i < motifs.length; i ++ )  // each elt of motifs
          { // motifs[0..i-1] is sorted (up) on |V|
            Graph sg_i = ((Graph) sgV.elt(i)) .canonical();
            int sz = sg_i.vSize();
            int j = i;
            for( /* j=i above */ ; j > 0; j -- )
             { if( motifs[j-1].vSize() <= sz ) break;
               motifs[j] = motifs[j-1];  // move 'j-1' up to 'j'
             }// now, j == 0 || sg[j-1].vSize() <= sz
            motifs[j] = sg_i;
          }//for i

         if( motifs.length > 0 )
          { sgMinV = motifs[0].vSize();
            sgMaxV = motifs[motifs.length-1].vSize();
            if( sgMinV < MotifD.this.sgMinV
             || sgMinV > sgMaxV
             || sgMaxV > MotifD.this.sgMaxV )
               error( "sizes of given motifs out of range" );
          }
         else { sgMinV = 1; sgMaxV = 1; }
       }//constructor


      /** ss = {@link #stats(Vector,int,int) stats(ds,lo,hi)}, requires
          that ss is ds, the Vector (data-set) of Graphs itself. */
      public double nlLH(Value ss)                                  // nlLH(ss)
       { return sumNlPr((Vector) ss); }

      /** The negative log probability of datum (Graph)&nbsp;G.
          Any two instances may not share any Vertices of G. */
      public double nlPr(Value G)                                    // nlPr(G)
       { final Graph g = (Graph) G;           // g
         final int   V = g.vSize(),           // |V|
                     E = g.eSize();           // |E|
         final int maxInst = motifs.length == 0 ? 0 : V / sgMinV;
         int g_E = E, g_Area = g.maxEdges();  // still unexplained by instances
         BitSet coVers = new BitSet(V);       // instance *Vertices* in g
         int instances = 0;                   // count instances of motifs[.]
         Graph.SubGraphs g_sgs = g.subGraphs(sgMinV, sgMaxV);

         while( g_sgs.hasSome() )   // simplest, dumbest, greedy search...
          { final Graph.Induced instnc = g_sgs.elt();
            if( ! coVers.intersects(g_sgs.vertices()) ) // share no Vertex
             { final int instnc_V = instnc.vSize(),
                         instnc_E = instnc.eSize();
               Graph instnc_C = null; // for now
               for( int i = 0; i < motifs.length; i ++ )
                { final Graph motifs_i = motifs[i];
                  if( instnc_V == motifs_i.vSize()
                  &&  instnc_E == motifs_i.eSize() ) // hopeful
                   { if( instnc_C == null ) instnc_C = instnc.canonical();
                     if( motifs_i.edgesCorrespond(instnc_C) )  // match
                      { instances ++ ;
                        coVers.or(g_sgs.vertices());
                        g_E    -= instnc_E;          // unexplained Edges, &
                        g_Area -= instnc.maxEdges(); // A[,]'s unexplained area
                      }//if
                   }//if
                }//for i
             }//if

            g_sgs.advance();  // next subGraph of g
          }//while g_sgs

         final double // here are M's assumptions made explicit ...
            costV     = mdlV.nlPr_n(V),         // note, |V| > 0
            costNinst = Math.log(1 + maxInst),  // uniform(#instances)
            costInsts = motifs.length > 1
                      ? instances * Math.log(motifs.length) // which instances
                      : 0,
            costNumb = logFactorial(V)           // numbering...
                     - logFactorial(V - coVers.cardinality()) // ...
                     - logFactorial(instances),  // ... some saving.
      //    costNumb = log_nCk(V, coVers.cardinality()), // numbering  ???
            costE = mdlE.nlLH(Vector.doubles(new double[]{g_Area - g_E, g_E}));
            // NB. we give mdlE suff stats for a binomial ^^^^^^^^^^^^^^^^^

         final double nlPr = costV + costNinst + costInsts + costNumb + costE;

         return nlPr;
       }//nlPr(G)


      /** Generate a random Graph according to 'this' {@link M Model}. */
      public Graph random()                                         // random()
       { int V = mdlV.random().n();           // vSize
         int [] [] arr = new int [V] [V];     // arr[,]
         Series.Discrete eSer = mdlE.randomSeries();
         for( int r = 0; r < V; r  ++ )
            for( int c = 0; c < V; c  ++ )
             { arr[r][c] = eSer.elt_n(); eSer.advance(); } // arr[,]
         if( ! gType.isDirected ) // UnDirected
            for( int r = 0; r < V; r  ++ )
               for( int c = 0; c < r; c ++ )
                  arr[r][c] = arr[c][r];      // arr[,] is symmetric
         if( ! gType.selfLoops )
            for( int r = 0; r < V; r  ++ )
               arr[r][r] = 0;
         // Next, overlay with zero or more instances of motifs...
         int nInstnc = new Discretes          // #Inst, uniform(possible)
                   . Uniform(pair(zero, new Int((int) (V/sgMinV))))
                   . Mdl . random_n();
         if( verbosity > 0 ) // ZZZ, tracing, debugging
            System.out.print( " {MotifD.random():" + nInstnc + "_instances} " );
         int offset = 0;  // position along main diagonal
         for( int i = 0; i < nInstnc; i ++ )
          { final int r = (int) (RNG.nextDouble() * motifs.length); // uniform
            Graph m = motifs[r];          // chosen motif
            final int m_V = m.vSize();
            if( offset + m_V > V ) break; // won't fit
            for( int v = 0; v < m_V; v ++ )
               for( int w = 0; w < m_V; w ++ ) // isEdge(,) not adjacent(,) ...
                  arr[offset+v][offset+w] = m.isEdge(v, w) ? 1 : 0;
            offset += m.vSize();
          }//for i
         Matrix A = Matrix.ints(arr);
         Graph  R = Graph.dense(gType, A);
         return R; // easier to see when debugging
         // XXX debug ...
         // int [] perm = new int [V];  // shuffle Vertices...
         // new Permutation.Uniform(new Int(V)). Mdl . random(perm);
         // return Graph.dense(gType, A).renumbered(perm);
       }//random()

    }//MotifD.M class, the fully parameterised Model
   // -------------------------------------------------------------------------

   /** Display x compactly with "reasonable" accuracy. */
   private static String x2S(double x) { return Util.x2String(x, 4); }

   /** Temporary routine, debugging, testing. */
   private static void compare(Graph g, Model m1, Model m2)
    { double b1 = m1.nl2Pr(g), b2 = m2.nl2Pr(g);
      System.out.println( "|V|=" + g.vSize() + ", |E|=" + g.eSize() + ", "
                        + x2S(b1) + (b1 < b2 ? " < " : " > ")
                        + x2S(b2)  + " bits" );
    }//compare(,,)
   // ..........................................

   /** Developer's debugging tests; main() will "go away" one day.
       Also see {@link Test Test}.main(argv) */
   public static void main(String[] argv)                               // main
    { System.out.println("-- test MotifD --");

      graph.Type gType = Directed.unlabelled;
      mml.Discretes Poisson1 = mml.MML.Poisson0.shifted(1);  // int |V|>=1
      Cts meanV = real(25.0);
      double pEdge = 0.1;      // density, pr(Edge(v,w)) exists
      Vector alpha = Vector.doubles(new double [] {1-pEdge, pEdge}),
            alpha2 = Vector.doubles(new double [] {1-2*pEdge, 2*pEdge});

      Graphs.GERfixed  UGER = new Graphs.GERfixed(pair(gType, Poisson1));
      Graphs.GERfixed.M GER = UGER.sp2Model(0, 0, pair(meanV, alpha2));  // GER

      Graphs.GERadaptive  UGER2 = new Graphs.GERadaptive(
         triple(gType, Poisson1, Vector.doubles(new double[]{16,4})));
      Graphs.GERadaptive.M GER2 = UGER2.sp2Model(0, 0, meanV);          // GER2

      // a small world -ish Model...
      Graphs.Skewed  USW = new Graphs.Skewed(pair(gType, Poisson1));
      Graphs.Skewed.M SW = USW.sp2Model(0, 0, meanV);                     // SW

      Graph [] motifs = new Graph [] {
        Directed.dense(new int[][]{{0,1,1,1}, {1,0,1,1}, {1,1,0,1}, {1,1,1,0}})
        };
      MotifD  UMM = new MotifD(triple(gType, Poisson1, pair(three, four)));
      MotifD.M MM = UMM.new M(0, 0,
                      pair(meanV, Vector.values(motifs)));  // MM

      Graphs.M rival = GER2;  // say
      // Graphs.M rival = SW;  // ???
      System.out.println( "   " + rival + "\nv. " + MM + "\n" );
      for( int i = 0; i < 3; i ++ )
         compare(rival.random(), rival, MM);  // rival should win, usually
      System.out.println();
      for( int i = 0; i < 3; i ++ )
         compare(MM.random(),  rival, MM);    // MM should win, usually

      Vector ds = MM.random(2);
la.util.Timer timer = new la.util.Timer(); timer.start();
      MotifD.M MMest = (M) UMM.estimator(fiveR).ds2Model( ds );         // MMest
System.out.println( "-- " + (double)timer.lastTime()/1000 + " secs --" );

      System.out.println( "\n|ds| = " + ds.nElts() + "\n"
       + MMest.msgBits() + "=" + MMest.msg1bits() + "+" + MMest.msg2bits() +
            " bits\n   Inferred: " + MMest + "\n"
       + GER.nl2LH(GER.stats(ds))   + " by " + GER  + ",\n"
       + GER2.nl2LH(GER2.stats(ds)) + " by " + GER2 + ",\n"
       + SW.nl2LH(GER2.stats(ds))  + " by " + SW );

      System.out.println("-- done MotifD --");
    }//main

 }//MotifD class, the UnParameterised Model

// (c)Lloyd Allison, Sep 2014, Mon 11 Dec 2017. Licence to 'MotifD' is Gnu GPL.
