//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.la.Type;
import static la.maths.Maths.*;  // for nlPrs2odds(nlPr)
import la.maths.Vector;          // for Vector etc.
import la.maths.Vector.Slice;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The UnParameterised NaiveBayes FunctionModel; the fully parameterised
    FunctionModel is {@link M NaiveBayes.M}. For a Discrete Bounded
    variable, od:O, and arbitrary variable(s), id:I,
    NaiveBayes I&rarr;O is based on a Dependent Model,
    {@link #dpndt dpndt}:O&times;I,
    [<A HREF="http://www.allisons.org/ll/Images/People/Bayes/"
        TARGET="LL">Bayes's&nbsp;theorem</A>], and the (naive) assumption
    that id's components are independent of each other. */
public class NaiveBayes extends UPFunctionModel                   // NaiveBayes
 {
   /** The Type of the output variable, od:Otype. */
   private final Type.Discrete Otype;

   /** lwb and upb, the bounds on the output datum, od. */
   public final Value lwb, upb;

   /** lwb_n and upb_n the {@link #lwb bounds} on the
       output datum, od, as ints. */
   public final int lwb_n, upb_n;

   /** The (UnParameterised) Dependent Model, made of
       &lang;O,&nbsp;O&rarr;I&rang;, that we are going to turn around
       (invert) to get I&rarr;O as desired.  Also see {@link M#dpndt_m}.  */
   public final Dependent dpndt;

   /** Given definition parameter, dp, being an UnParameterised
       {@link Dependent Dependent} Model, dp:O&times;I, construct
       an UnParameterised NaiveBayes FunctionModel of I&rarr;O. */
   public NaiveBayes(Value dp)
    { super(dp);
      dpndt = (Dependent) dp;
      Discretes.Bounded Iupm = (Discretes.Bounded) dpndt.upm;
      Otype = (Type.Discrete) Iupm.lwb().type();
      lwb = Iupm.lwb();  lwb_n = lwb.n();
      upb = Iupm.upb();  upb_n = upb.n();
    }//constructor

   /** Given a data-set ds, return statistics ss=ds.[lo,hi) the data itself.
       (Note {@link M#nlLH M.nlLH(ss)} really needs ss to be the raw data.) */
   public Vector stats(Vector ds, int lo, int hi) { return ds.slice(lo, hi); }

   /** Combine {@link #stats(Vector,int,int) stats} ss0 and ss1, either
       additively (add=true) or negatively (add=false). */
   public Vector stats(boolean add, Value ss0, Value ss1)
    { return ((Vector) ss0).combine(add, (Slice) ss1); }

   /** Return the Estimator for NaiveBayes; any parameter(s), ps,
       is passed to the {@link #dpndt Dependent}'s Estimator.
       <I>It seems that msg1 should be smaller than here:</I>
       The condModels of I got from the O&rarr;I need not be so precise
       because the Is are actually common knowledge?  */
   public Est estimator(final Value ps)
    { return new Est(ps)
       { /** Given statistics, ss=ds itself, of a
             data-set ds, return an {@link M M}. */
         public M ss2Model(Value ss)
          { final Dependent.M dpndt_m = (Dependent.M)
               dpndt.estimator(ps).ds2Model(
                  ((Vector)ss).cols(new int[]{1, 0}) ); // NB. cols flipped
            M fm = new M(0, 0, dpndt_m.statParams());
            return (M) fm.asGiven( dpndt_m.msg1(), /* msg2= */ fm.nlLH(ss) );
            // ... remember, dpndt codes id and od, but M only codes od.
          }//ss2Model(ss)
    }; }

   /** Given two-part message lengths, and statistical parameters
       sp, return a fully parameterised {@link M model}. */
   public M sp2Model(double msg1, double msg2, Value sps)
    { return new M( msg1, msg2, sps ); }
   // ..........................................

   /** The fully parameterised NaiveBayes FunctionModel I&rarr;O;
       the UnParameterised FunctionModel is {@link NaiveBayes}.
       pr(od|id)&nbsp;=&nbsp;pr(id&amp;od)/pr(id) = pr(od).pr(id|od)/pr(id),
       so pr(od|id)&nbsp;&prop;&nbsp;pr(od).pr(id|od);
       calculate for each case of od and normalise.  */
   public class M extends UPFunctionModel.M                  // NaiveBayes.M
    {
      /** The "backwards" Dependent Model of (O&times;I), that
          we are turning around, I&rarr;O;  it is made of
          &lang;O,&nbsp;O&rarr;I&rang; and we want I&rarr;O.  */
      public final Dependent.M dpndt_m;

      /** {@link #dpndt_m dpndt_m}'s Model of O. */
      public final Discretes.Bounded.M Omdl;

      /** {@link #dpndt_m dpndt_m's} "backwards" FunctionModel of O&rarr;I. */
      public final FunctionModel O2I;

      /** sp is for {@link #dpndt_m dpndt_m}
                  = {@link #dpndt   dpndt}.apply(sp). */
      public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp);
         dpndt_m = (Dependent.M) dpndt.apply(sp);
         Omdl = (Discretes.Bounded.M) dpndt_m.im;
         O2I  = dpndt_m.fm;
       }//constructor

      /** Assumes {@link #stats(Vector,int,int) statistics} 'ss'
          are the data-set itself. It all hinges on
          {@link #condModel(Value) condModel}.  */
      public double nlLH(Value ss)
       { return sumNlPr((Vector) ss); }//nlLH(ss)

      /** Return a {@link MultiState MultiState} Model of output datum od
          conditional on given input datum id.  Try every possible value, v,
          for od, calculate joint probability pr(v&amp;id), normalise these
          probabilities, which yields pr(v|id).  */
      public Model condModel(Value id)
       { double [] nlPr = new double [upb_n - lwb_n + 1];
         double []   pr = new double [upb_n - lwb_n + 1];
         for( int od_n = lwb_n; od_n <= upb_n; od_n ++ ) // upb-lwb+1 times
          { final Discrete od = Otype.int2value(od_n);   // id -> od
            nlPr[od_n-lwb_n] = Omdl.nlPr_n(od_n) + O2I.condNlPr(od, id);
         // nlPr(od|id) ~ nlPr(od) + nlPr(id|od)
          }
         return dpndt.upm.apply( Vector.doubles(nlPrs2odds(nlPr)) );
       }//condModel(id)

    }//NaiveBayes.M, the FunctionModel class

 }//NaiveBayes class, the UPFunctionModel

// (c) Lloyd Allison, 2012, Tue 31 Oct 2017. Licence to 'NaiveBayes' is Gnu GPL.
