//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.maths.Vector;
import la.maths.Vector.Slice;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The UnParameterised NearInverse Model of positive reals,
    (0,&nbsp;&infin;).  We would sometimes like to use a prior,
    <NOBR>pdf(x) ~ 1/x, x&gt;0,</NOBR> (then, e.g., 0.1+/-.005, 1.0+/-.05,
    10.+/-.5, etc., would all have the same probability) but it is improper
    because 1/x integrates to infinity;.  NearInverse is proper and close
    to 1/x, the smaller delta, &gt;&nbsp;0, is the closer.  There is
    no estimator &mdash; the intended use is as a prior close to 1/x.
    See the fully parameterised model {@link NearInverse.M}.  */
public class NearInverse extends Continuous                      // NearInverse
 { public NearInverse(Value t)
    { super(t); if( t != triv ) error( ", t!=triv" ); } //constructor

   /** Sufficient statistics, <NOBR>ss = stats(ds,lo,hi),</NOBR> of elements
       [lo,hi) of a data-set ds; here <NOBR>ss = ds.[lo,hi)</NOBR>.
       More on stats {@link UPModel#stats here}. */
   public Vector stats(Vector ds, int lo, int hi)
    { return ds.slice(lo, hi); }

   public Vector stats(boolean add, Value ss0, Value ss1)
    { return ((Vector) ss0).combine(add, (Slice) ss1); }

   public M sp2Model(double msg1, double msg2, Value sp)
    { return new M(msg1, msg2, sp); }
   // ...............................

    /** The fully parameterised NearInverse Model of positive reals.
        The "statistical parameter" is&nbsp;&delta;.
        Also see the UnParameterised model {@link NearInverse}.  */
    public class M extends Continuous.M                        // NearInverse.M
     { /** The "statistical parameter". */
       protected final Value.Cts delta;
       /** The "statistical parameter" as a double. */
       private final double      delta_x;
       /** The log of the normalising constant. */
       public  final double logIntegral;

      /** The standard constructor for NearInverse Model;
          statistical parameter delta must be a small Value.Cts in
          (0,&nbsp;1) such as&nbsp;0.1.  The integral,
          <NOBR>&int; = &int;<SUB>[0,1]</SUB> {1/x<SUP>(1-&delta;)</SUP>}
           + &int;<SUB>[1,&infin;]</SUB> {1/x<SUP>(1+&delta;)</SUP>} </NOBR>
             <NOBR> = { (x<SUP>&delta;</SUP>)  / &delta;}[0,1]
                    + {-(x<SUP>-&delta;</SUP>) / &delta;}[1,&infin;] </NOBR>
             =&nbsp;2/&delta;
          gives the normalising constant.   */
      public M(double msg1, double msg2, Value delta)
       { super(msg1, msg2, delta);
         this.delta   = (Value.Cts) delta;
         this.delta_x = this.delta.x();
         assert delta_x > 0 && delta_x < 1 : this.delta + " not in (0,1)";
         logIntegral = Math.log(2.0 / delta_x);
       }//constructor

      /** A possible use is for &sigma;'s prior in {@link MML#Normal}. */
      public M(double delta_x)
       { this(0, 0, Value.real(delta_x)); }

      /** Nearly -&nbsp;log(1/x), that is +&nbsp;log(x), but not quite,
          corresponding to a pdf(x) of ~1/x, but not quite
          (as always, <NOBR>x &gt; 0,</NOBR> of course).
          Note, nlPdf(x) is continuous but its derivative has a
          jump at x&nbsp;=&nbsp;1. */
      public double nlPdf_x(double x)
       { return logIntegral
              + ((x >= 1 ? 1+delta_x : 1-delta_x) * Math.log(x));
       }//nlPdf_x(x)

      /** Assumes that {@link NearInverse#stats(Vector,int,int) statistics}
          are the data-set itself. */
      public double nlLH(Value ss)
       { return sumNlPr((Vector) ss); }

    }//NearInverse.M class, the Model

 }//NearInverse class, the UPModel

// (c)Lloyd Allison 2012, Sat 11 Nov 2017. Licence to 'NearInverse' is Gnu GPL.
