//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Lexical;
import la.util.Util;   // standard functions
import la.la.Value;
import static la.maths.Maths.*;
import la.maths.Vector;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The UnParameterised Poisson0 Model. For most purposes {@link MML#Poisson0}
    and {@link Poisson0UPM.M} should be enough, with little need to
    call upon Poisson0UPM directly? (The '0' is to remind us that
    the Model is for integers &ge;&nbsp;0.) */
public class Poisson0UPM extends Discretes                       // Poisson0UPM
 { public Poisson0UPM(Value t /* = triv */)
    { super(t); if( t != triv ) error( ", t=" + t ); }//constructor

   /** Given m1, m2 and &alpha;, return a fully parameterised
       {@link M Poisson0}. */
   public M sp2Model(double m1, double m2, Value alpha)
    { return new M(m1, m2, alpha); }

   /** Calculate the {@link #stats(Vector) sufficient} statistics,
       <NOBR>[N, sum, sum log factorials]</NOBR>, of a data-set ds.
       Note that N and sum are sufficient to estimate {@link M#alpha &alpha;}
       but sum log factorials is needed for {@link M#nlLH nlLH}
       (and hence msg2). */
   public Vector stats(Vector ds, int lo, int hi)
    { double N = 0, sum = 0, sumLogFact = 0;
      for( int i = lo; i < hi; i ++ )
       { double ni = ds.elt(i).x(); double wi = ds.wt(i);
         N += wi; sum += wi*ni; sumLogFact += wi*logFactorial(ni);
       }
      return Vector.doubles(new double[]{N, sum, sumLogFact});
    }//stats(ds,lo,hi)

   public Vector stats(boolean add, Value ss0, Value ss1)
    { return (Vector)
         ((Vector) ss0).bOp(add?Lexical.plus:Lexical.minus, (Vector) ss1);
    }

   /** The Estimator has a parameter, AA, the parameter and mean
       of the {@link ExponentialUPM.M Exponential} prior.  */
   public Est estimator(final Value AA /* Cts */)
    { return new Est(AA)
       { /** 'A' the {@link #priorPdf prior}'s parameter, as a double. */
         public final double A;
         { A = AA.x(); }//constructor

         /** The pdf of the <I>prior</I> on the Poisson0's &alpha;,
             <NOBR>h(&alpha;) = (1/A).e<SUP>-&alpha;/A</SUP>;</NOBR> the
             prior itself has mean&nbsp;A so it is usually quite easy to
             think up a suitable value. */
         protected double priorPdf(double alpha)
          { return Math.exp( - alpha / A) / A; }

         /** The MML estimator is
             <NOBR>&alpha; ~ (sum+1/2) / (N+1/A)</NOBR>. */
         public M ss2Model(Value ss)
          { // Model prior = new Exponential.apply(Value.real(A));
            Vector v = (Vector) ss;  // i.e. [N, sum, ...]
            double N = v.elt(0).x(), sum = v.elt(1).x();
            if( N <= 0 ) return new M(0, 0, Value.real(A));  // !
            double alpha = (sum + 0.5) / (N + 1/A); // MML estimate
            double halfLogF = (Math.log(N) - Math.log(alpha)) / 2.0;
            Value.Cts alphaR = Value.real(alpha);
            // Value.Cts alphaR = Value.cts(alpha,
            //           Math.exp(-halfLogF-Math.log(latticeConstant(1))/2));
            double msg1 = - Math.log(priorPdf(alpha))        // pdf
                        + halfLogF
                        + Math.log(latticeConstant(1)) / 2;
            double msg2 = new M(0,0,alphaR).nlLH(ss) + 0.5;  // i.e., + D/2
            return new M(msg1, msg2, alphaR);
          }//ss2Model(ss)
       };//new Estimator
    }//estimator(AA)

   public String toString()
    { return this == Poisson0 ? "Poisson0" : super.toString(); }
   // ............................................

   /** The fully parameterised Poisson0 Model (probability distribution)
       on integers in <NOBR>[0, &infin;);</NOBR> its one statistical
       parameter, {@link #alpha &alpha;}, is both the mean and variance.
       Also see the UnParameterised {@link MML#Poisson0}, and
       [<A HREF="http://www.allisons.org/ll/MML/Discrete/Integers/Poisson/"
           TARGET="LL">www</A>].  */
   public class M extends Discretes.M                          // Poisson0UPM.M
    { /** The mean (and variance) of the Poisson0 distribution.  */
      public  final double alpha;
      private final double log_alpha;

      public M(double msg1, double msg2, Value alpha)
       { super(msg1, msg2, alpha);
         this.alpha = ((Value.Cts) alpha). x();
         log_alpha = Math.log(this.alpha);
       }//constructor

      /** The negative log of <NOBR> pr(n|&alpha;) =
          (e<SUP>-&alpha;</SUP>).(&alpha;<SUP>n</SUP>) / n!, </NOBR>
          integer <NOBR>n &ge; 0</NOBR>. */
      public double nlPr_n(int n)
       { if( n < 0 ) error( ".nlPr_n(n): n=" + n + " < 0" );
         return alpha - n * log_alpha + logFactorial(n);
       }//nlPr(n)

      /** Given sufficient statistics, ss = stats(ds), of a data-set ds,
          return the negative log LikeliHood of ds. */
      public double nlLH(Value ss)
       { Vector v = (Vector) ss;  // i.e. [N, sum]
         double N = v.elt(0).x(), sum = v.elt(1).x(),
                sumLogFact = v.elt(2).x();
         return N*alpha - sum*log_alpha + sumLogFact;
       }//nlLH(ss) where ss = stats(ds) of a data-set ds

      /** Return a random number from 'this' Poisson distribution;
          calls {@link #random_n()}. */
      public Value.Int random()
       { return new Value.Int(random_n()); }

      /** ??? Poisson0.random_n() could be made much(!) more efficient!!
          This implementation is just for completeness.  */
      public int random_n()
       { int i;  double rndm = Math.random();  // 0 <= rndm < 1, uniformly
         for( i = 0; rndm >= 0 && i < alpha*100; i++ ) rndm -= pr_n( i );
         return i-1;
       }//random_int()

    }//Poisson0.M class, the fully parameterised Model

 }//Poisson0UPM, the UPModel

// (c) Lloyd Allison 2012, Wed 13 May 2015. Licence to 'Poisson0UPM' is Gnu GPL.
