//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.util.Series;
import la.la.Value;
import la.maths.Vector;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** Fully parameterised (Time-) Series Models of data Series (Vectors).
    Most instances of SeriesModel will likely be instances of
    {@link UPSeriesModel.M}. Note that all
    {@link UPSeriesModel.Length Length} SeriesModels
    have an explicit sub-Model of length, others not if, for example, a
    data Series ends in an 'end' element (think Unix strings and&nbsp;'\0').
    Also see {@link Model Model}, and {@link FunctionModel FunctionModel}. */
public abstract class SeriesModel extends Model                // * SeriesModel
 { public SeriesModel(double msg1, double msg2, Value sp)
    { super(msg1, msg2, sp); }//constructor

   /** The negative log probability of datum <I>Vector</I> vec. */
   public abstract double nlPr(Value vec);

   /** Create (start) an {@link Analysis Analysis} of Scannable sv. */
   public abstract Analysis analysis(Scannable sv);           // * analysis(sv)
   // .........................................................................

   /** An Analysis of a Scannable Value (Series).  Roughly, for a
       Series of data {@link #elt produce} a Series of Models. */
   public static abstract class Analysis extends Series           // * Analysis
    { /** The Series produced by the given data Series. */
      protected final Series ds;  // the data Series

      /** Construct (prepare for) the Analysis of
          a given Scannable Value sv. */
      public Analysis(Scannable sv)
       { this.ds = sv.toSeries(); }//constructor

      /** Return ds.hasSome() where ds is the given data Series. */
      public boolean hasSome() { return ds.hasSome();  }           // hasSome()

      /** Return ds.position() where ds is in the given data Series.
          When hasNone(), position() is the length of the data Series. */
      public int    position() { return ds.position(); }          // position()

      /** Return the Model of the current {@link #eltD data element};
          also see {@link #elt} and {@link #finalModel()}. */
      public abstract Model eltM();                                 // * eltM()

      /** If and when the data Series is exhausted, return
          the Model of a future, hypothetical data element.
          Also see {@link #eltM()}. */
      public Model finalModel()                                 // finalModel()
       { if( hasSome() )
            throw new RuntimeException( this + ".finalModel()" );
         else return eltM();
       }//finalModel()

      /** Return the current element of the data Series;
          also see {@link #elt}. */
      public Value eltD() { return ds.elt(); }                        // eltD()

      /** Return a pair being the Model for the current
          data element and that data element, that is
          &lang;{@link #eltM() eltM()},&nbsp;{@link #eltD() eltD()}&rang;.
          Throw a RuntimeException if there is no data element.
          Also see {@link #finalModel()}. */
      public Value.Tuple elt()                                         // elt()
       { if( hasNone() )
            throw new RuntimeException( this + ".elt(), I am exhausted!" );
         return Value.pair(eltM(), eltD());
       }//elt()

      /** The probability of the current data element. */
      public double    pr() { return eltM().pr(eltD()); }

      /** The negative log probability of the current data element.
          Also see {@link #remainingNlPr()}. */
      public double  nlPr() { return eltM().nlPr(eltD()); }

      /** The neg log (base 2) probability of the current data element. */
      public double nl2Pr() { return eltM().nl2Pr(eltD()); }

      /** The total nlPr() of any remaining elements of the Series,
          including the current element if there is one. */
      public double remainingNlPr()
       { double ans = 0;
         while( hasSome() ) { ans += nlPr(); advance(); }
         return ans;
       }//remainingNlPr()

      /** Advance to the next {@link #eltM() Model} and the
          next {@link #eltD() data} element. */
      public abstract void advance();                            // * advance()

    }//SeriesModel.Analysis class
   // .........................................................................

   /** Calls {@link #asGiven(double,double) #asGiven(0, msg2)}. */
   public SeriesModel asGiven(double msg2)
    { return (SeriesModel) super.asGiven(0, msg2); }

   /** Return a clone of 'this' SeriesModel but with msg1 and
       msg2 as specified.
       Also see {@link Model#asGiven(double,double) asGiven(,)}. */
   public SeriesModel asGiven(double msg1, double msg2)
    { return (SeriesModel) super.asGiven(msg1, msg2); }

   /** Treat 'this' fully parameterised <I>SeriesModel</I> as an
       UnParameterised one.  Also see {@link Model#asUPModel}. */
   public UPSeriesModel asUPModel()
    { final SeriesModel sm = this; // enclosing SeriesModel
      return new UPSeriesModel(Value.triv)
       { public SeriesModel sp2Model(double msg1, double msg2, Value sp)
          { if( msg1 != 0 ) error( ".sp2Model("+msg1+","+msg2+","+sp+")" );
            return sm.asGiven(msg2); // !
          }//sp2Model(,,)
         public Value stats(Vector ds, int lo, int hi)
          { return sm.stats(ds, lo, hi); }
         public Value stats(boolean add, Value ss0, Value ss1)
          { return sm.stats(add, ss0, ss1); }
         public Estimator estimator(Value ps/* triv */)
          { return sm.asEstimator(ps); } // "estimates" sm
         public String toString() { return sm + ".asUPModel()"; }
    }; }//asUPModel()

 }//SeriesModel class

// (c) Lloyd Allison, 2012, Tues 3 December 2013. Licence is Gnu GPL.
