//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import java.util.Random;  // Is it good?

import la.util.*;   // standard functions, and also class Series
import la.la.Value;
import static la.la.Value.*;
import la.maths.*;  // for Vector etc.

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference:
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** {@link #main} runs a few <I>simple</I> tests on {@link README mml}.
    This is nowhere near an exhaustive set of tests.  */
public class Test                                                       // Test
 {
   /** Display x compactly with "reasonable" accuracy. */
   private static String x2S(double x) { return Util.x2String(x, 4); }

   /** Print a line of 'n' dashes.  */
   private static void dashed(int n)
    { for( int i = 0; i < n; i ++ ) System.out.print("-");
      System.out.println();
    }

   private static int ttyWidth   = 80;
   private static int ttyWidth_1 = ttyWidth - 1;
   private static Value.Bool tt = ttrue, ff = ffalse;
   // -------------------------------------------------------------------------

   /** Use Estimator 'est' to fit a Model 'm' to data 'ds', then
       {@link #exercise(mml.Model,Vector) exercise(m,ds)}. */
   public static void exercise(Estimator est, Vector ds)       // exercise(...)
    { System.out.println( "Estimator: " + est + " ..." );
      Model m = est.apply(ds);
      exercise(m, ds);
    }//exercise(e,ds)

   /** Run Model 'm' through a few simple tests on data-set 'ds'. Also tries
       {@link Model#random() m.random()} (which not all Models implement).
       Also see {@link #exercise(Estimator,Vector)}.  */
   public static void exercise(Model m, Vector ds)
    { System.out.print( "Model: " + m + " : " + m.type()
         + "\n   .msg() = " + x2S(m.msg())
         + " nits = " + x2S(m.msg1()) + "+" + x2S(m.msg2())
         + "\n          = " + x2S(m.msgBits())
         + " bits = " + x2S(m.msg1bits())+"+"+x2S(m.msg2bits())
         + (m.msg1() <= 0 ? "  (given?)" : "") + "\n   " );
      double prSum = 0, bitSum = 0;
      for( int i = 0; i < ds.nElts(); i ++ )
       { Value di = ds.elt(i);                        // di, the i-th datum
         double pr_di = m.pr(di), b_di = m.nl2Pr(di);
         prSum += pr_di;  bitSum += b_di;
         if( i < 10 )
          { System.out.print( "pr(" + di + ")=" + x2S(pr_di)
                            + "~" + x2S(b_di) + " bits, " );
            if( m instanceof HasPdf )
               System.out.print( "pdf=" + x2S(((HasPdf) m) .pdf(di)) + ", " );
          }
         else if( i == 10 ) System.out.print( " ... " );
       }

      double nl2LH = m.nl2LH(m.stats(ds)), m1 = m.msg1bits();
      System.out.println( "\n   (sum_pr=" + x2S(prSum)
           + "), sum_bits="  + x2S(bitSum)
           + " =?= nl2LH="+ x2S(nl2LH) + " bits," +
           (m1 > 0 ? " (+" + x2S(m1) + " = " + x2S(nl2LH+m1) + ")," : "") );

      System.out.print( "   random():" );
      try{ System.out.print( m.random(10) + "..." ); }
      catch(UnsupportedOperationException e)
       { System.out.print( " " + e.getMessage() ); }
      System.out.println( );
      dashed(ttyWidth_1);
    }//exercise(m,ds)
   // -------------------------------------------------------------------------

   /** Do a number of 'trials', each time using Model 'gen' to generate
       a data-set of 'N' values, and then fitting Models by
       estimators 'e1' and 'e2' to that data-set to see which
       one wins in the minimum message length challenge.  Also see
       {@link #compete(int,int,double,Model,Estimator,Estimator)}. */
   public static void compete(int trials, int N, Model gen,     // compete(...)
                              Estimator e1, Estimator e2)
    { if( gen instanceof Continuous.M )
         throw new RuntimeException( "wrong version of compete" );
      compete(trials, N, /* unused */ -1.0, gen, e1, e2);
    }//compete(...)

   /** Version of {@link #compete} with a specified 'AoM', thus
       allowing for competition between {@link Continuous.M} Models. */
   public static void compete(int trials, int N, double AoM, Model gen,
                              Estimator e1, Estimator e2)
    { System.out.println( "compete(" + trials + "x, N=" + N + ", AoM=" + AoM
                        + ", gen=" + gen + ", e1=" + e1 + ", e2=" + e2 + ")" );
      int w1=0, w2=0, dr=0;
      for(int i = 0; i < trials; i ++ )                        // each trial
       { Vector ds = gen instanceof Continuous.M               // data-set
                        ? ((Continuous.M) gen) .random(N, AoM)
                        : gen.random(N);
         Model m1 = e1.apply(ds), m2 = e2.apply(ds);           // m1 & m2
         double ml1 = m1.msgBits(), ml2 = m2.msgBits();        // NB. bits
         if( ml1 < ml2 ) w1 ++ ; else if( ml2 < ml1 ) w2 ++ ; else dr ++ ;
         char ch1 = ml1 < ml2 ? '*' : ' ',
              ch2 = ml2 < ml1 ? '*' : ' ';
         System.out.println( i + " : " + ch1 + x2S(ml1) + " bits " + m1
                           +  "\n v. " + ch2 + x2S(ml2) + " bits " + m2 );
       }
      System.out.println( "m1:draw:m2 = " + w1 + ":" + dr + ":" + w2 );
      dashed(ttyWidth_1);
    }//compete(...)
   // -------------------------------------------------------------------------

   /** Run a few, very(!) simple tests. */
   public static void main(String[] argv)                               // main
    { System.out.println("-- run mml.Test --");

      for( int i = 1; i < 4; i ++ )
         System.out.print("   +inform("+i+")=" + x2S(informativeIncrement(i)));
      System.out.println( " nits");

      // slightly test Java's random number generator, class Random
      Random[] rndm = new Random[]{new Random(), new Random()};
      for( int k = 0; k < 2; k ++ )
       { int[] r = new int[2];
         for( int j = 0; j < 2; j ++ )
          { System.out.print( "   Random_" + j + ":" );
            for( int i = 0; i < 5; i ++ )
             { r[j] = rndm[j].nextInt();
               System.out.print(" " + r[j] + ",");
             }
            System.out.println( "..." );
            rndm[j].setSeed(109);  // now both rndm[.] should be identical
          }
         if( k >= 1 ) break;
         if( r[0] == r[1] )  // gnu Java used to fail the test (!) ...
            System.out.println( "*** Warning: rndm[0] is rndm[1] ***" );
       }
      dashed(ttyWidth_1);

      Vector nonNegInts = Vector.ints(
         new int[] {0, 1,2,4,8,16, 2,3,4,5,10} );   // all >= 0

      Vector posCts= Vector.doubles(
         new double[]{1.0, 2.0, 4.0, 8.0, 16.0, 0.5, 1.414, 2.7, 3.14, 10.0},
         /* AoM= */ 0.1 );

      // The following data was actually generated by an N(0,1)
      Vector N_0_1_ish = Vector.doubles(new double[]
         {-0.75, -2.04, -1.04, 0.0, 1.45, 0.32, -0.44, 1.24, -0.97, 1.17},
         0.01 );

      // Following "columns" may be combined in various ways
      Vector gender  = Vector.values( new Value[]         // 4 women; 4 men
                               {ff, ff, ff, ff,    tt, tt, tt, tt});
      Vector height  = Vector.doubles(new double[]  // w tend to be shorter
                               {1.5,1.6,1.6,1.7,  1.6,1.7,1.7,1.8}, 0.1);
      Vector mass    = Vector.doubles(new double[]  // w tend to be lighter
                               {51, 66, 57, 62,   64, 71, 66, 69},  1.0);
      Vector age     = Vector.doubles(new double[]
                               {33, 21, 51, 25,    30, 20, 40, 50}, 1.0);
      Vector leftwing= Vector.values( new Value[]  // young tend to be left
                               {tt, tt, ff, tt,    ff, tt, ff, ff});
      Vector people  = Value.tuple( new Vector[]
                               {gender, height, age, leftwing}
                           ).zip();

      Value.Tuple zero_3 = Value.pair(zero, three),
                   one_4 = Value.pair(one, four);
      Vector     oneTo4s = Vector.ints(new int[]{1,2,3,4,2,4});  // [1,4]*

// Discretes:
      Model U_1_4 = new Discretes.Uniform(one_4) .apply(triv);  // of [1..4]
      exercise( U_1_4, oneTo4s );
      Model M_1_4 = new MultiState(one_4).estimator(triv).apply(oneTo4s);
      exercise( M_1_4, oneTo4s );
      exercise( new MultiState(one_4).estimator(triv), oneTo4s );

// Discretes, Adaptive:
      Model A_1_4 = new Adaptive(
                   Value.pair(one_4,  // bounds
                              Vector.doubles(new double[]{2,2,2,2}))).Mdl;
      exercise( A_1_4, oneTo4s );

// Simplex:
      Model sm = new Simplex.Uniform(three).apply(triv);
      exercise( sm, Vector.values(new Value[]{ M_1_4.statParams() }) );

// Ints:
      exercise( logStar0,    nonNegInts );
      exercise( WallaceInt0, nonNegInts );
      exercise( Geometric0 .estimator(Value.real(4)), nonNegInts );
      exercise( Poisson0   .estimator(Value.real(4)), nonNegInts );
        // NB. use of same prior (estimator param) gives slight mu change :-)
      exercise( Poisson0.shifted(-1).estimator(Value.real(4)), nonNegInts );

// Continuous:
      Vector ordersOfMag = Vector.values( new Value[]
         {Value.cts(100.0, 10.0),
          Value.cts( 10.0,  1.0),           // these data should
          Value.cts(  1.0,  0.1),           // have roughly same nlPr
          Value.cts(  0.1,  0.01),          // under NearInverse...
          Value.cts(  0.01, 0.001) } );
      exercise(new NearInverse(triv).apply(Value.real(0.1)), ordersOfMag);
     exercise(new HeavyTail.Over_x1(triv).apply(Value.real(1.1)), ordersOfMag);

      exercise(new Continuous.Uniform(
         Value.pair(Value.real(-5), Value.real(5)))  // [-5.0, 5.0]
         .estimator(triv), N_0_1_ish);

      exercise(Exponential.apply(Value.real(2)), posCts);

      exercise(Gamma.apply(pair(/*k*/two, /*theta*/two)), posCts);

// Normals:
      final Value.Tuple q_10_10_pt1_ten
       = Value.quad(negTenR, tenR,   // mu bounds
                    point1, tenR);   // sig bounds

      Estimator nEst = Normal.estimator( q_10_10_pt1_ten );
      exercise( nEst, N_0_1_ish );

      exercise( new NormalMu(Value.real(0.0))  // mu = 0.0
                  .estimator(Value.pair(point1, tenR)),
         N_0_1_ish );

// log-Normal:  /** See {@link MML#logNormal} */
      Estimator logNEst = logNormal.estimator( q_10_10_pt1_ten );
      exercise( logNEst,
                N_0_1_ish.map(la.la.Library.exp) );  // !

      // the two following are equivalent probability distributions...
      exercise( Normal.transform(la.la.Library.log)
                      .apply(Value.pair(zero,one)),   // (i)
                N_0_1_ish.map(la.la.Library.exp) );
      exercise( Normal.apply(Value.pair(zero,one))
                      .transform(la.la.Library.log),  // (ii)
                N_0_1_ish.map(la.la.Library.exp) );

// Laplace:
      Model L01 = Laplace.apply(Value.pair(zeroR, oneR));
      Estimator lEst = Laplace.estimator(Value.quad(negTenR, tenR, // mu range
                                             point1, tenR));        // b
      exercise(  lEst, N_0_1_ish );

compete(10, 20, 0.1, N01, nEst, lEst);
compete(10, 20, 0.1, L01, nEst, lEst);

// BestOf, i.e., choose one or the other
     BestOf N_or_L = new BestOf( pair(Normal, Laplace) );
     exercise( N_or_L.estimator( pair( Value.quad(negTenR, tenR,point1, tenR),
                                       Value.quad(negTenR, tenR,point1, tenR)
                               )     ),
               N_0_1_ish );

// Direction, Vector:
      Vector r3s =  // 3D Vectors, in R^3
         Vector.values(
            new Vector[]        // norms are roughly 3.8+/-0.1 ...
             { Vector.doubles(new double[]{2.2, 2.2, 2.2}, 0.001),
               Vector.doubles(new double[]{2.0, 2.0, 2.5}, 0.001),
               Vector.doubles(new double[]{2.0, 2.5, 2.0}, 0.001),
               Vector.doubles(new double[]{2.5, 2.0, 2.0}, 0.001),
               Vector.doubles(new double[]{3.7, 0,   0  }, 0.001),
               Vector.doubles(new double[]{0,   3.9, 0  }, 0.001),
               Vector.doubles(new double[]{0,   0,   3.8}, 0.001)
             }        );

      Vector r3tricky = Vector.values( // to check the effect of...
            new Vector[]               // ...normalisation on AoM...
             { Vector.doubles(new double[]{0, 1, 0}, 0.001),
               Vector.doubles(new double[]{0, 1, 0}, 0.001).normalised(),
               Vector.doubles(new double[]{0, 4, 0}, 0.001),
               Vector.doubles(new double[]{0, 4, 0}, 0.001).normalised(),
               Vector.doubles(new double[]{1, 1, 1}, 0.001).normalised()
             }        );
      // NB. Direction only! ...
      exercise( new Direction.Uniform(Value.three).Mdl, r3tricky );
      exercise( vMF3.estimator(triv),                   r3tricky );

      exercise( new R_D.NrmDir(Value.pair(Normal,     // Magnitude and...
                                          new Direction.Uniform(three)))
                   .estimator(Value.pair(Value.quad(threeR, fourR, point1, twoR),
                                         triv)),
                r3s);
         // 3.8 the distance between adjacent C_alphas in a protein backbone

      exercise( new R_D.NrmDir(Value.pair(Normal,     // Magnitude and...
                                          vMF3))      // ...direction
                   .estimator(Value.pair(
                       Value.quad(threeR, fourR, point1, oneR), // mu, sig, bounds
                       triv)),
                r3s);

// Multivariates...

//  Bivariate, Cartesian (2D Vectors in R^2) and Polar ...
      Vector cartesians = Vector.values(
         new Vector []
          { Vector.doubles(new double [] { 1.0,  0.0}, 0.1),     // E
            Vector.doubles(new double [] { 0.6,  0.8}, 0.1),     // NNE
            Vector.doubles(new double [] {-0.6,  0.8}, 0.1),     // NNW
            Vector.doubles(new double [] {-0.6, -0.8}, 0.1),     // SSW
            Vector.doubles(new double [] { 0.6, -0.8}, 0.1),     // SSE
            Vector.doubles(new double [] { 0.0,  1.0}, 0.1),     // N
            Vector.doubles(new double [] {-1.0,  0.0}, 0.1),     // W
            Vector.doubles(new double [] { 0.0, -1.0}, 0.1),     // S
            Vector.doubles(new double [] { 2.0,  0.0}, 0.1),     // EE
            Vector.doubles(new double [] { 0.5,  0.0}, 0.1) } ); // E/2
      Vector polars = cartesians.map(la.la.Library.cartesian2polar);  // polar
      System.out.println( "cartesians = " + cartesians +
                        "\n    polars = " + polars  +
                 "\n back = " + polars.map(la.la.Library.polar2cartesian));
      R_D.Independent NxN = new R_D.Independent(Value.pair(
                                Normal, two ));  // i.e., Normal x Normal
      Estimator NxNest = NxN.estimator(Vector.values(
                    new Value.Tuple [] { q_10_10_pt1_ten, q_10_10_pt1_ten } ));
      System.out.println( "Model of cartesians:" );
      exercise( NxNest, cartesians );
      UPModel Polar = NxN.transform( la.la.Library.polar2cartesian );
      Estimator PolarEst = Polar.estimator(Vector.values(
                    new Value.Tuple [] { q_10_10_pt1_ten, q_10_10_pt1_ten } ));
      System.out.println( "Transformed to a model of polars:" );
      exercise( PolarEst, polars ); 

// Multinomial...
      Vector trial = Vector.ints(new int[]
         {2,     2,     2,     10,    10}   );  // input data (trials)
      Matrix freqs = Matrix.ints(new int[][]
         {{0,2}, {1,1}, {2,0}, {1,9}, {5,5}});  // output data (frequencies)
      Multinomial mnml = new Multinomial(two);  // 2 categories
      exercise( mnml.estimator(triv), trial.zip( freqs ) );

// Independent:
//              person = gender x height x age x leftwing;  people = person*
      MultiState boolUPM = new MultiState(Value.pair(ff, tt));  // bool UPModel
      Value.Tuple ht_bnds = Value.quad(oneR,threeR, point1,oneR),// 1.0..3.0 m
                  ag_bnds = Value.quad(zeroR,CR,    oneR,tenR);  // 0.0..100y
      // Model (g, ht, ag, lw) by (2state x N x N x 2state) ...
      exercise(new Independent(Value.tuple(
                   new UPModel[]{boolUPM, Normal, Normal, boolUPM}))
               .estimator(Value.tuple(new Value[]
                             {triv, ht_bnds, ag_bnds, triv})),
               people);

// extra test -- that of a derived function-model
      Independent.M pm    // pm is parameterised
        = (Independent.M)
          new Independent(Value.tuple(
                 new UPModel[]{boolUPM, Normal, Normal, boolUPM}))
            . estimator(Value.tuple(new Value[]
                        {triv, ht_bnds, ag_bnds, triv}))
            .ds2Model(people);
      System.out.println( "pm.functionModel(3,...) for 3:leftwing "
                        + "(note Indep. so shouldn't help)..." );
      exercise( pm.functionModel(3, la.la.Type.BOOL),  // ... .functionModel(,)
                Vector.zip(new Vector[]{ people.cols(new int[]{0, 1, 2}),
                                         people.col(3) } )
              );

// CPT: (generalised) Conditional Probability Table, leftwing -> gha
      Vector     gha = people.cols(new int[]{0, 1, 2});  // (g,ht,ag)*
      Vector l_x_gha = people.col(3).zip( gha );      // (l,(g,ht,ag))*

      // Model l -> (g,ht,ag) by (2state x N x N) per case of leftwing...
      CPT l2gha = new CPT(Value.triple( ff, tt,  // input datum bounds
                          new Independent(Value.triple(boolUPM,Normal,Normal))
                         )            );
      Value gha_estParams = Value.triple(triv, ht_bnds, ag_bnds);
      exercise(l2gha.estimator(gha_estParams), l_x_gha);


// Mixture:
      Vector b03 = Vector.ints(new int[]{0, 3});
      UPModel upMS03 = new MultiState(zero_3);
      Model mx = new Mixture(upMS03).apply(
                           Value.pair(
                              Vector.doubles(new double[]{0.5, 0.5}), // wts
                              Vector.values (new Value []         {   // sps
            Vector.doubles(new double[]{0.5, 0.25, 0.125, 0.125}) ,
            Vector.doubles(new double[]{0.125, 0.125, 0.25, 0.5}) }
                     ) ) );
      exercise(mx, Vector.ints(new int[]{0, 1, 2, 3}));

      UPModel Clss = new Independent(Value.pair(new NormalUPM(triv), boolUPM));
      UPModel upMx = new Mixture(Clss);
      exercise( upMx.estimator(Value.pair(
                   Value.quad(negTenR, tenR, point1, tenR),  // mu, sig bounds
                   triv )),
                Vector.doubles(new double[]
                  {-1.0, -1.1, -1.2, -1.0, -1.1, -1.2,
                       0.2, 0.1, 0.3, 0.2, 0.1, 0.3, 0.2,
                         1.0, 1.1, 1.2, 1.0, 1.1, 1.2 }, 0.1)  // 3 clusters?
                .zip(
                Vector.values(new  Value[]
                  {  ff,   ff,   ff,    ff,  tt,   ff,
                       tt,  tt,  ff,  tt,  ff,  tt,  tt,
                          tt,  tt,  tt,  ff,  tt,  tt })
            ).weight(2) );

// Intervals: divide input-space into intervals, and Model 'od' per interval,
//            pretty obvious where cut-points are in 'id'
      Vector id = Vector.doubles( new double[]
                  {1.1, 1.2, 1.3,  3.1, 3.2, 3.3,  2.1, 2.2, 2.3}, 0.1),
             od = Vector.doubles( new double[]
                  {1.4, 1.5, 1.4,  2.4, 2.5, 2.6,  3.4, 3.5, 3.6}, 0.1);
      exercise(new Intervals(Normal)
              .estimator(Value.quad(oneR,fourR,point01,oneR)), /* mu sig bnds*/
         id.zip(od) );

// Dependent: estimate a joint probability distribution ...
      // (note, l2gha defined earlier)
      // Model (l,(g,ht,ag)) by (2state x (bool->(2state x N x N)) ...
      Dependent l_l2gha = new Dependent(Value.pair(boolUPM, l2gha));
      Value  estParams = Value.pair(triv, gha_estParams);
      exercise(l_l2gha.estimator(estParams), l_x_gha);

// NaiveBayes: reverse a Dependent (Y x Y->X), to give X->Y ...
      Vector gha_x_l = gha.zip(people.col(3));  // ((g,ht,ag),l)*
      // Model (g,ht,ag)->l by using (2state x (bool->(2statexNxN))
      exercise(new NaiveBayes(l_l2gha).estimator(estParams), gha_x_l);

// Linear1: linear regression function-model, height -> mass
      Vector hxm = height.zip(mass);
      // System.out.println(" *** hxm = " + hxm );
      Linear1 ul1 = MML.Linear1;
      exercise(ul1.estimator(Value.quad(negCR,CR, point1,tenR)), hxm);

// Series:
      Vector vs = Vector.values( new Value[] {
           Vector.ints( new int[]{1, 3, 6, 2} ),
           Vector.ints( new int[]{2, 3, 5} ),
           Vector.ints( new int[]{4, 3, 1, 2, 2} ) } );
      UPModel GPu = new UPSeriesModel.K(Value.pair(Geometric0,  // lengths
                                        Poisson0)  ); // elements
      exercise(GPu .estimator(Value.pair(Value.tenR, Value.tenR)), vs);

      // Markov Model...
      Vector v_mm1 = Vector.values( new Value[] {       // a data-set
           Vector.ints( new int[]{1, 2, 1, 2, 2, 1, 2, 1, 1, 2} ),
           Vector.ints( new int[]{1, 2, 1, 2, 1, 1, 2, 2} ) } );
      UPModel mm1 = new Markov(Value.quad(Geometric0, one, one,two));
      exercise(mm1.estimator(Value.tenR), v_mm1);

// Sequences (rep. by Vectors):
      Discretes.Uniform Uni_1_2 = new Discretes.Uniform(pair(one, two));
      MultiState MS_1_2 = new MultiState(pair(one, two));
      UPModel Seq = new Sequences.K(Value.pair(Geometric0,  // lengths
                                               MS_1_2)  ); // elements
      exercise(Seq.estimator(Value.pair(tenR, triv)), v_mm1);

// Permutations
      Permutation.Uniform perm10 = new Permutation.Uniform(ten);
      exercise( perm10.Mdl, perm10.Mdl.random(8) );

// Missing Values,   Maybe.Just(v) and Maybe.None :
      Vector mvs = Vector.values( new Value [] {
         just(cts(0.0, 0.01)), just(cts(1.0, 0.01)), just(cts(-1.0, 0.01)),
         None} );
      Missing mUpm = new Missing( MML.Normal );
      exercise( mUpm.estimator(pair(triv,
                               quad(negTenR,tenR, point1,tenR))),  mvs );

      System.out.println("-- done mml.Test --");
    }//main

 }//Test class   https://www.cantab.net/users/mmlist/MML/A/

// (c) Lloyd Allison, 2012, 2017, 3 Mar 2020. Licence to 'mml.Test' is Gnu GPL.
