//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.util.*;   // standard functions, and also class Series
import la.la.Value;
import la.maths.Vector;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** UnParameterised (abstract) Function Model (aka regression) with
    input datum (independent variable), 'id', and output datum
    (dependent variable), 'od'. Fully parameterised Function Models
    are {@link UPFunctionModel.M} and {@link FunctionModel}. Also see
    {@link UPModel UPModel} and {@link UPSeriesModel UPSeriesModel}. */
public abstract class UPFunctionModel extends UPModel      // * UPFunctionModel
 {
   /** 'dp' is <I>given</I>, problem-defing parameters.  */
   public UPFunctionModel(Value dp)
    { super(dp); }//constructor

   /** Return the fully parameterised {@link M} with
       statistical parameters sp. */
   public FunctionModel apply(Value sp) { return sp2Model(0, 0, sp); }

   /** Given two-part message lengths, msg1 and msg2, and statistical
       parameter(s), sp, return a fully parameterised
       Function {@link M Model}. */
   public abstract FunctionModel sp2Model(double m1, double m2, Value sp);

   /** A class of Estimator of <I>FunctionModel</I>, that uses
       information from its enclosing {@link UPFunctionModel}. */
   public abstract class Est extends UPModel.Est    // * UPFunctionModel.Est
    { public Est(Value ps)
       { super(ps); }//constructor
      /** Given a data-set, ds, return a fully parameterised {@link M M}. */
      public FunctionModel ds2Model(Value ds)                // ds2Model(ds)
       { return ss2Model(stats((Vector) ds)); }
      /** Synonym for {@link #ds2Model ds2Model(ds)}. */
      public final FunctionModel ds2FunctionModel(Value ds)
       { return ds2Model(ds); }
      /** Given statistics ss&nbsp;=&nbsp;{@link #stats stats(ds)} of
          a data-set, ds, return a fully parameterised {@link M M}. */
      public abstract FunctionModel ss2Model(Value ss);    // * ss2Model(ss)
      /** Synonym for {@link #ss2Model ss2Model(ds)}. */
      public final FunctionModel ss2FunctionModel(Value ss)
       { return ss2Model(ss); }
    }//Est class
   // .........................................................................

    /** M, the (abstract) fully parameterised Function Model class of
        an {@link UPFunctionModel UnParameterised} Function Model. */
   public abstract class M extends FunctionModel         // * UPFunctionModel.M
    {
      public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp); }//constructor

   // public abstract Model condModel(Value id), * from class FunctionModel.

      /** Return sufficient statistics, ss&nbsp;=&nbsp;stats(ds,lo,hi),
          of elements [lo,hi) of a data-set ds as per the enclosing
          UPFunctionModel's
          {@link UPFunctionModel#stats(Vector,int,int) stats(ds,lo,hi)}.
          More on stats {@link UPModel#stats here}. */
      public Value stats(Vector ds, int lo, int hi)
       { return UPFunctionModel.this.stats(ds, lo, hi); }

      /** Use the enclosing UPFunctionModel's
      {@link UPFunctionModel#stats(boolean,Value,Value) stats(add,ss0,ss1)}
          to combine statisticses ss0 and ss1. */
      public Value stats(boolean add, Value ss0, Value ss1)
       { return UPFunctionModel.this.stats(add, ss0, ss1); }

      /** Calls {@link #asGiven(double,double) asGiven(0,msg2)}. */
      public FunctionModel asGiven(double msg2) { return asGiven(0, msg2); }

      /** Enables setting the first- and second-part message lengths, msg1
          and msg2, after having estimated the statistical parameter(s),
          say. Also see {@link #asGiven(double) asGiven(msg2)}. */
      public FunctionModel asGiven(double msg1, double msg2)
       { return sp2Model(msg1, msg2, statParams()); }

      /** Return a String representation of 'this' FunctionModel. */
      public String toString()
       { return "(" + UPFunctionModel.this + " " + statParams() + ")"; }

    }//UPFunctionModel.M (abstract) class, the Function Model
   // -------------------------------------------------------------------------

   /** A very simple UnParameterised Function Model that uses a
       single Model for every output datum, od, regardless of
       the input datum, id.  Also see the fully parameterised
       {@link K.M K.M}.  */
   public static class K extends UPFunctionModel           // UPFunctionModel.K
    {
      /** upm, the UnParameterised Model, to be fully
          {@link K.M#mdl parameterised} and then applied to the
          output datum in every input "case" by {@link K.M K.M}.  */
      public final UPModel upm;

      /** The problem-defining parameter 'upm' is used to set {@link #upm}. */
      public K(Value upm)
       { super(upm);  this.upm = (UPModel)upm; }//constructor

      /** Two-part message lengths m1 and m2, and statistical parameter
          'sp', for fully parameterised {@link K.M K.M}. */
      public M sp2Model(double m1, double m2, Value sp)
       { return new M(m1, m2, sp); }

      /** Given a data-set, 'ds', sufficient statistics,
          ss&nbsp;=&nbsp;upm.stats(ds.col(1),lo,hi), on the
          output datum <I>only</I>, are as per {@link #upm upm}.
          More on stats {@link UPModel#stats here}. */
      public Value stats(Vector ds, int lo, int hi)
       { return upm.stats(ds.col(1), lo, hi); /* output data values */ }

      /** {@link #upm upm}.stats(add,.,.).  */
      public Value stats(boolean add, Value ss0, Value ss1)
       { return upm.stats(add, ss0, ss1); }

      /** The Estimator gets the Model for the (every)
          output datum, od, from {@link #upm}.estimator(ps). */
      public Est estimator(final Value ps)
       { return new Est(ps)
          { public M ss2Model(Value ss)
             { Model om = upm.estimator(ps).ss2Model(ss); // Model / "case"
               return new M(om.msg1(), om.msg2(), om.statParams());
             }//ss2Model(ss)
       }; }//estimator(ps)
      // ......................................................................

      /** The fully parameterised Konstant function-model.
          It models the output datum with {@link #mdl mdl},
          regardless of the input data value.
          Also see the unparameterised&nbsp;{@link K K}.  */
      public class M extends UPFunctionModel.M           // UPFunctionModel.K.M
       {
         /** The Model of the output datum, od (for every input, id).
             It is {@link #upm upm}, parameterised. */
         public final Model mdl;

         /** Statistical parameter sp is for
         {@link #mdl mdl}&nbsp;=&nbsp;{@link #upm upm}.apply(sp). */
         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp); this.mdl = upm.apply(sp); }//constructor

         /** Return {@link #mdl mdl}; input datum id is ignored in K. */
         public Model condModel(Value id) { return mdl; }

         /** Given a data-set, 'ds', statistics,
             ss&nbsp;=&nbsp;mdl.stats(ds.col(1),lo,hi), on the
             output datum <I>only</I>, are as per {@link #mdl mdl}.
             More on stats {@link UPModel#stats here}. */
         public Value stats(Vector ds, int lo, int hi)
          { return mdl.stats(ds.col(1), lo, hi); /* output data values */ }

         /** {@link #mdl mdl}.stats(add,.,.).  */
         public Value stats(boolean add, Value ss0, Value ss1)
          { return mdl.stats(add, ss0, ss1); }

         /** Negative log likelihood; note, statistics are
             <KBD>ss&nbsp;=&nbsp;{@link K#stats stats(ds)}</KBD>. */
         public double nlLH(Value ss)
          { return mdl.nlLH(ss); }

       }//UPFunctionModel.K.M class, the Function Model

    }//UPFunctionModel.K class, the UPFunctionModel
   // -------------------------------------------------------------------------

 }//UPFunctionModel (abstract) class, the UPFunctionModel(s)

// (c)Lloyd Allison 2012, Thu 7 Dec 2017. Licence to UPFunctionModel is Gnu GPL.
